//  WXFeeds Server
//
//  Garry Lockyer
//  Garry@lockyer.ca
//  C: +1.250.689.0686
//
//  This is intended to work with WXFeeds - a service extension to WeeWX.
//  It provides authentication and routes to create a proxy server to 
//  hide various API access keys and secrets.
//
//  Supplied credentials are authenticated against credentials in '.env' 
//  environment variables and a JSON Web Token is returned for use in 
//  subsequent requests to the server.
//
//  The username (WX_FEEDS_USERNAME) is stored in plain text.
//  The password (WX_FEEDS_PASSWORD) is stored as a bcrypt hash.
//
//  Use a service such as https://passwordhashing.com/BCrypt to generate
//  a hashed password (to be stored as an environment variable).
//
//  WX_FEEDS_ACCESS_TOKEN_SECRET was generated using node command:
//
//      require('crypto').randomBytes(64).toString('hex')
//

"use strict";

import Express, { Router, json } from "express";
import { createServer } from "https";
import { readFileSync } from "fs";
import {} from 'dotenv/config.js';
import jwt from "jsonwebtoken";
import verify from "jsonwebtoken";
import get from "axios";
import { parse } from "url";
import { createClient } from "redis";
import bcrypt from 'bcrypt';

const port      = process.env.WX_FEEDS_SERVER_PORT;
//console.log(`port: ${port}`);
const keyFile   = process.env.WX_FEEDS_KEY_FILE;
const certFile  = process.env.WX_FEEDS_CERT_FILE;
const username  = process.env.WX_FEEDS_USERNAME;
//console.log(`username: ${username}`)
const password  = process.env.WX_FEEDS_PASSWORD;

const REDIS_PORT = process.env.WX_FEEDS_REDIS_PORT;

const CacheMissResponseStatusCode = 298
const CacheHitResponseStatusCode = 299

const app = Express()
//const express = Express()
//const router = app.Router
//app.use("/", router)
app.use(json())
app.use(Express.json());
//const jwt = jsonwebtoken();


const redisClient = createClient(REDIS_PORT)

//  Standard listener.  Kept in caae we need to debug using http.
//app.listen( PORT )

//  Create an https server.

createServer(
    {
      key: readFileSync(keyFile),
      cert: readFileSync(certFile),
    },
    app
  )
  .listen(port, function (err) {
    if (err) {
      console.log(
        "There was an error starting the WxFeedsServer!  Err: ",
        err
      );
      return;
    }

    console.log(`WxFeedsServer Listening On Port: ${port}.`);
  });

//  Route to log user in.

app.post("/login", async (req, res) =>
  {
    //  Authenticate the user.

    //console.log("In WXFeedsServer/login. . .")
    //console.log(`username: ${username}`)
  
    var userAuthenticated = false;
  
    if (
         username == null ||
         username == "" ||
         password == null ||
         password == ""
       )
    {
      return res.status(400).send("Cannot find username and/or password!");
    }

    //console.log(`req.body.username: ${req.body.username}`)

    if (req.body.username === username)
    {
      try
      {
        if ((await bcrypt.compare(req.body.password, password)) == true)
        {
          userAuthenticated = true;
        } else {
                 return res.send("Login Failed");
               }
      }
      catch {
        return res.status(500).json({ error: "bcrypt exception!" });
      }
    } else {
             return res.json({ error: "Username does not match!" });
           }
  
    if (userAuthenticated == true)
    {
      //  The user has been authenticated - send them an access token.
  
      const user = { name: username };
  
      const accessToken = jwt.sign(
          user,
        process.env.WX_FEEDS_ACCESS_TOKEN_SECRET, { expiresIn: process.env.WX_FEEDS_TOKEN_EXPIRY_SECONDS * 1 }
      );
  
      res.json({ loginStatus: "Login Successful", accessToken: accessToken, tokenExpiry: process.env.WX_FEEDS_TOKEN_EXPIRY_SECONDS });
    }
  }
);

//  Function to authenticate a JSON Web Token.
//
//  Called by all routes except Login.

function authenticateToken(req, res, next) {
  //console.log(`req.url: ${req.url}`);

  const query = parse(req.url, true).query;
  const token = query.accessToken;
  //console.log(`token: ${token}`);

  if (token == null) return res.sendStatus(401);

    jwt.verify(token, process.env.WX_FEEDS_ACCESS_TOKEN_SECRET, (err, user) => {
    //console.log(`Token: ${token}`);

    if (err) return res.sendStatus(403);

    // Token is good!

    //console.log("Token Is Good!");
    //console.log(`decode: ${decode}`);

    next();
  });
}

//  Handler to verify JWT token.

app.get("/verify", authenticateToken, (req, res) => {
  //console.log("Access Token Verified OK!");

  res.send("Login Successful");
});

//  Handler for WXFeeds proxy request(s).
//
//  This handler handles all WXFeeds proxy requests.
//
//  Example URL: 'https://rqleo158.hostpapavps.net:61712/WXFeedsProxy?ContentType=json
//                                                                   &Service=GoogleMapsGeocoding
//                                                                   &CacheKey=Lockyers
//                                                                   &CacheData=True
//                                                                   &CacheExpiry_minutes=1440
//                                                                   &ForceCacheMiss=False
//                                                                   &url=https://maps.googleapis.com/maps/api/geocode/json%3Faddress=101+-+2450+Radio+Tower+Road,+Oliver+BC,+V0H+1T1,+Canada%26components=country:CA
//                                                                   &accessToken=eyJhbGciOi. . . .3P9LrNOPs5MYLbitb-ZzpIU7A50'

app.get("/WXFeeds", authenticateToken, (req, res, next) => {
  //console.log(`url: ${req.url}`);

  // The request was authenticated - make the call.

  const query = parse(req.url, true).query;

  const ContentType = query.ContentType;
  //console.log(`ContentType: ${ContentType}`);
  const Service = query.Service;
  //console.log(`Service: ${Service}`);
  const CacheKey_str = query.CacheKey;
  //console.log(`CacheKey: ${CacheKey_str}`);
  const CacheData_str = query.CacheData;
  //console.log(`CacheData: ${CacheData_str}`);
  const CacheExpiry_minutes_str = query.CacheExpiry_minutes;
  //console.log(`CacheExpiry_minutes: ${CacheExpiry_minutes_str}`);
  const ForceCacheMiss_str = query.ForceCacheMiss;
  //console.log(`ForceCacheMiss: ${ForceCacheMiss_str}`);
  var URL = query.url;
  //console.log(`URL: ${URL}`);

  var cacheKey = Service + ":" + CacheKey_str
  cacheKey = cacheKey.replace(/ /g, "");
  cacheKey = cacheKey.replace(/,/g, "");
  cacheKey = cacheKey.replace(/'/g, "");
  //console.log(`cacheKey: ${cacheKey}`);
 
  var contentIsJSON = false;
  if (ContentType != undefined &&
      ContentType.toUpperCase() == "JSON")
  {
    contentIsJSON = true;  
  }

  var setDataCache = false
  if (CacheData_str != undefined &&
      CacheData_str.toUpperCase() == "TRUE")
  {
    setDataCache = true;  
  }

  var expiry = -1
  if (CacheExpiry_minutes_str != undefined)
  {
    expiry = parseInt(CacheExpiry_minutes_str);
  }
  //console.log(`expiry: ${expiry}`);

  var forceCacheMiss = false;
  if (ForceCacheMiss_str != undefined &&
      ForceCacheMiss_str.toUpperCase() == "TRUE")
  {
    forceCacheMiss = true;
  }
  //console.log(`forceCacheMiss: ${forceCacheMiss}`);

  switch( Service )
  {
    case "GoogleMapsGeocoding":
    case "GoogleMapsPlaces":
    case "GoogleMapsMaps":

      URL += "&key=" + process.env.WX_FEEDS_GOOGLE_MAPS_API_KEY;

      break;
      
    case "AirPortCodesAirportInfo":
          
        URL += "&key=" + process.env.WX_FEEDS_AIR_PORT_CODES_API_KEY;
        URL += "&secret=" + process.env.WX_FEEDS_AIR_PORT_CODES_API_SECRET;

        break;
    
    case "WSDOTBorderCrossings":
          
      URL += "&AccessCode=" + process.env.WX_FEEDS_WS_DOT_SERVICES_ACCESS_CODE;

      break;
    
    case "NOAANWSPoint":
    case "NOAANWSAWCMETAR":
    case "NOAANWSForecast":
    case "NOAANWSHourlyForecast":
    case "NOAANWSSPC":
    case "NOAANWSNHC":
    case "ECCityWeather":
    case "ECCityAlert":
    case "ECPublicAlert":
    case "ECMarineForecast":
    case "DriveBCEvents":
    case "511AlbertaRoadConditions":

      break;

    default:
      
      console.log("Unknown WXFeeds Service!");
      return res.status(999);
  }
 
  //console.log(`URL: ${URL}`);

  const getService = async () => {
    try
    {
      // Try to get the data from the cache.

        redisClient.get(cacheKey, async (err, cacheData) => {
        if (cacheData && forceCacheMiss == false) {
          //  Cache Hit - data was found in the cache so return it to the caller.

          //console.log( "Cache Hit!")

          if (contentIsJSON == true)
          {
            return res.status(CacheHitResponseStatusCode).send(JSON.parse(cacheData));
          } else
          {
            return res.status(CacheHitResponseStatusCode).send(cacheData);
          }
        } else {
          //  Cache Miss - the data was not found in the cache so get it from the source.

          //console.log( "Cache Miss!")

          try
          {
            const response = await get(URL);

            //  We should have good data - cache it and return it to caller.

            if ( setDataCache == true &&
                 expiry > 0 )
            {
              //console.log("Caching data . . .");

              expiry *= 60;

              const expiry_str = expiry.toString();

              if (contentIsJSON == true)
              {
                const JSON_str = JSON.stringify(response.data);
                redisClient.setex(cacheKey, BigInt(expiry_str), JSON_str);
              } else
              {
                redisClient.setex(cacheKey, BigInt(expiry_str), response.data);
              }
            }

            if (contentIsJSON == true)
            {
              res.status(CacheMissResponseStatusCode).json(response.data);
            } else
            {
              res.send(response.data);
            }
          } catch (error) {
            if (error.response) {
              // The request was made and the server responded with a status code
              // that falls out of the range of 2xx.

              //console.log("error.response")
              //console.log(`error.response.data: ${error.response.data}`);
              console.log(`/WXFeeds/service - error.response.status: ${error.response.status}`);
              //console.log(`error.response.headers: ${error.response.headers}`);

              //return res.status(error.response.status);
              res.status(error.response.status);
            } else if (error.request) {
              // The request was made but no response was received
              // `error.request` is an instance of XMLHttpRequest in the browser and an instance of
              // http.ClientRequest in node.js
              //console.log("error.request");
              console.log(`error.request:)${error.request}`);
            } else {
              // Something happened in setting up the request that triggered an Error
              //console.log("error.message");
              console.log("error.message", error.message);

              return res.status(998);
            }
            console.log("error.config");
            console.log(`error.config: {error.config}`);

            return res.status(999);
          }
        }
      });
    } catch (error) {
      console.log(`redis error: ${error}`);

      return res.status(997);
    }
  };

  getService();
});
