"""
# WXFeeds.py
#
# Copyright (c) 2021 Garry Lockyer <Garry@lockyer.ca>
#
"""
WXFEEDS_VERSION = "Version 0.10 - February 25, 2021"
"""
# Version 0.10
#
# - deploy to Area 27
#
# - integration of WXFeeds authorization and servers 
#   into a single server.
#
# Version 0.09
#
# - fix error handling for NOAANWSSPC:
#   - received apparently goof return with 0 entries so
#     references to data elements fail, but validation 
#     function returned True!?!?
#
# - add a few endpoints to WXFeeds server.
#
# Version 0.08
#
# - more cache controls for several endpoints.
#
# Version 0.07
#
# - added cache controls for several endpoints.
#
# Verson 0.06
#
# - Fix error handle for DriveBC areas.
#
# - Fix file deletion for Cheetah workaround.
#
# Version 0.05
#
# - rewrite of  511 Alberta caching.
#
# Version 0.04
#
# - Got rid of all html and js cache files.
#   - now use actual file modification time to determine if stale.
#
# Version 0.03
#
# - Works with WeewX 4.4.0.
#
# - Works with Belchertown Skin 1.2 but does not require Belchertown.
#
# -All feeds are now cached.
#  - but caching controls do not yet exist for all feeds!
# 
# - 511Alberta Road Conditions caching
#
# Version 0.02 - January 28, 2021"
#
# - back to Raspberry Pi OS for development
#  
# - removed use of polylabel
#   - couldn't find shapely module for RPi
#   - and polylabel calculation very "expensive"
#
# -511Alberta Road Conditions greatly improved
#  - before caching
#
# Version 0.01 - January 7, 2021"
#
# Initial Version - convert Belchertown extension to WeeWX service.
#
# Works with WeeWX V4.3 and Belchertown Skin V1.2.
#
# Developed on Ubuntu Raspberry Pi Desktop 20.10.
#
# Deployed to:
# - Raspberry Pi OS Lite on Raspberry Pi Zero W
#
# Distributed under the terms of the GNU GENERAL PUBLIC LICENSE
#
# This program pulls weather (wx) information and other information 
# from multiple sources (feeds).
#
# It currently supports:
# - ATOM feeds from Environment Canada
#   - weather conditions
#     - forecasts
#     - alerts
#   - marine forecasts
#   - public alerts
#
# - METARs and TAFS from NOAA NWS Aviation Weather Center
# - Forecasts from NOAA NWS
# - Alerts from NOAA NWS Storm Prediction Center (land and sea)
# - Hurricane information from NOAA NWS National Hurricane Center
# - Road conditions and other data from:
#   - Drive BC
#   - 511 Alberta
#   - Washington State - Department of Transportion
"""

import datetime
import weewx
from weewx.engine import StdService
import logging
from pathlib import Path
import pathlib
import sys
import os
import gc
import time
import json
import requests
import feedparser
import dateutil
from dateutil import parser, tz
import glob
import pytz
from pytz import timezone
import isodate
import polyline
import haversine
from bisect import bisect_left
import pickle
import random
from typing import Callable

# Some global constants:

WXFEEDS_KEY                                                 = "WXFeeds"
SETUP_KEY                                                   = "Setup"
REPORTS_KEY                                                 = "Reports"
FEEDS_KEY                                                   = "Feeds"
OUTPUT_KEY                                                  = "Output"
OPTIONS_KEY                                                 = "Options"
FILE_KEY                                                    = "file"
FILE_SEQUENCE_KEY                                           = "file_sequence"

CHEETAH_MEMORY_LEAK_WORKAROUND_KEY                          = "cheetah_memory_leak_workaround"

TAKE_OUT_THE_TRASH_KEY                                      = "take_out_the_trash"

FILE_EXPIRY_MINUTES_KEY                                     = "file_expiry_minutes"

WXFEEDS_ROOT_KEY                                            = "WXFeeds_root"
WXFEEDS_ICON_BASE_KEY                                       = "WXFeeds_icon_base"

WXFEEDS_SERVER_USE_KEY                                      = "WXFeeds_server_use"
WXFEEDS_SERVER_URL_KEY                                      = "WXFeeds_server_url"
WXFEEDS_SERVER_USERNAME_KEY                                 = "WXFeeds_server_username"
WXFEEDS_SERVER_PASSWORD_KEY                                 = "WXFeeds_server_password"

HTTP_RESPONSE_CODE_OK                                       = 200   # Standard HTTP OK Response Code.
WXFEEDS_SERVER_RESPONSE_CODE_CACHE_MISS                     = 298   # This must match CacheMissResponseStatusCode in 'WXFeedsServer.py'!
WXFEEDS_SERVER_RESPONSE_CODE_CACHE_HIT                      = 299   # This must match CacheHitResponseStatusCode in 'WXFeedsServer.py'!

GOOGLE_MAPS_USE_KEY                                         = "googlemaps_use"
GOOGLE_MAPS_API_KEY_KEY                                     = "googlemaps_api_key"

GOOGLE_MAPS_MAPS_USE_KEY                                    = "googlemaps_maps_use"
GOOGLE_MAPS_MAPS_SERVER_USE_KEY                             = "googlemaps_maps_server_use"

GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT              = 10080 # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM              = 43200 # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.
GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT             = 10080 # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM             = 43200 # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.

GOOGLE_MAPS_GEOCODING_USE_KEY                               = "googlemaps_geocoding_use"
GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_USE_KEY                   = "googlemaps_geocoding_local_cache_use"
GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_EXPIRY_MINUTES_KEY        = "googlemaps_geocoding_local_cache_expiry_minutes"
GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_FORCE_MISS_KEY            = "googlemaps_geocoding_local_cache_force_miss"
GOOGLE_MAPS_GEOCODING_SERVER_USE_KEY                        = "googlemaps_geocoding_server_use"
GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_USE_KEY                  = "googlemaps_geocoding_remote_cache_use"
GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_EXPIRY_MINUTES_KEY       = "googlemaps_geocoding_remote_cache_expiry_minutes"
GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_FORCE_MISS_KEY           = "googlemaps_geocoding_remote_cache_force_miss"

GOOGLE_MAPS_PLACES_USE_KEY                                  = "googlemaps_places_use"
GOOGLE_MAPS_PLACES_LOCAL_CACHE_USE_KEY                      = "googlemaps_places_local_cache_use"
GOOGLE_MAPS_PLACES_LOCAL_CACHE_EXPIRY_MINUTES_KEY           = "googlemaps_places_local_cache_expiry_minutes"
GOOGLE_MAPS_PLACES_LOCAL_CACHE_FORCE_MISS_KEY               = "googlemaps_places_local_cache_force_miss"
GOOGLE_MAPS_PLACES_SERVER_USE_KEY                           = "googlemaps_places_server_use"
GOOGLE_MAPS_PLACES_REMOTE_CACHE_USE_KEY                     = "googlemaps_places_remote_cache_use"
GOOGLE_MAPS_PLACES_REMOTE_CACHE_EXPIRY_MINUTES_KEY          = "googlemaps_places_remote_cache_expiry_minutes"
GOOGLE_MAPS_PLACES_REMOTE_CACHE_FORCE_MISS_KEY              = "googlemaps_places_remote_cache_force_miss"

GOOGLE_MAPS_GEOCODING_SERVICE_URL                           = "https://maps.googleapis.com/maps/api/geocode/json"
GOOGLE_MAPS_PLACES_SERVICE_URL                              = "https://maps.googleapis.com/maps/api/place/findplacefromtext/json"
GOOGLE_MAPS_MAPS_SERVICE_URL                                = "https://maps.googleapis.com/maps/api/js?callback=initMap"

AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT           = 10080  # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM           = 43200  # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.
AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT          = 10080 # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM          = 43200 # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.

AIR_PORT_CODES_USE_KEY                                      = "air-port-codes_use"
AIR_PORT_CODES_API_KEY_KEY                                  = "air-port-codes_api_key"
AIR_PORT_CODES_API_SECRET_KEY                               = "air-port-codes_api_secret"
AIR_PORT_CODES_LOCAL_CACHE_USE_KEY                          = "air-port-codes_local_cache_use"
AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_KEY               = "air-port-codes_local_cache_expiry_minutes"
AIR_PORT_CODES_LOCAL_CACHE_FORCE_MISS_KEY                   = "air-port-codes_local_cache_force_miss"
AIR_PORT_CODES_SERVER_USE_KEY                               = "air-port-codes_server_use"
AIR_PORT_CODES_REMOTE_CACHE_USE_KEY                         = "air-port-codes_remote_cache_use"
AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_KEY              = "air-port-codes_remote_cache_expiry_minutes"
AIR_PORT_CODES_REMOTE_CACHE_FORCE_MISS_KEY                  = "air-port-codes_remote_cache_force_miss"

AIR_PORT_CODES_AIRPORT_INFO_SERVICE_URL                     = "https://www.air-port-codes.com/api/v1/single"

NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT          = 10080 # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM          = 43200 # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.
NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT         = 10080 # 7 30 days in minutes = 60 min /hr * 24 hr/day * 7 day = 10,80.
NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM         = 43200 # 30 days in minutes = 60 min /hr * 24 hr/day * 30 day = 43,200.

NOAA_NWS_POINTS_USE_KEY                                     = "noaa_nws_forecasts_use"
NOAA_NWS_POINTS_LOCAL_CACHE_USE_KEY                         = "noaa_nws_forecasts_local_cache_use"
NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY              = "noaa_nws_forecasts_local_cache_expiry_minutes"
NOAA_NWS_POINTS_LOCAL_CACHE_FORCE_MISS_KEY                  = "noaa_nws_forecasts_local_cache_force_miss"
NOAA_NWS_POINTS_SERVER_USE_KEY                              = "noaa_nws_forecasts_server_use"
NOAA_NWS_POINTS_REMOTE_CACHE_USE_KEY                        = "noaa_nws_forecasts_remote_cache_use"
NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY             = "noaa_nws_forecasts_remote_cache_expiry_minutes"
NOAA_NWS_POINTS_REMOTE_CACHE_FORCE_MISS_KEY                 = "noaa_nws_forecasts_remote_cache_force_miss"

NOAA_NWS_POINTS_SERVICE_URL                                 = "https://api.weather.gov/points/"

NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT           = 10
NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM           = 120
NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT          = 5
NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM          = 120

NOAA_NWS_FORECASTS_USE_KEY                                      = "noaa_nws_forecasts_use"
NOAA_NWS_FORECASTS_LOCAL_CACHE_USE_KEY                          = "noaa_nws_forecasts_local_cache_use"
NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY               = "noaa_nws_forecasts_local_cache_expiry_minutes"
NOAA_NWS_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY                   = "noaa_nws_forecasts_local_cache_force_miss"
NOAA_NWS_FORECASTS_SERVER_USE_KEY                               = "noaa_nws_forecasts_server_use"
NOAA_NWS_FORECASTS_REMOTE_CACHE_USE_KEY                         = "noaa_nws_forecasts_remote_cache_use"
NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY              = "noaa_nws_forecasts_remote_cache_expiry_minutes"
NOAA_NWS_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY                  = "noaa_nws_forecasts_remote_cache_force_miss"

NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT    = 10
NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM    = 120
NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT   = 5
NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM   = 120

NOAA_NWS_HOURLY_FORECASTS_USE_KEY                               = "noaa_nws_hourly_forecasts_use"
NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_USE_KEY                   = "noaa_nws_hourly_forecasts_local_cache_use"
NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY        = "noaa_nws_hourly_forecasts_local_cache_expiry_minutes"
NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY            = "noaa_nws_hourly_forecasts_local_cache_force_miss"
NOAA_NWS_HOURLY_FORECASTS_SERVER_USE_KEY                        = "noaa_nws_hourly_forecasts_server_use"
NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_USE_KEY                  = "noaa_nws_hourly_forecasts_remote_cache_use"
NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY       = "noaa_nws_hourly_forecasts_remote_cache_expiry_minutes"
NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY           = "noaa_nws_hourly_forecasts_remote_cache_force_miss"

WS_DOT_ACCESS_CODE_KEY                                      = "ws_dot_access_code"
WS_DOT_SERVICES_USE_KEY                                     = "ws_dot_services_use"
WS_DOT_SERVICES_ACCESS_CODE_KEY                             = "ws_dot_services_access_code"
WS_DOT_SERVICES_USE_LOCAL_CACHE_KEY                         = "ws_dot_services_use_local_cache"
WS_DOT_SERVICES_CACHE_EXPIRY_MINUTES_DEFAULT                = 10080
WS_DOT_SERVICES_CACHE_EXPIRY_MINUTES_KEY                    = "ws_dot_services_cache_expiry_minutes"
WS_DOT_SERVICES_FORCE_CACHE_MISS_KEY                        = "ws_dot_services_force_cache_miss"

NAME_KEY                                                    = "name"

SETTINGS_KEY                                                = "settings"
SKIN_KEY                                                    = "skin"
GENERATED_FILES_ROOT_KEY                                    = "generated_files_root"
HTML_FILES_ROOT_KEY                                         = "html_files_root"
CUSTOM_CSS_PATH_KEY                                         = "custom_css_path"
DISABLE_WXFEEDS_CSS_KEY                                     = "disable_WXFeeds_css"
INCLUDE_FILENAMES_TO_DELETE_KEY                             = "include_filenames_to_delete"
OTHER_FILENAMES_TO_DELETE_KEY                               = "other_filenames_to_delete"

ADD_BACK_TO_TOP_BUTTON_KEY                                  = "add_back_to_top_button"
BACK_TO_TOP_BUTTON_ADDED_KEY                                = "back_to_top_button_added"
GENERATED_FILE_KEY                                          = "generated_file"
CHANGING_GENERATED_FILE_KEY                                 = "changing_generated_file"
CURRENT_GENERATED_FILENAME_KEY                              = "current_generated_filename"
CURRENT_GENERATED_FILE_SEQUENCE_KEY                         = "current_generated_file_sequence"
CURRENT_GENERATED_FILE_TYPE_KEY                             = "current_generated_file_type"
NEW_GENERATED_FILENAME_KEY                                  = "new_generated_filename"
NEW_GENERATED_FILE_SEQUENCE_KEY                             = "new_generated_file_sequence"
NEW_GENERATED_FILE_TYPE_KEY                                 = "new_generated_file_type"

FEED_REPORTS_KEY                                            = "reports"
FEED_GENERATED_FILES_KEY                                    = "generated_files"
FEED_GENERATED_FILE_SEQUENCES_KEY                           = "generated_file_sequences"

SKIN_ROOT_KEY                                               = "SKIN_ROOT"
HTML_ROOT_KEY                                               = "HTML_ROOT"

WEEWX_ROOT                                                  = "WEEWX_ROOT"
SKIN_ROOT                                                   = "SKIN_ROOT"
HTML_ROOT                                                   = "HTML_ROOT"

STATION_KEY                                                 = "Station"
LOCATION_KEY                                                = "location"
LATITUDE_KEY                                                = "latitude"
LONGITUDE_KEY                                               = "longitude"
STATION_URL_KEY                                             = "station_url"
STD_ARCHIVE_KEY                                             = "StdArchive"
ARCHIVE_INTERVAL_KEY                                        = "archive_interval"
STD_REPORT_KEY                                              = "StdReport"

NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT      = 5
NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM      = 60
NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT     = 1
NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM     = 60

NOAA_NWS_AWC_METARS_LOCAL_CACHE_USE_KEY                     = "noaa_nws_awc_metars_local_cache_use"
NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_KEY          = "noaa_nws_awc_metars_local_cache_expiry_minutes"
NOAA_NWS_AWC_METARS_LOCAL_CACHE_FORCE_MISS_KEY              = "noaa_nws_awc_metars_local_cache_force_miss"
NOAA_NWS_AWC_METARS_SERVER_USE_KEY                          = "noaa_nws_awc_metars_server_use"
NOAA_NWS_AWC_METARS_REMOTE_CACHE_USE_KEY                    = "noaa_nws_awc_metars_remote_cache_use"
NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_KEY         = "noaa_nws_awc_metars_remote_cache_expiry_minutes"
NOAA_NWS_AWC_METARS_REMOTE_CACHE_FORCE_MISS_KEY             = "noaa_nws_awc_metars_remote_cache_force_miss"

NOAA_NWS_AWC_METARS_SERVICE_URL                             = "https://www.aviationweather.gov/adds/metars"

NOAA_NWS_AWC_METARS                                         = "NOAA NWS - Aviation Weather Center - METARs"
NOAA_NWS_AWC_METARS_KEY                                     = "NOAA NWS AWC - METARs"
NOAA_NWS_AWC_METARS_URL                                     = "https://www.aviationweather.gov/metar"
STATIONS_KEY                                                = "Stations"
GET_METARS_KEY                                              = "get_metars"
GET_TAFS_KEY                                                = "get_tafs"
TRANSLATE_KEY                                               = "translate"
PAST_HOURS_KEY                                              = "past_hours"
STATION_WEBSITE_URL_KEY                                     = "station_website_url"
LOCATION_WEBSITE_URL_KEY                                    = "location_website_url"

STANDARD                                                    = "standard"
TRANSLATED                                                  = "translated"

DISPLAY_LINK_TO_NAVCAN_AWWS_KEY                             = "display_link_to_navcan_awws"
NAVCAN_AWWS_URL                                             = "https://flightplanning.navcanada.ca/"

DISPLAY_LINK_TO_US_AWC_BRIEFING_KEY                         = "display_link_to_us_awc_briefing"
US_AWC_BRIEFING_URL                                         = "https://www.aviationweather.gov/briefing/"

NOAA_NWS_SPC_USE_LOCAL_CACHE                                = "noaa_nws_spc_use_local_cache"
NOAA_NWS_SPC_CACHE_EXPIRY_MINUTES_DEFAULT                   = 0
NOAA_NWS_SPC_CACHE_EXPIRY_MINUTES                           = "noaa_nws_spc_cache_expiry_minutes"
NOAA_NWS_SPC_FORCE_CACHE_MISS                               = "noaa_nws_spc_force_cache_miss"

NOAA_NWS_NHC_USE_LOCAL_CACHE                                = "noaa_nws_nhc_use_local_cache"
NOAA_NWS_NHC_CACHE_EXPIRY_MINUTES_DEFAULT                   = 0
NOAA_NWS_NHC_CACHE_EXPIRY_MINUTES                           = "noaa_nws_nhc_cache_expiry_minutes"
NOAA_NWS_NHC_FORCE_CACHE_MISS                               = "noaa_nws_nhc_force_cache_miss"

SEARCH_CRITERIA_KEY                                         = "search_criteria"
ZOOM_KEY                                                    = "zoom"
MAP_ZOOM_DEFAULT                                            = 10

ENVIRONMENT_CANADA                                          = "Environment Canada"
ENVIRONMENT_CANADA_KEY                                      = "Environment Canada"
ENVIRONMENT_CANADA_URL                                      = "https://www.canada.ca/en/services/environment/weather.html"
CITYS_KEY                                                   = "Citys"
ALERTS_FIRST_KEY                                            = "alerts_first"
INLINE_FORECAST_KEY                                         = "inline_forecast"
GET_WEATHER_KEY                                             = "get_weather"
GET_ALERTS_KEY                                              = "get_alerts"
MARINE_FORECASTS_KEY                                        = "Marine Forecasts"
PUBLIC_ALERTS_KEY                                           = "Public Alerts"
NO_ALERTS_IN_EFFECT                                         = "No alerts in effect"
WARNING                                                     = "WARNING"
WATCH                                                       = "WATCH"
SPECIAL_AIR_QUALITY_STATEMENT                               = "SPECIAL AIR QUALITY STATEMENT"
SPECIAL_WEATHER_STATEMENT                                   = "SPECIAL WEATHER STATEMENT"
ENDED                                                       = "ENDED"
WARNINGS_AND_WATCHES                                        = "Warnings and Watches"
CURRENT_CONDITIONS                                          = "Current Conditions"
WEATHER_FORECASTS                                           = "Weather Forecasts"
NO_WATCHES_OR_WARNINGS_IN_EFFECT                            = "No watches or warnings in effect"
MARINE_FORECASTS                                            = "Marine Forecasts"
SHORT_TERM_MARINE_FORECAST_PHRASE                           = "Forecast for "
EXTENDED_MARINE_FORECAST_PHRASE                             = "Extended Forecast - "
MARINE_WARNINGS_AND_WATCHES                                 = "Marine Warnings and Watches"

ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT      = 10
ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM      = 120
ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT     = 5
ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM     = 120

ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_USE_KEY                     = "environment_canada_city_weather_local_cache_use"
ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_KEY          = "environment_canada_city_weather_local_cache_expiry_minutes"
ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_FORCE_MISS_KEY              = "environment_canada_city_weather_local_cache_force_miss"
ENVIRONMENT_CANADA_CITY_WEATHER_SERVER_USE_KEY                          = "environment_canada_city_weather_server_use"
ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_USE_KEY                    = "environment_canada_city_weather_remote_cache_use"
ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_KEY         = "environment_canada_city_weather_remote_cache_expiry_minutes"
ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_FORCE_MISS_KEY             = "environment_canada_city_weather_remote_cache_force_miss"

ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT       = 5
ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM       = 60
ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT      = 2
ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM      = 60

ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_USE_KEY                      = "environment_canada_city_alerts_local_cache_use"
ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY           = "environment_canada_city_alerts_local_cache_expiry_minutes"
ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_FORCE_MISS_KEY               = "environment_canada_city_alerts_local_cache_force_miss"
ENVIRONMENT_CANADA_CITY_ALERTS_SERVER_USE_KEY                           = "environment_canada_city_alerts_server_use"
ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_USE_KEY                     = "environment_canada_city_alerts_remote_cache_use"
ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY          = "environment_canada_city_alerts_remote_cache_expiry_minutes"
ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_FORCE_MISS_KEY              = "environment_canada_city_alerts_remote_cache_force_miss"

ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT     = 5
ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM     = 60
ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT    = 2
ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM    = 60

ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_USE_KEY                    = "environment_canada_public_alerts_local_cache_use"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY         = "environment_canada_public_alerts_local_cache_expiry_minutes"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_FORCE_MISS_KEY             = "environment_canada_public_alerts_local_cache_force_miss"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_SERVER_USE_KEY                         = "environment_canada_public_alerts_server_use"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_USE_KEY                   = "environment_canada_public_alerts_remote_cache_use"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY        = "environment_canada_public_alerts_remote_cache_expiry_minutes"
ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_FORCE_MISS_KEY            = "environment_canada_public_alerts_remote_cache_force_miss"

ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT  = 10
ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM  = 120
ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT = 5
ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM = 120

ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_USE_KEY                 = "environment_canada_marine_forecasts_local_cache_use"
ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY      = "environment_canada_marine_forecasts_local_cache_expiry_minutes"
ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY          = "environment_canada_marine_forecasts_local_cache_force_miss"
ENVIRONMENT_CANADA_MARINE_FORECASTS_SERVER_USE_KEY                      = "environment_canada_marine_forecasts_server_use"
ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_USE_KEY                = "environment_canada_marine_forecasts_remote_cache_use"
ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY     = "environment_canada_marine_forecasts_remote_cache_expiry_minutes"
ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY         = "environment_canada_marine_forecasts_remote_cache_force_miss"

DRIVE_BC                                                    = "Drive BC"
DRIVE_BC_KEY                                                = "Drive BC"
DRIVE_BC_URL                                                = "https://www.drivebc.ca/"
DRIVE_BC_AREAS_KEY                                          = "Areas"
DRIVE_BC_ENTIRE_PROVINCE_KEY                                = "Entire Province"
DRIVE_BC_ENTIRE_PROVINCE_FEED_URL                           = "https://www.drivebc.ca/api/events?format=rss"
DRIVE_BC_LOWER_MAINLAND_KEY                                 = "Lower Mainland"
DRIVE_BC_LOWER_MAINLAND_FEED_URL                            = "https://www.drivebc.ca/api/events/district/mainland?format=rss"
DRIVE_BC_VANCOUVER_ISLAND_KEY                               = "Vancouver Island"
DRIVE_BC_VANCOUVER_ISLAND_FEED_URL                          = "https://www.drivebc.ca/api/events/region/vanisland?format=rss"
DRIVE_BC_KOOTENAYS_KEY                                      = "Kootenays"
DRIVE_BC_KOOTENAYS_FEED_URL                                 = "https://www.drivebc.ca/api/events/region/kootenays?format=rss"
DRIVE_BC_THOMPSON_OKANAGAN_KEY                              = "Thompson Okanagan"
DRIVE_BC_THOMPSON_OKANAGAN_FEED_URL                         = "https://www.drivebc.ca/api/events/region/thomok?format=rss"
DRIVE_BC_CARIBOO_KEY                                        = "Cariboo"
DRIVE_BC_CARIBOO_FEED_URL                                   = "https://www.drivebc.ca/api/events/region/cariboo?format=rss"
DRIVE_BC_NORTH_CENTRAL_KEY                                  = "North Central"
DRIVE_BC_NORTH_CENTRAL_FEED_URL                             = "https://www.drivebc.ca/api/events/region/northcentral?format=rss"
DRIVE_BC_NORTH_EAST_KEY                                     = "North East"
DRIVE_BC_NORTH_EAST_FEED_URL                                = "https://www.drivebc.ca/api/events/region/northeast?format=rss"
DRIVE_BC_NORTH_WEST_KEY                                     = "North West"
DRIVE_BC_NORTH_WEST_FEED_URL                                = "https://www.drivebc.ca/api/events/region/northwest?format=rss"

DRIVE_BC_MAP_VIEW_URL                                       = "https://www.drivebc.ca/#mapView&ll=54.8833%2C-122.6667&z=5"

DRIVE_BC_MAP_VIEW_URL_BASE                                  = "https://www.drivebc.ca/#mapView&ll="
DRIVE_BC_MAP_VIEW_URL_LATITUDE                              = "54.8833"
DRIVE_BC_MAP_VIEW_URL_LAT_LON_SEPARATOR                     = "%2C"
DRIVE_BC_MAP_VIEW_URL_LONGITUDE                             = "-122.6667"
DRIVE_BC_MAP_VIEW_URL_ZOOM_SEPARATOR                        = "&z="
DRIVE_BC_MAP_VIEW_URL_ZOOM                                  = "5"

DRIVE_BC_CONDITIONS_AND_EVENTS_URL                          = "https://www.drivebc.ca/#listView"

DRIVE_BC_DISPLAY_LINKS_TO_DRIVE_BC_KEY                      = "display_links_to_drive_bc"
DRIVE_BC_DISTRICT_FILTER_KEY                                = "dbc_district_filter"
DRIVE_BC_ROUTE_FILTER_KEY                                   = "dbc_route_filter"
DRIVE_BC_TYPE_FILTER_KEY                                    = "dbc_type_filter"
DRIVE_BC_SEVERITY_FILTER_KEY                                = "dbc_severity_filter"

DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT                 = 20
DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM                 = 120
DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT                = 10
DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM                = 120

DRIVE_BC_LOCAL_CACHE_USE_KEY                                = "drive_bc_local_cache_use"
DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_KEY                     = "drive_bc_local_cache_expiry_minutes"
DRIVE_BC_LOCAL_CACHE_FORCE_MISS_KEY                         = "drive_bc_local_cache_force_miss"
DRIVE_BC_SERVER_USE_KEY                                     = "drive_bc_server_use"
DRIVE_BC_REMOTE_CACHE_USE_KEY                               = "drive_bc_remote_cache_use"
DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_KEY                    = "drive_bc_remote_cache_expiry_minutes"
DRIVE_BC_REMOTE_CACHE_FORCE_MISS_KEY                        = "drive_bc_remote_cache_forece_miss"

ALBERTA_511                                                 = "511 Alberta"
ALBERTA_511_KEY                                             = "511 Alberta"
ALBERTA_511_URL                                             = "https://511.alberta.ca/"
ALBERTA_511_RESOURCES_KEY                                   = "Resources"
ALBERTA_511_ROAD_CONDITIONS_KEY                             = "Road Conditions"
ALBERTA_511_ROAD_CONDITIONS_REQUEST_URL                     = "https://511.alberta.ca/api/v2/get/winterroads?format=json&lang=en"
ALBERTA_511_CAMERAS_KEY                                     = "Cameras"
ALBERTA_511_CAMERAS_REQUEST_URL                             = "https://511.alberta.ca/api/v2/get/cameras?format=json&lang=en"
ALBERTA_511_GROUPED_CAMERAS_KEY                             = "Grouped Cameras"
ALBERTA_511_GROUPED_CAMERAS_REQUEST_URL                     = "https://511.alberta.ca/api/v2/get/groupedcameras?format=json&lang=en"
ALBERTA_511_PARKS_KEY                                       = "Parks"
ALBERTA_511_PARKS_REQUEST_URL                               = "https://511.alberta.ca/api/v2/get/parks?format=json&lang=en"
ALBERTA_511_FERRIES_KEY                                     = "Ferries"
ALBERTA_511_FERRIES_REQUEST_URL                             = "https://511.alberta.ca/api/v2/get/ferryterminals?format=json&lang=en"
ALBERTA_511_EVENTS_KEY                                      = "Events"
ALBERTA_511_EVENTS_REQUEST_URL                              = "https://511.alberta.ca/api/v2/get/event?format=json&lang=en"
ALBERTA_511_ALERTS_KEY                                      = "Alerts"
ALBERTA_511_ALERTS_REQUEST_URL                              = "https://511.alberta.ca/api/v2/get/alerts?format=json&lang=en"

ALBERTA_511_ROAD_CONDITIONS_USE_LOCAL_CACHE                 = "511_alberta_road_conditions_use_local_cache"
ALBERTA_511_ROAD_CONDITIONS_CACHE_EXPIRY_MINUTES_DEFAULT    = 0
ALBERTA_511_ROAD_CONDITIONS_CACHE_EXPIRY_MINUTES            = "511_alberta_road_conditions_cache_expiry_minutes"
ALBERTA_511_ROAD_CONDITIONS_FORCE_CACHE_MISS                = "511_alberta_road_conditions_force_cache_miss"

AREA_NAME_FILTER                                            = "area_name_filter"
ROADWAY_NAME_FILTER                                         = "roadway_name_filter"
EVENT_TYPE_FILTER                                           = "event_type_filter"
PARK_NAME_FILTER                                            = "park_name_filter"
PARK_NAME_FILTER_CONDITION                                  = "park_name_filter_condition"
DISPLAY_DISABLED_CAMERAS                                    = "display_disabled_cameras"
EQUAL                                                       = "equal"
CONTAINS                                                    = "contains"

WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION               = "Washington State - Department of Transportation"
WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_KEY           = "Washington State - Department of Transportation"
WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_URL           = "https://wsdot.wa.gov/"
WS_DOT_TRAVELER_INFORMATION_KEY                             = "Traveler Information"
WS_DOT_BORDER_CROSSINGS_KEY                                 = "Border Crossings"
WS_DOT_BORDER_CROSSINGS_REQUEST_URL                         = "http://www.wsdot.com/Traffic/api/BorderCrossings/BorderCrossingsREST.svc/GetBorderCrossingsAsJson"

WS_DOT_BORDER_CROSSINGS_USE_LOCAL_CACHE                     = "ws_dot_border_crossings_use_local_cache"
WS_DOT_BORDER_CROSSINGS_CACHE_EXPIRY_MINUTES_DEFAULT        = 0
WS_DOT_BORDER_CROSSINGS_CACHE_EXPIRY_MINUTES                = "ws_dot_border_crossings_cache_expiry_minutes"
WS_DOT_BORDER_CROSSINGS_FORCE_CACHE_MISS                    = "ws_dot_border_crossings_force_cache_miss"

BEFORE                                                      = "before"
AFTER                                                       = "after"
BOTH                                                        = "both"

NOAA_NWS_FORECASTS                                          = "NOAA National Weather Service - Forecasts"
NOAA_NWS_FORECASTS_KEY                                      = "NOAA NWS - Forecasts"
NOAA_NWS_FORECASTS_HEADER                                   = "NOAA NWS - Forecasts"
NOAA_NWS_FORECASTS_URL                                      = "https://www.weather.gov/"
LOCATIONS_KEY                                               = "Locations"
NOAA_NWS_STATION_LOCATION_KEY                               = "<<Station Location>>"
GET_FORECAST_KEY                                            = "get_forecast"
GET_HOURLY_FORECAST_KEY                                     = "get_hourly_forecast"
HOURLY_FORECAST_HOURS_KEY                                   = "hourly_forecast_hours"
HOURLY_FORECAST_HOURS_DEFAULT                               = 48
GET_FORECAST_GRID_DATA_KEY                                  = "get_forecast_grid_data"
GET_OBSERVATION_STATIONS_KEY                                = "get_observation_stations"
GET_FORECAST_OFFICE_DATA_KEY                                = "get_forecast_office_data"

UNITS_KEY                                                   = "units"
UNITS_US                                                    = "us"
UNITS_SI                                                    = "si"

NOAA_NWS_SPC                                                = "NOAA National Weather Service - Storm Prediction Center"
NOAA_NWS_SPC_KEY                                            = "NOAA NWS - Storm Prediction Center"
NOAA_NWS_SPC_ALERTS_KEY                                     = "Alerts"
NOAA_NWS_SPC_URL                                            = "https://www.spc.noaa.gov/"
DISPLAY_LINKS_TO_ALL_CURRENT_ALERTS                         = "display_links_to_all_current_alerts"
NOAA_NWS_SPC_NO_ACTIVE_ALERTS                               = "There are no active watches, warnings or advisories"

NOAA_NWS_NHC                                                = "NOAA National Weather Service - National Hurricane Center"
NOAA_NWS_NHC_KEY                                            = "NOAA NWS - National Hurricane Center"
NOAA_NWS_NHC_URL                                            = "https://www.nhc.noaa.gov/"
BASINS_KEY                                                  = "Basins"

YES                                         = "yes"
NO                                          = "no"
ON                                          = "on"
OFF                                         = "off"
TRUE                                        = "true"
FALSE                                       = "false"
OK                                          = "OK"

BAD_LATITUDE                                = 1313.0
BAD_LONGITUDE                               = 1313.0
BAD_ELEVATION                               = -131313.0

FAKE_ASTERISK                               = "{{FAKE_ASTERISK}}"

DISPLAY_ATTRIBUTES_KEY                      = "display_attributes"
DISPLAY_ATTRIBUTE_LISTS_KEY                 = "display_attribute_lists"

CONTENT_JSON                                = "json"
CONTENT_XML                                 = "xml"
CONTENT_HTML                                = "html"

BELCHERTOWN_INDEX_HOOK_AFTER_STATION_INFO_INC   = "index_hook_after_station_info.inc"
BELCHERTOWN_INDEX_HOOK_AFTER_FORECAST_INC       = "index_hook_after_forecast.inc"
BELCHERTOWN_INDEX_HOOK_AFTER_SNAPSHOT_INC       = "index_hook_after_snapshot.inc"
BELCHERTOWN_INDEX_HOOK_AFTER_CHARTS_INC         = "index_hook_after_charts.inc"

class PositionLatLon:

    def __init__(
                  self,
                  Latitude      : float,
                  Longitude     : float
                ):

        self.Latitude           = Latitude
        self.Longitude          = Longitude
 
class GoogleMapsMarker:

    def __init__(
                    self
                  , Type              : str
                  , Latitude          : float
                  , Longitude         : float
                  , Text              : str
                  , InfoWindowContent : str
                ):

        self.Type               = Type
        self.Latitude           = Latitude
        self.Longitude          = Longitude
        self.Text               = Text
        self.InfoWindowContent  = InfoWindowContent

class SelectedEventsItem:

    def __init__(
                  self,
                  ID            : str,
                  Severity      : str,
                  Route         : str,
                  TypeIndex     : int,
                  LastUpdated   : datetime,
                  NextUpdate    : datetime,
                  Latitude      : float,
                  Longitude     : float,
                  CacheHit      : bool,
                  Index         : int
                ):

        self.ID             = ID
        self.Severity       = Severity
        self.Route          = Route
        self.TypeIndex      = TypeIndex
        self.LastUpdated    = LastUpdated
        self.NextUpdate     = NextUpdate
        self.Latitude       = Latitude
        self.Longitude      = Longitude
        self.CacheHit       = CacheHit
        self.Index          = Index

    def GetSeverity( e ):
        return e.Severity

    def GetRoute( e ):
        return e.Route

    def GetTypeIndex( e ):
        return e.TypeIndex

    def GetLastUpdated( e ):
        return e.LastUpdated

    def GetNextUpdate( e ):
        return e.NextUpdate

class Selected511AlbertaRoadConditionItem:

    def __init__(
                  self,
                  ID                : str,
                  AreaName          : str,
                  RoadwayName       : str,
                  Index             : int
                ):

        self.ID                 = ID
        self.AreaName           = AreaName
        self.RoadwayName        = RoadwayName
        self.Index              = Index

    def GetID( e ):
        return e.ID

    def GetAreaName( e ):
        return e.AreaName

    def GetRoadwayName( e ):
        return e.RoadwayName

    def GetIndex( self, e ):
        return e.Index

class SortList:

    def __init__(
                    self
                  , Key : str
                  , Index : int
                ):

        self.Key    = Key
        self.Index  = Index

    def GetKey( e ):
        return e.Key

# Inherit from the base class StdService:

class WXFeeds( StdService ):
    """
    WeeWX service that reads many RSS/Atom feeds and create files 
    that can be used by WeeWX skins such as Belchertown.
    """
    
    def __init__(
                  self,
                  engine, 
                  config_dict
                ) -> None:

        # Pass the initialization information on to my superclass:

        super( WXFeeds, self ).__init__( engine, config_dict )

        self.log = logging.getLogger( __name__ )
        assert self.log != None

#        self.log.debug( "WXFeeds::__init__(). . ." )

        self.bind( weewx.NEW_ARCHIVE_RECORD, self.new_archive_record )
        
        self.log.info( "WXFeeds: Version: %s" % WXFEEDS_VERSION )

        assert config_dict != None                          # Must exist for weewx to work.
#        self.log.debug( "config_dict: %s" % config_dict )


        self.Config_dict = config_dict

        WeeWXRoot = self.Config_dict.get( WEEWX_ROOT )
        assert WeeWXRoot != None                            # Pretty much must exist for WeeWX to work.
        self.log.info("WeeWXRoot: %s" % WeeWXRoot)
        assert WeeWXRoot[ 0 ] == "/"                        # Must be absolute path.

        # Get the modification time of the WeeWX configuration file.

        WeeWXConfigPath         = WeeWXRoot + "/weewx.conf"
        self.WeeWXConfig_mtime  = 0

        p = Path( WeeWXConfigPath )
        assert p != None
        assert p.exists() == True

        self.WeeWXConfig_mtime = p.stat().st_mtime
        self.log.info( "weewx.conf Modification Time: %s" % datetime.datetime.fromtimestamp(self.WeeWXConfig_mtime) )

        # Get the modification time of the WXFeeds service.

        WXFeedsPyPath        = WeeWXRoot + "/bin/user/WXFeeds.py"
        self.WXFeedsPy_mtime = 0

        p = Path( WXFeedsPyPath )
        assert p != None
        assert p.exists() == True

        self.WXFeedsPy_mtime = p.stat().st_mtime
        self.log.info( "WXFeeds.py Modification Time: %s" % datetime.datetime.fromtimestamp(self.WXFeedsPy_mtime) )

        # Get some information about this station.

        Station_dict = self.Config_dict[ STATION_KEY ]
        assert Station_dict != None
#        self.log.debug( "Station_dict: %s" % Station_dict )

        self.StationLocation = Station_dict.get( LOCATION_KEY, "" )
        if self.StationLocation != "":
            self.log.info( "StationLocation: %s" % self.StationLocation )
        else:
           self.log.warning( "Station location NOT set!" )

        self.StationLatitude = BAD_LATITUDE
        if "latitude" in Station_dict:
            self.StationLatitude = float( Station_dict.get( LATITUDE_KEY ) )
#            self.log.debug( "StationLatitude: %s" % self.StationLatitude )
        else:
           self.log.warning( "Station latitude NOT set!" )

        self.StationLongitude = BAD_LONGITUDE
        if "longitude" in Station_dict:
            self.StationLongitude = float( Station_dict.get( LONGITUDE_KEY ) )
#            self.log.debug( "StationLongitude: %s" % self.StationLongitude )
        else:
           self.log.warning( "Station longitude NOT set!" )

        self.StationURL = Station_dict.get( "station_url", "" )
        if self.StationURL != "":
            self.log.info( "Station URL: %s" % self.StationURL )
        else:
           self.log.warning( "Station URL NOT set!" )

        # Get archive_interval.

        StdArchive_dict = self.Config_dict[ STD_ARCHIVE_KEY ]
        assert StdArchive_dict != None
#        self.log.debug( "StdArchive_dict: %s" % StdArchive_dict )

        self.ArchiveInterval_seconds = 0
        if ARCHIVE_INTERVAL_KEY in StdArchive_dict:
            self.ArchiveInterval_seconds = int( StdArchive_dict.get( ARCHIVE_INTERVAL_KEY ) )
        self.log.info( "ArchiveInterval_seconds: %d" % self.ArchiveInterval_seconds )

        # Get [StdReport] dictionary.

        self.StdReport_dict = self.Config_dict[ STD_REPORT_KEY ]
        assert self.StdReport_dict != None
#        self.log.debug( "self.StdReport_dict: %s" % self.StdReport_dict )

        # Get WeeWX HTML_ROOT.

        self.WeeWXHTMLRoot = self.StdReport_dict[ HTML_ROOT_KEY ]
        assert self.WeeWXHTMLRoot != ""
        self.log.info( "WeeWXHTMLRoot: %s" % self.WeeWXHTMLRoot )

        # Create the WXFeeds HTML ROOT.

        self.WXFeedsHTMLRoot = str( Path( self.WeeWXHTMLRoot ).joinpath( "WXFeeds" ) )
        self.log.info( "WXFeedsHTMLRoot: %s" % self.WXFeedsHTMLRoot )

        p = Path( self.WXFeedsHTMLRoot )
        assert p != None

        if p.exists() == False:
            self.log.info( "Creating WXFeeds HTML Directory. . ." )

            os.makedirs( self.WXFeedsHTMLRoot )

        # Get the [WXFeeds] dictionary.

        WXFeeds_dict = None

        try:
            WXFeeds_dict = self.Config_dict[ WXFEEDS_KEY ]
        except:
           self.log.warning( "[%s] Not Found In WeeWX Configuration File!  Aborting WXFeeds Initialization." % WXFEEDS_KEY )
           return

       # Get the [WXFeeds][[Setup]] dictionary.

        Setup_dict = None

        try:
            Setup_dict = WXFeeds_dict[ SETUP_KEY ]
        except:
           self.log.warning( "[[%s]] Not Found In [%s] In WeeWX Configuration File!  Aborting WXFeeds Initialization." % ( SETUP_KEY, WXFEEDS_KEY ) )
           return

        # File Expiry - Minutes

        self.FileExpiry_minutes = 0
        if FILE_EXPIRY_MINUTES_KEY in Setup_dict:
            self.FileExpiry_minutes = int( Setup_dict.get( FILE_EXPIRY_MINUTES_KEY ) )
        self.log.info( "FileExpiry_minutes: %d" % self.FileExpiry_minutes )

        # TAKE_OUT_THE_TRASH_KEY

        self.TakeOutTheTrash = self.YesNoTrueFalse( Setup_dict, TAKE_OUT_THE_TRASH_KEY, False )
        self.log.info( "TakeOutTheTrash: '%s'" % self.TakeOutTheTrash )

        # Get WXFeeds root.
        #
        # Files in WXFeedsRoot:
        # - cached files common to all skins.
        #   - files for tracking file modification time.

        WXFeedsRoot = Setup_dict.get( WXFEEDS_ROOT_KEY, "WXFeeds" )
#        self.log.info("WXFeedsRoot (config): %s" % WXFeedsRoot)

        if WXFeedsRoot == "": 
            self.log.warning( "%s can not be ''!  Setting it to 'WXFeeds'" % WXFEEDS_ROOT_KEY )

            WXFeedsRoot = "WXFeeds"
#            self.log.info("WXFeedsRoot (''): %s" % WXFeedsRoot)

        p = Path( WXFeedsRoot )
        if p.is_absolute() == False:
            # WXFeedsRoot is not absolute - make it relative to WeeWx root.

#            self.log.info("WXFeedsRoot (absolute): %s" % WXFeedsRoot)
            WXFeedsRoot = str( Path( WeeWXRoot ).joinpath( WXFeedsRoot ) )

        self.log.info("WXFeedsRoot: %s" % WXFeedsRoot)

        #   Create the WXFeeds directory.

        p = Path( WXFeedsRoot )
        assert p != None

        if p.exists() == False:
            self.log.info( "Creating WXFeeds Root Directory. . ." )

            os.makedirs( WXFeedsRoot )

        #   Create the WXFeeds cache directory.

        #self.WXFeedsLocalCacheRoot = WXFeedsRoot + "/cache"
        self.WXFeedsLocalCacheRoot = str( Path( WXFeedsRoot ).joinpath( "cache" ) )
        self.log.info("WXFeedsLocalCacheRoot: %s" % self.WXFeedsLocalCacheRoot)

        p = Path( self.WXFeedsLocalCacheRoot )
        assert p != None

        if p.exists() == False:
            self.log.info( "Creating WXFeeds Local Cache Directory. . ." )

            os.makedirs( self.WXFeedsLocalCacheRoot )

        #   Create the WXFeeds temporary directory.

        #self.WXFeedsTemporaryRoot = WXFeedsRoot + "/tmp"
        self.WXFeedsTemporaryRoot = str( Path( WXFeedsRoot ).joinpath( "tmp" ) )
        self.log.info("WXFeedsTemporaryRoot: %s" % self.WXFeedsTemporaryRoot)

        p = Path( self.WXFeedsTemporaryRoot )
        assert p != None

        if p.exists() == False:
            self.log.info( "Creating WXFeeds Temporary Directory. . ." )

            os.makedirs( self.WXFeedsTemporaryRoot )

        #   Get WXFeeds Icon Base.

        self.WXFeedsIconBase = Setup_dict.get( WXFEEDS_ICON_BASE_KEY, "WXFeeds/images" )
        self.log.info( "WXFeedsIconBase: %s" % self.WXFeedsIconBase )

        if self.WXFeedsIconBase == "":
            self.log.warning( "WXFeeds_icon_base can not be ''!  Setting it to '%s/images'" % WXFeedsRoot )

            #self.WXFeedsIconBase = WXFeedsRoot + "/images"
            self.WXFeedsIconBase = str( Path( WXFeedsRoot ).joinpath( "images" ) )

        #   If WXFeedsIconBase is not a URL, make it absolute.

        if self.WXFeedsIconBase.find( "http" ) != 0:
            # Not a URL.

            p = Path( self.WXFeedsIconBase )
            if p.is_absolute() == False:
                # self.WXFeedsIconBase is not absolute - make it relative to WeeWx root.

                #self.WXFeedsIconBase = self.WeeWXRoot + "/" + self.WXFeedsIconBase
                self.WXFeedsIconBase = str( Path( WXFeedsRoot ).joinpath( "images" ) )

        # Setup for WXFeeds Servers.

        self.WXFeedsServerUse                       = False
        self.WXFeedsServerURL                       = ""
        self.WXFeedsServerAccessToken               = ""
        self.WXFeedsServerAccessTokenExpiry_seconds = 0

        self.WXFeedsServerUse = self.YesNoTrueFalse( Setup_dict, WXFEEDS_SERVER_USE_KEY, False )
        self.log.info( "WXFeedsServerUse: '%s'" % self.WXFeedsServerUse )

        if self.WXFeedsServerUse == True:
            self.WXFeedsServerURL = Setup_dict.get( WXFEEDS_SERVER_URL_KEY, "" )
            self.log.info( "WXFeedsServerURL: '%s'" % self.WXFeedsServerURL )

            self.WXFeedsServerUsername = Setup_dict.get( WXFEEDS_SERVER_USERNAME_KEY, "" )
            self.log.info( "WXFeedsServerUsername: %s" % self.WXFeedsServerUsername )
            self.WXFeedsServerPassword = Setup_dict.get( WXFEEDS_SERVER_PASSWORD_KEY, "" )
            self.log.info( "WXFeedsServerPassword: %s" % self.WXFeedsServerPassword )

            # If we're going to use a WXFeeds server, log in to it.
            #
            # This is really just a test to see if we can login.  No feeds are accessed but 
            # flags for feeds are cleared if we cannot login.
            # We will login each archive cycle and then access feeds.
    
            if self.WXFeedsServerUse == True:

                ReturnedStatus                              = False
                self.WXFeedsServerAccessToken               = ""
                self.WXFeedsServerAccessTokenExpiry_seconds = -1

                ReturnedStatus,         \
                self.WXFeedsServerAccessToken,                  \
                self.WXFeedsServerAccessTokenExpiry_seconds = self.WXFeedsServerLogin(
                                                                                       self.WXFeedsServerURL,
                                                                                       self.WXFeedsServerUsername,
                                                                                       self.WXFeedsServerPassword
                                                                                     )
                if ReturnedStatus == False:
                    self.log.error( "WXFeeds: Server Login NOT Successful!  Use of WXFeeds Server will be disabled." )

                    self.WXFeedsServerUse = False
                else:
                    assert self.WXFeedsServerAccessToken != ""
                    assert self.WXFeedsServerAccessTokenExpiry_seconds >= 0

                    self.log.info( "WXFeeds: Server Login Successful." )

        # Setup for Google Maps Services.

        self.GoogleMapsUse                                  = False
        self.GoogleMapsAPIKey                               = ""

        self.GoogleMapsMapsUse                              = False
        self.GoogleMapsMapsServerUse                        = False

        self.GoogleMapsGeocodingUse                         = False
        self.GoogleMapsGeocodingLocalCacheUse               = False
        self.GoogleMapsGeocodingLocalCacheExpiry_minutes    = int( GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT )
        self.GoogleMapsGeocodingLocalCacheForceMiss         = False
        self.GoogleMapsGeocodingServerUse                   = False
        self.GoogleMapsGeocodingRemoteCacheUse              = False
        self.GoogleMapsGeocodingRemoteCacheExpiry_minutes   = int( GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT )
        self.GoogleMapsGeocodingRemoteCacheForceMiss        = False

        self.GoogleMapsPlacesUse                            = False
        self.GoogleMapsPlacesLocalCacheUse                  = False
        self.GoogleMapsPlacesLocalCacheExpiry_minutes       = int( GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT )
        self.GoogleMapsPlacesLocalCacheForceMiss            = False
        self.GoogleMapsPlacesServerUse                      = False
        self.GoogleMapsPlacesRemoteCacheUse                 = False
        self.GoogleMapsPlacesRemoteCacheExpiry_minutes      = int( GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT )
        self.GoogleMapsPlacesRemoteCacheForceMiss           = False

        # GoogleMaps Services.

        self.GoogleMapsUse      = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_USE_KEY, self.GoogleMapsUse )
        self.GoogleMapsAPIKey   = Setup_dict.get( GOOGLE_MAPS_API_KEY_KEY, self.GoogleMapsAPIKey )
 
        self.log.info( "GoogleMapsUse: %s" % self.GoogleMapsUse )
        self.log.info( "GoogleMapsAPIKey: %s" % self.GoogleMapsAPIKey )

        # GoogleMaps Maps Service.

        self.GoogleMapsMapsUse          = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_MAPS_USE_KEY, self.GoogleMapsMapsUse )
        self.GoogleMapsMapsServerUse    = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_MAPS_SERVER_USE_KEY, self.GoogleMapsMapsServerUse )

        if self.GoogleMapsUse == True and               \
           self.GoogleMapsMapsUse == True and           \
           self.GoogleMapsMapsServerUse == True and     \
           self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use GoogleMaps Maps via a server!  It will be set to not use a server." )
            self.GoogleMapsMapsServerUse = False

        if self.GoogleMapsMapsServerUse == False and self.GoogleMapsAPIKey == "":
            self.log.error( "A Google Maps API key is required to use GoogleMaps Maps without a server!  It will be disabled." )
            self.GoogleMapsMapsUse = False

        self.log.info( "GoogleMapsMapsUse: %s" % self.GoogleMapsMapsUse )
        self.log.info( "GoogleMapsMapsServerUse: %s" % self.GoogleMapsMapsServerUse )

        # GoogleMaps Geocoding Service.

        self.GoogleMapsGeocodingUse                         = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_USE_KEY, self.GoogleMapsGeocodingUse )
        self.GoogleMapsGeocodingLocalCacheUse               = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_USE_KEY, self.GoogleMapsGeocodingLocalCacheUse )
        self.GoogleMapsGeocodingLocalCacheExpiry_minutes    = int( Setup_dict.get( GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.GoogleMapsGeocodingLocalCacheExpiry_minutes ) ) )
        self.GoogleMapsGeocodingLocalCacheForceMiss         = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_LOCAL_CACHE_FORCE_MISS_KEY, self.GoogleMapsGeocodingLocalCacheForceMiss )
        self.GoogleMapsGeocodingServerUse                   = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_SERVER_USE_KEY, self.GoogleMapsGeocodingServerUse )
        self.GoogleMapsGeocodingRemoteCacheUse              = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_USE_KEY, self.GoogleMapsGeocodingRemoteCacheUse )
        self.GoogleMapsGeocodingRemoteCacheExpiry_minutes   = int( Setup_dict.get( GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.GoogleMapsGeocodingRemoteCacheExpiry_minutes ) )
        self.GoogleMapsGeocodingRemoteCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_GEOCODING_REMOTE_CACHE_FORCE_MISS_KEY, self.GoogleMapsGeocodingRemoteCacheForceMiss )

        if self.GoogleMapsGeocodingUse == True:
            if self.GoogleMapsGeocodingServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds Server to use GoogleMaps Geocoding via a server!  It will be set to not use a  server." )
                self.GoogleMapsGeocodingServerUse = False

            if self.GoogleMapsGeocodingServerUse == False and self.GoogleMapsAPIKey == "":
                self.log.error( "A Google Maps API key is required to use GoogleMaps Geocoding without a server!  It will be disabled." )
                self.GoogleMapsGeocodingUse = False

            if self.GoogleMapsGeocodingUse == True:
                if self.GoogleMapsGeocodingLocalCacheUse == True:
                    if self.GoogleMapsGeocodingLocalCacheExpiry_minutes < 0 or self.GoogleMapsGeocodingLocalCacheExpiry_minutes > GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "Geocoding Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.GoogleMapsGeocodingLocalCacheExpiry_minutes, GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.GoogleMapsGeocodingLocalCacheExpiry_minutes = GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT

                if self.GoogleMapsGeocodingServerUse == True and self.GoogleMapsGeocodingRemoteCacheUse == True:
                    if self.GoogleMapsGeocodingRemoteCacheExpiry_minutes < 0 or self.GoogleMapsGeocodingRemoteCacheExpiry_minutes > GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "Geocoding Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.GoogleMapsGeocodingRemoteCacheExpiry_minutes,GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM,GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.GoogleMapsGeocodingRemoteCacheExpiry_minutes = GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                                               
        self.log.info( "GoogleMapsGeocodingUse: %s" % self.GoogleMapsGeocodingUse )
        self.log.info( "GoogleMapsGeocodingLocalCacheUse: %s" % self.GoogleMapsGeocodingLocalCacheUse )
        self.log.info( "GoogleMapsGeocodingLocalCacheExpiry_minutes: %s" % self.GoogleMapsGeocodingLocalCacheExpiry_minutes )
        self.log.info( "GoogleMapsGeocodingLocalCacheForceMiss: %s" % self.GoogleMapsGeocodingLocalCacheForceMiss )
        self.log.info( "GoogleMapsGeocodingServerUse: %s" % self.GoogleMapsGeocodingServerUse )
        self.log.info( "GoogleMapsGeocodingRemoteCacheUse: %s" % self.GoogleMapsGeocodingRemoteCacheUse )
        self.log.info( "GoogleMapsGeocodingRemoteCacheExpiry_minutes: %s" % self.GoogleMapsGeocodingRemoteCacheExpiry_minutes )
        self.log.info( "GoogleMapsGeocodingRemoteCacheForceMiss: %s" % self.GoogleMapsGeocodingRemoteCacheForceMiss )

        # GoogleMaps Places Service.

        self.GoogleMapsPlacesUse                        = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_USE_KEY, self.GoogleMapsPlacesUse )
        self.GoogleMapsPlacesLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_LOCAL_CACHE_USE_KEY,self.GoogleMapsPlacesLocalCacheUse )
        self.GoogleMapsPlacesLocalCacheExpiry_minutes   = int( Setup_dict.get( GOOGLE_MAPS_PLACES_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.GoogleMapsPlacesLocalCacheExpiry_minutes ) ) )
        self.GoogleMapsPlacesLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_LOCAL_CACHE_FORCE_MISS_KEY, self.GoogleMapsPlacesLocalCacheForceMiss )
        self.GoogleMapsPlacesServerUse             = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_SERVER_USE_KEY, self.GoogleMapsPlacesServerUse )
        self.GoogleMapsPlacesRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_REMOTE_CACHE_USE_KEY, self.GoogleMapsPlacesRemoteCacheUse )
        self.GoogleMapsPlacesRemoteCacheExpiry_minutes  = int( Setup_dict.get( GOOGLE_MAPS_PLACES_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.GoogleMapsPlacesRemoteCacheExpiry_minutes ) )
        self.GoogleMapsPlacesRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, GOOGLE_MAPS_PLACES_REMOTE_CACHE_FORCE_MISS_KEY, self.GoogleMapsPlacesRemoteCacheForceMiss )

        if self.GoogleMapsPlacesUse == True:
            if self.GoogleMapsPlacesServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds server to use GoogleMaps Places via a server! !  It will be set to not use a server." )
                self.GoogleMapsPlacesServerUse = False

            if self.GoogleMapsPlacesServerUse == False and self.GoogleMapsAPIKey == "":
                self.log.error( "A Google Maps API key is required to use GoogleMaps Places without a server!  It will be disabled." )
                self.GoogleMapsPlacesUse = False

            if self.GoogleMapsPlacesUse == True:
                if self.GoogleMapsPlacesLocalCacheUse == True:
                    if self.GoogleMapsPlacesLocalCacheExpiry_minutes < 0 or self.GoogleMapsPlacesLocalCacheExpiry_minutes > GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "Places Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.GoogleMapsPlacesLocalCacheExpiry_minutes, GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.GoogleMapsPlacesLocalCacheExpiry_minutes = GOOGLE_MAPS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
                if self.GoogleMapsPlacesServerUse == True and self.GoogleMapsPlacesRemoteCacheUse == True:
                    if self.GoogleMapsPlacesRemoteCacheExpiry_minutes < 0 or self.GoogleMapsPlacesRemoteCacheExpiry_minutes > GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "Places Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.GoogleMapsPlacesRemoteCacheExpiry_minutes, GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.GoogleMapsPlacesRemoteCacheExpiry_minutes = GOOGLE_MAPS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT

        self.log.info( "GoogleMapsPlacesUse: %s" % self.GoogleMapsPlacesUse )
        self.log.info( "GoogleMapsPlacesLocalCacheUse: %s" % self.GoogleMapsPlacesLocalCacheUse )
        self.log.info( "GoogleMapsPlacesLocalCacheExpiry_minutes: %s" % self.GoogleMapsPlacesLocalCacheExpiry_minutes )
        self.log.info( "GoogleMapsPlacesLocalCacheForceMiss: %s" % self.GoogleMapsPlacesLocalCacheForceMiss )
        self.log.info( "GoogleMapsPlacesServerUse: %s" % self.GoogleMapsPlacesServerUse )
        self.log.info( "GoogleMapsPlacesRemoteCacheUse: %s" % self.GoogleMapsPlacesRemoteCacheUse )
        self.log.info( "GoogleMapsPlacesRemoteCacheExpiry_minutes: %s" % self.GoogleMapsPlacesRemoteCacheExpiry_minutes )
        self.log.info( "GoogleMapsPlacesRemoteCacheForceMiss: %s" % self.GoogleMapsPlacesRemoteCacheForceMiss )

        # Setup for AIR-PORT-CODES Airport Information Service.

        self.AirPortCodesUse                        = False
        self.AirPortCodesAPIKey                     = ""
        self.AirPortCodesAPISecret                  = ""
        self.AirPortCodesLocalCacheUse              = False
        self.AirPortCodesLocalCacheExpiry_minutes   = AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.AirPortCodesLocalCacheForceMiss        = False
        self.AirPortCodesServerUse             = False
        self.AirPortCodesRemoteCacheUse             = False
        self.AirPortCodesRemoteCacheExpiry_minutes  = AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.AirPortCodesRemoteCacheForceMiss       = False

        self.AirPortCodesUse                        = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_USE_KEY, self.AirPortCodesUse )
        self.AirPortCodesAPIKey                     = Setup_dict.get( AIR_PORT_CODES_API_KEY_KEY, self.AirPortCodesAPIKey )
        self.AirPortCodesAPISecret                  = Setup_dict.get( AIR_PORT_CODES_API_SECRET_KEY, self.AirPortCodesAPISecret )
        self.AirPortCodesLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_LOCAL_CACHE_USE_KEY, self.AirPortCodesLocalCacheUse )
        self.AirPortCodesLocalCacheExpiry_minutes   = int( Setup_dict.get( AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.AirPortCodesLocalCacheExpiry_minutes ) ) )
        self.AirPortCodesLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_LOCAL_CACHE_FORCE_MISS_KEY, self.AirPortCodesLocalCacheForceMiss )
        self.AirPortCodesServerUse             = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_SERVER_USE_KEY, self.AirPortCodesServerUse )
        self.AirPortCodesRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_REMOTE_CACHE_USE_KEY, self.AirPortCodesRemoteCacheUse )
        self.AirPortCodesRemoteCacheExpiry_minutes  = int( Setup_dict.get( AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.AirPortCodesRemoteCacheExpiry_minutes ) )
        self.AirPortCodesRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, AIR_PORT_CODES_REMOTE_CACHE_FORCE_MISS_KEY, self.AirPortCodesRemoteCacheForceMiss )
                    
        if self.AirPortCodesUse == True:
            if self.AirPortCodesServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds Server to use AIR-PORT-CODES via a server!  It will be set to not use a server." )
                self.AirPortCodesServerUse = False

            if self.AirPortCodesServerUse == False and ( self.AirPortCodesAPIKey == "" or self.AirPortCodesAPISecret == "" ):                       
                self.log.error( "AIR-PORT_CODES API and Secret keys are required to use AIR-PORT-CODES without a server!  It will be disabled." )
                self.AirPortCodesUse = False

            if self.AirPortCodesUse == True:
                if self.AirPortCodesLocalCacheUse == True:
                    if self.AirPortCodesLocalCacheExpiry_minutes < 0 or self.AirPortCodesLocalCacheExpiry_minutes > AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "AIR-PORT-CODES Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.AirPortCodesLocalCacheExpiry_minutes, AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.AirPortCodesLocalCacheExpiry_minutes = AIR_PORT_CODES_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
                if self.AirPortCodesServerUse == True and self.AirPortCodesRemoteCacheUse == True:
                    if self.AirPortCodesRemoteCacheExpiry_minutes < 0 or self.AirPortCodesRemoteCacheExpiry_minutes > AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "AIR-PORT-CODES Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.AirPortCodesRemoteCacheExpiry_minutes, AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.AirPortCodesRemoteCacheExpiry_minutes = AIR_PORT_CODES_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "AirPortCodesUse: %s" % self.AirPortCodesUse )
        self.log.info( "AirPortCodesAPIKey: %s" % self.AirPortCodesAPIKey )
        self.log.info( "AirPortCodesAPISecret: %s" % self.AirPortCodesAPISecret )
        self.log.info( "AirPortCodesLocalCacheUse: %s" % self.AirPortCodesLocalCacheUse )
        self.log.info( "AirPortCodesLocalCacheExpiry_minutes: %s" % self.AirPortCodesLocalCacheExpiry_minutes )
        self.log.info( "AirPortCodesLocalCacheForceMiss: %s" % self.AirPortCodesLocalCacheForceMiss )
        self.log.info( "AirPortCodesServerUse: %s" % self.AirPortCodesServerUse )
        self.log.info( "AirPortCodesRemoteCacheUse: %s" % self.AirPortCodesRemoteCacheUse )
        self.log.info( "AirPortCodesRemoteCacheExpiry_minutes: %s" % self.AirPortCodesRemoteCacheExpiry_minutes )
        self.log.info( "AirPortCodesRemoteCacheForceMiss: %s" % self.AirPortCodesRemoteCacheForceMiss )

        # Setup for NOAA NWS Points Service.

        self.NOAANWSPointsUse                        = True     # For consistency with other services.  We do not allow it to be set.
        self.NOAANWSPointsLocalCacheUse              = False
        self.NOAANWSPointsLocalCacheExpiry_minutes   = NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSPointsLocalCacheForceMiss        = False
        self.NOAANWSPointsServerUse             = False
        self.NOAANWSPointsRemoteCacheUse             = False
        self.NOAANWSPointsRemoteCacheExpiry_minutes  = NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSPointsRemoteCacheForceMiss       = False

        #self.NOAANWSPointsUse                        = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_USE_KEY, self.NOAANWSPointsUse )
        self.NOAANWSPointsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_LOCAL_CACHE_USE_KEY, self.NOAANWSPointsLocalCacheUse )
        self.NOAANWSPointsLocalCacheExpiry_minutes   = int( Setup_dict.get( NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.NOAANWSPointsLocalCacheExpiry_minutes ) ) )
        self.NOAANWSPointsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_LOCAL_CACHE_FORCE_MISS_KEY, self.NOAANWSPointsLocalCacheForceMiss )
        self.NOAANWSPointsServerUse             = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_SERVER_USE_KEY, self.NOAANWSPointsServerUse )
        self.NOAANWSPointsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_REMOTE_CACHE_USE_KEY, self.NOAANWSPointsRemoteCacheUse )
        self.NOAANWSPointsRemoteCacheExpiry_minutes  = int( Setup_dict.get( NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.NOAANWSPointsRemoteCacheExpiry_minutes ) )
        self.NOAANWSPointsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_POINTS_REMOTE_CACHE_FORCE_MISS_KEY, self.NOAANWSPointsRemoteCacheForceMiss )
                    
        if self.NOAANWSPointsUse == True:
            if self.NOAANWSPointsServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds Server to use NOAA NWS Points via a server!  It will be set to not use a server." )
                self.NOAANWSPointsServerUse = False

            if self.NOAANWSPointsUse == True:
                if self.NOAANWSPointsLocalCacheUse == True:
                    if self.NOAANWSPointsLocalCacheExpiry_minutes < 0 or self.NOAANWSPointsLocalCacheExpiry_minutes > NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "NOAA NWS Points Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSPointsLocalCacheExpiry_minutes, NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.NOAANWSPointsLocalCacheExpiry_minutes = NOAA_NWS_POINTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
                if self.NOAANWSPointsServerUse == True and self.NOAANWSPointsRemoteCacheUse == True:
                    if self.NOAANWSPointsRemoteCacheExpiry_minutes < 0 or self.NOAANWSPointsRemoteCacheExpiry_minutes > NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                        self.log.error( "NOAA NWS Points Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSPointsRemoteCacheExpiry_minutes, NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                        self.NOAANWSPointsRemoteCacheExpiry_minutes = NOAA_NWS_POINTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "NOAANWSPointsUse: %s" % self.NOAANWSPointsUse )
        self.log.info( "NOAANWSPointsLocalCacheUse: %s" % self.NOAANWSPointsLocalCacheUse )
        self.log.info( "NOAANWSPointsLocalCacheExpiry_minutes: %d" % self.NOAANWSPointsLocalCacheExpiry_minutes )
        self.log.info( "NOAANWSPointsLocalCacheForceMiss: %s" % self.NOAANWSPointsLocalCacheForceMiss )
        self.log.info( "NOAANWSPointsServerUse: %s" % self.NOAANWSPointsServerUse )
        self.log.info( "NOAANWSPointsRemoteCacheUse: %s" % self.NOAANWSPointsRemoteCacheUse )
        self.log.info( "NOAANWSPointsRemoteCacheExpiry_minutes: %d" % self.NOAANWSPointsRemoteCacheExpiry_minutes )
        self.log.info( "NOAANWSPointsRemoteCacheForceMiss: %s" % self.NOAANWSPointsRemoteCacheForceMiss )

        # Setup for NOAA NWS Forecasts.

        self.NOAANWSForecastsLocalCacheUse              = False
        self.NOAANWSForecastsLocalCacheExpiry_minutes   = NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSForecastsLocalCacheForceMiss        = False
        self.NOAANWSForecastsServerUse                  = False
        self.NOAANWSForecastsRemoteCacheUse             = False
        self.NOAANWSForecastsRemoteCacheExpiry_minutes  = NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSForecastsRemoteCacheForceMiss       = False

        self.NOAANWSForecastsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_FORECASTS_LOCAL_CACHE_USE_KEY, self.NOAANWSForecastsLocalCacheUse )
        self.NOAANWSForecastsLocalCacheExpiry_minutes   = int( Setup_dict.get( NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.NOAANWSForecastsLocalCacheExpiry_minutes ) ) )
        self.NOAANWSForecastsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY, self.NOAANWSForecastsLocalCacheForceMiss )
        self.NOAANWSForecastsServerUse                  = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_FORECASTS_SERVER_USE_KEY, self.NOAANWSForecastsServerUse )
        self.NOAANWSForecastsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_FORECASTS_REMOTE_CACHE_USE_KEY, self.NOAANWSForecastsRemoteCacheUse )
        self.NOAANWSForecastsRemoteCacheExpiry_minutes  = int( Setup_dict.get( NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.NOAANWSForecastsRemoteCacheExpiry_minutes ) )
        self.NOAANWSForecastsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY, self.NOAANWSForecastsRemoteCacheForceMiss )

        if self.NOAANWSForecastsServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds Server to use NOAA NWS Forecasts via a server!  It will be set to not use a server." )
                self.NOAANWSForecastsServerUse = False

        if self.NOAANWSForecastsLocalCacheUse == True:
            if self.NOAANWSForecastsLocalCacheExpiry_minutes < 0 or self.NOAANWSForecastsLocalCacheExpiry_minutes > NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS Points Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSForecastsLocalCacheExpiry_minutes, NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSForecastsLocalCacheExpiry_minutes = NOAA_NWS_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.NOAANWSForecastsServerUse == True and self.NOAANWSForecastsRemoteCacheUse == True:
            if self.NOAANWSForecastsRemoteCacheExpiry_minutes < 0 or self.NOAANWSForecastsRemoteCacheExpiry_minutes > NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS Points Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSForecastsRemoteCacheExpiry_minutes, NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSForecastsRemoteCacheExpiry_minutes = NOAA_NWS_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "NOAANWSForecastsLocalCacheUse: %s" % self.NOAANWSForecastsLocalCacheUse )
        self.log.info( "NOAANWSForecastsLocalCacheExpiry_minutes: %s" % self.NOAANWSForecastsLocalCacheExpiry_minutes )
        self.log.info( "NOAANWSForecastsLocalCacheForceMiss: %s" % self.NOAANWSForecastsLocalCacheForceMiss )
        self.log.info( "NOAANWSForecastsServerUse: %s" % self.NOAANWSForecastsServerUse )
        self.log.info( "NOAANWSForecastsRemoteCacheUse: %s" % self.NOAANWSForecastsRemoteCacheUse )
        self.log.info( "NOAANWSForecastsRemoteCacheExpiry_minutes: %s" % self.NOAANWSForecastsRemoteCacheExpiry_minutes )
        self.log.info( "NOAANWSForecastsRemoteCacheForceMiss: %s" % self.NOAANWSForecastsRemoteCacheForceMiss )

        # Setup for NOAA NWS Hourly Forecasts.

        self.NOAANWSHourlyForecastsLocalCacheUse              = False
        self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes   = NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSHourlyForecastsLocalCacheForceMiss        = False
        self.NOAANWSHourlyForecastsServerUse                  = False
        self.NOAANWSHourlyForecastsRemoteCacheUse             = False
        self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes  = NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSHourlyForecastsRemoteCacheForceMiss       = False

        self.NOAANWSHourlyForecastsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_USE_KEY, self.NOAANWSHourlyForecastsLocalCacheUse )
        self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes   = int( Setup_dict.get( NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes ) ) )
        self.NOAANWSHourlyForecastsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY, self.NOAANWSHourlyForecastsLocalCacheForceMiss )
        self.NOAANWSHourlyForecastsServerUse                  = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_HOURLY_FORECASTS_SERVER_USE_KEY, self.NOAANWSHourlyForecastsServerUse )
        self.NOAANWSHourlyForecastsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_USE_KEY, self.NOAANWSHourlyForecastsRemoteCacheUse )
        self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes  = int( Setup_dict.get( NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes ) )
        self.NOAANWSHourlyForecastsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY, self.NOAANWSHourlyForecastsRemoteCacheForceMiss )
                    
        if self.NOAANWSHourlyForecastsServerUse == True and self.WXFeedsServerUse == False:
                self.log.error( "You must use WXFeeds Server to use NOAA NWS Hourly Forecasts via a server!  It will be set to not use a server." )
                self.NOAANWSHourlyForecastsServerUse = False

        if self.NOAANWSHourlyForecastsLocalCacheUse == True:
            if self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes < 0 or self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes > NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS Points Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes, NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes = NOAA_NWS_HOURLY_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.NOAANWSHourlyForecastsServerUse == True and self.NOAANWSHourlyForecastsRemoteCacheUse == True:
            if self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes < 0 or self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes > NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS Points Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes, NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes = NOAA_NWS_HOURLY_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "NOAANWSHourlyForecastsLocalCacheUse: %s" % self.NOAANWSHourlyForecastsLocalCacheUse )
        self.log.info( "NOAANWSHourlyForecastsLocalCacheExpiry_minutes: %s" % self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes )
        self.log.info( "NOAANWSHourlyForecastsLocalCacheForceMiss: %s" % self.NOAANWSHourlyForecastsLocalCacheForceMiss )
        self.log.info( "NOAANWSHourlyForecastsServerUse: %s" % self.NOAANWSHourlyForecastsServerUse )
        self.log.info( "NOAANWSHourlyForecastsRemoteCacheUse: %s" % self.NOAANWSHourlyForecastsRemoteCacheUse )
        self.log.info( "NOAANWSHourlyForecastsRemoteCacheExpiry_minutes: %s" % self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes )
        self.log.info( "NOAANWSHourlyForecastsRemoteCacheForceMiss: %s" % self.NOAANWSHourlyForecastsRemoteCacheForceMiss )

        # Setup for NOAA NWS AWS METARs.

        self.NOAANWSAWCMETARsLocalCacheUse              = False
        self.NOAANWSAWCMETARsLocalCacheExpiry_minutes   = NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSAWCMETARsLocalCacheForceMiss        = False
        self.NOAANWSAWCMETARsServerUse                  = False
        self.NOAANWSAWCMETARsRemoteCacheUse             = False
        self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes  = NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.NOAANWSAWCMETARsRemoteCacheForceMiss       = False

        self.NOAANWSAWCMETARsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_AWC_METARS_LOCAL_CACHE_USE_KEY, self.NOAANWSAWCMETARsLocalCacheUse )
        self.NOAANWSAWCMETARsLocalCacheExpiry_minutes   = int( Setup_dict.get( NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.NOAANWSAWCMETARsLocalCacheExpiry_minutes ) ) )
        self.NOAANWSAWCMETARsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_AWC_METARS_LOCAL_CACHE_FORCE_MISS_KEY, self.NOAANWSAWCMETARsLocalCacheForceMiss )
        self.NOAANWSAWCMETARsServerUse                  = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_AWC_METARS_SERVER_USE_KEY, self.NOAANWSAWCMETARsServerUse )
        self.NOAANWSAWCMETARsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_AWC_METARS_REMOTE_CACHE_USE_KEY, self.NOAANWSAWCMETARsRemoteCacheUse )
        self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes  = int( Setup_dict.get( NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes ) )
        self.NOAANWSAWCMETARsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, NOAA_NWS_AWC_METARS_REMOTE_CACHE_FORCE_MISS_KEY, self.NOAANWSAWCMETARsRemoteCacheForceMiss )
                    
        if self.NOAANWSAWCMETARsServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use NOAA NWS AWC METARs via a server!  It will be set to not use a server." )
            self.NOAANWSAWCMETARsServerUse = False

        if self.NOAANWSAWCMETARsLocalCacheUse == True:
            if self.NOAANWSAWCMETARsLocalCacheExpiry_minutes < 0 or self.NOAANWSAWCMETARsLocalCacheExpiry_minutes > NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS AWC METARs Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSAWCMETARsLocalCacheExpiry_minutes, NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSAWCMETARsLocalCacheExpiry_minutes = NOAA_NWS_AWC_METARS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.NOAANWSAWCMETARsServerUse == True and self.NOAANWSAWCMETARsRemoteCacheUse == True:
            if self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes < 0 or self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes > NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "NOAA NWS AWC METARs Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes, NOAA_NWS_AWC_METARS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, NOAA_NWS_AWC_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes = NOAA_NWS_AWC_METARs_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "NOAANWSAWCMETARsLocalCacheUse: %s" % self.NOAANWSAWCMETARsLocalCacheUse )
        self.log.info( "NOAANWSAWCMETARsLocalCacheExpiry_minutes: %s" % self.NOAANWSAWCMETARsLocalCacheExpiry_minutes )
        self.log.info( "NOAANWSAWCMETARsLocalCacheForceMiss: %s" % self.NOAANWSAWCMETARsLocalCacheForceMiss )
        self.log.info( "NOAANWSAWCMETARsServerUse: %s" % self.NOAANWSAWCMETARsServerUse )
        self.log.info( "NOAANWSAWCMETARsRemoteCacheUse: %s" % self.NOAANWSAWCMETARsRemoteCacheUse )
        self.log.info( "NOAANWSAWCMETARsRemoteCacheExpiry_minutes: %s" % self.NOAANWSAWCMETARsRemoteCacheExpiry_minutes )
        self.log.info( "NOAANWSAWCMETARsRemoteCacheForceMiss: %s" % self.NOAANWSAWCMETARsRemoteCacheForceMiss )

        # Setup for Environment Canada.

        # Environment Canada - City Weather.

        self.ECCityWeatherLocalCacheUse              = False
        self.ECCityWeatherLocalCacheExpiry_minutes   = ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECCityWeatherLocalCacheForceMiss        = False
        self.ECCityWeatherServerUse             = False
        self.ECCityWeatherRemoteCacheUse             = False
        self.ECCityWeatherRemoteCacheExpiry_minutes  = ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECCityWeatherRemoteCacheForceMiss       = False

        self.ECCityWeatherLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_USE_KEY, self.ECCityWeatherLocalCacheUse )
        self.ECCityWeatherLocalCacheExpiry_minutes   = int( Setup_dict.get( ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.ECCityWeatherLocalCacheExpiry_minutes ) ) )
        self.ECCityWeatherLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_FORCE_MISS_KEY, self.ECCityWeatherLocalCacheForceMiss )
        self.ECCityWeatherServerUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_WEATHER_SERVER_USE_KEY, self.ECCityWeatherServerUse )
        self.ECCityWeatherRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_USE_KEY, self.ECCityWeatherRemoteCacheUse )
        self.ECCityWeatherRemoteCacheExpiry_minutes  = int( Setup_dict.get( ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.ECCityWeatherRemoteCacheExpiry_minutes ) )
        self.ECCityWeatherRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_FORCE_MISS_KEY, self.ECCityWeatherRemoteCacheForceMiss )
                    
        if self.ECCityWeatherServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use Environment Canada City Weather via a server!  It will be set to not use a server." )
            self.ECCityWeatherServerUse = False

        if self.ECCityWeatherLocalCacheUse == True:
            if self.ECCityWeatherLocalCacheExpiry_minutes < 0 or self.ECCityWeatherLocalCacheExpiry_minutes > ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada City Weather Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d. I t will be set to default: %d" %
                                        ( self.ECCityWeatherLocalCacheExpiry_minutes, ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECCityWeatherLocalCacheExpiry_minutes = ENVIRONMENT_CANADA_CITY_WEATHER_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.ECCityWeatherServerUse == True and self.ECCityWeatherRemoteCacheUse == True:
            if self.ECCityWeatherRemoteCacheExpiry_minutes < 0 or self.ECCityWeatherRemoteCacheExpiry_minutes > ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada City Weather Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECCityWeatherRemoteCacheExpiry_minutes, ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_CITY_WEATHER_EMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECCityWeatherRemoteCacheExpiry_minutes = ENVIRONMENT_CANADA_CITY_WEATHER_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "ECCityWeatherLocalCacheUse: %s" % self.ECCityWeatherLocalCacheUse )
        self.log.info( "ECCityWeatherLocalCacheExpiry_minutes: %s" % self.ECCityWeatherLocalCacheExpiry_minutes )
        self.log.info( "ECCityWeatherLocalCacheForceMiss: %s" % self.ECCityWeatherLocalCacheForceMiss )
        self.log.info( "ECCityWeatherServerUse: %s" % self.ECCityWeatherServerUse )
        self.log.info( "ECCityWeatherRemoteCacheUse: %s" % self.ECCityWeatherRemoteCacheUse )
        self.log.info( "ECCityWeatherRemoteCacheExpiry_minutes: %s" % self.ECCityWeatherRemoteCacheExpiry_minutes )
        self.log.info( "ECCityWeatherRemoteCacheForceMiss: %s" % self.ECCityWeatherRemoteCacheForceMiss )

        # Environment Canada - City Alerts.

        self.ECCityAlertsLocalCacheUse              = False
        self.ECCityAlertsLocalCacheExpiry_minutes   = ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECCityAlertsLocalCacheForceMiss        = False
        self.ECCityAlertsServerUse             = False
        self.ECCityAlertsRemoteCacheUse             = False
        self.ECCityAlertsRemoteCacheExpiry_minutes  = ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECCityAlertsRemoteCacheForceMiss       = False

        self.ECCityAlertsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_USE_KEY, self.ECCityAlertsLocalCacheUse )
        self.ECCityAlertsLocalCacheExpiry_minutes   = int( Setup_dict.get( ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.ECCityAlertsLocalCacheExpiry_minutes ) ) )
        self.ECCityAlertsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_FORCE_MISS_KEY, self.ECCityAlertsLocalCacheForceMiss )
        self.ECCityAlertsServerUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_ALERTS_SERVER_USE_KEY, self.ECCityAlertsServerUse )
        self.ECCityAlertsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_USE_KEY, self.ECCityAlertsRemoteCacheUse )
        self.ECCityAlertsRemoteCacheExpiry_minutes  = int( Setup_dict.get( ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.ECCityAlertsRemoteCacheExpiry_minutes ) )
        self.ECCityAlertsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_FORCE_MISS_KEY, self.ECCityAlertsRemoteCacheForceMiss )
                    
        if self.ECCityAlertsServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use Environment Canada City Alerts via a server!  It will be set to not use a server." )
            self.ECCityAlertsServerUse = False

        if self.ECCityAlertsLocalCacheUse == True:
            if self.ECCityAlertsLocalCacheExpiry_minutes < 0 or self.ECCityAlertsLocalCacheExpiry_minutes > ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environemnt Canada City Alerts Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECCityAlertsLocalCacheExpiry_minutes, ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECCityAlertsLocalCacheExpiry_minutes = ENVIRONMENT_CANADA_CITY_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.ECCityAlertsServerUse == True and self.ECCityAlertsRemoteCacheUse == True:
            if self.ECCityAlertsRemoteCacheExpiry_minutes < 0 or self.ECCityAlertsRemoteCacheExpiry_minutes > ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada City Alerts Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECCityAlertsRemoteCacheExpiry_minutes, ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_CITY_ALERTS_EMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECCityAlertsRemoteCacheExpiry_minutes = ENVIRONMENT_CANADA_CITY_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "ECCityAlertsLocalCacheUse: %s" % self.ECCityAlertsLocalCacheUse )
        self.log.info( "ECCityAlertsLocalCacheExpiry_minutes: %s" % self.ECCityAlertsLocalCacheExpiry_minutes )
        self.log.info( "ECCityAlertsLocalCacheForceMiss: %s" % self.ECCityAlertsLocalCacheForceMiss )
        self.log.info( "ECCityAlertsServerUse: %s" % self.ECCityAlertsServerUse )
        self.log.info( "ECCityAlertsRemoteCacheUse: %s" % self.ECCityAlertsRemoteCacheUse )
        self.log.info( "ECCityAlertsRemoteCacheExpiry_minutes: %s" % self.ECCityAlertsRemoteCacheExpiry_minutes )
        self.log.info( "ECCityAlertsRemoteCacheForceMiss: %s" % self.ECCityAlertsRemoteCacheForceMiss )

        # Environment Canada - Public Alerts.

        self.ECPublicAlertsLocalCacheUse              = False
        self.ECPublicAlertsLocalCacheExpiry_minutes   = ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECPublicAlertsLocalCacheForceMiss        = False
        self.ECPublicAlertsServerUse             = False
        self.ECPublicAlertsRemoteCacheUse             = False
        self.ECPublicAlertsRemoteCacheExpiry_minutes  = ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECPublicAlertsRemoteCacheForceMiss       = False

        self.ECPublicAlertsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_USE_KEY, self.ECPublicAlertsLocalCacheUse )
        self.ECPublicAlertsLocalCacheExpiry_minutes   = int( Setup_dict.get( ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.ECPublicAlertsLocalCacheExpiry_minutes ) ) )
        self.ECPublicAlertsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_FORCE_MISS_KEY, self.ECPublicAlertsLocalCacheForceMiss )
        self.ECPublicAlertsServerUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_PUBLIC_ALERTS_SERVER_USE_KEY, self.ECPublicAlertsServerUse )
        self.ECPublicAlertsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_USE_KEY, self.ECPublicAlertsRemoteCacheUse )
        self.ECPublicAlertsRemoteCacheExpiry_minutes  = int( Setup_dict.get( ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.ECPublicAlertsRemoteCacheExpiry_minutes ) )
        self.ECPublicAlertsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_FORCE_MISS_KEY, self.ECPublicAlertsRemoteCacheForceMiss )
                    
        if self.ECPublicAlertsServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use Environment Canada Public Alerts via a server!  It will be set to not use a server." )
            self.ECPublicAlertsServerUse = False

        if self.ECPublicAlertsLocalCacheUse == True:
            if self.ECPublicAlertsLocalCacheExpiry_minutes < 0 or self.ECPublicAlertsLocalCacheExpiry_minutes > ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada Public Alerts Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECPublicAlertsLocalCacheExpiry_minutes, ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECPublicAlertsLocalCacheExpiry_minutes = ENVIRONMENT_CANADA_PUBLIC_ALERTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.ECPublicAlertsServerUse == True and self.ECPublicAlertsRemoteCacheUse == True:
            if self.ECPublicAlertsRemoteCacheExpiry_minutes < 0 or self.ECPublicAlertsRemoteCacheExpiry_minutes > ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada Public Alerts Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECPublicAlertsRemoteCacheExpiry_minutes, ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_PUBLIC_ALERTS_EMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECPublicAlertsRemoteCacheExpiry_minutes = ENVIRONMENT_CANADA_PUBLIC_ALERTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "ECPublicAlertsLocalCacheUse: %s" % self.ECPublicAlertsLocalCacheUse )
        self.log.info( "ECPublicAlertsLocalCacheExpiry_minutes: %s" % self.ECPublicAlertsLocalCacheExpiry_minutes )
        self.log.info( "ECPublicAlertsLocalCacheForceMiss: %s" % self.ECPublicAlertsLocalCacheForceMiss )
        self.log.info( "ECPublicAlertsServerUse: %s" % self.ECPublicAlertsServerUse )
        self.log.info( "ECPublicAlertsRemoteCacheUse: %s" % self.ECPublicAlertsRemoteCacheUse )
        self.log.info( "ECPublicAlertsRemoteCacheExpiry_minutes: %s" % self.ECPublicAlertsRemoteCacheExpiry_minutes )
        self.log.info( "ECPublicAlertsRemoteCacheForceMiss: %s" % self.ECPublicAlertsRemoteCacheForceMiss )

        # Environment Canada - Marine Forecasts.

        self.ECMarineForecastsLocalCacheUse              = False
        self.ECMarineForecastsLocalCacheExpiry_minutes   = ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECMarineForecastsLocalCacheForceMiss        = False
        self.ECMarineForecastsServerUse                  = False
        self.ECMarineForecastsRemoteCacheUse             = False
        self.ECMarineForecastsRemoteCacheExpiry_minutes  = ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.ECMarineForecastsRemoteCacheForceMiss       = False

        self.ECMarineForecastsLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_USE_KEY, self.ECMarineForecastsLocalCacheUse )
        self.ECMarineForecastsLocalCacheExpiry_minutes   = int( Setup_dict.get( ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.ECMarineForecastsLocalCacheExpiry_minutes ) ) )
        self.ECMarineForecastsLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_FORCE_MISS_KEY, self.ECMarineForecastsLocalCacheForceMiss )
        self.ECMarineForecastsServerUse                  = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_MARINE_FORECASTS_SERVER_USE_KEY, self.ECMarineForecastsServerUse )
        self.ECMarineForecastsRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_USE_KEY, self.ECMarineForecastsRemoteCacheUse )
        self.ECMarineForecastsRemoteCacheExpiry_minutes  = int( Setup_dict.get( ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.ECMarineForecastsRemoteCacheExpiry_minutes ) )
        self.ECMarineForecastsRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_FORCE_MISS_KEY, self.ECMarineForecastsRemoteCacheForceMiss )
                    
        if self.ECMarineForecastsServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use Environment Canada Marine Forecasts via a server!  It will be set to not use a server." )
            self.ECMarineForecastsServerUse = False

        if self.ECMarineForecastsLocalCacheUse == True:
            if self.ECMarineForecastsLocalCacheExpiry_minutes < 0 or self.ECMarineForecastsLocalCacheExpiry_minutes > ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada Marine Forecasts Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECMarineForecastsLocalCacheExpiry_minutes, ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECMarineForecastsLocalCacheExpiry_minutes = ENVIRONMENT_CANADA_MARINE_FORECASTS_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.ECMarineForecastsServerUse == True and self.ECMarineForecastsRemoteCacheUse == True:
            if self.ECMarineForecastsRemoteCacheExpiry_minutes < 0 or self.ECMarineForecastsRemoteCacheExpiry_minutes > ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Environment Canada Marine Forecasts Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.ECMarineForecastsRemoteCacheExpiry_minutes, ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, ENVIRONMENT_CANADA_MARINE_FORECASTS_EMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.ECMarineForecastsRemoteCacheExpiry_minutes = ENVIRONMENT_CANADA_MARINE_FORECASTS_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "ECMarineForecastsLocalCacheUse: %s" % self.ECMarineForecastsLocalCacheUse )
        self.log.info( "ECMarineForecastsLocalCacheExpiry_minutes: %s" % self.ECMarineForecastsLocalCacheExpiry_minutes )
        self.log.info( "ECMarineForecastsLocalCacheForceMiss: %s" % self.ECMarineForecastsLocalCacheForceMiss )
        self.log.info( "ECMarineForecastsServerUse: %s" % self.ECMarineForecastsServerUse )
        self.log.info( "ECMarineForecastsRemoteCacheUse: %s" % self.ECMarineForecastsRemoteCacheUse )
        self.log.info( "ECMarineForecastsRemoteCacheExpiry_minutes: %s" % self.ECMarineForecastsRemoteCacheExpiry_minutes )
        self.log.info( "ECMarineForecastsRemoteCacheForceMiss: %s" % self.ECMarineForecastsRemoteCacheForceMiss )

        # Setup for Drive BC.

        self.DriveBCLocalCacheUse              = False
        self.DriveBCLocalCacheExpiry_minutes   = DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
        self.DriveBCLocalCacheForceMiss        = False
        self.DriveBCServerUse             = False
        self.DriveBCRemoteCacheUse             = False
        self.DriveBCRemoteCacheExpiry_minutes  = DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
        self.DriveBCRemoteCacheForceMiss       = False

        self.DriveBCLocalCacheUse              = self.YesNoTrueFalse( Setup_dict, DRIVE_BC_LOCAL_CACHE_USE_KEY, self.DriveBCLocalCacheUse )
        self.DriveBCLocalCacheExpiry_minutes   = int( Setup_dict.get( DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_KEY, str( self.DriveBCLocalCacheExpiry_minutes ) ) )
        self.DriveBCLocalCacheForceMiss        = self.YesNoTrueFalse( Setup_dict, DRIVE_BC_LOCAL_CACHE_FORCE_MISS_KEY, self.DriveBCLocalCacheForceMiss )
        self.DriveBCServerUse             = self.YesNoTrueFalse( Setup_dict, DRIVE_BC_SERVER_USE_KEY, self.DriveBCServerUse )
        self.DriveBCRemoteCacheUse             = self.YesNoTrueFalse( Setup_dict, DRIVE_BC_REMOTE_CACHE_USE_KEY, self.DriveBCRemoteCacheUse )
        self.DriveBCRemoteCacheExpiry_minutes  = int( Setup_dict.get( DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_KEY, self.DriveBCRemoteCacheExpiry_minutes ) )
        self.DriveBCRemoteCacheForceMiss       = self.YesNoTrueFalse( Setup_dict, DRIVE_BC_REMOTE_CACHE_FORCE_MISS_KEY, self.DriveBCRemoteCacheForceMiss )
                    
        if self.DriveBCServerUse == True and self.WXFeedsServerUse == False:
            self.log.error( "You must use WXFeeds Server to use Drive BC via a server!  It will be set to not use a server." )
            self.DriveBCServerUse = False

        if self.DriveBCLocalCacheUse == True:
            if self.DriveBCLocalCacheExpiry_minutes < 0 or self.DriveBCLocalCacheExpiry_minutes > DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Drive BC Local Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.DriveBCLocalCacheExpiry_minutes, DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_MAXIMUM, DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.DriveBCLocalCacheExpiry_minutes = DRIVE_BC_LOCAL_CACHE_EXPIRY_MINUTES_DEFAULT
                                        
        if self.DriveBCServerUse == True and self.DriveBCRemoteCacheUse == True:
            if self.DriveBCRemoteCacheExpiry_minutes < 0 or self.DriveBCRemoteCacheExpiry_minutes > DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM:
                self.log.error( "Drive BC Remote Cache Expiry (minutes): %d is less than 0 or greater than maximum: %d.  It will be set to default: %d" %
                                        ( self.DriveBCRemoteCacheExpiry_minutes, DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_MAXIMUM, DRIVE_BC_EMOTE_CACHE_EXPIRY_MINUTES_DEFAULT ) )
                self.DriveBCRemoteCacheExpiry_minutes = DRIVE_BC_REMOTE_CACHE_EXPIRY_MINUTES_DEFAULT
                                           
        self.log.info( "DriveBCLocalCacheUse: %s" % self.DriveBCLocalCacheUse )
        self.log.info( "DriveBCLocalCacheExpiry_minutes: %s" % self.DriveBCLocalCacheExpiry_minutes )
        self.log.info( "DriveBCLocalCacheForceMiss: %s" % self.DriveBCLocalCacheForceMiss )
        self.log.info( "DriveBCServerUse: %s" % self.DriveBCServerUse )
        self.log.info( "DriveBCRemoteCacheUse: %s" % self.DriveBCRemoteCacheUse )
        self.log.info( "DriveBCRemoteCacheExpiry_minutes: %s" % self.DriveBCRemoteCacheExpiry_minutes )
        self.log.info( "DriveBCRemoteCacheForceMiss: %s" % self.DriveBCRemoteCacheForceMiss )

        # Setup for WS-DOT.

        self.WSDOTServicesUse                               = False
        self.WSDOTServicesAccessCode                        = ""
        self.WSDOTServicesUseLocalCache                     = False
        self.WSDOTServicesCacheExpiry_minutes               = WS_DOT_SERVICES_CACHE_EXPIRY_MINUTES_DEFAULT
        self.WSDOTServicesForceCacheMiss                    = False

        # WS DOT services controls.

        self.WSDOTServicesUse = self.YesNoTrueFalse( Setup_dict,
                                                     WS_DOT_SERVICES_USE_KEY,
                                                     self.WSDOTServicesUse )
        self.log.info( "WSDOTServicesUse: %s" % self.WSDOTServicesUse )

        self.WSDOTServicesAccessCode = Setup_dict.get( WS_DOT_SERVICES_ACCESS_CODE_KEY, self.WSDOTServicesAccessCode )
        self.log.info( "WSDOTServicesAccessCode: %s" % self.WSDOTServicesAccessCode )

        # If we're not going to use a server and we're going to use WS DOT services 
        # we need an access code.

        if self.WXFeedsServerUse == False and      \
           self.WSDOTServicesUse == True and            \
           self.WSDOTServicesAccessCode == "":
            self.log.error( "A WS DOT Access Code is required to use WS DOT API services without a server!  WS DOT services will be disabled." )

            self.WSDOTServicesUse = False

        self.WSDOTServicesUseLocalCache = self.YesNoTrueFalse( Setup_dict,
                                                               WS_DOT_SERVICES_USE_LOCAL_CACHE_KEY,
                                                               self.WSDOTServicesUseLocalCache )
        self.log.info( "WSDOTServicesUseLocalCache: %s" % self.WSDOTServicesUseLocalCache )

        self.WSDOTServicesCacheExpiry_minutes = int( Setup_dict.get( WS_DOT_SERVICES_CACHE_EXPIRY_MINUTES_KEY,
                                                                     WS_DOT_SERVICES_CACHE_EXPIRY_MINUTES_DEFAULT ) )
        self.log.info( "WSDOTServicesCacheExpiry_minutes: %d" % self.WSDOTServicesCacheExpiry_minutes )

        self.WSDOTServicesForceCacheMiss = self.YesNoTrueFalse( Setup_dict,
                                                                WS_DOT_SERVICES_FORCE_CACHE_MISS_KEY,
                                                                self.WSDOTServicesForceCacheMiss )
        self.log.info( "WSDOTServicesForceCacheMiss: %s" % self.WSDOTServicesForceCacheMiss )

        # If we're going to use local caching, we need a path.

        if self.WSDOTServicesUse == True and                \
           self.WSDOTServicesUseLocalCache == True and      \
           self.WXFeedsLocalCacheRoot == "":
            self.log.error( "\'%s\' not set! Local caching will be disabled." % LOCAL_CACHE_PATH_KEY )

            self.WSDOTServicesUseLocalCache = False

        # NOAA NWS - Storm Prediction Center.

        self.NOAANWSSPCUseLocalCache         = False
        self.NOAANWSSPCCacheExpiry_minutes   = 0
        self.NOAANWSSPCForceCacheMiss        = False

        self.NOAANWSSPCUseLocalCache = self.YesNoTrueFalse( Setup_dict,
                                                            NOAA_NWS_SPC_USE_LOCAL_CACHE,
                                                            self.NOAANWSSPCUseLocalCache )
        self.log.info( "NOAANWSSPCUseLocalCache: %s" % self.NOAANWSSPCUseLocalCache )

        self.NOAANWSSPCCacheExpiry_minutes = int( Setup_dict.get( NOAA_NWS_SPC_CACHE_EXPIRY_MINUTES,
                                                                  NOAA_NWS_SPC_CACHE_EXPIRY_MINUTES_DEFAULT ) )
        self.log.info( "NOAANWSSPCCacheExpiry_minutes: %d" % self.NOAANWSSPCCacheExpiry_minutes )

        self.NOAANWSSPCForceCacheMiss = self.YesNoTrueFalse( Setup_dict,
                                                             NOAA_NWS_SPC_FORCE_CACHE_MISS,
                                                             self.NOAANWSSPCForceCacheMiss )
        self.log.info( "NOAANWSSPCForceCacheMiss: %s" % self.NOAANWSSPCForceCacheMiss )

        # If we're going to use local caching, we need a path.

        if self.NOAANWSSPCUseLocalCache == True and      \
           self.WXFeedsLocalCacheRoot == "":
            self.log.error( "\'%s\' not set! Local caching will be disabled." % LOCAL_CACHE_PATH_KEY )

            self.NOAANWSSPCUseLocalCache = False

         # NOAA NWS - National Hurricane Center.

        self.NOAANWSNHCUseLocalCache         = False
        self.NOAANWSNHCCacheExpiry_minutes   = 0
        self.NOAANWSNHCForceCacheMiss        = False

        self.NOAANWSNHCUseLocalCache = self.YesNoTrueFalse( Setup_dict,
                                                            NOAA_NWS_NHC_USE_LOCAL_CACHE,
                                                            self.NOAANWSNHCUseLocalCache )
        self.log.info( "NOAANWSNHCUseLocalCache: %s" % self.NOAANWSNHCUseLocalCache )

        self.NOAANWSNHCCacheExpiry_minutes = int( Setup_dict.get( NOAA_NWS_NHC_CACHE_EXPIRY_MINUTES,
                                                                  NOAA_NWS_NHC_CACHE_EXPIRY_MINUTES_DEFAULT ) )
        self.log.info( "NOAANWSNHCCacheExpiry_minutes: %d" % self.NOAANWSNHCCacheExpiry_minutes )

        self.NOAANWSNHCForceCacheMiss = self.YesNoTrueFalse( Setup_dict,
                                                             NOAA_NWS_NHC_FORCE_CACHE_MISS,
                                                             self.NOAANWSNHCForceCacheMiss )
        self.log.info( "NOAANWSNHCForceCacheMiss: %s" % self.NOAANWSNHCForceCacheMiss )

        # If we're going to use local caching, we need a path.

        if self.NOAANWSNHCUseLocalCache == True and      \
           self.WXFeedsLocalCacheRoot == "":
            self.log.error( "\'%s\' not set! Local caching will be disabled." % LOCAL_CACHE_PATH_KEY )

            self.NOAANWSNHCUseLocalCache = False

        # Washington State - Department of Transportion.

        # Border Crossings.

        self.WSDOTBorderCrossingsUseLocalCache         = False
        self.WSDOTBorderCrossingsCacheExpiry_minutes   = 0
        self.WSDOTBorderCrossingsForceCacheMiss        = False

        self.WSDOTBorderCrossingsUseLocalCache = self.YesNoTrueFalse( Setup_dict,
                                                                      WS_DOT_BORDER_CROSSINGS_USE_LOCAL_CACHE,
                                                                      self.WSDOTBorderCrossingsUseLocalCache )
        self.log.info( "WSDOTBorderCrossingsUseLocalCache: %s" % self.WSDOTBorderCrossingsUseLocalCache )

        self.WSDOTBorderCrossingsCacheExpiry_minutes = int( Setup_dict.get( WS_DOT_BORDER_CROSSINGS_CACHE_EXPIRY_MINUTES,
                                                                            WS_DOT_BORDER_CROSSINGS_CACHE_EXPIRY_MINUTES_DEFAULT ) )
        self.log.info( "WSDOTBorderCrossingsCacheExpiry_minutes: %d" % self.WSDOTBorderCrossingsCacheExpiry_minutes )

        self.WSDOTBorderCrossingsForceCacheMiss = self.YesNoTrueFalse( Setup_dict,
                                                                       WS_DOT_BORDER_CROSSINGS_FORCE_CACHE_MISS,
                                                                       self.WSDOTBorderCrossingsForceCacheMiss )
        self.log.info( "WSDOTBorderCrossingsForceCacheMiss: %s" % self.WSDOTBorderCrossingsForceCacheMiss )

        # If we're going to use local caching, we need a path.

        if self.WSDOTBorderCrossingsUseLocalCache == True and      \
           self.WXFeedsLocalCacheRoot == "":
            self.log.error( "\'%s\' not set! Local caching will be disabled." % LOCAL_CACHE_PATH_KEY )

            self.WSDOTBorderCrossingsUseLocalCache = False

        # 511 Alberta.

        # Road Conditions.

        self.Alberta511RoadConditionsUseLocalCache         = False
        self.Alberta511RoadConditionsCacheExpiry_minutes   = 0
        self.Alberta511RoadConditionsForceCacheMiss        = False

        self.Alberta511RoadConditionsUseLocalCache = self.YesNoTrueFalse( Setup_dict,
                                                                          ALBERTA_511_ROAD_CONDITIONS_USE_LOCAL_CACHE,
                                                                          self.Alberta511RoadConditionsUseLocalCache )
        self.log.info( "Alberta511RoadConditionsUseLocalCache: %s" % self.Alberta511RoadConditionsUseLocalCache )

        self.Alberta511RoadConditionsCacheExpiry_minutes = int( Setup_dict.get( ALBERTA_511_ROAD_CONDITIONS_CACHE_EXPIRY_MINUTES,
                                                                                ALBERTA_511_ROAD_CONDITIONS_CACHE_EXPIRY_MINUTES_DEFAULT ) )
        self.log.info( "Alberta511RoadConditionsCacheExpiry_minutes: %d" % self.Alberta511RoadConditionsCacheExpiry_minutes )

        self.Alberta511RoadConditionsForceCacheMiss = self.YesNoTrueFalse( Setup_dict,
                                                                           ALBERTA_511_ROAD_CONDITIONS_FORCE_CACHE_MISS,
                                                                           self.Alberta511RoadConditionsForceCacheMiss )
        self.log.info( "Alberta511RoadConditionsForceCacheMiss: %s" % self.Alberta511RoadConditionsForceCacheMiss )

        # If we're going to use local caching, we need a path.

        if self.Alberta511RoadConditionsUseLocalCache == True and      \
           self.WXFeedsLocalCacheRoot == "":
            self.log.error( "\'%s\' not set! Local caching will be disabled." % LOCAL_CACHE_PATH_KEY )

            self.Alberta511RoadConditionsUseLocalCache = False

       # Set up WS-DOT dictionary.

        self.WSDOTDictionary = {
                                 WS_DOT_BORDER_CROSSINGS_KEY : WS_DOT_BORDER_CROSSINGS_REQUEST_URL
                               }

       # Set up Drive BC dictionary.

        self.DriveBCDictionary = {
                                   DRIVE_BC_ENTIRE_PROVINCE_KEY : DRIVE_BC_ENTIRE_PROVINCE_FEED_URL,
                                   DRIVE_BC_LOWER_MAINLAND_KEY : DRIVE_BC_LOWER_MAINLAND_FEED_URL,
                                   DRIVE_BC_VANCOUVER_ISLAND_KEY : DRIVE_BC_VANCOUVER_ISLAND_FEED_URL,
                                   DRIVE_BC_KOOTENAYS_KEY : DRIVE_BC_KOOTENAYS_FEED_URL,
                                   DRIVE_BC_THOMPSON_OKANAGAN_KEY : DRIVE_BC_THOMPSON_OKANAGAN_FEED_URL,
                                   DRIVE_BC_CARIBOO_KEY : DRIVE_BC_CARIBOO_FEED_URL,
                                   DRIVE_BC_NORTH_CENTRAL_KEY : DRIVE_BC_NORTH_CENTRAL_FEED_URL,
                                   DRIVE_BC_NORTH_EAST_KEY : DRIVE_BC_NORTH_EAST_FEED_URL,
                                   DRIVE_BC_NORTH_WEST_KEY : DRIVE_BC_NORTH_WEST_FEED_URL
                                 }

       # Set up 511 Alberta dictionary.

        self.Alberta511Dictionary = {
                                      ALBERTA_511_ROAD_CONDITIONS_KEY : ALBERTA_511_ROAD_CONDITIONS_REQUEST_URL,
                                      ALBERTA_511_CAMERAS_KEY :  ALBERTA_511_CAMERAS_REQUEST_URL,
                                      ALBERTA_511_GROUPED_CAMERAS_KEY : ALBERTA_511_GROUPED_CAMERAS_REQUEST_URL,
                                      ALBERTA_511_PARKS_KEY : ALBERTA_511_PARKS_REQUEST_URL,
                                      ALBERTA_511_FERRIES_KEY : ALBERTA_511_FERRIES_REQUEST_URL,
                                      ALBERTA_511_EVENTS_KEY : ALBERTA_511_EVENTS_REQUEST_URL,
                                      ALBERTA_511_ALERTS_KEY : ALBERTA_511_ALERTS_REQUEST_URL
                                    }

        # 511 Alberta Road Condition Cache.
        # 
        # 511 Alberta road conditions do not have position (latitude and longitude) for 
        # each road condition.  They do have a polyline to define the road 
        # condition from which we can calculate the latitude and longitude for the road 
        # condition but that is an expensive calculation, so cache the result.  And there 
        # can be humdreds of entries so we only process a fraction of them per archive 
        # cycle - we store some state variable is the cache.  See Alberta511RoadConditions() 
        # for the format of the cache.

        self.Alberta511RoadConditionCache = {}
    
        # Report (skin) specific setup.

        self.ReportSettings = []

        SetupReports_dict = Setup_dict.get( REPORTS_KEY )
        assert SetupReports_dict != None                          # Must exist for WXFeeds to work.
#        self.log.debug("SetupReports_dict: %s" % SetupReports_dict)

        for name, value in SetupReports_dict.items():
            # name should be a skin name, value should be its setup dictionary. 

            ReportName            = name
            ReportSettings_dict   = value

#            self.log.debug( "Report: %s, Settings: %s" % ( ReportName, ReportSettings_dict ) )

            skin = self.GetSkinSetting( ReportName, SKIN_KEY )
#            self.log.debug("%s: %s: %s" % ( ReportName, SKIN_KEY, skin ) )

            custom_css_path = ReportSettings_dict.get( CUSTOM_CSS_PATH_KEY, "" )
#            self.log.debug("%s: %s: %s" % ( ReportName, CUSTOM_CSS_PATH_KEY, custom_css_file ) )

            disable_WXFeeds_css = self.YesNoTrueFalse( ReportSettings_dict,
                                                       DISABLE_WXFEEDS_CSS_KEY )
#            self.log.debug( "disable_WXFeeds_css: %s" % disable_WXFeeds_css )

            include_filenames_to_delete = ReportSettings_dict.get( INCLUDE_FILENAMES_TO_DELETE_KEY, "" )
#            self.log.debug("%s: %s: %s" % ( ReportName, INCLUDE_FILENAMES_TO_DELETE_KEY, include_filenames_to_delete ) )

            other_filenames_to_delete = ReportSettings_dict.get( OTHER_FILENAMES_TO_DELETE_KEY, "" )
#            self.log.debug("%s: %s: %s" % ( ReportName, OTHER_FILENAMES_TO_DELETE_KEY, other_filenames_to_delete ) )

            add_back_to_top_button  = self.YesNoTrueFalse( ReportSettings_dict, ADD_BACK_TO_TOP_BUTTON_KEY )
#            self.log.debug("%s: %s: %s" % ( ReportName, ADD_BACK_TO_TOP_BUTTON_KEY, add_back_to_top_button ) )

            # Get settings that might be in this skin or an ancestor.

            SkinRoot = self.GetSkinSetting( ReportName, SKIN_ROOT )
#            self.log.debug("%s: %s: %s" % ( ReportName, SKIN_ROOT, SkinRoot ) )

            HTMLRoot = self.GetSkinSetting( ReportName, HTML_ROOT )
#            self.log.debug("%s: %s: %s" % ( ReportName, HTML_ROOT, HTMLRoot ) )

            generated_files_root = SkinRoot + "/" + skin

            html_files_root = HTMLRoot

            # Make relative roots/files absolute.

            if Path( generated_files_root ).is_absolute() == False:
                # Make it relative to WEEWX_ROOT.

                generated_files_root = WeeWXRoot + "/" + generated_files_root

            if Path( html_files_root ).is_absolute() == False:
                # Make it relative to WEEWX_ROOT.

                html_files_root = WeeWXRoot + "/" + html_files_root

            if custom_css_path != "" and    \
               Path( custom_css_path ).is_absolute() == False:
                # Make it relative to include_files_root.

                custom_css_path = generated_files_root + "/" + custom_css_path

            # Cheetah memory leak workaround.
            #
            # As of 2021/01/28, the Cheetah generator has a memeory leak which 
            # can be worked around by.

            cheetah_memory_leak_workaround = self.YesNoTrueFalse( ReportSettings_dict,
                                                                  CHEETAH_MEMORY_LEAK_WORKAROUND_KEY,
                                                                  False )
#            self.log.debug( "cheetah_memory_leak_workaround: %s" % cheetah_memory_leak_workaround )

            # Create the settings dictionary entry for this skin.

            self.ReportSettings.append(
                                        {
                                          NAME_KEY     : ReportName,
                                          SETTINGS_KEY : 
                                                         {
                                                           SKIN_KEY                             : skin,
                                                           GENERATED_FILES_ROOT_KEY             : generated_files_root,
                                                           HTML_FILES_ROOT_KEY                  : html_files_root,
                                                           CUSTOM_CSS_PATH_KEY                  : custom_css_path,
                                                           DISABLE_WXFEEDS_CSS_KEY              : disable_WXFeeds_css,
                                                           INCLUDE_FILENAMES_TO_DELETE_KEY      : include_filenames_to_delete,
                                                           OTHER_FILENAMES_TO_DELETE_KEY        : other_filenames_to_delete,
                                                           ADD_BACK_TO_TOP_BUTTON_KEY           : add_back_to_top_button,
                                                           CHEETAH_MEMORY_LEAK_WORKAROUND_KEY   : cheetah_memory_leak_workaround,
                                                           BACK_TO_TOP_BUTTON_ADDED_KEY         : False,
                                                           GENERATED_FILE_KEY                   : None,
                                                           CHANGING_GENERATED_FILE_KEY          : False,
                                                           CURRENT_GENERATED_FILENAME_KEY       : "",
                                                           CURRENT_GENERATED_FILE_SEQUENCE_KEY  : "",
                                                           CURRENT_GENERATED_FILE_TYPE_KEY      : "",
                                                           NEW_GENERATED_FILENAME_KEY           : "",
                                                           NEW_GENERATED_FILE_SEQUENCE_KEY      : "",
                                                           NEW_GENERATED_FILE_TYPE_KEY          : ""
                                                         }
                                        }
                                      )

#            self.log.debug( "WXFeeds: Report Dictionary: %s" % self.ReportSettings[ len( self.ReportSettings ) - 1 ] )

        # Get [WXFeeds][[Feeds]] dictionary.

        self.Feeds_dict = WXFeeds_dict.get( FEEDS_KEY )
        assert self.Feeds_dict != None

        # Report (skin) specific setup.

        for r in self.ReportSettings:
#            self.log.debug( "ReportSetting: %s" % r )
#            self.log.debug( "ReportSetting: DISABLE_WXFEEDS_CSS_KEY: %s" % r[ SETTINGS_KEY ].get( DISABLE_WXFEEDS_CSS_KEY ) )
#            self.log.debug( "ReportSetting: CUSTOM_CSS_PATH_KEY: %s" % r[ SETTINGS_KEY ].get( CUSTOM_CSS_PATH_KEY ) )

            # Create the report (skin) cache directory.
            # This must be done before any files are created.

            assert self.WXFeedsLocalCacheRoot != ""

            #CachePath = self.WXFeedsLocalCacheRoot + "/" + r[ SETTINGS_KEY ][ SKIN_KEY ]
            CachePath = str( Path( self.WXFeedsLocalCacheRoot ).joinpath( r[ SETTINGS_KEY ][ SKIN_KEY ] ) )
#            self.log.debug( "WXFeeds: CachePath: %s" % CachePath )

            p = Path( CachePath )
            assert p != None

            if p.exists() == False:
#                self.log.debug( "Creating Local Cache Directory: '%s'" % CachePath )

                os.makedirs( CachePath )

            # Create the WXFeeds HTML and JavaScript directories relative to html_files_root.

            HTMLFilesRoot = r[ SETTINGS_KEY ].get( HTML_FILES_ROOT_KEY )

            WXFeedsHTMLDirectory = HTMLFilesRoot + "/WXFeeds"
#            self.log.debug("WXFeeds: HTML Directory: %s" % WXFeedsHTMLDirectory )

            p = Path( WXFeedsHTMLDirectory )
            assert p != None

            if p.exists() == False:
#                self.log.debug( "Creating WXFeeds HTML Directory. . ." )

                os.makedirs( WXFeedsHTMLDirectory )

            WXFeedsJavaScriptDirectory = WXFeedsHTMLDirectory + "/js"
#            self.log.debug("WXFeeds: JavaScript Directory: %s" % WXFeedsJavaScriptDirectory )

            p = Path( WXFeedsJavaScriptDirectory )
            assert p != None

            if p.exists() == False:
#                self.log.debug( "Creating WXFeeds Java Script Directory. . ." )

                os.makedirs( WXFeedsJavaScriptDirectory )

            # Create custom CSS file(s).

            if r[ SETTINGS_KEY ][ CUSTOM_CSS_PATH_KEY ] != "":
                self.CreateCustomCSS(
                                      r[ SETTINGS_KEY ][ SKIN_KEY ],
                                      r[ SETTINGS_KEY ][ CUSTOM_CSS_PATH_KEY ],
                                      r[ SETTINGS_KEY ][ DISABLE_WXFEEDS_CSS_KEY ]
                                    )


            # Delete files.
            #
            # Generated files are files created (generated) by WXFeeds.
            #
            # They are complete HTML files with all required elements or 
            # 'include' files with HTML elements to be included in 
            # another HTML file.
            #
            # If set up correctly, they may be deleted at start up.

            GeneratedFilesRoot = r[ SETTINGS_KEY ][ GENERATED_FILES_ROOT_KEY ]

            FilesToDelete = r[ SETTINGS_KEY ][ INCLUDE_FILENAMES_TO_DELETE_KEY ]
            assert FilesToDelete != None

            FilesToDeleteList = FilesToDelete.split( "|" )

            for FileToDelete in FilesToDeleteList:

                FileToDelete = FileToDelete.strip()

                if FileToDelete != "":
                    FileToDeletePath = FileToDelete
                    
                    p = Path( FileToDelete )
                    if p.is_absolute() == False:
                        FileToDeletePath = GeneratedFilesRoot + "/" + FileToDelete

#                        self.log.debug( "WXFeeds: Include FileToDeletePath: '%s'" % FileToDeletePath )

                        p = Path( FileToDeletePath )
                        if p.exists() == True:
                            os.remove( FileToDeletePath )

                        if r[ SETTINGS_KEY ][ CHEETAH_MEMORY_LEAK_WORKAROUND_KEY ] == True:
                            FileToDeletePath = GeneratedFilesRoot + "/WXFeeds-" + FileToDelete

#                            self.log.debug( "WXFeeds: Include FileToDeletePath: '%s'" % FileToDeletePath )

                            p = Path( FileToDeletePath )
                            if p.exists() == True:
                                os.remove( FileToDeletePath )

            FilesToDelete = r[ SETTINGS_KEY ][ OTHER_FILENAMES_TO_DELETE_KEY ]
            assert FilesToDelete != None

            FilesToDeleteList = FilesToDelete.split( "|" )

            for FileToDelete in FilesToDeleteList:

                FileToDelete = FileToDelete.strip()

                if FileToDelete != "":
                    FileToDeletePath = ""

                    p = Path( FileToDelete )
                    assert p != None

                    if p.is_absolute == True:
                        FileToDeletePath = FileToDelete
                    else:
                        FileToDeletePath = GeneratedFilesRoot + "/" + FileToDelete

#                    self.log.debug( "WXFeeds: Other FileToDeletePath: '%s'" % FileToDeletePath )

                    p = Path( FileToDeletePath )
                    if p.exists() == True:
                        os.remove( FileToDeletePath )
                        p = Path( FileToDeletePath )
                        if p.exists() == True:
                            self.log.warning( "WXFeeds: Other File: '%s' was not deleted!" % FileToDeletePath )

        # Seed the random number generator.

        random.seed()

        # All Done!

    def new_archive_record(self, event):
        """Gets called on a new archive record event."""

        #DEBUG = True
        DEBUG = False
        
        assert self.Feeds_dict != None  # We need some feeds!

        if self.TakeOutTheTrash == True:
            # Force garbage collection.

            assert gc.isenabled() == True

            if DEBUG == True:
                self.log.debug( "WXFeeds.new_archive_record(): Entering . . ." )
                self.log.debug( "WXFeeds.new_archive_record(): Process ID: %d" % os.getpid() )
                self.log.debug( "WXFeeds.new_archive_record(): Memory use before gc.collect():" )
                x, y, z = gc.get_count()
                self.log.debug( "WXFeeds.new_archive_record(): gc.get_count(): %d, %d, %d" % ( x, y, z ) )
                self.log.debug( "WXFeeds.new_archive_record(): sys.getallocatedblocks(): %d" % sys.getallocatedblocks() )

            gc.collect()

            if DEBUG == True:
                self.log.debug( "WXFeeds.new_archive_record(): Memory use after gc.collect():" )
                x, y, z = gc.get_count()
                self.log.debug( "WXFeeds.new_archive_record(): gc.get_count(): %d, %d, %d" % ( x, y, z ) )
                self.log.debug( "WXFeeds.new_archive_record(): sys.getallocatedblocks(): %d" % sys.getallocatedblocks() )
                self.log.debug( "WXFeeds.new_archive_record()" )

        # If we're going to use a WXFeeds server, log in to it else the AccessToken will expire!
        #
        # We tested server login previously so if we still intend to use a server, a loging failure 
        # here should be transient.  We don't chenge any flags so things may blow up later 
        # (because we don't have a token)!
    
        if self.WXFeedsServerUse == True:

            ReturnedStatus                              = False
            self.WXFeedsServerAccessToken               = ""
            self.WXFeedsServerAccessTokenExpiry_seconds = -1

            ReturnedStatus,                                 \
            self.WXFeedsServerAccessToken,                  \
            self.WXFeedsServerAccessTokenExpiry_seconds = self.WXFeedsServerLogin(
                                                                                    self.WXFeedsServerURL,
                                                                                    self.WXFeedsServerUsername,
                                                                                    self.WXFeedsServerPassword
                                                                                    )
            if ReturnedStatus == False:
                self.log.error( "WXFeeds: Server Login NOT Successful!" )
            else:
                assert self.WXFeedsServerAccessToken != ""
                assert self.WXFeedsServerAccessTokenExpiry_seconds >= 0

                self.log.info( "WXFeeds: Server Login Successful." )

        if self.GoogleMapsGeocodingUse == True:
            # Testing for GoogleMapsGeocoding 'cause we don't actually use geocoding anywhere - yet!

            Location    = "Lockyers"
            Address     = "101 - 2450 Radio Tower Road, Oliver BC, V0H 1T1, Canada"
            Country     = "CA"

            ReturnedStatus, Latitude, Longitude = self.GoogleMapsGeocoding(
                                                                            Location,
                                                                            Address,
                                                                            "country:%s" % Country
                                                                          )
            if DEBUG == True:
                assert ReturnedStatus == True

            #self.log.info( "WXFeeds: '%s', '%s', '%s', Lat: %f, Lon: %f" % ( Location, Address, Country, Latitude, Longitude ) )

        t1 = time.time()
        ngen = 0

        # Initialize self.ReportSettings.

        for r in self.ReportSettings:

            # Check to see that all generated files were closed after 
            # previous cycle.

            assert r[ SETTINGS_KEY][ GENERATED_FILE_KEY ] == None

            r[ SETTINGS_KEY][ BACK_TO_TOP_BUTTON_ADDED_KEY ]        = False
            r[ SETTINGS_KEY][ CURRENT_GENERATED_FILENAME_KEY ]      = ""
            r[ SETTINGS_KEY][ CURRENT_GENERATED_FILE_SEQUENCE_KEY ] = ""
            r[ SETTINGS_KEY][ CURRENT_GENERATED_FILE_TYPE_KEY ]     = ""
            r[ SETTINGS_KEY][ NEW_GENERATED_FILENAME_KEY ]          = ""
            r[ SETTINGS_KEY][ NEW_GENERATED_FILE_SEQUENCE_KEY ]     = ""
            r[ SETTINGS_KEY][ NEW_GENERATED_FILE_TYPE_KEY ]         = ""

        # Get initial active reports.

        self.SetNewGeneratedFiles( self.Feeds_dict )

        # Process feeds . . .

        for name, value in self.Feeds_dict.items():
#            self.log.debug("Feeds_dict item: name=%s" % name)
#            self.log.debug("Feeds_dict item: value=%s" % value)

            # Settings we've already processed.

            if name == TAKE_OUT_THE_TRASH_KEY or                \
               name == OUTPUT_KEY:
                continue

            if name.find( NOAA_NWS_AWC_METARS_KEY ) == 0:
#                self.log.debug( "%s Key found!" % NOAA_NWS_AWC_METARS_KEY )

                get_metars_default      = True
                get_tafs_default        = False
                translate_default       = True
                past_hours_default_str  = "1"

                aviation_weather_center_metar_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( aviation_weather_center_metar_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, NOAA_NWS_AWC_METARS_URL )
                assert ReturnStatus == True

                display_link_to_navcan_awws     = False
                display_link_to_us_awc_briefing = False

                options_dict = aviation_weather_center_metar_dict.get( OPTIONS_KEY )
                if options_dict != None:
                    # Get options.

                    display_link_to_navcan_awws     = self.YesNoTrueFalse( options_dict, DISPLAY_LINK_TO_NAVCAN_AWWS_KEY )
                    display_link_to_us_awc_briefing = self.YesNoTrueFalse( options_dict, DISPLAY_LINK_TO_US_AWC_BRIEFING_KEY )

                for name, value in aviation_weather_center_metar_dict.items():
#                    self.log.debug("item: name=%s" % name)
#                    self.log.debug("item: value=%s" % value)

                    # Settings we've already processed.

                    if name == OUTPUT_KEY or        \
                       name == OPTIONS_KEY:
                        continue

                    if name == STATIONS_KEY:
#                        self.log.debug("Stations Key found!")

                        stations_dict = value

                        # Modify option defaults for subsequent list of Stations.

                        stations_get_metars = self.YesNoTrueFalse( stations_dict, GET_METARS_KEY, get_metars_default )
#                        self.log.debug("stations_get_metars: %s" % stations_get_metars)

                        stations_get_tafs = self.YesNoTrueFalse( stations_dict, GET_TAFS_KEY, get_tafs_default )
#                        self.log.debug("stations_get_tafs: %s" % stations_get_tafs)

                        stations_translate = self.YesNoTrueFalse( stations_dict, TRANSLATE_KEY, translate_default )
#                        self.log.debug("stations_translate: %s" % stations_translate)

                        stations_past_hours_str = stations_dict.get(PAST_HOURS_KEY, past_hours_default_str)
#                        self.log.debug("stations_past_hours: %s" % stations_past_hours_str)

                        # We put the returned METAR/TAF information in a <table>.

                        HTML = "<table class=\"WXFeedsNOAANWSAWCMETARTable\">\n<tbody>\n<tr>\n"
                        self.WriteGeneratedFiles( HTML )

                        if display_link_to_navcan_awws == True:
                            HTML =   "<td>\n"                                                                         \
                                   +     "<a href=\"%s\">Nav Canada</a>\n" % NAVCAN_AWWS_URL                          \
                                   +     "&nbsp;"                                                                     \
                                   +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % NAVCAN_AWWS_URL   \
                                   +     "<br/>\n"                                                                    \
                                   +     "<a href=\"%s\">AWWS</a>\n" % NAVCAN_AWWS_URL                                \
                                   +     "&nbsp;"                                                                     \
                                   +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>" % NAVCAN_AWWS_URL     \
                                   + "</td>\n" 
                            self.WriteGeneratedFiles( HTML )

                        if display_link_to_us_awc_briefing == True:
                            HTML =   "<td>\n"                                                                               \
                                   +     "<a href=\"%s\">US AWC</a>\n" % US_AWC_BRIEFING_URL                                \
                                   +     "&nbsp;"                                                                           \
                                   +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % US_AWC_BRIEFING_URL     \
                                   +     "<br/>\n"                                                                          \
                                   +     "<a href=\"%s\">Briefing</a>\n" % US_AWC_BRIEFING_URL                              \
                                   +     "&nbsp;"                                                                           \
                                   +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>" % US_AWC_BRIEFING_URL       \
                                   + "</td>\n" 
                            self.WriteGeneratedFiles( HTML )

                        for name, value in stations_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            # If we get an option just continue 'cause we got them above.

                            if name == GET_METARS_KEY or    \
                               name == GET_TAFS_KEY or      \
                               name == TRANSLATE_KEY or     \
                               name == PAST_HOURS_KEY:
                                continue

                            # This item should be a Station.

                            station = name
#                            self.log.debug("station: %s" % station)

                            station_dict = value

                            location = station_dict.get( LOCATION_KEY, "" )
#                            self.log.debug( "location: %s" % location )

                            # Get options for this Station.

                            get_metars = self.YesNoTrueFalse( station_dict, GET_METARS_KEY, stations_get_metars )
#                            self.log.debug("get_metars: %s" % get_metars)

                            get_tafs = self.YesNoTrueFalse( station_dict, GET_TAFS_KEY, stations_get_tafs )
#                            self.log.debug("get_tafs: %s" % get_tafs)

                            translate = self.YesNoTrueFalse( station_dict, TRANSLATE_KEY, stations_translate )
#                            self.log.debug("translate: %s" % translate)

                            past_hours_str = station_dict.get(PAST_HOURS_KEY, stations_past_hours_str)
                            past_hours = int( past_hours_str )
#                            self.log.debug("past_hours: %s" % past_hours)

                            station_website_url = station_dict.get( STATION_WEBSITE_URL_KEY, "" )
#                            self.log.debug("station_website_url: %s" % station_website_url)

                            location_website_url = station_dict.get( LOCATION_WEBSITE_URL_KEY, "" )
#                            self.log.debug("location_website_url: %s" % location_website_url)

                            # Get position optionals.
                            #
                            # We don't provide a station to lookup because Google Maps Places doesn't 
                            # find the right place all the time.  For example, it gets 'CYKA' wrong.
                            #
                            # NOAANWSAWCMETAR() can use an airport data base before Places so we 
                            # let it look up station position.
                            #
                            # Position and zoom may still be set for each station though.

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                self.GetPositionOptionals( station_dict )
#                            self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                    ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                          )

                            ReturnedStatus = self.NOAANWSAWCMETAR(
                                                                   station,
                                                                   location,
                                                                   latitude,
                                                                   longitude,
                                                                   PositionSpecified,
                                                                   zoom,
                                                                   get_metars,
                                                                   get_tafs,
                                                                   translate,
                                                                   past_hours,
                                                                   station_website_url,
                                                                   location_website_url
                                                                 )
                            if ReturnedStatus == True:
                               ngen += 1
                            else:
                                self.log.error("NOAANWSAWCMETARs returned False for Station: %s, Location: %s" % (station, location))

                        # End the <table>.

                        self.WriteGeneratedFiles( "</tr>\n</tbody>\n</table>\n<hr>\n" )
                    else:
                        self.log.error("Unexpected Aviation Weather Center METAR Key Returned! Key = %s" % name)    # We received a key for a feed within 
                                                                                                                    # an Aviation Weather Center METAR 
                                                                                                                    # stanza that is unkown to us!
            elif name.find( ENVIRONMENT_CANADA_KEY ) == 0:
#                self.log.debug("Environment Canada Key found!")

                alerts_first_default    = True
#                self.log.debug("EC: alerts_first_default: %s" % alerts_first_default)
                inline_forecast_default = False
#                self.log.debug("EC: inline_forecast_default: %s" % inline_forecast_default)
                get_weather_default     = True
#                self.log.debug("EC: get_weather_default: %s" % get_weather_default)
                get_alerts_default      = True
#                self.log.debug("EC: get_alerts_default: %s" % get_alerts_default)

                environment_canada_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( environment_canada_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, ENVIRONMENT_CANADA_URL )
                assert ReturnStatus == True

                for name, value in environment_canada_dict.items():
#                    self.log.debug("item: name='%s'" % name)
#                    self.log.debug("item: value='%s'" % value)

                    # Settings we've already processed.

                    if name == OUTPUT_KEY:
                        continue

                    if name == CITYS_KEY:
#                        self.log.debug("Citys Key found!")

                        citys_dict = value

                        # Modify option defaults for subsequent list of Citys.

                        alerts_first_default = self.YesNoTrueFalse( citys_dict, ALERTS_FIRST_KEY, alerts_first_default )
#                        self.log.debug("Citys: alerts_first_default: %s" % alerts_first_default)
                        inline_forecast_default = self.YesNoTrueFalse( citys_dict, INLINE_FORECAST_KEY, inline_forecast_default )
#                        self.log.debug("Citys: inline_forecast_default: %s" % inline_forecast_default)
                        get_weather_default = self.YesNoTrueFalse( citys_dict, GET_WEATHER_KEY, get_weather_default )
#                        self.log.debug("Citys: get_weather_default: %s" % get_weather_default)
                        get_alerts_default = self.YesNoTrueFalse( citys_dict, GET_ALERTS_KEY, get_alerts_default )
#                        self.log.debug("Citys: get_alerts_default: %s" % get_alerts_default)

                        for name, value in citys_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            # If we get any options just continue 'cause we got them above.

                            if name == ALERTS_FIRST_KEY or                \
                               name == INLINE_FORECAST_KEY or             \
                               name == GET_WEATHER_KEY or                 \
                               name == GET_ALERTS_KEY:
                                continue

                            # This item should be a City.

                            city = name
#                            self.log.debug("city: %s" % city)

                            city_dict = value

                            feed_url = city_dict.get('feed_url')
#                            self.log.debug("feed_url: %s" % feed_url)

                            # Get options for this City.

                            alerts_first = self.YesNoTrueFalse( city_dict, ALERTS_FIRST_KEY, alerts_first_default )
#                            self.log.debug("City: alerts_first: %s" % alerts_first)
                            inline_forecast = self.YesNoTrueFalse( city_dict, INLINE_FORECAST_KEY, inline_forecast_default )
#                            self.log.debug("City: inline_forecast: %s" % inline_forecast)
                            get_weather = self.YesNoTrueFalse( city_dict, GET_WEATHER_KEY, get_weather_default )
#                            self.log.debug("City: get_weather: %s" % get_weather)
                            get_alerts = self.YesNoTrueFalse( city_dict, GET_ALERTS_KEY, get_alerts_default )
#                            self.log.debug("City: get_alerts: %s" % get_alerts)

                            # Get position optionals.

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                    self.GetPositionOptionals( city_dict, city )
#                            self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                          )

                            ReturnedStatus = False

                            if alerts_first == True:
                                if get_alerts == True:
                                    ReturnedStatus = self.ECCityAlert(
                                                                       city,
                                                                       feed_url,
                                                                       latitude,
                                                                       longitude,
                                                                       PositionSpecified,
                                                                       zoom
                                                                     )
                                    if ReturnedStatus == True:
                                        ngen += 1
                                    else:
                                        self.log.error("ECCityAlert Returned False For City: %s, Feed: %s" % (city, feed_url))

                                if get_weather == True:
                                    ReturnedStatus = self.ECCityWeather(
                                                                         city,
                                                                         feed_url,
                                                                         inline_forecast,
                                                                         latitude,
                                                                         longitude,
                                                                         PositionSpecified,
                                                                         zoom
                                                                      )
                                    if ReturnedStatus == True:
                                        ngen += 1
                                    else:
                                        self.log.error("ECCityWeather Returned False For City: %s, Feed: %s" % (city, feed_url))
                            else:
                                if get_weather == True:
                                    ReturnedStatus = self.ECCityWeather(
                                                                         city,
                                                                         feed_url,
                                                                         inline_forecast,
                                                                         latitude,
                                                                         longitude,
                                                                         PositionSpecified,
                                                                         zoom
                                                                       )
                                    if ReturnedStatus == True:
                                        ngen += 1
                                    else:
                                        self.log.error("ECCityWeather Returned False For City: %s, Feed: %s" % (city, feed_url))

                                if get_alerts == True:
                                    ReturnedStatus = self.ECCityAlert(
                                                                       city,
                                                                       feed_url,
                                                                       latitude,
                                                                       longitude,
                                                                       PositionSpecified,
                                                                       zoom
                                                                     )
                                    if ReturnedStatus == True:
                                        ngen += 1
                                    else:
                                        self.log.error("ECCityAlert Returned False For City: %s, Feed: %s" % (city, feed_url))
                            self.WriteGeneratedFiles( "<hr>\n" )
                    elif name == PUBLIC_ALERTS_KEY:
#                        self.log.debug("Public Alerts Key found!")

                        public_alerts_dict = value

                        for name, value in public_alerts_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            public_alerts_area = name
#                            self.log.debug("public_alerts_area: %s" % public_alerts_area)

                            public_alerts_dict = value

                            feed_url = public_alerts_dict.get('feed_url')
#                            self.log.debug("feed_url: %s" % feed_url)

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                    self.GetPositionOptionals(
                                                               public_alerts_dict,
                                                               public_alerts_area
                                                             )
#                            self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                   ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                          )

                            ReturnedStatus = self.ECPublicAlert(
                                                                 public_alerts_area,
                                                                 feed_url,
                                                                 latitude,
                                                                 longitude,
                                                                 PositionSpecified,
                                                                 zoom
                                                               )
                            if ReturnedStatus == True:
                                ngen += 1
                            else:
                                self.log.error("ECPublicAlert Returned False For: '%s', Feed: %s" % (public_alerts_area, feed_url))

                            self.WriteGeneratedFiles( "<hr>\n" )
                    elif name == MARINE_FORECASTS_KEY:
#                        self.log.debug("Marine Forecasts Key found!")

                        marine_areas_dict = value

                        for name, value in marine_areas_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            marine_area = name
#                            self.log.debug("marine_area: %s" % marine_area)

                            marine_area_dict = value

                            feed_url = marine_area_dict.get('feed_url')
#                            self.log.debug("feed_url: %s" % feed_url)

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =         \
                                    self.GetPositionOptionals(
                                                               marine_area_dict,
                                                               marine_area
                                                             )
#                            self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                          )

                            ReturnedStatus = self.ECMarineForecast(
                                                                    marine_area, 
                                                                    feed_url,
                                                                    latitude,
                                                                    longitude,
                                                                    PositionSpecified,
                                                                    zoom
                                                                  )
                            if ReturnedStatus == True:
                                ngen += 1
                            else:
                                self.log.error( "WXFeeds: ECMarineForecast() Returned False For: '%s', Feed: %s" % ( marine_area, feed_url ) )

                            self.WriteGeneratedFiles( "<hr>\n" )
                    else:
                        # We received a key for a feed within an Environment Canada stanza that is unkown to us!

                        self.log.error( "WXFeeds: Unexpected Environment Canada Key Returned! Key = %s" % name )
            elif name.find( DRIVE_BC_KEY ) == 0:
#                self.log.debug( "'%s' Key found: '%s'" % ( DRIVE_BC_KEY, name ) )

                drive_bc_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( drive_bc_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, DRIVE_BC_URL )
                assert ReturnStatus == True

                latitude                    = BAD_LATITUDE
                longitude                   = BAD_LONGITUDE
                PositionSpecified           = False
                zoom                        = MAP_ZOOM_DEFAULT
                search_criteria             = ""
                SearchLocation              = ""
                Name                        = ""
                Icon                        = ""
                display_links_to_drive_bc   = ""
                drive_bc_district_filter    = ""
                drive_bc_route_filter       = ""
                drive_bc_type_filter        = ""
                drive_bc_severity_filter    = ""
                display_attributes          = False
                display_attribute_lists     = False

                options_dict = drive_bc_dict.get( OPTIONS_KEY )
                if options_dict != None:
                    # Get options.

                    # Get position optionals.

                    latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                            self.GetPositionOptionals( options_dict )
    #                self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
    #                                    ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
    #                              )

                    display_links_to_drive_bc = options_dict.get( DRIVE_BC_DISPLAY_LINKS_TO_DRIVE_BC_KEY, NO )
    #                self.log.debug( "display_links_to_drive_bc = %s" % display_links_to_drive_bc );

                    drive_bc_district_filter = options_dict.get( DRIVE_BC_DISTRICT_FILTER_KEY, "" )
    #                self.log.debug( "drive_bc_district_filter = %s" % drive_bc_district_filter );

                    drive_bc_route_filter = options_dict.get( DRIVE_BC_ROUTE_FILTER_KEY, "" )
    #                self.log.debug( "drive_bc_route_filter = %s" % drive_bc_route_filter );

                    drive_bc_type_filter = options_dict.get( DRIVE_BC_TYPE_FILTER_KEY, "" )
    #                self.log.debug( "drive_bc_type_filter = %s" % drive_bc_type_filter );

                    drive_bc_severity_filter = options_dict.get( DRIVE_BC_SEVERITY_FILTER_KEY, "" )
    #                self.log.debug( "drive_bc_severity_filter = %s" % drive_bc_severity_filter );

                    display_attributes = self.YesNoTrueFalse( options_dict, DISPLAY_ATTRIBUTES_KEY, display_attributes )
#                    self.log.debug( "display_attributes = %s" % display_attributes );

                    display_attribute_lists = self.YesNoTrueFalse( options_dict, DISPLAY_ATTRIBUTE_LISTS_KEY, display_attribute_lists )
#                    self.log.debug( "display_attribute_lists = %s" % display_attribute_lists )

                for name, value in drive_bc_dict.items():
#                    self.log.debug("item: name=%s" % name)
#                    self.log.debug("item: value=%s" % value)

                    if name == OUTPUT_KEY or        \
                       name == OPTIONS_KEY:
                        continue

                    if name == DRIVE_BC_AREAS_KEY:

#                        self.log.debug("Areas Key found!")

                        areas_dict = value

                        # We're going to aggregate areas into a single list.

                        Areas       = ""
                        FeedURLs    = ""

                        for name, value in areas_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            # This item should be an area.

                            area = name
#                            self.log.debug( "area: %s" % area )

                            area_dict = value

                            # Get feed URL from Drive BC dictionary.

                            feed_url = self.DriveBCDictionary.get( name, "" )
#                            self.log.debug("Drive BC dictionary: area = %s, url = %s" % ( area, feed_url ) )

                            if feed_url == "":
                                self.log.warning( "Area: %s NOT in Drive BC Dictionary!" % area )

                            # Allow user to override dictionary entry.
                            #
                            # If the get below fails, the dictionary entry (if any) 
                            # will be used.

                            feed_url = area_dict.get( 'feed_url', feed_url )
#                            self.log.debug("Final feed_url: %s" % feed_url)

                            assert area != ""
                            assert feed_url != ""

                            if Areas != "":
                                Areas += "|"

                            Areas += area

                            if FeedURLs != "":
                                FeedURLs +="|"

                            FeedURLs += feed_url

                        if Areas != "" and FeedURLs != "":
                            ReturnedStatus = self.DriveBCAreas(
                                                                Areas,
                                                                FeedURLs,
                                                                latitude,
                                                                longitude,
                                                                PositionSpecified,
                                                                zoom,
                                                                display_links_to_drive_bc,
                                                                drive_bc_district_filter,
                                                                drive_bc_route_filter,
                                                                drive_bc_type_filter,
                                                                drive_bc_severity_filter,
                                                                display_attributes,
                                                                display_attribute_lists
                                                              )
                            if ReturnedStatus == True:
                               ngen += 1
                            else:
                                self.log.error("DriveBCAreas returned False for Area: %s" % Areas )
                    else:
                        self.log.error( "Unexpected Drive BC Key returned! Key = %s" % name )    # We received a key for a feed within 
                                                                                                 # a Drive BC stanza that is unkown to us!
            elif name.find( NOAA_NWS_FORECASTS_KEY ) == 0:
#                self.log.debug( "%s Key found!" % NOAA_NWS_FORECASTS_KEY )

                noaa_nws_forecast_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( noaa_nws_forecast_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, NOAA_NWS_FORECASTS_URL )
                assert ReturnStatus == True

                locations_units_default                 = UNITS_US
                locations_get_forecast_default          = False
                locations_get_hourly_forecast_default   = False
                locations_hourly_forecast_hours_default = HOURLY_FORECAST_HOURS_DEFAULT

                options_dict = noaa_nws_forecast_dict.get( OPTIONS_KEY )
                if options_dict != None:
                    # Get option defaults for subsequent list of locations.

                    locations_units_default = options_dict.get( UNITS_KEY, locations_units_default )
#                    self.log.debug( "NOAA NWS Forecasts: locations_units_default: %s" % locations_units_default )

                    locations_get_forecast_default = self.YesNoTrueFalse( options_dict, GET_FORECAST_KEY, locations_get_forecast_default )
#                    self.log.debug( "NOAA NWS Forecasts: locations_get_forecast_default: %s" % locations_get_forecast_default)

                    locations_get_hourly_forecast_default = self.YesNoTrueFalse( options_dict, GET_HOURLY_FORECAST_KEY, locations_get_hourly_forecast_default )
#                    self.log.debug( "NOAA NWS Forecasts: locations_get_hourly_forecast_default: %s" % locations_get_hourly_forecast_default)

                    temp = options_dict.get( HOURLY_FORECAST_HOURS_KEY, "" )
                    if temp != "":
                        locations_hourly_forecast_hours_default = int( temp )
#                        self.log.debug( "NOAA NWS Forecasts: locations_hourly_forecast_hours_default: %s" % locations_hourly_forecast_hours_default )

                for name, value in noaa_nws_forecast_dict.items():
#                    self.log.debug("item: name=%s" % name)
#                    self.log.debug("item: value=%s" % value)

                    if name == OUTPUT_KEY or        \
                       name == OPTIONS_KEY:
                        continue

                    if name == LOCATIONS_KEY:
#                       self.log.debug("Locations Key found!")

                        locations_dict = value

                        for name, value in locations_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            # This item should be a location.

                            location = name
#                            self.log.debug("NOAA NWS Forecasts: location: %s" % location)

                            location_dict = value

                             # Get options for this location.

                            latitude            = BAD_LATITUDE
                            longitude           = BAD_LONGITUDE
                            PositionSpecified   = False

                            if location == NOAA_NWS_STATION_LOCATION_KEY:
#                                self.log.debug( "NOAA NWS Forecasts: %s Found!" % NOAA_NWS_STATION_LOCATION_KEY )

                                location            = self.StationLocation
                                latitude            = float( self.StationLatitude )
                                longitude           = float( self.StationLongitude )
                                if latitude != BAD_LATITUDE and     \
                                   longitude != BAD_LONGITUDE:
                                    PositionSpecified = True
                                else:
                                    PositionSpecified = False

                            location_units = location_dict.get( UNITS_KEY, locations_units_default )
#                            self.log.debug( "NOAA NWS Forecasts: location_units: %s" % location_units )

                            location_get_forecast = self.YesNoTrueFalse( location_dict, GET_FORECAST_KEY, locations_get_forecast_default )
#                               self.log.debug( "NOAA NWS Forecasts: location_get_forecast: %s" % location_get_forecast )

                            location_get_hourly_forecast = self.YesNoTrueFalse( location_dict, GET_HOURLY_FORECAST_KEY, locations_get_hourly_forecast_default )
#                            self.log.debug( "NOAA NWS Forecasts: location_get_hourly_forecast: %s" % location_get_hourly_forecast )

                            location_hourly_forecast_hours = locations_hourly_forecast_hours_default
                            temp = location_dict.get( HOURLY_FORECAST_HOURS_KEY, "" )
                            if temp != "":
                                location_hourly_forecast_hours = int( temp )
#                            self.log.debug( "NOAA NWS Forecasts: location_hourly_forecast_hours: %d" % location_hourly_forecast_hours )

                            assert location_hourly_forecast_hours >= 0

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                        self.GetPositionOptionals( location_dict, location )
#                                self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                    ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                              )

                            ReturnedStatus = self.NOAANWSForecast( 
                                                                   location,
                                                                   latitude,
                                                                   longitude,
                                                                   PositionSpecified,
                                                                   zoom,
                                                                   location_units,
                                                                   location_get_forecast,
                                                                   location_get_hourly_forecast,
                                                                   location_hourly_forecast_hours
                                                                 )
                            if ReturnedStatus == True:
                                ngen += 1
                            else:
                                self.log.error( "NOAA NWS Forecasts: NOAANWSForecast() returned False for Location: %s" % location )
                    else:
                        self.log.error( "NOAA NWS Forecasts: Unexpected NOAA NWS Forecast Key Returned! Key = %s" % name )    # We received a key for a feed within 
                                                                                                                              # a NOAA NWS Forecast stanza that is unkown to us!
            elif name.find( NOAA_NWS_SPC_KEY ) == 0:
#                self.log.debug( "%s Key found!" % NOAA_NWS_SPC )

                noaa_nws_spc_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( noaa_nws_spc_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, NOAA_NWS_SPC_URL )
                assert ReturnStatus == True

                display_links_to_all_current_alerts = ""
#                   self.log.debug( "display_links_to_all_current_alerts: %s" % display_links_to_all_current_alerts )

                options_dict = noaa_nws_spc_dict.get( OPTIONS_KEY )
                if options_dict != None:
                    # Get options.

                    display_links_to_all_current_alerts = options_dict.get( DISPLAY_LINKS_TO_ALL_CURRENT_ALERTS )
#                    self.log.debug( "display_links_to_all_current_alerts: %s" % display_links_to_all_current_alerts )

                HTML =   "<p class=\"WXFeedsP1\">\n"                                                                              \
                       +     "<b>\n"                                                                                                         \
                       +         "<a href=\"%s\">All Current US Alerts</a>\n" % "https://alerts.weather.gov/cap/us.php?x=1"                  \
                       +         "&nbsp;\n"                                                                                                  \
                       +         "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % "https://alerts.weather.gov/cap/us.php?x=1"      \
                       +         "&nbsp; - &nbsp;\n"                                                                                         \
                       +         "<a href=\"%s\">All Current US Marine Zone Alerts</a>\n" % "https://alerts.weather.gov/cap/mzus.php?x=1"    \
                       +         "&nbsp;\n"                                                                                                  \
                       +         "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % "https://alerts.weather.gov/cap/mzus.php?x=1"    \
                       +     "</b>\n"                                                                                                        \
                       + "</p>\n"                                                                                                            \
                        + "<hr>\n"

                for name, value in noaa_nws_spc_dict.items():
#                    self.log.debug( "item: name=%s" % name )
#                    self.log.debug( "item: value=%s" % value )

                    # If we get any options just continue 'cause we got them above.

                    if name == OUTPUT_KEY or        \
                       name == OPTIONS_KEY:
                        continue

                    if name == NOAA_NWS_SPC_ALERTS_KEY:
#                        self.log.debug("Alerts Key found!")

                        alerts_dict = value

                        if display_links_to_all_current_alerts == BEFORE or    \
                           display_links_to_all_current_alerts == BOTH:

                            self.WriteGeneratedFiles( HTML )

                        for name, value in alerts_dict.items():
#                            self.log.debug( "item: name=%s" % name )
#                            self.log.debug( "item: value=%s" % value )

                            # This item should be a location.

                            location = name
#                            self.log.debug( "location: %s" % location )

                            location_dict = value

                            feed_url = location_dict.get( "feed_url" )
#                                self.log.debug( "feed_url: %s" % feed_url )

                            latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                    self.GetPositionOptionals(
                                                               location_dict,
                                                               location
                                                             )
#                            self.log.debug( "WXFeeds: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                               ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                          )

                            ReturnedStatus = self.NOAANWSSPC(
                                                              location,
                                                              feed_url,
                                                              latitude,
                                                              longitude,
                                                              PositionSpecified,
                                                              zoom
                                                            )
                            if ReturnedStatus == True:
                                ngen += 1
                            else:
                                self.log.error( "NOAANWSSPC Returned False For Location: %s, Feed URL: %s" % ( location, feed_url ) )

                        if display_links_to_all_current_alerts == AFTER or    \
                           display_links_to_all_current_alerts == BOTH:

                            self.WriteGeneratedFiles( HTML )
                    else:
                        self.log.error("Unexpected NOAA NWS SPC Key Returned! Key = %s" % name)    # We received a key within a NOAA National Weather Service Storm 
            elif name.find( ALBERTA_511_KEY ) == 0:
#                self.log.debug( "'%s' Key found: '%s'" % ( ALBERTA_511_KEY, name ) )

                alberta_511_dict = value

                # Get active reports for this feed, and subsequent feeds if they're not changed.

                self.SetNewGeneratedFiles( alberta_511_dict )

                ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                assert ReturnStatus == True

                ReturnedStatus = self.WriteFeedHeaders( name, ALBERTA_511_URL )
                assert ReturnStatus == True

                for name, value in alberta_511_dict.items():
#                    self.log.debug("item: name=%s" % name)
#                    self.log.debug("item: value=%s" % value)

                    if name == OUTPUT_KEY:
                        continue

                    if name == ALBERTA_511_RESOURCES_KEY:
#                        self.log.debug( "Resources Key found!" )

                        resources_dict = value

                        for name, value in resources_dict.items():
#                            self.log.debug("item: name=%s" % name)
#                            self.log.debug("item: value=%s" % value)

                            # This item should be a resource.

                            resource = name
#                            self.log.debug( "Alberta 511 Resource: %s" % resource )

                            resource_dict = value

                            # Get feed URL from Alberta 511 dictionary.

                            feed_url = self.Alberta511Dictionary.get( resource, "" )
#                            self.log.debug("Alberta 511 Dictionary: resource = %s, url = %s" % ( resource, feed_url ) )

                            if feed_url == "":
                                self.log.warning( "Resource: %s NOT in ALberta 511 Dictionary!" % resource )

                                # Allow user to override dictionary entry.
                                #
                                # If the get below fails, the dictionary entry (if any) will be used.

                                feed_url = resource_dict.get( 'feed_url', feed_url )
#                                self.log.debug("Final feed_url: %s" % feed_url)

                            ValidResource   = True
                            ReturnedStatus  = True

                            if resource == ALBERTA_511_ROAD_CONDITIONS_KEY:
                                area_name_filter = resource_dict.get( AREA_NAME_FILTER, "" )
#                                self.log.debug( "area_name_filter: %s" % area_name_filter )
                                roadway_name_filter = resource_dict.get( ROADWAY_NAME_FILTER, "" )
#                                self.log.debug( "roadway_name_filter: %s" % roadway_name_filter )
                                display_attribute_lists = resource_dict.get( DISPLAY_ATTRIBUTE_LISTS_KEY, NO )
#                                self.log.debug( "display_attribute_lists = %s" % display_attribute_lists )

                                latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                        self.GetPositionOptionals( resource_dict, "" )

                                ReturnedStatus = self.Alberta511RoadConditions(
                                                                                feed_url,
                                                                                area_name_filter,
                                                                                roadway_name_filter,
                                                                                zoom,
                                                                                display_attribute_lists
                                                                               )
                            elif resource == ALBERTA_511_ALERTS_KEY:

                                ReturnedStatus = self.Alberta511Alerts( feed_url )
                            elif resource == ALBERTA_511_EVENTS_KEY:
                                event_type_filter = resource_dict.get( EVENT_TYPE_FILTER, "" )
                                roadway_name_filter = resource_dict.get( ROADWAY_NAME_FILTER, "" )

                                ReturnedStatus = self.Alberta511Events( feed_url, event_type_filter, roadway_name_filter )
                            elif resource == ALBERTA_511_FERRIES_KEY:

                                ReturnedStatus = self.Alberta511Ferries( feed_url )
                            elif resource == ALBERTA_511_PARKS_KEY:
                                park_name_filter_condition = resource_dict.get( PARK_NAME_FILTER_CONDITION, EQUAL )
                                park_name_filter = resource_dict.get( PARK_NAME_FILTER, "" )
                                    
                                ReturnedStatus = self.Alberta511Parks( feed_url, park_name_filter_condition, park_name_filter )
                            elif resource == ALBERTA_511_CAMERAS_KEY:
                                display_disabled_cameras = self.YesNoTrueFalse( resource_dict, DISPLAY_DISABLED_CAMERAS, False )
                                roadway_name_filter = resource_dict.get( ROADWAY_NAME_FILTER, "" )

                                ReturnedStatus = self.Alberta511Cameras( feed_url, display_disabled_cameras, roadway_name_filter )
                            elif resource == ALBERTA_511_GROUPED_CAMERAS_KEY:
                                display_disabled_cameras = self.YesNoTrueFalse( resource_dict, DISPLAY_DISABLED_CAMERAS, False )
                                roadway_name_filter = resource_dict.get( ROADWAY_NAME_FILTER, "" )

                                ReturnedStatus = self.Alberta511GroupedCameras( feed_url, display_disabled_cameras, roadway_name_filter )
                            else:
                                self.log.error( "Unexpected Alberta511 Resource: '%s'" % resource )
                                ValidResource = False

                            if ValidResource == True:
                                if ReturnedStatus == True:
                                    ngen += 1
                                else:
                                    self.log.error("An Alberta511 Resource() returned False for Resource: %s" % resource )
                    else:
                        self.log.error( "Unexpected Alberta 511 Key returned! Key = %s" % name )    # We received a key for a feed within 
                                                                                                    # a Alberta 511 stanza that is unkown to us!
            elif    name.find( WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_KEY ) == 0:
#                    self.log.debug( "%s Key found!" % WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_KEY )

                    ws_dot_dict = value

                    # Get active reports for this feed, and subsequent feeds if they're not changed.

                    self.SetNewGeneratedFiles( ws_dot_dict )

                    ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                    assert ReturnStatus == True

                    ReturnedStatus = self.WriteFeedHeaders( name, WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_URL )
                    assert ReturnStatus == True

                    for name, value in ws_dot_dict.items():
#                        self.log.debug("item: name=%s" % name)
#                        self.log.debug("item: value=%s" % value)

                        if name == OUTPUT_KEY or                \
                           name == WS_DOT_ACCESS_CODE_KEY:
                            continue

                        if name == WS_DOT_TRAVELER_INFORMATION_KEY:
#                            self.log.debug("WS-DOT Traveler Information Key found!")

                            traveler_info_dict = value

                            for name, value in traveler_info_dict.items():
#                                self.log.debug("item: name=%s" % name)
#                                self.log.debug("item: value=%s" % value)

                                # This item should be a traveler information subject.

                                traveler_info_subject = name
#                                self.log.debug( "WA-DOT Traveler Information Subject: %s" % traveler_info_subject )

                                traveler_info_subject_dict = value

                                ReturnedStatus = False

                                if traveler_info_subject == WS_DOT_BORDER_CROSSINGS_KEY:

                                    # Get position optionals.

                                    latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon =     \
                                    self.GetPositionOptionals(
                                                               traveler_info_subject_dict
                                                             )
#                                    self.log.debug( "WXFeeds: WS DOT Border Crossings: Position Optionals: Lat: %f, Lon: %f, Pos: %s, Z: %d, SC: %s, SL: %s, N: %s, I: %s" %     \
#                                                    ( latitude, longitude, PositionSpecified, zoom, search_criteria, SearchLocation, Name, Icon )
#                                                  )

                                    ReturnedStatus = self.WSDOTBorderCrossings( zoom )

                                if ReturnedStatus == True:
                                   ngen += 1
                                else:
                                    self.log.error("A WS-DOT Traveler Information() returned False for: %s" % traveler_info_subject )
                        else:
                            self.log.error( "Unexpected Washington State - DOT Key returned! Key = %s" % name )    # We received a key for a feed within 
                                                                                                                   # a Washington State - DOT stanza that 
                                                                                                                   # is unkown to us!
            elif    name.find( NOAA_NWS_NHC_KEY ) == 0:
#                    self.log.debug( "%s Key found!" % NOAA_NWS_NHC )

                    noaa_nhc_dict = value

                    # Get active reports for this feed, and subsequent feeds if they're not changed.

                    self.SetNewGeneratedFiles( noaa_nhc_dict )

                    ReturnStatus = self.OpenGeneratedFiles( name, "", "" )
                    assert ReturnStatus == True

                    ReturnedStatus = self.WriteFeedHeaders( name, NOAA_NWS_NHC_URL )
                    assert ReturnStatus == True

                    for name, value in noaa_nhc_dict.items():
#                        self.log.debug( "item: name=%s" % name )
#                        self.log.debug( "item: value=%s" % value )

                        if name == OUTPUT_KEY:
                            continue

                        if name == BASINS_KEY:
#                            self.log.debug("%s Key found!" % BASINS_KEY )

                            basins_dict = value

                            for name, value in basins_dict.items():
#                                self.log.debug( "item: name=%s" % name )
#                                self.log.debug( "item: value=%s" % value )

                                # This item should be a basin.

                                basin = name
#                                self.log.debug( "basin: %s" % basin )

                                basin_dict = value

                                feed_url = basin_dict.get( "feed_url" )
#                                self.log.debug( "feed_url: %s" % feed_url )

                                ReturnedStatus = self.NOAANWSNHCBasin(
                                                                         basin
                                                                       , feed_url
                                                                     )
                                if ReturnedStatus == True:
                                   ngen += 1
                                else:
                                    self.log.error( "NOAANHCBasin Returned False For Basin: %s, Feed URL: %s" % ( basin, feed_url ) )
                        else:
                            self.log.error("Unexpected NOAA NWS SPC Key Returned! Key = %s" % name)    # We received a key within a NOAA 
                                                                                                       # National Hurricane Center that is 
                                                                                                       # unkown to us!

            else:
                self.log.error("Unexpected Key Returned! Key = %s" % name)    # We received a key for a feed that is unkown to us, or not 
                                                                              # implemented yet!

        # Close any open generated files by specifying no new generated files and no header information.

        self.SetNewGeneratedFiles()
        self.OpenGeneratedFiles()

        for r in self.ReportSettings:
            assert r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] == None

        t2 = time.time()

        self.log.info("WXFeeds: %d report(s) were created in %.2f seconds" % ( ngen, t2 - t1 ) )

        if self.TakeOutTheTrash == True:
            # Force garbage collection.

            assert gc.isenabled() == True

            if DEBUG == True:
                self.log.debug( "WXFeeds.new_archive_record(): Exiting . . ." )
                self.log.debug( "WXFeeds.new_archive_record(): Process ID: %d" % os.getpid() )
                self.log.debug( "WXFeeds.new_archive_record(): Memory use before gc.collect():" )
                x, y, z = gc.get_count()
                self.log.debug( "WXFeeds.new_archive_record(): gc.get_count(): %d, %d, %d" % ( x, y, z ) )
                self.log.debug( "WXFeeds.new_archive_record(): sys.getallocatedblocks(): %d" % sys.getallocatedblocks() )

            gc.collect()

            if DEBUG == True:
                self.log.debug( "WXFeeds.new_archive_record(): Memory use after gc.collect():" )
                x, y, z = gc.get_count()
                self.log.debug( "WXFeeds.new_archive_record(): gc.get_count(): %d, %d, %d" % ( x, y, z ) )
                self.log.debug( "WXFeeds.new_archive_record(): sys.getallocatedblocks(): %d" % sys.getallocatedblocks() )
                self.log.debug( "WXFeeds.new_archive_record()" )

        return

    def WriteFeedHeaders(
                          self,
                          FeedSource        : str,
                          FeedWebpageURL    : str
                       ) -> bool:
                       #    ReturnStatus from WriteGeneratedFiles()

        d        = datetime.datetime.utcnow()
        d        = d.replace(tzinfo=datetime.timezone.utc) # Convert it to an aware datetime object in UTC time.
        utc_dt   = d
        d        = d.astimezone()                          # Convert local timezone (still aware).
        local_dt = d

        HTML =   "<hr>\n"                                 \
               + "<p class=\"WXFeedsP0\">\n"

        if FeedWebpageURL == "":
            HTML += "<b>%s</b>\n" % FeedSource
        else:
            HTML +=   "<a href=\"%s\">" % FeedWebpageURL          \
                    +     "<b>%s</b>" % FeedSource                \
                    + "</a>\n"                                    \
                    + "&nbsp;\n"                                  \
                    + "<a href=\"%s\">" % FeedWebpageURL          \
                    +     "<span class=\"icon-play\"></span>"     \
                    + "</a>\n"

        HTML +=   "<br/>\n"                                                                                \
                +     "Report Generated: "                                                                 \
                +     local_dt.strftime("<b>%H:%M %Z</b> %a %d %b %Y / ")                                  \
                +     utc_dt.strftime("<b>%H:%M %Z</b> %a %d %b %Y")                                       \
                +     "&nbsp;\n"                                                                           \
                +     "<a href=\"javascript:history.go(0)\">Refresh Page</a>\n"                            \
                +     "&nbsp;\n"                                                                           \
                +     "<a href=\"javascript:history.go(0)\"><span class=\"icon-refresh\"></span></a>\n"    \
                + "</p>\n"                                                                                 \
                + "<hr>\n"

        return self.WriteGeneratedFiles( HTML )

    def WriteGeneratedFiles(
                            self,
                            Text : str
                           ) -> bool:
                           #    ReturnStatus

        ReturnStatus = True    # Assume success.

        for r in self.ReportSettings:
#            self.log.debug( "WriteGeneratedFiles: File: %s" % r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] )

            if r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] != None:
                # We have an open file so write to it.

                try:
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( Text )
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].flush()
                    os.fsync( r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].fileno() )
                except:
                    self.log.error( "Write To Generated File Failed!" )
                    ReturnStatus = False

        return ReturnStatus

    def WriteGeneratedFilesRelativeLink(
                                         self,
                                         FileName : str
                                       ) -> bool:
                                       #    ReturnStatus

        # FileName should be an HTML file.
        #
        # We develop a relative link based on ReportSetting.

        ReturnStatus = True    # Assume success.

        # A report's HTML files root is where the WeeWx skin will write its HTML files, 
        # most notably index.html.
        #
        # WXFeeds write it's HTML file to self.WXFeedsHTMLRoot so FileName must be there.

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "WriteGeneratedFilesRelativeLink: WXFeedsHTMLRoot: '%s'" % self.WXFeedsHTMLRoot )

        WXFeedsHTMLRoot_Path = Path( self.WXFeedsHTMLRoot )
        assert WXFeedsHTMLRoot_Path != None
        if DEBUG == True:
            self.log.debug( "WriteGeneratedFilesRelativeLink: WXFeedsHTMLRoot: Num Parents: %d" % len( WXFeedsHTMLRoot_Path.parents ) )
            self.log.debug( "WriteGeneratedFilesRelativeLink: WXFeedsHTMLRoot: name: '%s'" % WXFeedsHTMLRoot_Path.name )

        for r in self.ReportSettings:
#            self.log.debug( "WriteGeneratedFilesRelativeLink: r: %s" % r )

            if r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] != None:
                # We have an open file so write to it.

                if DEBUG == True:
                    self.log.debug( "WriteGeneratedFilesRelativeLink: %s: '%s'" % ( HTML_FILES_ROOT_KEY, r[ SETTINGS_KEY ][ HTML_FILES_ROOT_KEY ] ) )

                ReportPath = r[ SETTINGS_KEY ][ HTML_FILES_ROOT_KEY ]
                if DEBUG == True:
                    self.log.debug( "WriteGeneratedFilesRelativeLink: ReportPath: '%s'" % ReportPath )

                p = Path( ReportPath )
                if DEBUG == True:
                    self.log.debug( "WriteGeneratedFilesRelativeLink: ReportPath: Num Parents: %d" % len( p.parents ) )

                # DeltaNumParents is the difference in number of parents of ReportPath and WXFeedsHTMLRoot.
                # It + 1 is the number of directory level above ReportPath to WXFeedsHTMLRoot.

                DeltaNumParents = len( p.parents ) - len( WXFeedsHTMLRoot_Path.parents ) + 1

                RelativeLink = ""
                
                for i in range( DeltaNumParents ):
                    RelativeLink += "../"

                RelativeLink += str( Path( WXFeedsHTMLRoot_Path.name ).joinpath( FileName ) )
                if DEBUG == True:
                    self.log.debug( "WriteGeneratedFilesRelativeLink: RelativeLink: '%s'" % RelativeLink )

                try:
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( RelativeLink )
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].flush()
                    os.fsync( r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].fileno() )
                except:
                    self.log.error( "Write To Generated File Failed!" )
                    ReturnStatus = False

        return ReturnStatus

    def GetDivTag(
                    self
                   , FeedSource : str
                   , FeedType : str
                   , Location : str
                 ):

        FeedID = ""

        if FeedSource != "":
            FeedID = "-" + FeedSource

        if FeedType != "":
            FeedID += "-" + FeedType

        if Location != "":
            FeedID += "-" + Location

        FeedID = FeedID.replace(" ", "")

        HTML = "<div class=\"WXFeeds%s\">\n" % FeedID

        return HTML

    def SetNewGeneratedFiles(
                              self,
                              Dictionary = None
                            ) -> bool:
                            #    ReturnStatus

        ReturnStatus = True     # Assume success.

        #DEBUG = True
        DEBUG = False

        # Initialize new generated file settings in ReportSettings.

        if Dictionary != None:
            Output_dict = Dictionary.get( OUTPUT_KEY )

            if Output_dict != None:
                if DEBUG == True:
                    self.log.debug( "SetNewGeneratedFiles: Output: %s" % Output_dict )

                for name, value in Output_dict.items():
                    Report      = name
                    Report_dict = value
                    if DEBUG == True:
                        self.log.debug( "SetNewGeneratedFiles: Report: '%s', Report_dict: '%s'" % ( Report, Report_dict ) )

                    if Report_dict != None:

                        NewGeneratedFileName        = Report_dict.get( FILE_KEY )
                        NewGeneratedFileSequence    = Report_dict.get( FILE_SEQUENCE_KEY )

                        if NewGeneratedFileName != None and         \
                           NewGeneratedFileSequence != None:

                            if DEBUG == True:
                                self.log.debug( "SetNewGeneratedFiles: NewGeneratedFileName: '%s', NewGeneratedFileSequence: '%s'" % ( NewGeneratedFileName, NewGeneratedFileSequence ) )

                        for r in self.ReportSettings:
                            if DEBUG == True:
                                self.log.debug( "SetNewGeneratedFiles: ReportSettings.name: '%s'" % r[ NAME_KEY ] )

                            if r[ NAME_KEY ] == Report:
                                if DEBUG == True:
                                    self.log.debug( "SetNewGeneratedFiles: Report Found!" )

                                r[ SETTINGS_KEY ][ NEW_GENERATED_FILENAME_KEY ]         = NewGeneratedFileName
                                r[ SETTINGS_KEY ][ NEW_GENERATED_FILE_SEQUENCE_KEY ]    = NewGeneratedFileSequence

                                p = Path( r[ SETTINGS_KEY ][ NEW_GENERATED_FILENAME_KEY ] )
                                r[ SETTINGS_KEY ][ NEW_GENERATED_FILE_TYPE_KEY ] = p.suffix[ 1: ]

                                break
                        else:
                            self.log.error( "WXFeeds: file *and* file_sequence must both be specified for 'Output'! file: '%s', file_sequence: '%s'" % ( NewFile, NewFileSequence ) )
                            ReturnStatus = False
        else:
            # 'Dictionary' == None - set all new generated file settings to "".

            for r in self.ReportSettings:
                r[ SETTINGS_KEY ][ NEW_GENERATED_FILENAME_KEY ]         = ""
                r[ SETTINGS_KEY ][ NEW_GENERATED_FILE_SEQUENCE_KEY ]    = ""

        return ReturnStatus

    def OpenGeneratedFiles(
                            self,
                            FeedSource  = "",
                            FeedType    = "",
                            Location    = "",
                        ) -> bool:
                        #    ReturnStatus

        # Open generated files.
        #
        # If a file is currently open, close it.

        ReturnStatus = True     # Assume success.

        #DEBUG = True
        DEBUG = False

        # Close open files if we're going to change to a different generated file.

        for r in self.ReportSettings:

            if r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILENAME_KEY ] != r[ SETTINGS_KEY ][ NEW_GENERATED_FILENAME_KEY ]:
                # Need to change the file we will be writing to.

                if DEBUG == True:
                    self.log.debug( "OpenGeneratedFiles: Changing Generated File . . ." )

                if r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] != None:
                    # A file is currently open so we have to close any open tags and close the file.

                    # Close the <div> tag.

                    if DEBUG == True:
                        self.log.debug( "OpenGeneratedFiles: Writing </div> tag." )

                    HTML = "</div>\n"
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )

                    if r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_TYPE_KEY ] == "html":
                        # Close the <body> and <html> tags.

                        if DEBUG == True:
                            self.log.debug( "OpenGeneratedFiles: Writing </body></html> tags." )

                        HTML = "</body>\n"
                        r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )

                    if DEBUG == True:
                        self.log.debug( "OpenGeneratedFiles: Closing File." )

                    # Close the file.
                    
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].close()
                    r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] = None

                # Move NEW to CURRENT.

                if DEBUG == True:
                    self.log.debug( "OpenGeneratedFiles: Moving NEW to CURRENT." )

                r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILENAME_KEY ]         = r[ SETTINGS_KEY ][ NEW_GENERATED_FILENAME_KEY ]
                r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_SEQUENCE_KEY ]    = r[ SETTINGS_KEY ][ NEW_GENERATED_FILE_SEQUENCE_KEY ]
                r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_TYPE_KEY ]        = r[ SETTINGS_KEY ][ NEW_GENERATED_FILE_TYPE_KEY ]

        # Open files for reports without an open file.

        for r in self.ReportSettings:
            if r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] == None:
                # A file is not open - we might have to open one.

                if r[ SETTINGS_KEY][ CURRENT_GENERATED_FILENAME_KEY ] != "":
                    # A file to generate is specified - open it.

                    if DEBUG == True:
                        self.log.debug( "OpenGeneratedFiles: Opening New File . . ." )

                    # Cheetah Memory Leak Workaround.
                    #
                    # The problem is that if include files are regenerated (are dynamic) 
                    # WeeWX memory usage will increase every archive period.
                    #
                    # This is a memory leak in the Cheetah Generator.
                    #
                    # The workaround is to create an include file that is static and only 
                    # contains a Cheetah directive to include the dynamic file.

                    FileName = ""
            
                    if r[ SETTINGS_KEY ][ CHEETAH_MEMORY_LEAK_WORKAROUND_KEY ] == True:
                        # Create the static file if it does not exist.

                        FileName = "WXFeeds-" + r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILENAME_KEY ] # Dynamic file.  This is the file 
                                                                                                    # HTML will be written to.

                        # Make the relative static filename absolute.

                        StaticFileName  = r[ SETTINGS_KEY ][ GENERATED_FILES_ROOT_KEY ] + "/" + r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILENAME_KEY ]

                        p = Path( StaticFileName )
                        assert p != None

                        if p.exists() == False:
                            # Create the static file.

                            try:
                                StaticFile = open( StaticFileName, "wt", 1 )
                                if DEBUG == True:
                                    self.log.debug( "OpenGeneratedFiles: Static File: '%s' Opened OK" % StaticFileName )

                                StaticFile.write( "#include raw \"%s\"" % FileName )
                                StaticFile.close()
                            except:
                                self.log.error( "OpenGeneratedFiles: Static Include File: '%s' Open Failed!" % StaticFileName )
                                ReturnStatus = False
                    else:
                        FileName = r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILENAME_KEY ]

                    # Create the dynamic file.

                    # Make relative filename absolute.

                    FileName = r[ SETTINGS_KEY ][ GENERATED_FILES_ROOT_KEY ] + "/" + FileName

                    FileMode = "w"         # For "only" or "first" include files, we write a new file.

                    if r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_SEQUENCE_KEY ] == "subsequent" or    \
                       r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_SEQUENCE_KEY ] == "last":
                        # For "subsequent" or "last" include files, we append to an existing file.

                        FileMode = "a"

                    FileMode += "t"

                    if DEBUG == True:
                        self.log.debug("OpenGeneratedFiles: FileName = %s, FileMode = %s" % ( FileName, FileMode ) )

                    try:
                        r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] = open( FileName, FileMode, 1 )
                        if DEBUG == True:
                            self.log.debug( "OpenGeneratedFiles: After open(): File: %s" % r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] )
                    except:
                        self.log.error("Open Include File Failed!")
                        ReturnStatus = False

                        # Continue for all ReportSettings even if we get an error.

                    if r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ] != None:
                        # File is open.

                        if DEBUG == True:
                            self.log.debug( "OpenGeneratedFiles: r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_TYPE_KEY ]: %s" % r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_TYPE_KEY ] )

                        if r[ SETTINGS_KEY ][ CURRENT_GENERATED_FILE_TYPE_KEY ] == "html":
                            # Write HTML preamble.

                            CSSFileName = Path( r[ SETTINGS_KEY ][ CUSTOM_CSS_PATH_KEY ] ).name
                            CSSVersion  = datetime.datetime.now().timestamp()
                            if DEBUG == True:
                                self.log.debug( "OpenGeneratedFiles: CSSFileName: %s, CSSFileVersion: %d" % ( CSSFileName, CSSVersion ) )

                            HTML =   "<!DOCTYPE html>\n"                                                                    \
                                   + "<html lang=\"en\">\n"                                                                 \
                                   + "  <head>\n"                                                                           \
                                   + "    <meta charset=\"utf-8\">\n"                                                       \
                                   + "    <title>\n"                                                                        \
                                   + "      %s - %s - %s\n" % ( FeedSource, FeedType, Location )                            \
                                   + "    </title>\n"                                                                       \
                                   + "    <link rel='stylesheet' href='./%s?ver=%d' type='text/css' media='all' />" %       \
                                                ( CSSFileName, CSSVersion )                                                 \
                                   + "  </head>\n"                                                                          \
                                   + "  <body>\n"
                            r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )

                        # Add 'back to top button' if required.

                        if r[ SETTINGS_KEY ][ ADD_BACK_TO_TOP_BUTTON_KEY ] == True and    \
                           r[ SETTINGS_KEY ][ BACK_TO_TOP_BUTTON_ADDED_KEY ] == False:

                            # Add "back to top" button.

                            HTML =   "<button onclick=\"topFunction()\" id=\"BackToTopButton\" title=\"Go to Top\">Top</button>\n"    \
                                   + "<script>\n"                                                                                     \
                                   + "    //Get the button.\n"                                                                        \
                                   + "\n"                                                                                             \
                                   + "    var mybutton = document.getElementById( \"BackToTopButton\" );\n"                           \
                                   + "\n"                                                                                             \
                                   + "    // When the user scrolls down 20px from the top of the document, show the button.\n"        \
                                   + "\n"                                                                                             \
                                   + "    window.onscroll = function()\n"                                                             \
                                   + "    {\n"                                                                                        \
                                   + "        scrollFunction()\n"                                                                     \
                                   + "    };\n"                                                                                       \
                                   + "\n"                                                                                             \
                                   + "    function scrollFunction()\n"                                                                \
                                   + "    {\n"                                                                                        \
                                   + "        if ( document.body.scrollTop > 20 || document.documentElement.scrollTop > 20 )\n"       \
                                   + "        {\n"                                                                                    \
                                   + "            mybutton.style.display = \"block\";\n"                                              \
                                   + "        } else {\n"                                                                             \
                                   + "                   mybutton.style.display = \"none\";\n"                                        \
                                   + "               }\n"                                                                             \
                                   + "    }\n"                                                                                        \
                                   + "\n"                                                                                             \
                                   + "    // When the user clicks on the button, scroll to the top of the document.\n"                \
                                   + "\n"                                                                                             \
                                   + "    function topFunction()\n"                                                                   \
                                   + "    {\n"                                                                                        \
                                   + "        document.body.scrollTop = 0;              // For Safari\n"                              \
                                   + "        document.documentElement.scrollTop = 0;   // For Chrome, Firefox, IE and Opera\n"       \
                                   + "    }\n"                                                                                        \
                                   + "</script>\n"
                            r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )

                            r[ SETTINGS_KEY ][ BACK_TO_TOP_BUTTON_ADDED_KEY ] = True

                        # Write a <div> tag.

                        HTML = self.GetDivTag( FeedSource, FeedType, Location )
                        r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )
            else:
                # We're not changing to a new file so just close the <div> and open a new one.

                HTML = "</div>\n"
                HTML += self.GetDivTag( FeedSource, FeedType, Location )
                r[ SETTINGS_KEY ][ GENERATED_FILE_KEY ].write( HTML )

        return ReturnStatus

    def GetSkinSetting(
                        self,
                        SkinName   : str,
                        Setting    : str
                      ) -> str:

        assert self.Config_dict != None
        assert self.StdReport_dict != None

        SettingValue = ""

        # Find the skin in [StdReport].

        if SkinName in self.StdReport_dict:
            # Skin name exists.

            Skin_dict = self.StdReport_dict.get( SkinName )
            assert Skin_dict != None

            if Setting in Skin_dict:
                # Setting exists - use it.

                SettingValue = Skin_dict.get( Setting )
            else:
                # Setting does not exist - try skin's parent, which is [StdReport].

                if Setting in self.StdReport_dict:
                    # Setting exists in [StdReport] - use it.

                    SettingValue = self.StdReport_dict.get( Setting )
                else:
                    # Setting does not exist - try [StdParent]'s parent, which is the top level configuration.

                    if Setting in self.Config_dict:
                        # Setting exists - use it.

                        SettingValue = self.Config_dict.get( Setting )
        else:
            self.log.error( "WXFeeds: GetSkinSetting: Skin '%s' NOT in [StdReport]!" )

#        self.log.debug( "Setting: %s = %s" % ( Setting, SettingValue ) )

        return SettingValue

    def YesNoTrueFalse(
                        self,
                        Dictionary    : dict,
                        Key           : str,
                        Default = False
                      ) -> bool:

        ReturnValue = Default

        assert Dictionary != None

        temp = Dictionary.get( Key )

        if temp != None:
            temp = temp.upper()

            if temp == "YES" or \
               temp == "TRUE":
                ReturnValue = True
            else:
                ReturnValue = False

        return ReturnValue

    def CreateCustomCSS(
                         self,
                         Skin                       : str,
                         CustomCSSPath              : str,
                         DisableWXFeedsCustomCSS    : bool
                       ) -> bool:
                       #    ReturnStatus

        # If a custom CSS file does not exist, create it with our 
        # custom CSS.
        #
        # If a custom CSS files already exists, open it and search 
        # contents for our class name prefix.
        #
        # If the prefix exists, update it with our custom CSS.
        #
        # If the prefix does not exist, append our custom CSS to 
        # the existing file.

        ReturnStatus = True     # Assume success.

        #DEBUG = True    # Enable/Disable debug logging.
        DEBUG = False

        # For really detailed debugging . . .

        #DEBUGDEBUG = True    # Enable/Disable debug logging.
        DEBUGDEBUG = False

        IconBase = self.WXFeedsIconBase

        if DEBUG == True:
            self.log.debug( "CreateCustomCSS: ****** Start of DEBUG ******" )
            self.log.debug( "CreateCustomCSS: Skin: %s" % Skin )
            self.log.debug( "CreateCustomCSS: CustomCSSPath: %s" % CustomCSSPath )
            self.log.debug( "CreateCustomCSS: DisableWXFeedsCustomCSS: %s" % DisableWXFeedsCustomCSS )
            self.log.debug( "CreateCustomCSS: IconBase: %s" % IconBase )

        p = Path( CustomCSSPath )
        FileName = p.name
        if DEBUG == True:
            self.log.debug( "CreateCustomCSS: FileName: '%s'" % FileName )

        CustomCSSCachePath  = self.GetCachePath( FileName, Skin )
 
        if DEBUG == True:
            self.log.debug( "CreateCustomCSS: CustomCSSCachePath: %s" % CustomCSSCachePath )

        UpdateCustomCSS = False

        if self.IsFileStale( CustomCSSPath, CustomCSSCachePath ) == True:
            if DEBUG == True:
                self.log.debug( "CreateCustomCSS: %s will be updated" % CustomCSSPath )

            OurCustomCSSPrefix = "WXFeeds"

            START_OF_OUR_CUSTOM_CSS_KEY = "/* ****** Start of CSS for WXFeeds Extension ****** */\n"
            END_OF_OUR_CUSTOM_CSS_KEY   = "/* ****** End of CSS for WXFeeds Extension ****** */\n"

            OurCustomCSS = START_OF_OUR_CUSTOM_CSS_KEY

            if DisableWXFeedsCustomCSS == False:
                OurCustomCSS +=   ".%sP0 { margin-left: 0; }\n" % OurCustomCSSPrefix                                                    \
                                + ".%sP1 { margin-left: 1em; }\n" % OurCustomCSSPrefix                                                  \
                                + ".%sP2 { margin-left: 2em; }\n" % OurCustomCSSPrefix                                                  \
                                + ".%sP3 { margin-left: 3em; }\n" % OurCustomCSSPrefix                                                  \
                                + "\n"                                                                                                  \
                                + ".%sT0 { margin-left: 0; }\n" % OurCustomCSSPrefix                                                    \
                                + ".%sT1 { margin-left: 1em; }\n" % OurCustomCSSPrefix                                                  \
                                + ".%sT2 { margin-left: 2em; }\n" % OurCustomCSSPrefix                                                  \
                                + ".%sT3 { margin-left: 3em; }\n" % OurCustomCSSPrefix                                                  \
                                + "\n"                                                                                                  \
                                + ".%sNoWatchesOrWarnings { background-color: lime; color: black; }\n" % OurCustomCSSPrefix             \
                                + ".%sWatch { background-color: yellow; color: black; }\n" % OurCustomCSSPrefix                         \
                                + ".%sWarning { background-color: red; color: white; }\n" % OurCustomCSSPrefix                          \
                                + ".%sSpecialAirQualityStatement { background-color: grey; color: white; }\n" % OurCustomCSSPrefix      \
                                + ".%sSpecialWeatherStatement { background-color: grey; color: white; }\n" % OurCustomCSSPrefix         \
                                + ".%sUnknownAlert { background-color: blue; color: black; }\n" % OurCustomCSSPrefix                    \
                                + ".%sNoActiveAlerts { background-color: lime; color: black; }\n" % OurCustomCSSPrefix                  \
                                + ".%sMinor { background-color: yellow; color: black; }\n" % OurCustomCSSPrefix                         \
                                + ".%sModerate { background-color: yellow; color: red; }\n" % OurCustomCSSPrefix                        \
                                + ".%sSevere { background-color: orange; color: black; }\n" % OurCustomCSSPrefix                        \
                                + ".%sExtreme { background-color: red; color: white; }\n" % OurCustomCSSPrefix                          \
                                + ".%sDBCSeverityNormal { color: black; }\n" % OurCustomCSSPrefix                                       \
                                + ".%sDBCSeverityMajor { color: red; }\n" % OurCustomCSSPrefix                                          \
                                + "\n"                                                                                                  \
                                + "#BackToTopButton {\n"                                                                                \
                                + "                     display: none;\n"                                                               \
                                + "                     position: fixed;\n"                                                             \
                                + "                     bottom: 20px;\n"                                                                \
                                + "                     right: 30px;\n"                                                                 \
                                + "                     z-index: 99;\n"                                                                 \
                                + "                     font-size: 18px;\n"                                                             \
                                + "                     border: none;\n"                                                                \
                                + "                     outline: none;\n"                                                               \
                                + "                     background-color: red;\n"                                                       \
                                + "                     color: white;\n"                                                                \
                                + "                     cursor: pointer;\n"                                                             \
                                + "                     padding: 15px;\n"                                                               \
                                + "                     border-radius: 4px;\n"                                                          \
                                + "                 }\n"                                                                                \
                                + "\n"                                                                                                  \
                                + "#BackToTopButton:hover {\n"                                                                          \
                                + "                           background-color: #555;\n"                                                \
                                + "                       }\n"                                                                          \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Warning Sign Icon\n"                                                                              \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_road__sign__alert__danger__2992209.png\n"                              \
                                + "  Resized Icon:    iconfinder_road__sign__alert__danger__2992209-20.png\n"                           \
                                + "  Simplified Icon: WXFeeds-warning.png\n"                                                            \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-warning-sign    {\n"                                                                           \
                                + "                          background-image: url('" + IconBase + "/WXFeeds-warning-sign.png');\n"     \
                                + "                          height: 20px;\n"                                                           \
                                + "                          width: 20px;\n"                                                            \
                                + "                          position: relative;\n"                                                     \
                                + "                          top: 4px;\n"                                                               \
                                + "                          left: 0;\n"                                                                \
                                + "                          display: inline-block;\n"                                                  \
                                + "                      }\n"                                                                           \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Play Icon\n"                                                                                      \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_13_3049253-play.png\n"                                                 \
                                + "  Resized Icon:    iconfinder_13_3049253-play-20.png\n"                                              \
                                + "  Simplified Icon: WXFeeds-play.png\n"                                                               \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-play    {\n"                                                                                   \
                                + "                  background-image: url('" + IconBase + "/WXFeeds-play.png');\n"                     \
                                + "                  height: 20px;\n"                                                                   \
                                + "                  width: 20px;\n"                                                                    \
                                + "                  position: relative;\n"                                                             \
                                + "                  top: 4px;\n"                                                                       \
                                + "                  left: 0;\n"                                                                        \
                                + "                  display: inline-block;\n"                                                          \
                                + "              }\n"                                                                                   \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Google Maps Icon\n"                                                                               \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_google_4975305-map.png\n"                                              \
                                + "  Resized Icon:    iconfinder_google_4975305-map-20.png\n"                                           \
                                + "  Simplified Icon: WXFeeds-google-map.png\n"                                                         \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-google-map  {\n"                                                                               \
                                + "                      background-image: url('" + IconBase + "/WXFeeds-google-map.png');\n"           \
                                + "                      height: 20px;\n"                                                               \
                                + "                      width: 20px;\n"                                                                \
                                + "                      position: relative;\n"                                                         \
                                + "                      top: 4px;\n"                                                                   \
                                + "                      left: 0;\n"                                                                    \
                                + "                      display: inline-block;\n"                                                      \
                                + "                  }\n"                                                                               \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Map Icon\n"                                                                                       \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_map_285662.png\n"                                                      \
                                + "  Resized Icon:    iconfinder_map_285662-20.png\n"                                                   \
                                + "  Simplified Icon: WXFeeds-map.png\n"                                                                \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-map {\n"                                                                                       \
                                + "              background-image: url('" + IconBase + "/WXFeeds-map.png');\n"                          \
                                + "              height: 20px;\n"                                                                       \
                                + "              width: 20px;\n"                                                                        \
                                + "              position: relative;\n"                                                                 \
                                + "              top: 4px;\n"                                                                           \
                                + "              left: 0;\n"                                                                            \
                                + "              display: inline-block;\n"                                                              \
                                + "          }\n"                                                                                       \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Refresh Icon\n"                                                                                   \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_refresh_48523-24.png\n"                                                \
                                + "  Resized Icon:    iconfinder_refresh_48523-24.png\n"                                                \
                                + "  Simplified Icon: WXFeeds-refresh.png\n"                                                            \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-refresh {\n"                                                                                   \
                                + "                  background-image: url('" + IconBase + "/WXFeeds-refresh.png');\n"                  \
                                + "                  height: 24px;\n"                                                                   \
                                + "                  width: 24px;\n"                                                                    \
                                + "                  position: relative;\n"                                                             \
                                + "                  top: 4px;\n"                                                                       \
                                + "                  left: 0;\n"                                                                        \
                                + "                  display: inline-block;\n"                                                          \
                                + "              }\n"                                                                                   \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   Checklist Icon\n"                                                                                 \
                                + "\n"                                                                                                  \
                                + "  Base Icon:       iconfinder_Artboard_1_1790658.png\n"                                              \
                                + "  Resized Icon:    iconfinder_Artboard_1_1790658-20.png\n"                                           \
                                + "  Simplified Icon: WXFeeds-checklist.png\n"                                                          \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-checklist   {\n"                                                                               \
                                + "                      background-image: url('" + IconBase + "/WXFeeds-checklist.png');\n"            \
                                + "                      height: 20px;\n"                                                               \
                                + "                      width: 20px;\n"                                                                \
                                + "                      position: relative;\n"                                                         \
                                + "                      top: 4px;\n"                                                                   \
                                + "                      left: 0;\n"                                                                    \
                                + "                      display: inline-block;\n"                                                      \
                                + "                  }\n"                                                                               \
                                + "\n"                                                                                                  \
                                + "/*\n"                                                                                                \
                                + "   NOAA NWS Forecast Icon\n"                                                                         \
                                + "\n"                                                                                                  \
                                + "  Style icons referenced in the NOAA NWS Forecast feed.\n"                                           \
                                + "*/\n"                                                                                                \
                                + "\n"                                                                                                  \
                                + ".icon-noaa-nws-forecast   {\n"                                                                       \
                                + "                             width: 86px;\n"                                                         \
                                + "                             height: 86px;\n"                                                        \
                                + "                             position: relative;\n"                                                  \
                                + "                             top: 50%;\n"                                                            \
                                + "                             left: 0;\n"                                                             \
                                + "                             display: inline-block;\n"                                               \
                                + "                          }\n"                                                                       \
                                + "\n"                                                                                                  \
                                + ".icon-noaa-nws-hourly-forecast    {\n"                                                               \
                                + "                                      width: 56px;\n"                                                \
                                + "                                      height: 56px;\n"                                               \
                                + "                                      position: relative;\n"                                         \
                                + "                                      top: 50%;\n"                                                   \
                                + "                                      left: 0;\n"                                                    \
                                + "                                      display: inline-block;\n"                                      \
                                + "                                  }\n"
            else:
                OurCustomCSS += "/* ****** WXFeeds Custom CSS Disabled ****** */\n"

            OurCustomCSS += END_OF_OUR_CUSTOM_CSS_KEY

            CustomCSSFile = None

            try:
                CustomCSSFile = open( CustomCSSPath, "rt" )

                if DEBUG == True:
                    self.log.debug( "Open 1: Custom CSS File: %s opened OK." % CustomCSSPath )
            except IOError as e:
                # This is not really an error as the file may simply have never existed.
                # We'll create it below.

                #self.log.error( "open 1: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
                pass

            if CustomCSSFile == None:
                # Custom CSS file does not exists - create it.

                if DEBUG == True:
                    self.log.debug( "Custom CSS File: %s does not exist." % CustomCSSPath )

                try:
                    if DEBUG == True:
                        self.log.debug( "Creating Custom CSS File: %s" % CustomCSSPath )

                    CustomCSSFile = open( CustomCSSPath, "wt" )

                    if DEBUG == True:
                        self.log.debug( "Open 2: Custom CSS File: %s opened OK." % CustomCSSPath )
                except IOError as e:
                    self.log.error( "Custom CSS File: %s was not created!" % CustomCSSPath )
                    self.log.error( "Open 2: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )

                assert CustomCSSFile != None

                try:
                    CustomCSSFile.write( OurCustomCSS )

                    if DEBUG == True:
                        self.log.debug( "Write 1: Write to Custom CSS File: %s OK." % CustomCSSPath )
                except IOError as e:
                    self.log.error( "Write 1: Writing OurCustomCSS to new custom CSS file: %s failed!" % CustomCSSPath )
                    self.log.error( "Write 1: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
 
                CustomCSSFile.close()
                CustomCSSFile = None
            else:
                # Custom CSS file is open and ready to read.

                assert CustomCSSFile != None

                CurrentCustomCSS = {}

                try:
                    CurrentCustomCSS = CustomCSSFile.readlines()

                    if DEBUG == True:
                        self.log.debug( "readlines returned: %d lines." % len( CurrentCustomCSS ) )
                except io.UnsupportedOperation as e:
                    self.log.error( "readlines returned 'io.UnsupportedOperation' error." )

                CustomCSSFile.close()
                CustomCSSFile = None

                # Open new custom CSS file.

                CustomCSSFile = open( CustomCSSPath, "wt" )
                assert CustomCSSFile != None

                if len( CurrentCustomCSS ) == 0:
                    if DEBUG == True:
                        self.log.debug( "No current Custom CSS - going to write our Custom CSS. . ." )

                    try:
                        if DEBUG == True:
                            self.log.debug( "   Inserting our Custom CSS: %s" % OurCustomCSS )
                        assert CustomCSSFile != None
                        CustomCSSFile.write( OurCustomCSS )
                        CustomCSSFile.close()
                        CustomCSSFile = None
                    except IOError as e:
                        self.log.error( "Write 3: Inserting OurCustomCSS in new custom CSS file: %s failed!" % CustomCSSFileName )
                        self.log.error( "Write 3: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
                else:
                    # There is an existing custom CSS file so process it.

                    if DEBUG == True:
                        self.log.debug( "Processing current Custom CSS . . ." )

                    OurCustomCSSFound = False

                    ReadingOurCustomCSS = False

                    for Line in CurrentCustomCSS:
                        if DEBUG == True:
                            self.log.debug( "Current Custom CSS Line: %s" % Line )

                        if Line == START_OF_OUR_CUSTOM_CSS_KEY:
                            if DEBUG == True:
                                self.log.debug( "   Start of our custom CSS." )

                            OurCustomCSSFound = True

                            ReadingOurCustomCSS = True

                            continue

                        if Line == END_OF_OUR_CUSTOM_CSS_KEY:
                            if DEBUG == True:
                                self.log.debug( "   End of our Custom CSS." )

                            ReadingOurCustomCSS = False

                            # Insert our custom CSS.

                            try:
                                if DEBUGDEBUG == True:
                                    self.log.debug( "   Inserting our Custom CSS: %s" % OurCustomCSS )
                                assert CustomCSSFile != None
                                CustomCSSFile.write( OurCustomCSS )
                            except IOError as e:
                                self.log.error( "Write 2: Inserting OurCustomCSS in new custom CSS file: %s failed!" % CustomCSSPath )
                                self.log.error( "Write 2: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )

                            continue

                        if ReadingOurCustomCSS == False:
                            # This line is not in our custom CSS so write it.

                            if DEBUGDEBUG == True:#
#                                self.log.debug( "      Writing other Custom CSS Line: %s" % Line )
                                self.log.debug( "      Writing other Custom CSS Line" )

                            try:
                                assert CustomCSSFile != None
                                CustomCSSFile.write( Line )
                            except IOError as e:
                                self.log.error( "Write 3: Writing other Custom CSS to new custom CSS file: %s failed!" % CustomFileName )
                                self.log.error( "Write 3: Failed! e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
                        else:
                            if DEBUGDEBUG == True:
#                                self.log.debug( "      NOT writing current Custom CSS Line: %s" % Line )
                                self.log.debug( "      NOT writing current Custom CSS Line." )

                    if OurCustomCSSFound == False:
                        # We have an existing custom CSS file without out custom CSS so
                        # write our custom CSS.

                        assert CustomCSSFile != None
                        CustomCSSFile.write( OurCustomCSS )

                    assert CustomCSSFile != None
                    CustomCSSFile.close()
                    CustomCSSFile = None

                    if CustomCSSCachePath != "":       # We have a cache path.
                        # Write a custom CSS cache file.

                        CustomCSSCacheFile = None

                        try:
                            CustomCSSCacheFile = open( CustomCSSCachePath, "wt" )

                            if DEBUG == True:
                                self.log.debug( "Custom CSS Cache File: %s opened OK." % CustomCSSCachePath )

                            CustomCSSCacheFile.close()
                            CustomCSSCacheFile = None
                        except IOError as e:
                            self.log.error( "CreateCustomCSS: File open for custom CSS cache file: '%s' failed!" % CustomCSSCachePath )
                            self.log.error( "CreateCustomCSS: e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
        else:
            if DEBUG == True:
                self.log.debug( "CreateCustomCSS: custom CSS will NOT be updated." )

        if DEBUG == True:
            self.log.debug( "CreateCustomCSS: ****** End of DEBUG ******" )

        return ReturnStatus

    def GetCachePath(
                      self,
                      FileName  : str,
                      Skin = ""
                    ) -> str:

        assert FileName != ""
        assert self.WXFeedsLocalCacheRoot != ""

        CachePath = self.WXFeedsLocalCacheRoot

        if Skin != "":
            #CachePath += "/" + Skin
            CachePath = str( Path( CachePath ).joinpath( Skin ) )

        #CachePath += "/" + FileName + "-cache"
        CachePath = str( Path( CachePath ).joinpath( FileName + "-cache" ) )

        return CachePath

    def IsFileStale(
                     self,
                     FilePath               : str,
                     CacheFilePath          : str,
                     HasAccessToken = False
                   ) -> bool:
                   #    IsStale

        # Many files can be considered 'static' - once they are 
        # created, they don't change until WeeWX is restarted.
        #
        # Examples include:
        # - custom.css
        # - 'simple' Google Maps maps:
        #   - both the js (containing markers) and html files are static
        # - more 'complex' Google Maps maps:
        #   - the js (containing markers) may be dynamic but the html file is static.
        #
        # But!
        #
        # If the WXFeeds server is being used, files such as Google Maps map 
        # HTML files may contain an access token which is volatile - it will expire 
        # after some time set by the WXFeeds authentication server.
        #
        # This code determines if a file is stale by comparing its modification 
        # date with the modification dates for weewx.conf and WXFeeds.py.
        #
        # A file is stale if:
        # - it doesn't exist,
        # - its modification date precedes weewx.conf's modification date,
        #   - changes to weewx.conf could cause a file to be created or changed.
        # - its modification date precedes WXFeeds.py's modification date or
        #   - changes to WXFeeds.py could cause a file to be created or changed.
        # - its modification date precedes its saved (cached) modification date.
        #   - the file may have been changed manually (without changes to weewx.conf or WXFeeds.
        #     - custom.css is an example of a file than might be changed manually.
        # - it has an Access Token and the token has expired.

        # NOTE: A 'cache file' in this context is only used to "save" a files 
        # modification date so that can tell if it was changed outside WXFeeds.
        # It does NOT contain cached data!

        IsStale = False    # Assume FilePath is NOT stale.

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "IsFileStale: FilePath: '%s'" % FilePath )
            self.log.debug( "IsFileStale: CacheFilePath: '%s'" % CacheFilePath )

        Now_dt                  = datetime.datetime.now()

        p = Path( FilePath )
        assert p != None

        if p.exists() == False:
            # File does not exist - it's stale.

            if DEBUG == True:
                self.log.debug( "IsFileStale: File: %s Does NOT Exist - it is stale." % FilePath )

            IsStale = True
        else:
            assert p.exists() == True

            # File exists.

            if DEBUG == True:
                self.log.debug( "IsFileStale: File: %s Exists." % FilePath )

            # Get its modification time.

            FilePath_mtime = p.stat().st_mtime

            if DEBUG == True:
                self.log.debug( "IsFileStale:              File Modification Time: %s  %s" % (
                                                                                               datetime.datetime.fromtimestamp(FilePath_mtime),
                                                                                               FilePath
                                                                                             )
                                  )
                self.log.debug( "IsFileStale: WeeWX Config File Modification Time: %s" % datetime.datetime.fromtimestamp(self.WeeWXConfig_mtime) )
                self.log.debug( "IsFileStale:        WXFeeds.py Modification Time: %s" % datetime.datetime.fromtimestamp(self.WXFeedsPy_mtime) )

            if self.WeeWXConfig_mtime > FilePath_mtime:
                # The WeeWX confguration file is more recent than FilePath - FilePath is stale.

                if DEBUG == True:
                    self.log.debug( "IsFileStale: The WeeWX configuration file is more recent than %s - it is stale." % FilePath )

                IsStale = True
            else:
                if self.WXFeedsPy_mtime > FilePath_mtime:
                    # The WXFeeds extension is more recent than FilePath - FilePath is stale.

                    if DEBUG == True:
                        self.log.debug( "IsFileStale: The WXFeeds extension is more recent than %s - it is stale." % FilePath )

                    IsStale = True

            if IsStale == False:
                # FilePath is not stale so far.

                # Check to see if FilePath has been modified since it was last updated.

                c = Path( CacheFilePath )
                assert c != None

                if c.exists() == True:
                    # Get the cache file's modification time.

                    Cache_mtime = c.stat().st_mtime

                    if DEBUG == True:
                        self.log.debug(
                                        "%s st_mtime:  %s" % (
                                                               CacheFilePath,
                                                               datetime.datetime.fromtimestamp(Cache_mtime)
                                                             )
                                      )

                    if FilePath_mtime > Cache_mtime:
                        # The current file is more recent than the cache file - need to update cache file.

                        if DEBUG == True:
                            self.log.debug( "The current file is more recent than it's cache file - it will be updated." )

                        IsStale = True

            if IsStale == False:
                # FilePath is not stale so far.

                if HasAccessToken == True:
                    # Check to see if the Access Token has expired.
                    # We allow a margin equal to 2 X the station archive period.
                    # This is to avoid having an access token expire in the 
                    # period just before the next cycle and having to wait 
                    # another cycle for files with a valid token to be created.

                    AccessTokenExpiry_dt    = datetime.datetime.fromtimestamp( FilePath_mtime )                                         \
                                                + datetime.timedelta( seconds = ( self.WXFeedsServerAccessTokenExpiry_seconds      \
                                                - ( 2 * self.ArchiveInterval_seconds ) ) )

                    if DEBUG == True:
                        self.log.debug( "IsFileStale: Now_dt:               %s" % Now_dt )
                        self.log.debug( "IsFileStale: AccessTokenExpiry_dt: %s" % AccessTokenExpiry_dt )

                    if AccessTokenExpiry_dt < Now_dt:
                        # Access token is about to expire - FilePath is stale.

                        if DEBUG == True:
                            self.log.debug( "The AccessToken is about to expire or has expired - the file is stale." )

                        IsStale = True

#                if IsStale == False:
#                    # FilePath is not stale so far.
#
#                    # General purpose / catch all condition - file has not been modified 
#                    # in X minutes,so it's stale.
#
#                    Jitter = random.randint( -20, +20 )
#                    Jitter = float ( float( Jitter ) / 100.0 )
#                    if DEBUG == True:
#                        self.log.debug( "IsFileStale: Jitter = %f" % Jitter )
#
#                    Expiry_minutes = int( ( 1.0 + Jitter ) * self.FileExpiry_minutes )
#                    if DEBUG == True:
#                        self.log.debug( "IsFileStale: Expiry_minutes = %d" % Expiry_minutes )
#
#                    Expiry_dt   = datetime.datetime.fromtimestamp( FilePath_mtime ) + datetime.timedelta( minutes = Expiry_minutes )
#
#                    if DEBUG == True:
#                        self.log.debug( "IsFileStale: Now_dt:    %s" % Now_dt )
#                        self.log.debug( "IsFileStale: Expiry_dt: %s" % Expiry_dt )
#
#                    if Expiry_dt < Now_dt:
#                        # FilePath is stale.
#
#                        if DEBUG == True:
#                            self.log.debug( "The file is more than %d minutes old - it is stale." % self.FileExpiry_minutes )
#
#                        IsStale = True

        if DEBUG == True:
            if IsStale == True:
                self.log.debug( "IsFileStale: %s is stale." % FilePath )
            else:
                self.log.debug( "IsFileStale: %s is NOT stale." % FilePath )

        return IsStale

    def GetPositionOptionals(
                              self,
                              Dictionary     : dict,
                              Location = ""
                            ) -> ( float,    float,     bool,              int,  str,            str,            str,  str ):
                            #      Latitude, Longitude, PositionSpecified, Zoom, SearchCriteria, SearchLocation, Name, Icon

        # Method to get optional position controls and lookup latitude and longitude if neccesary.
        #
        # Dictionary is required.
        # Location is optional.
        #
        # We attempt to get:
        # - Latitude
        # - Longitude
        # - Zoom and
        # - SearchCriteria
        #
        # If latitude and longitude are not specified, we will 
        # attempt to get them using Google Maps Places.
        #
        # If SearchCriteria is specified we will use it, 
        # otherwise we will use Location (if it's specified).

        assert Dictionary != None

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "GetPositionOptionals: %s" % Dictionary )
            self.log.debug( "GetPositionOptionals: Location: '%s'" % Location )

        Latitude            = BAD_LATITUDE
        Longitude           = BAD_LONGITUDE
        PositionSpecified   = False
        Zoom                = MAP_ZOOM_DEFAULT
        SearchCriteria      = ""
        SearchLocation      = ""
        Name                = ""
        Icon                = ""

        Latitude_str = Dictionary.get( LATITUDE_KEY )
        if Latitude_str != None:
            Latitude = float( Latitude_str )

        Longitude_str = Dictionary.get( LONGITUDE_KEY )
        if Longitude_str != None:
            Longitude = float( Longitude_str )

        if Latitude != BAD_LATITUDE and Longitude != BAD_LONGITUDE:
            PositionSpecified = True

        Zoom_str = Dictionary.get( ZOOM_KEY )
        if Zoom_str != None:
            Zoom = int( Zoom_str )

        SearchCriteria = Dictionary.get( SEARCH_CRITERIA_KEY, "" )

        if DEBUG == True:
            self.log.debug(
                            "GetPositionOptionals: Lat: %f, Lon: %f, Z: %d, SC: %s" % (
                                                                                        Latitude,
                                                                                        Longitude,
                                                                                        Zoom,
                                                                                        SearchCriteria
                                                                                      )
                          )

        if PositionSpecified == False:
            SearchLocation = Location   # Default to Location.

            if SearchCriteria != "":
                # Use search criteria to look up position.

               SearchLocation = SearchCriteria
            
            if SearchLocation != "":
                # Look up position using Google Maps Places.

                ReturnedStatus      = False
                ReturnedName        = "* Not Found *"
                ReturnedLatitude    = BAD_LATITUDE
                ReturnedLongitude   = BAD_LONGITUDE
                ReturnedIcon        = ""

                try:
                    if DEBUG == True:
                        self.log.debug( "GetPositionOptionals: Calling GoogleMapsPlaces(): SearchLocation: %s" % SearchLocation )

                    ReturnedStatus, ReturnedName, ReturnedLatitude, ReturnedLongitude, ReturnedIcon = self.GoogleMapsPlaces( SearchLocation )
                    if DEBUG == True:
                        self.log.debug( "GetPositionOptionals: Returned from GoogleMapsPlaces()." )
                except:
                    self.log.error( "GetPositionOptionals: GoogleMapsPlaces() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

                    ReturnedStatus = False

                if ReturnedStatus == True:
                    # Use the returned information.

                    if DEBUG == True:
                        self.log.debug( "GetPositionOptionals: GoogleMapsPlaces() returned: %s, %s, %f, %f, %s for %s." %   \
                                            (
                                              ReturnedStatus,
                                              ReturnedName,
                                              ReturnedLatitude,
                                              ReturnedLongitude,
                                              ReturnedIcon,
                                              SearchLocation
                                            )
                                      )

                    Latitude            = ReturnedLatitude
                    Longitude           = ReturnedLongitude
                    PositionSpecified   = True
                    Name                = ReturnedName
                    Icon                = ReturnedIcon

        return Latitude, Longitude, PositionSpecified, Zoom, SearchCriteria, SearchLocation, Name, Icon

#    def CopyHTMLFile(
#                      self,
#                      SourceFilePath : str,
#                      DeleteSource = False
#                    ) -> bool:
#                    #    ReturnStatus
#
#        ReturnStatus = True     # Assume success.
#
#        self.log.debug( "CopyHTMLFile: self.ReportSettings: %s" % self.ReportSettings )
#
#        for r in self.ReportSettings:
#            if r[ SETTINGS_KEY ][ SELECTED_KEY ] == True:
#                # Selected report.
#
#                p = Path( SourceFilePath )
#                assert p != None
#                assert p.exists() == True
#
#                FileName = p.name
#                self.log.debug( "CopyHTMLFile: FileName: '%s'" % FileName )
#
#                DestinationFilePath = r[ SETTINGS_KEY ][ HTML_FILES_ROOT_KEY ] + "/WXFeeds/" + FileName
#                self.log.debug( "CopyHTMLFile: DestinationFilePath: '%s'" % DestinationFilePath )
#
#                p = Path( DestinationFilePath )
#                assert p != None
#
#                if p.exists() == True: os.remove( DestinationFilePath )
#
#                shutil.copy2( SourceFilePath, DestinationFilePath )
#
#        if DeleteSource == True:
#            os.remove( SourceFilePath )
#
#        return ReturnStatus

    def NOAANWSAWCMETAR(
                         self,
                         StationIDs             : str,    # ICAO and/or IATA Code for airport or weather station to generate METAR for.
                         Location               : str,
                         Latitude               : float,
                         Longitude              : float,
                         PositionSpecified      : bool,
                         Zoom                   : int,
                         GetMETARs              : bool,
                         GetTAFs                : bool,
                         Translate              : bool,
                         PastHours              : int,
                         StationWebsiteURL      : str,
                         LocationWebsiteURL     : str
                       ) -> bool:
                       #    ReturnStatus

        # Get METAR(s) and TAF(s).

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        # StationIDs can include an ICAO code (4 characters) and an IATA 
        # code (3 characters) separated by a "-".
        #
        # The first code will be displayed on the webpage.
        #
        # An IATA code will be used to get information from Air-Port-Codes.
        #
        # If Air-Port-Codes can not provide information, the ICAO and IATA 
        # codes will be used to get information from the Google Maps Places 
        # API with the ICAO code (in the belief the longer ICAO code 
        # is more specific then the shorter IATA code).

        StationID     = ""
        StationIDIATA = ""
        StationIDICAO = ""

        SplitStationsIDs = StationIDs.split( "-" )
        assert len( SplitStationsIDs ) > 0 and len( SplitStationsIDs ) <= 2

        StationID = SplitStationsIDs[ 0 ]

        for Station in SplitStationsIDs:
            if len( Station ) == 3:
                StationIDIATA = Station
                continue

            if len( Station ) == 4:
                StationIDICAO = Station

        if StationIDICAO == "":
            self.log.error( "NOAANWSAWCMETAR: ICAO Station Code has not been specified - it is requrired to get METARs/TAFs!" )

        if DEBUG == True:
            self.log.debug( "NOAANWSAWCMETAR: StationID: %s, StationIDIATA: %s, StationIDICAO: %s" % ( StationID, StationIDIATA, StationIDICAO ) )

        # If Location or position (Latitude and Longitude) are not specified, try to 
        # get the data from Air-Port-Codes or Google Maps.

        if ( Location == "" or                  \
             PositionSpecified == False or      \
             LocationWebsiteURL == "" ) and     \
           StationIDIATA != "" and              \
           self.AirPortCodesUse == True:

            ReturnedStatus      = False
            ReturnedName        = "! Not Found !"
            ReturnedLatitude    = BAD_LATITUDE
            ReturnedLongitude   = BAD_LONGITUDE
            ReturnedElevation   = BAD_ELEVATION
            ReturnedWebsite     = ""
            ReturnedIcon        = ""

            try:
                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: Calling AirPortCodesGetAirportInfo(): StationIDIATA: %s" % StationIDIATA )

                ReturnedStatus,         \
                ReturnedName,           \
                ReturnedLatitude,       \
                ReturnedLongitude,      \
                ReturnedElevation,      \
                ReturnedWebsite = self.AirPortCodesGetAirportInfo( StationIDIATA )
                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: Returned from AirPortCodesGetAirportInfo()." )
            except:
                self.log.error( "NOAANWSAWCMETAR: AirPortCodesGetAirportInfo() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

                ReturnedStatus = False

            if ReturnedStatus == True:
                # Use the returned information.

                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: AirPortCodesGetAirportInfo() returned: %s, %s, %f, %f, %f, %s for %s" %  \
                                    (
                                      ReturnedStatus,
                                      ReturnedName,
                                      ReturnedLatitude,
                                      ReturnedLongitude,
                                      ReturnedElevation,
                                      ReturnedWebsite,
                                      StationIDIATA
                                    )
                                  )

                if Location == "":
                    Location = ReturnedName

                if PositionSpecified == False:
                    Latitude            = ReturnedLatitude
                    Longitude           = ReturnedLongitude
                    PositionSpecified   = True

                if LocationWebsiteURL == "":
                    LocationWebsiteURL = ReturnedWebsite

        if ( Location == "" or                      \
             PositionSpecified == False ) and       \
           StationIDICAO != "" and                  \
           self.GoogleMapsPlacesUse == True:

            ReturnedStatus      = False
            ReturnedName        = "* Not Found *"
            ReturnedLatitude    = BAD_LATITUDE
            ReturnedLongitude   = BAD_LONGITUDE
            ReturnedElevation   = BAD_ELEVATION
            ReturnedWebsite     = ""
            ReturnedIcon        = ""

            try:
                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: Calling GoogleMapsPlaceSearch(): StationIDICAO: %s" % StationIDICAO )

                ReturnedStatus,         \
                ReturnedName,           \
                ReturnedLatitude,       \
                ReturnedLongitude,      \
                ReturnedIcon = self.GoogleMapsPlaces( StationIDICAO )

                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: Returned from GoogleMapsPlaces()" )
            except:
                self.log.error( "NOAANWSAWCMETAR: GoogleMapsPlaces() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

                ReturnedStatus = False

            if ReturnedStatus == True:
                # Use the returned information.

                if DEBUG == True:
                    self.log.debug( "NOAANWSAWCMETAR: GoogleMapsPlaces() returned: %s, %s, %f, %f, %s for %s." %   \
                                    (
                                      ReturnedStatus,
                                      ReturnedName,
                                      ReturnedLatitude,
                                      ReturnedLongitude,
                                      ReturnedIcon,
                                      StationIDICAO
                                    )
                                  )

                if Location == "":
                    Location = ReturnedName

                if PositionSpecified == False:
                    Latitude            = ReturnedLatitude
                    Longitude           = ReturnedLongitude
                    PositionSpecified   = True

        if Location == "":
            Location = StationID + " - ** Not Found **"

        # Convert True/False input for GetMETARs and GetTAFs to "on" / "off" required by call to AWC.

        CheckMETARs = OFF
        if GetMETARs == True:
            CheckMETARs = ON
#        self.log.debug("GetMETARs = %s, CheckMETARs = %s" % (GetMETARs, CheckMETARs))

        CheckTAFs = OFF
        if GetTAFs == True:
            CheckTAFs = ON
#        self.log.debug("GetTAFs = %s, CheckTAFs = %s" % (GetTAFs, CheckTAFs))

        # Convert True/False input for Translate to "translate" / "standard" required by call to AWC.

        StandardTranslated = STANDARD
        if Translate == True:
            StandardTranslated = TRANSLATED
#        self.log.debug("Translate = %s, StandardTranslated = %s" % (Translate, StandardTranslated))

        METARsURL = NOAA_NWS_AWC_METARS_SERVICE_URL
        METARsURL += "?station_ids=%s&chk_metars=%s&chk_tafs=%s&std_trans=%s&hoursStr=%s" % ( StationID, CheckMETARs, CheckTAFs, StandardTranslated, PastHours )

        if DEBUG == True:
            self.log.debug( "NOAANWSAWCMETAR: METARsURL: '%s'" % METARsURL )

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        METARData       = {}

        try:
            ReturnedStatus, CacheHit, StatusCode, METARData = self.GetEndpointData(
                                                                                    EndpointURL                 = METARsURL,
                                                                                    ContentType                 = CONTENT_HTML,
                                                                                    ValidateDataFunction        = self.NOAANWSAWCMETARValidateData,
                                                                                    Service                     = "NOAANWSAWCMETAR",
                                                                                    CacheKey                    = StationID,
                                                                                    LocalCacheUse               = self.NOAANWSAWCMETARsLocalCacheUse,
                                                                                    LocalCacheExpiry_minutes    = self.NOAANWSAWCMETARsLocalCacheExpiry_minutes,
                                                                                    LocalCacheForceMiss         = self.NOAANWSAWCMETARsLocalCacheForceMiss,
                                                                                    ServerUse              = self.NOAANWSAWCMETARsServerUse,
                                                                                    RemoteCacheUse              = self.NOAANWSAWCMETARsLocalCacheUse,
                                                                                    RemoteCacheExpiry_minutes   = self.NOAANWSAWCMETARsLocalCacheExpiry_minutes,
                                                                                    RemoteCacheForceMiss        = self.NOAANWSAWCMETARsLocalCacheForceMiss
                                                                                  )
        except:
            self.log.error( "NOAANWSAWCMETAR: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 
 
        if DEBUG == True:
            self.log.debug( "NOAANWSAWCMETAR: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "NOAANWSAWCMETAR: GetEndpointData() returned False!" )

            ReturnStatus = False
            return ReturnStatus
 
        if StatusCode != 200:
            # GetEndpointData returned bad StatusCode so we're done!

            self.log.error( "NOAANWSAWCMETAR: GetEndpointData() returned bad StatusCode: %d!" % StatusCode )

            ReturnStatus = False
            return ReturnStatus
 
        self.log.info( "NOAANWSAWCMETAR: Received METAR/TAF for '%s'." % StationID )

        # We successfully got a METAR so all is well unless we get an error while 
        # processing the METAR.

        ReturnStatus = True 

        # Write the returned METAR/TAF to our METAR file.

        FileStemPrefix          = self.CleanupPathStem( "NOAANWSAWCMETAR-" + StationID )
        METARFileName           = FileStemPrefix + ".html"
        METARFilePathAbsolute   = self.WXFeedsHTMLRoot + "/" + METARFileName

        if DEBUG == True:
            self.log.debug( "NOAANWSAWCMETAR: FileStemPrefix: %s" % FileStemPrefix )
            self.log.debug( "NOAANWSAWCMETAR: METARFileName: %s" % METARFileName )
            self.log.debug( "NOAANWSAWCMETAR: METARFilePathAbsolute: %s" % METARFilePathAbsolute )

        if CacheHit == False:
            # We got a METAR but not from cache so it's new - write it!

            try:
                with open( METARFilePathAbsolute, 'wt' ) as f:
                    f.write( METARData )
            except:
                self.log.error( "NOAANWSAWCMETAR: Writing METAR file: %s Failed!" % METARFilePathAbsolute )
                ReturnStatus = False
        else:
            # We got a METAR from cache so it's file should exist.

            p = Path( METARFilePathAbsolute )
            assert p.exists() == True

#        self.log.debug( "NOAANWSAWCMETAR: len( Response.text ) = %d" % len( Response.text ) )
#        self.log.debug( "NOAANWSAWCMETAR: Response.text = %s" % Response.text )

        # It looks like we have a good response but it is possible the returned text (which is HTML) 
        # does NOT actually contain METAR/TAF data.  Such a response is (currently!) 797 characters in 
        # length.
        #
        # Check the length of the returned text to determine if we want to use it or not.

#        minResponseTextLength = 797

#        if len( METARData ) >= minResponseTextLength:
#            try:
#                with open( METARFilePathAbsolute, 'wt' ) as f:
#                    f.write( METARData )

#                ReturnStatus = True 
#            except:
#                self.log.error( "NOAANWSAWCMETAR: Writing METAR file: %s Failed!" % METARFilePathAbsolute )
#                ReturnStatus = False
#        else:
#            self.log.error( "NOAANWSAWCMETAR: Received METAR file length (%d) less than expected (%d)!" % ( len( METARData ), minResponseTextLength ) )
#            ReturnStatus = False

        # Create Google Maps map if enabled.

        # We should (MUST!) have Location, Latitude and Longitude by now.

        MapHTMLFileName = ""
 
        if self.GoogleMapsMapsUse == True and    \
           Location != "" and                       \
           PositionSpecified == True:

            MarkerInfoWindowContent = ""

            if LocationWebsiteURL != "":
                MarkerInfoWindowContent += "<a href=\"%s\">Location: %s</a>" % ( LocationWebsiteURL, Location )

            if DEBUG == True:
                self.log.debug( "NOAANWSAWCMETAR: 1: MarkerInfoWindowContent: %s" % MarkerInfoWindowContent )

            if StationWebsiteURL != "":
                if MarkerInfoWindowContent != "":
                    MarkerInfoWindowContent += "<br/><br/n>"

                MarkerInfoWindowContent += "<a href=\"%s\">Station: %s</a>" % ( StationWebsiteURL, StationID )

            if DEBUG == True:
                self.log.debug( "NOAANWSAWCMETAR: 2: MarkerInfoWindowContent: %s" % MarkerInfoWindowContent )

            if MarkerInfoWindowContent != "":
                MarkerInfoWindowContent += "<br/><br/n>"

            if DEBUG == True:
                self.log.debug( "NOAANWSAWCMETAR: 3: MarkerInfoWindowContent: %s" % MarkerInfoWindowContent )

            MarkerInfoWindowContent += "<a href=\"%s\">METAR: %s</a>" % ( METARFileName, StationID )

            if DEBUG == True:
                self.log.debug( "NOAANWSAWCMETAR: 4: MarkerInfoWindowContent: %s" % MarkerInfoWindowContent )

            # If we get here we have a valid METAR.  If we did NOT get a cache hit, the METAR is new so we 
            # want to force recreating the map files.  If we did get a cache hit then the METAR has not been 
            # updated so possibly no need to recreate map files but maybe they should be, especially if 
            # we're using the WXFeeds servers, so inform CreateGoogleMapsMap().
            #
            # In either case, CreateGoogleMapsMap returns valid status and filenames (regardless if the map 
            # files were (re-)created or not).

            # CreateGoogleMapsMap() returns an error, we just disable use of the map.

            ReturnedStatus, MapHTMLFileName, MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                                                 FilePathStemPrefix        = FileStemPrefix,
                                                                                                 ForceStaleFiles           = not CacheHit,
                                                                                                 Location                  = Location,
                                                                                                 Latitude                  = Latitude,
                                                                                                 Longitude                 = Longitude,
                                                                                                 PositionSpecified         = PositionSpecified,
                                                                                                 Zoom                      = Zoom,
                                                                                                 MarkerType                = "misc",
                                                                                                 MarkerInfoWindowContent   = MarkerInfoWindowContent,
                                                                                                 DisplayStationMarker      = True
                                                                                               )
            if ReturnedStatus == True:
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                # Disable use of "bad" map.

                MapHTMLFileName = ""

        # Write a default HTML file.
        #
        # We do this to delete any existing METAR file and to 
        # create a minimal file that says a METAR was not returned.

#        HTML =   "<!DOCTYPE html>\n"                                                                                            \
#               + "<html lang=\"en\">\n"                                                                                         \
#               + "  <head>\n"                                                                                                   \
#               + "    <meta charset=\"utf-8\">\n"                                                                               \
#               + "    <title>\n"                                                                                                \
#               + "      METAR(s) and/or TAF(s) Requested For: %s, %s\n" % ( StationID, Location )                               \
#               + "    </title>\n"                                                                                               \
#               + "  </head>\n"                                                                                                  \
#               + "  <body>\n"                                                                                                   \
#               +      "METAR/TAF Information Was NOT Returned For Station: %s, Location: %s!\n" % ( StationID, Location )       \
#               + "  </body>\n"                                                                                                  \
#               + "</html>\n"
#
#        try:
#            with open( METARFilePathAbsolute, 'wt' ) as f:
#                f.write( HTML )
#        except:
#           self.log.error( "NOAANWSAWCMETAR: Writing Temporary METAR File: %s Failed!" % METARFilePathAbsolute )
#           ReturnStatus = False
#           return ReturnStatus

        # Table entry will look something like this:
        #              Text                            | Icon
        # Top Line:    Location - Link to Location Web | Play Icon - Link to METAR
        #              or Link to METAR                |
        #                                              |
        # Bottom Line: StationID - Link to Station Web | Google Map Icon - Link To Map
        #              or Link to METAR                |

        HTML = "<td>\n"

        if LocationWebsiteURL != "":
            # Link to location website.

            HTML += "  <a href=\"%s\">%s</a>\n" % ( LocationWebsiteURL, Location )
            self.WriteGeneratedFiles( HTML )
        else:
            # Link to METAR.

            HTML += "  <a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( METARFileName )
            HTML = "\">%s</a>\n" % Location
            self.WriteGeneratedFiles( HTML )

        HTML = "&nbsp;<a href=\""
        self.WriteGeneratedFiles( HTML )
        self.WriteGeneratedFilesRelativeLink( METARFileName )
        HTML = "\"><span class=\"icon-play\"></span></a>\n"
        self.WriteGeneratedFiles( HTML )

        HTML = "<br/>\n"
        self.WriteGeneratedFiles( HTML )

        if StationWebsiteURL != "":
            # Link to station website.

            HTML = " <a href=\"%s\">%s</a>\n" % ( StationWebsiteURL, StationID )
            self.WriteGeneratedFiles( HTML )
        else:
            # Link to METAR.

            HTML = "  <a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( METARFileName )
            HTML = "\">%s</a>\n" % StationID
            self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
            HTML = "&nbsp;<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        HTML = "</td>\n"
        self.WriteGeneratedFiles( HTML )

        return ReturnStatus

    def NOAANWSAWCMETARValidateData(
                                     self,
                                     StatusCode     : int,
                                     METARData      : str
                                   ) -> bool:
                                   #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "NOAANWSAWCMETARValidateData: METARData: %s" % METARData )

        if StatusCode == 200:
#            self.log.debug( "NOAANWSAWCMETARValidateData: METARData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "NOAANWSAWCMETARValidateData: METARData is NOT valid!" )

            pass

        return ReturnStatus


    def CleanupPathStem(
                         self,
                         PathStem : str
                       ) -> str:
                       #    ReturnValue

        # Remove or replace characters we don't want in a filename.

        ReturnValue = PathStem.strip()
 
        # The linux and Windows characters:

        # Linux:

        ReturnValue = ReturnValue.replace( "\0", "" )
        ReturnValue = ReturnValue.replace( "/", "" )

        # Windows:

        #ReturnValue = ReturnValue.replace( "/", "" )
        ReturnValue = ReturnValue.replace( "<", "" )
        ReturnValue = ReturnValue.replace( ">", "" )
        ReturnValue = ReturnValue.replace( ":", "" )
        ReturnValue = ReturnValue.replace( "\"", "" )
        ReturnValue = ReturnValue.replace( "\\", "" )
        ReturnValue = ReturnValue.replace( "|", "-" )
        ReturnValue = ReturnValue.replace( "?", "" )
        ReturnValue = ReturnValue.replace( "*", "" )

        # WXFeeds:

        ReturnValue = ReturnValue.replace( " ", "" )
        ReturnValue = ReturnValue.replace( ",", "" )

        return ReturnValue

    def GoogleMapsGeocoding(
                             self,
                             Location       : str,
                             Address        : str,
                             Components     : str
                           ) -> ( bool,         float,    float ):
                           #      ReturnStatus, Latitude, Longitude

        ReturnStatus = False

        Latitude = BAD_LATITUDE
        Longitude = BAD_LONGITUDE

        #DEBUG = True
        DEBUG = False

        GoogleMapGeocodingsURL = GOOGLE_MAPS_GEOCODING_SERVICE_URL
        GoogleMapGeocodingsURL += "?address=" + Address.replace( " ", "+" )
        if Components != "":
            GoogleMapGeocodingsURL += "&components=" + Components

        if self.GoogleMapsGeocodingServerUse == False:
            assert self.GoogleMapsAPIKey != ""

            if GoogleMapGeocodingsURL.find( "?" ) < 0:
                GoogleMapGeocodingsURL += "?"
            else:
                GoogleMapGeocodingsURL += "&"

            GoogleMapGeocodingsURL += "key=" + self.GoogleMapsAPIKey

        if DEBUG == True:
            self.log.debug( "GoogleMapsGeocoding: GoogleMapGeocodingsURL: %s" % GoogleMapGeocodingsURL )
            self.log.debug( "GoogleMapsGeocoding: Location: %s" % Location )
            self.log.debug( "GoogleMapsGeocoding: Address: %s" % Address )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingLocalCacheUse: %s" % self.GoogleMapsGeocodingLocalCacheUse )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingLocalCacheExpiry_minutes: %d" % self.GoogleMapsGeocodingLocalCacheExpiry_minutes )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingLocalCacheForceMiss: %s" % self.GoogleMapsGeocodingLocalCacheForceMiss )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingServerUse: %s" % self.GoogleMapsGeocodingServerUse )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingServerExpiry_minutes: %d" % self.GoogleMapsGeocodingServerExpiry_minutes )
#            self.log.debug( "GoogleMapsGeocoding: self.GoogleMapsGeocodingServerCacheMiss: %s" % self.GoogleMapsGeocodingServerCacheMiss )

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        GeocodingData   = None

        try:
            ReturnedStatus, CacheHit, StatusCode, GeocodingData = self.GetEndpointData(
                                                                                        EndpointURL                 = GoogleMapGeocodingsURL,
                                                                                        ContentType                 = CONTENT_JSON,
                                                                                        ValidateDataFunction        = self.GoogleMapsGeocodingValidateData,
                                                                                        Service                     = "GoogleMapsGeocoding",
                                                                                        CacheKey                    = Location,
                                                                                        LocalCacheUse               = self.GoogleMapsGeocodingLocalCacheUse,
                                                                                        LocalCacheExpiry_minutes    = self.GoogleMapsGeocodingLocalCacheExpiry_minutes,
                                                                                        LocalCacheForceMiss         = self.GoogleMapsGeocodingLocalCacheForceMiss,
                                                                                        ServerUse                   = self.GoogleMapsGeocodingServerUse,
                                                                                        RemoteCacheUse              = self.GoogleMapsGeocodingRemoteCacheUse,
                                                                                        RemoteCacheExpiry_minutes   = self.GoogleMapsGeocodingRemoteCacheExpiry_minutes,
                                                                                        RemoteCacheForceMiss        = self.GoogleMapsGeocodingRemoteCacheForceMiss
                                                                                      )
#        except:
#            self.log.error( "GoogleMapsGeocoding: GetEndpointData() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )
        except UnboundLocalError as error:
            # Output expected UnboundLocalErrors.
            self.log.error( "GoogleMapsGeocoding: GetEndpointData() threw an '<UnboundLocalError>' exception!: %s" % error )
        except Exception as exception:
            # Output unexpected Exceptions.
            self.log.error( "GoogleMapsGeocoding: GetEndpointData() threw an exception!  Unexpected error: %s" % exception )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "GoogleMapsGeocoding: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "GoogleMapsGeocoding: GetEndpointData() returned False for Location: '%s'" % Location )

            ReturnStatus = False
            return ReturnStatus, Latitude, Longitude

        if DEBUG == True:
            self.log.debug( "GoogleMapsGeocoding: GetEndpointData() returned successully for Location: '%s'" % Location )

        if StatusCode != 200:
            if DEBUG == True:
                self.log.debug( "GoogleMapsGeocoding: GetEndpointData() returned BAD StatusCode for Location: '%s'" % Location )

            ReturnStatus = False
            return ReturnStatus

        # We should have valid Google Maps geocoding information (in json format).

        if GeocodingData[ 'status' ] == OK:
            Results = GeocodingData[ 'results' ]
            assert len( Results ) == 1

            Geometry = Results[ 0 ][ 'geometry' ]
            Location = Geometry[ 'location' ]
            Latitude_str = Location[ 'lat' ]
            Longitude_str = Location[ 'lng' ]

            Latitude    = float( Latitude_str )
            Longitude   = float( Longitude_str )

            if DEBUG == True:
                self.log.debug( "GoogleMapsGeocoding: Latitude = %f, Longitude = %f" % ( Latitude, Longitude ) )

            ReturnStatus = True
        else:
            self.log.error( "Google Maps Geocoding Service did NOT return \"'status': 'OK'\"!  Location: %s" % Location )

        return ReturnStatus, Latitude, Longitude

    def GoogleMapsGeocodingValidateData(
                                         self,
                                         StatusCode     : int,
                                         GeocodingData  : dict
                                       ) -> bool:
                                       #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "StatusCode: %d" % StatusCode )
            self.log.debug( "GeocodingData[ 'status' ]: %s" % GeocodingData[ 'status' ] )
            self.log.debug( "GeocodingData[ 'results' ][ 0 ]: %s" % GeocodingData[ 'results' ][ 0 ] )

        if StatusCode in range( 200, 300 ) and      \
           GeocodingData[ 'status' ] == OK and      \
           GeocodingData[ 'results' ] != None:

            if DEBUG == True:
                self.log.debug( "GoogleMapsGeocodingValidateData: GeocodingData is valid." )

            ReturnStatus = True

        return ReturnStatus

    def GoogleMapsPlaces(
                          self,
                          Place : str
                        ) -> ( bool,         str,  float,    float,     str ):
                        #      ReturnStatus, Name, Latitude, Longitude, Icon

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        Name        = "!!Not Found By GoogleMaps Places API!!"
        Latitude    = BAD_LATITUDE
        Longitude   = BAD_LONGITUDE
        Icon        = ""

        GoogleMapsPlacesURL = GOOGLE_MAPS_PLACES_SERVICE_URL
        GoogleMapsPlacesURL += "?inputtype=textquery&input=" + Place.replace( " ", "+" )
        GoogleMapsPlacesURL += "&fields=business_status,formatted_address,geometry,icon,name,place_id,plus_code,types"

        if self.GoogleMapsPlacesServerUse == False:
            assert self.GoogleMapsAPIKey != ""

            if GoogleMapsPlacesURL.find( "?" ) < 0:
                GoogleMapsPlacesURL += "?"
            else:
                GoogleMapsPlacesURL += "&"

            GoogleMapsPlacesURL += "key=" + self.GoogleMapsAPIKey

        if DEBUG == True:
            self.log.debug( "GoogleMapsPlaces: GoogleMapsPlacesURL: %s" % GoogleMapsPlacesURL )
#            self.log.debug( "GoogleMapsPlaces: self.GoogleMapsPlacesUseLocalCache: %s" % self.GoogleMapsPlacesUseLocalCache )
#            self.log.debug( "GoogleMapsPlaces: PlaceSearchCacheStem: %s" % LocalCacheStem )
#            self.log.debug( "GoogleMapsPlaces: self.GoogleMapsPlacesCacheExpiry_minutes: %d" % self.GoogleMapsPlacesCacheExpiry_minutes )
#            self.log.debug( "GoogleMapsPlaces: self.WXFeedsServerUse: %s" % self.WXFeedsServerUse )
#            self.log.debug( "GoogleMapsPlaces: self.GoogleMapsPlacesForceCacheMiss: %s" % self.GoogleMapsPlacesForceCacheMiss )

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        PlacesData      = {}

        try:
            ReturnedStatus, CacheHit, StatusCode, PlacesData = self.GetEndpointData(
                                                                                     EndpointURL                = GoogleMapsPlacesURL,
                                                                                     ContentType                = CONTENT_JSON,
                                                                                     ValidateDataFunction       = self.GoogleMapsPlacesValidateData,
                                                                                     Service                    = "GoogleMapsPlaces",
                                                                                     CacheKey                   = Place,
                                                                                     LocalCacheUse              = self.GoogleMapsPlacesLocalCacheUse,
                                                                                     LocalCacheExpiry_minutes   = self.GoogleMapsPlacesLocalCacheExpiry_minutes,
                                                                                     LocalCacheForceMiss        = self.GoogleMapsPlacesLocalCacheForceMiss,
                                                                                    ServerUse             = self.GoogleMapsPlacesServerUse,
                                                                                     RemoteCacheUse             = self.GoogleMapsPlacesRemoteCacheUse,
                                                                                     RemoteCacheExpiry_minutes  = self.GoogleMapsPlacesRemoteCacheExpiry_minutes,
                                                                                     RemoteCacheForceMiss       = self.GoogleMapsPlacesRemoteCacheForceMiss
                                                                                   )
        except:
            self.log.error( "GoogleMapsPlaces: GetEndpointData() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "GoogleMapsPlaces: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.warning( "GoogleMapsPlaces: GetEndpointData() returned False for Place: '%s'." % Place )

            ReturnStatus = False
            return ReturnStatus

        if DEBUG == True:
            self.log.debug( "GoogleMapsPlaces: GetEndpointData() returned successully for Place: '%s'" % Place )

        if StatusCode != 200:
            if DEBUG == True:
                self.log.debug( "GoogleMapsPlaces: GetEndpointData() returned BAD StatusCode for Place: '%s'" % Place )

            ReturnStatus = False
            return ReturnStatus

        # We should have good Google Maps Places information (in json format).

        Candidates = PlacesData[ 'candidates' ]

#        self.log.debug("%s" % Candidates )
        Geometry = Candidates[ 0 ][ 'geometry' ]
        assert Geometry != None

        Location = Geometry[ 'location' ]
        assert Location != None

        Latitude_str    = Location[ 'lat' ]
        Longitude_str   = Location[ 'lng' ]

        Latitude    = float( Latitude_str )
        Longitude   = float( Longitude_str )

        Name = Candidates[ 0 ][ 'name' ]

        Icon = Candidates[ 0 ][ 'icon' ]

        ReturnStatus = True

        if DEBUG == True:
            self.log.debug( "GoogleMapsPlaces: Going To Return: ReturnStatus = %s, Name = %s, Latitude = %f, Longitude = %f, Icon = %s" % \
                            (
                              ReturnStatus,
                              Name,
                              Latitude,
                              Longitude,
                              Icon
                            )
                          )

        return ReturnStatus, Name, Latitude, Longitude, Icon

    def GoogleMapsPlacesValidateData(
                                      self,
                                      StatusCode    : int,
                                      PlacesData    : dict
                                    ) -> bool:
                                    #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if StatusCode == 200 and                       \
           PlacesData[ 'status' ] == OK and            \
           PlacesData[ 'candidates' ] != None and      \
           len( PlacesData[ 'candidates' ] ) == 1:

            if DEBUG == True:
                self.log.debug( "GoogleMapsPlacesValidateData: PlacesData is valid." )

            ReturnStatus = True

        return ReturnStatus

    def GetEndpointData(
                         self,
                         EndpointURL                : str,
                         ContentType                : str,
                         ValidateDataFunction       : Callable,
                         Service                    = "",
                         CacheKey                   = "",
                         LocalCacheUse              = False,
                         LocalCacheExpiry_minutes   = -1,
                         LocalCacheForceMiss        = False,
                         ServerUse             = False,
                         RemoteCacheUse             = False,
                         RemoteCacheExpiry_minutes  = -1,
                         RemoteCacheForceMiss       = False

                       ) -> ( bool,         bool,     int,        dict ):
                       #      ReturnStatus, CacheHit, StatusCode, EndpointData

        # Implement caching for endpoint data.

        ReturnStatus    = False
        CacheHit        = False
        StatusCode      = -1
        EndpointData    = None

        #DEBUG = True
        DEBUG = False

        assert EndpointURL != ""
        assert ContentType == CONTENT_HTML or ContentType == CONTENT_JSON or ContentType == CONTENT_XML
        assert ValidateDataFunction != None

        if DEBUG == True:
            self.log.debug( "GetEndpointData: In GetEndpointData. . ." )
            self.log.debug( "GetEndpointData: EndpointURL: %s" % EndpointURL )
            self.log.debug( "GetEndpointData: ContentType: %s" % ContentType )
            self.log.debug( "GetEndpointData: Service: %s" % Service )
            self.log.debug( "GetEndpointData: CacheKey: %s" % CacheKey )
            self.log.debug( "GetEndpointData: LocalCacheUse: %s" % LocalCacheUse )
            self.log.debug( "GetEndpointData: LocalCacheExpiry_minutes: %d" % LocalCacheExpiry_minutes )
            self.log.debug( "GetEndpointData: LocalCacheForceMiss: %s" % LocalCacheForceMiss )
            self.log.debug( "GetEndpointData: ServerUse: %s" %ServerUse )
            self.log.debug( "GetEndpointData: RemoteCacheUse: %s" % RemoteCacheUse )
            self.log.debug( "GetEndpointData: RemoteCacheExpiry_minutes: %d" % RemoteCacheExpiry_minutes )
            self.log.debug( "GetEndpointData: RemoteCacheForceMiss: %s" % RemoteCacheForceMiss )

        LocalCachePath          = self.WXFeedsLocalCacheRoot + "/" + self.CleanupPathStem( Service + "-" + CacheKey ) + ".cache"
        CacheFile               = None
        WXFeedsCacheFileMarker  = "WXFeedsCacheFile"
        DateTimeFormat          = "%Y-%m-%d %H:%M:%S"

        if LocalCacheUse == True:
            # We're going to use local caching. . .

            if DEBUG == True:
                self.log.debug( "GetEndpointData: Local caching. . ." )

            if LocalCacheForceMiss == False:
                if DEBUG == True:
                    self.log.debug( "GetEndpointData: Reading cache file: %s" % LocalCachePath )

                # Try to get data from the local cache.

                # Open the cache file.

                try:
                    CacheFile = open( LocalCachePath, "rb" )
                    if DEBUG == True:
                        self.log.debug( "GetEndpointData: Cache file opened OK." )
                except FileNotFoundError as e:
                    if DEBUG == True:
                        self.log.debug( "GetEndpointData: Cache file open failed!  File Not Found: e = %s, sys.exc_info() = %s" % ( e, sys.exc_info() ) )
                except IOError as e:
                    self.log.error( "GetEndpointData: Cache file open failed!  Unexpected Error: e = %s, sys.exc_info() = %s" % ( e, sys.exc_info() ) )
                except:
                    self.log.error( "GetEndpointData: Unexpected Error: %s" % sys.exc_info()[0] )

                if CacheFile != None:
                    # We succesfully opened a cache file - read it.

                    # The first line should be a WXFeeds cache "marker".

                    try:
                        CacheFileMarker = pickle.load( CacheFile )
                    except:
                        self.log.error( "GetEndpointData: Unexpected error while unpickling WXFeedsCacheFileMarker: %s" % sys.exc_info()[0] )

                    if DEBUG == True:
                        self.log.debug( "GetEndpointData: CacheFileMarker = \"%s\"" % CacheFileMarker )

                    if CacheFileMarker == WXFeedsCacheFileMarker:

                        # The second line should be the date/time when the cache data expires.

                        try:
                            CacheExpiryDateTime_str = pickle.load( CacheFile )
                        except:
                            self.log.error( "GetEndpointData: Unexpected error while unpickling CacheExpiryDateTime_str: %s" % sys.exc_info()[0] )

                        if DEBUG == True:
                            self.log.debug( "GetEndpointData: CacheExpiryDateTime_str = \"%s\"" % CacheExpiryDateTime_str )

                        CacheExpiryDateTime_dt  = datetime.datetime.fromtimestamp( int( CacheExpiryDateTime_str ) )
                        Now_dt                  = datetime.datetime.now()
                        if DEBUG == True:
                            self.log.debug( "GetEndpointData: CacheDateTime_dt = %s, Now_dt = %s" %
                                                    (
                                                    CacheExpiryDateTime_dt.strftime( DateTimeFormat ),
                                                    Now_dt.strftime( DateTimeFormat )
                                                    )
                                        )

                        if Now_dt < CacheExpiryDateTime_dt:
                            # Possible cache hit.

                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: Cache not expired." )

                            # Read the status code.

                            try:
                                StatusCode = pickle.load( CacheFile )
                            except:
                                self.log.error( "GetEndpointData: Unexpected error while unpickling StatusCode: %s" % sys.exc_info()[0] )

                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: StatusCode = %d" % StatusCode )

                            # Read the cached data.

                            try:
                                CachedData = pickle.load( CacheFile )
                            except:
                                self.log.error( "GetEndpointData: Unexpected error while unpickling CachedData: %s" % sys.exc_info()[0] )

                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: CachedData: %s" % CachedData )

                            CacheFile.close()
                            CacheFile = None

                            EndpointDataOK = False
                            
                            if ContentType == CONTENT_JSON:
                                # Convert cached data to JSON.

                                try:
                                    EndpointData    = json.loads( CachedData )
                                    EndpointDataOK  = True
                                except:
                                    self.log.error( "GetEndpointData: json.loads() caused Unexpected Error: %s" % sys.exc_info()[0] )
                            elif ContentType == CONTENT_XML:
                                # Parse cached XML.

                                EndpointData    = feedparser.parse( CachedData )
                                EndpointDataOK  = True
                            elif ContentType == CONTENT_HTML:
                                # No conversion required.

                                EndpointData    = CachedData
                                EndpointDataOK  = True
                            else:
                                # Unknown Content Type!
                                #
                                # We should never get here because ContentType was checked before data was cached.

                                self.log.error( "GetEndpointData: Cached Data: Unknown 'ContentType': '%s'" % ContentType  )

                            assert EndpointDataOK == True

                            if EndpointDataOK == True:
                                if DEBUG == True:
                                    self.log.debug( "GetEndpointData: EndpointData: %s" % EndpointData )

                                if ValidateDataFunction( StatusCode, EndpointData ) == True:
                                    # Cached data is valid.

                                    if DEBUG == True:
                                        self.log.debug( "GetEndpointData: ValidateDataFunction() returned: True" )

                                    CacheHit        = True
                                    ReturnStatus    = True

                                    if DEBUG == True:
                                        self.log.debug( "GetEndpointData: Cache Hit!" )
                                        self.log.debug( "%s, %s, %d, %s" % ( ReturnStatus, CacheHit, StatusCode, EndpointData ) )
                                else:
                                    # Cached data is NOT (no longer!) valid.

                                    if DEBUG == True:
                                        self.log.debug( "GetEndpointData: ValidateDataFunction() returned: False" )
                                        self.log.debug( "GetEndpointData: Cache Miss!  Stale Cache - cached data not valid!" )

                                    assert CacheHit == False
                            else:
                                # Endpoint data conversion failed!

                                if DEBUG == True:
                                    self.log.debug( "GetEndpointData: Endpoint data conversion failed!" )

                                assert CacheHit == False
                        else:
                            # Stale Cache - past expiry data/time.

                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: Cache Miss!  Stale Cache - past expiry date/time!" )

                            assert CacheHit == False
                    else:
                        self.log.error( "GetEndpointData: Cache Miss!  WXFeeds cache marker not found!" )

                        assert CacheHit == False
                else:
                    # We were not able to open a cache file so we'll try to get data from the source.

                    if DEBUG == True:
                        self.log.debug( "GetEndpointData: Cache Miss!  Cache file open failed!" )

                    assert CacheHit == False
            else:
                # Force Cache Miss.

                self.log.debug( "GetEndpointData: Forced Cache Miss!" )

                assert CacheHit == False
        else:
            # We're not going to use local caching so delete the local cache file if it exists.

            if DEBUG == True:
                self.log.debug( "GetEndpointData: Not Using Local Caching - Deleting Cache File: '%s'. . ." % LocalCachePath )

            p = Path( LocalCachePath )
            if p.exists() == True:
                p.unlink()

        if CacheHit == False:
            if DEBUG == True:
                self.log.debug( "GetEndpointData: Fetching data. . ." )

            # Fetch data from the source.

            if DEBUG == True:
                self.log.debug( "GetEndpointData: EndpointURL: %s" % EndpointURL )

            GetDataURL = ""

            # If we're using a WXFeeds Server, we need to build an new URL.

            if ServerUse == True:
                assert self.WXFeedsServerUse == True
                assert self.WXFeedsServerURL != ""
                assert self.WXFeedsServerAccessToken != ""
                assert Service != ""
                assert CacheKey != ""

                GetDataURL = self.WXFeedsServerURL + "/WXFeeds"
                GetDataURL += "?ContentType=" + ContentType
                GetDataURL += "&Service=" + Service
                GetDataURL += "&CacheKey=" + CacheKey
                GetDataURL += "&CacheData=%s" % RemoteCacheUse
                GetDataURL += "&CacheExpiry_minutes=%s" % str( RemoteCacheExpiry_minutes )
                GetDataURL += "&ForceCacheMiss=%s" % RemoteCacheForceMiss
                TempEndpointURL = EndpointURL.replace( "?", "%3F" )
                TempEndpointURL = TempEndpointURL.replace( "&", "%26")
                GetDataURL += "&url=" + TempEndpointURL
                if DEBUG == True:
                    self.log.debug( "GetEndpointData: GetDataURL: '%s'" % GetDataURL )

                GetDataURL += "&accessToken=%s" % self.WXFeedsServerAccessToken
            else:
                GetDataURL = EndpointURL

            ResponseOK = False

            try:
                if DEBUG == True:
                    self.log.debug( "GetEndpointData: Fetching data using: '%s'" % GetDataURL )

                Response = requests.get( GetDataURL )
                ResponseOK = True

                if DEBUG == True:
                    self.log.debug( "GetEndpointData: Returned from fetching data." )
            except requests.exceptions.ConnectionError:
                self.log.error( "GetEndpointData: requests.get() returned ConnectionError!" )
            except requests.exceptions.HTTPError:
                self.log.error( "GetEndpointData: requests.get() returned HTTPError!" )
            except requests.exceptions.Timeout:
                self.log.error( "GetEndpointData: requests.get() returned Timeout!" )
            except requests.exceptions.TooManyRedirects:
                self.log.error( "GetEndpointData: requests.get() returned TooManyRedirects!" )
            except requests.exceptions.RequestException as e:
                self.log.error( "GetEndpointData: requests.get() returned RequestException!  e = %s, sys.exc_type = %s" % ( e, sys.exc_info() ) )
            except:
                self.log.error( "GetEndpointData: Unexpected error: %s" % sys.exc_info()[0] )

            if DEBUG == True:
                self.log.debug( "GetEndpointData: 'Reponse': %s" % ( Response ) )
                self.log.debug( "GetEndpointData: 'Response.status_code': %d" % ( Response.status_code ) )
                self.log.debug( "GetEndpointData: Checking 'status_code'. . ." )

            # It's possible to get a successful return with an unsuccessful status_code.
            #
            # One known case is NOAA NWS Forecasts - a 503 'status_code' indicates an expiry message.

            StatusCode = Response.status_code
            
            if DEBUG == True:
                self.log.debug( "GetEndpointData: StatusCode: %d" % StatusCode )

            if ResponseOK == True and                   \
               ( StatusCode in range( 200, 300 ) or     \
                 StatusCode == 503 ):
                if DEBUG == True:
                    self.log.debug( "GetEndpointData: 'StatusCode' = 200 - 299 or 503: %d" % StatusCode )

                # Response and its data should be good.

                EndpointDataOK = False

                if StatusCode == WXFEEDS_SERVER_RESPONSE_CODE_CACHE_MISS or        \
                   StatusCode == WXFEEDS_SERVER_RESPONSE_CODE_CACHE_HIT:
                    if StatusCode == WXFEEDS_SERVER_RESPONSE_CODE_CACHE_MISS:
                        CacheHit = False

                        if DEBUG == True:
                            self.log.debug( "GetEndpointData: WXFeeds Server Cache Miss!" )
                    else:
                        CacheHit = True

                        if DEBUG == True:
                            self.log.debug( "GetEndpointData: WXFeeds Server Cache Hit!" )

                    StatusCode = HTTP_RESPONSE_CODE_OK

                if ContentType == CONTENT_JSON:
                    EndpointData    = Response.json()
                    EndpointDataOK  = True
                elif ContentType == CONTENT_XML:
                    EndpointData    = feedparser.parse( Response.text )
                    EndpointDataOK  = True
                elif ContentType == CONTENT_HTML:
                    EndpointData    = Response.text
                    EndpointDataOK  = True
                else:
                    self.log.error( "GetEndpointData: Fetched Data: Unknown 'ContentType': '%s'" % ContentType  )

                if EndpointDataOK == True:
                    #if DEBUG == True:
                    #    self.log.debug( "GetEndpointData: Got endpoint data OK! %s" % EndpointData )

                    ReturnStatus = ValidateDataFunction( StatusCode, EndpointData )

                    if DEBUG == True:
                        self.log.debug( "GetEndpointData: ValidateDataFunction() returned: %s" % ReturnStatus )

                    # If the returned data is valid then we can return success, even if caching the data fails.

                    #if DEBUG == True:
                    #    self.log.debug( "%s, %s, %d, %s" % ( ReturnStatus, CacheHit, StatusCode, EndpointData ) )

                    if ReturnStatus == True:
                        if LocalCacheUse == True and        \
                           StatusCode == HTTP_RESPONSE_CODE_OK:
                            # Cache the returned data.
                            #
                            # We only cache data with 'StatusCode' == OK.  In other words, 
                            # do not cache data with 'StatusCode' == 503.

                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: Writing local cache file:  Cache file: %s" % LocalCachePath )

                            CachingOK = True        # Assume success.

                            try:
                                CacheFile = open( LocalCachePath, "wb" )
                            except IOError as e:
                                self.log.error( "GetEndpointData: Local cache file creation failed!  Unexpected Error: e = %s, sys.exc_info() = %s" % ( e, sys.exc_info() ) )
                                CachingOK = False
                            except:
                                self.log.error( "GetEndpointData: Unexpected error: %s" % sys.exc_info()[0] )
                                CachingOK = False

                            if CachingOK == True:
                                if DEBUG == True:
                                    self.log.debug( "GetEndpointData: Cache file opened OK." )

                                # Write WXFeeds cache file marker.

                                try:
                                    pickle.dump( WXFeedsCacheFileMarker, CacheFile )
                                except:
                                    self.log.error( "GetEndpointData: Unexpected error while pickling WXFeedsCacheFileMarker: %s" % sys.exc_info()[0] )
                                    CachingOK = False

                                if CachingOK == True:
                                    # Write cache file expiry date/time.

                                    if DEBUG == True:
                                        self.log.debug( "GetEndpointData: CacheExpiryPeriod_minutes = %d." % LocalCacheExpiry_minutes )

                                    Now_dt = datetime.datetime.now()
                                    Expiry_dt = Now_dt + datetime.timedelta( minutes = LocalCacheExpiry_minutes )

                                    Expiry_str = "%d" % Expiry_dt.timestamp()

                                    try:
                                        pickle.dump( Expiry_str, CacheFile )
                                    except:
                                        self.log.error( "GetEndpointData: Unexpected error while pickling Expiry_str: %s" % sys.exc_info()[0] )
                                        CachingOK = False

                                if CachingOK == True:
                                    # Write StatusCode.

                                    try:
                                        pickle.dump( StatusCode, CacheFile )
                                    except:
                                        self.log.error( "GetEndpointData: Unexpected error while pickling StatusCode: %s" % sys.exc_info()[0] )
                                        CachingOK = False

                                if CachingOK == True:
                                    # Write data.

                                    try:
                                        pickle.dump( Response.text, CacheFile )
                                    except:
                                        self.log.error( "GetEndpointData: Unexpected error while pickling Response.text: %s" % sys.exc_info()[0] )
                                        CachingOK = False

                                if CachingOK == True:
                                    if DEBUG == True:
                                        self.log.debug( "GetEndpointData: Local cache file created OK." )
                                else:
                                    self.log.error( "GetEndpointData: Local cache file creation FAILED!" )

                            CacheFile.close()
                            CacheFile = None
                        else:
                            if DEBUG == True:
                                self.log.debug( "GetEndpointData: Not using local cache." )

                            ReturnStatus = True
                    else:
                        # Returned data is not valid!

                        if DEBUG == True:
                            self.log.debug( "GetEndpointData: Returned data is not valid!" )

                        assert ReturnStatus == False
                else:
                    # Data conversion failed!

                    self.log.error( "GetEndpointData: Fetched Data: Data Conversion Failed: 'ContentType': '%s'" % ContentType )

                    assert ReturnStatus == False
            else:
                assert ReturnStatus == False
                assert CacheHit == False
                assert EndpointData == None

                if DEBUG == True:
                    self.log.debug( "GetEndpointData: Error: Response.status_code: %d" % Response.status_code )

                pass

        if CacheFile != None:
            CacheFile.close()
            CacheFile = None

        assert CacheFile == None

        if DEBUG == True:
            self.log.debug( "GetEndpointData: %s, %s, %d, %s" % ( ReturnStatus, CacheHit, StatusCode, EndpointData ) )
            self.log.debug( "GetEndpointData: Exiting . . ."  )

        return ReturnStatus, CacheHit, StatusCode, EndpointData

    def GetMapHTMLPaths(
                         self,
                         FilePathStemPrefix : str
                       ) -> ( str,          str,                  str,                  str ):
                       #      HTMLFileName, HTMLFilePathRelative, HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute

        #DEBUG = True
        DEBUG = False

        HTMLFileName                = self.CleanupPathStem( FilePathStemPrefix + "-GoogleMapsMap.html" )
        assert HTMLFileName != ""
        HTMLFilePathRelative        = "WXFeeds/" + HTMLFileName
        HTMLFilePathAbsolute        = self.WXFeedsHTMLRoot + "/" + HTMLFileName
        HTMLCacheFilePathAbsolute   = self.GetCachePath( HTMLFileName )

        if DEBUG == True:
            self.log.debug( "GetMapHTMLPaths: FilePathStemPrefix: %s" % FilePathStemPrefix )
            self.log.debug( "GetMapHTMLPaths: HTMLFileName: %s" % HTMLFileName )
            self.log.debug( "GetMapHTMLPaths: HTMLFilePathRelative: %s" % HTMLFilePathRelative )
            self.log.debug( "GetMapHTMLPaths: HTMLFilePathAbsolute: %s" % HTMLFilePathAbsolute )
            self.log.debug( "GetMapHTMLPaths: HTMLCacheFilePathAbsolute: %s" % HTMLCacheFilePathAbsolute )

        return HTMLFileName, HTMLFilePathRelative, HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute

    def GetMapJavaScriptPaths(
                               self,
                               FilePathStemPrefix : str
                             ) -> ( str,        str,                str,                str ):
                             #      JSFileName, JSFilePathRelative, JSFilePathAbsolute, JSCacheFilePathAbsolute

        #DEBUG = True
        DEBUG = False

        JSFileName              = self.CleanupPathStem( FilePathStemPrefix + "-GoogleMapsMap.js" )
        assert JSFileName != ""
        JSFilePathRelative      = "js/" + JSFileName
        JSFilePathAbsolute      = self.WXFeedsHTMLRoot + "/" + JSFilePathRelative
        JSCacheFilePathAbsolute = self.GetCachePath( JSFileName )

        if DEBUG == True:
            self.log.debug( "GetMapJavaScriptPaths: FilePathStemPrefix: %s" % FilePathStemPrefix )
            self.log.debug( "GetMapJavaScriptPaths: JSFileName: %s" % JSFileName )
            self.log.debug( "GetMapJavaScriptPaths: JSFilePathRelative: %s" % JSFilePathRelative )
            self.log.debug( "GetMapJavaScriptPaths: JSFilePathAbsolute: %s" % JSFilePathAbsolute )
            self.log.debug( "GetMapJavaScriptPaths: JSCacheFilePathAbsolute: %s" % JSCacheFilePathAbsolute )

        return JSFileName, JSFilePathRelative, JSFilePathAbsolute, JSCacheFilePathAbsolute

    def CreateGoogleMapsMap(
                             self,
                             FilePathStemPrefix         : str,
                             ForceStaleFiles            = False,
                             Location                   = "",
                             Latitude                   = BAD_LATITUDE,
                             Longitude                  = BAD_LONGITUDE,
                             PositionSpecified          = False,
                             MarkerType                 = "misc",
                             MarkerInfoWindowContent    = "",
                             Zoom                       = 12,
                             DisplayStationMarker       = False,
                             Markers                    = None,
                             MarkersAreStatic           = True
                           ) -> ( bool,         str,                  str ):
                           #      ReturnStatus, HTMLFileName, HTMLFilePathAbsolute

        # CreateGoogleMapsMap:
        #
        # - Lookup position (Latitude and Longitude) if it is not specified.
        # - Display a marker for Location at position.
        # - Display a marker for this station (optional).
        # - Display a list of markers (optional).
        #
        # The HTML and Java Script files will only be created if they 
        # are not stale or they are forced to be stale.
        #
        # If MarkersAreStatic == True, the Java Script file will be only be created if it is stale.
        #
        # If MarkersAreStatic == False, the Java Script file will be created.

        ReturnStatus            = False
        HTMLFilePathRelative    = ""

        #DEBUG = True
        DEBUG = False

        # We create:
        # - an HTML file with the HTML for a Google Maps map and 
        # - a Java Script file for any markers.

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsMap: FilePathStemPrefix: %s" % FilePathStemPrefix )
            self.log.debug( "CreateGoogleMapsMap: ForceStaleFiles: %s" % ForceStaleFiles )
            self.log.debug( "CreateGoogleMapsMap: Location: %s" % Location )
            self.log.debug( "CreateGoogleMapsMap: Latitude: %f" % Latitude )
            self.log.debug( "CreateGoogleMapsMap: Longitude: %f" % Longitude )
            self.log.debug( "CreateGoogleMapsMap: PositionSpecified: %s" % PositionSpecified )
            self.log.debug( "CreateGoogleMapsMap: MarkerType: %s" % MarkerType )
            self.log.debug( "CreateGoogleMapsMap: MarkerInfoWindowContent: %s" % MarkerInfoWindowContent )
            self.log.debug( "CreateGoogleMapsMap: Zoom: %d" % Zoom )
            self.log.debug( "CreateGoogleMapsMap: DisplayStationMarker: %s" % DisplayStationMarker )
            self.log.debug( "CreateGoogleMapsMap: Markers: %s" % Markers )
            self.log.debug( "CreateGoogleMapsMap: MarkersAreStatic: %s" % MarkersAreStatic )

        HTMLFileName, HTMLFilePathRelative, HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute =       \
                self.GetMapHTMLPaths( FilePathStemPrefix )
        JavaScriptFileName, JavaScriptFilePathRelative, JavaScriptFilePathAbsolute, JavaScriptCacheFilePathAbsolute =       \
                self.GetMapJavaScriptPaths( FilePathStemPrefix )

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsMap: HTMLFileName: %s" % HTMLFileName )
            self.log.debug( "CreateGoogleMapsMap: HTMLFilePathRelative: %s" % HTMLFilePathRelative )
            self.log.debug( "CreateGoogleMapsMap: HTMLFilePathAbsolute: %s" % HTMLFilePathAbsolute )
            self.log.debug( "CreateGoogleMapsMap: HTMLCacheFilePathAbsolute: %s" % HTMLCacheFilePathAbsolute )
            self.log.debug( "CreateGoogleMapsMap: JavaScriptFileName: %s" % JavaScriptFileName )
            self.log.debug( "CreateGoogleMapsMap: JavaScriptFilePathRelative: %s" % JavaScriptFilePathRelative )
            self.log.debug( "CreateGoogleMapsMap: JavaScriptFiePathAbsolute: %s" % JavaScriptFilePathAbsolute )
            self.log.debug( "CreateGoogleMapsMap: JavaScriptCacheFilePathAbsolute: %s" % JavaScriptCacheFilePathAbsolute )

        HTMLIsFileStale         = False
        JavaScriptIsFileStale   = False

        if ForceStaleFiles == True:
            HTMLIsFileStale         = True
            JavaScriptIsFileStale   = True
        else:
            HTMLIsFileStale = self.IsFileStale( HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute, self.WXFeedsServerUse )

            if MarkersAreStatic == False:
                # Markers are not static so create the JavaScript file.

                JavaScriptIsFileStale = True
            else:
                JavaScriptIsFileStale = self.IsFileStale( JavaScriptFilePathAbsolute, JavaScriptCacheFilePathAbsolute )

        if DEBUG == True:
            if HTMLIsFileStale == True:
                self.log.debug( "CreateGoogleMapsMap: HTML file is stale." )
            else:
                self.log.debug( "CreateGoogleMapsMap: HTML file is NOT stale." )

            if JavaScriptIsFileStale == True:
                self.log.debug( "CreateGoogleMapsMap: Java Script file is stale." )
            else:
                self.log.debug( "CreateGoogleMapsMap: Java Script file is NOT stale." )

        if HTMLIsFileStale == False and  \
           JavaScriptIsFileStale == False:
            #   HTML and JavaScript files are both NOT stale.

            if DEBUG == True:
                self.log.debug( "CreateGoogleMapsMap: Map files are NOT stale." )

            return True, HTMLFileName, HTMLFilePathAbsolute

        # The HTML file or the JavaScript file is stale so create one or both of them.

        assert HTMLIsFileStale == True or JavaScriptIsFileStale == True

        ReturnStatus = True     # Assume success.

        if HTMLIsFileStale == True:
            # Create the HTML map file.

            assert JavaScriptFilePathRelative != ""

            GoogleMapsMapsURL = GOOGLE_MAPS_MAPS_SERVICE_URL

            if self.GoogleMapsMapsServerUse == True:
                GoogleMapsMapsURL = GoogleMapsMapsURL.replace( "?", "%3F" )
                GoogleMapsMapsURL = GoogleMapsMapsURL.replace( "&", "%26")

                GoogleMapsMapsURL = self.WXFeedsServerURL + "/WXFeeds?Service=GoogleMapsMaps&url=" + GoogleMapsMapsURL
                GoogleMapsMapsURL += "&accessToken=" + self.WXFeedsServerAccessToken
            else:
                assert self.GoogleMapsAPIKey != ""

                GoogleMapsMapsURL += "&key=" + self.GoogleMapsAPIKey

            HTML =   "<!DOCTYPE html>\n"                                                                \
                   + "<html lang=\"en-CA\">\n"                                                          \
                   + "    <head>\n"                                                                     \
                   + "        <title>%s</title>\n" % Location                                           \
                   + "        <script src=\'%s\'></script>\n" % JavaScriptFilePathRelative              \
                   + "        <style>\n"                                                                \
                   + "            /* Set the size of the div element that contains the map */\n"        \
                   + "            #WXFeedsMap {\n"                                                      \
                   + "                height: 400px;  /* The height is 400 pixels */\n"                 \
                   + "                width: 100%;    /* The width is the width of the web page */\n"   \
                   + "            }\n"                                                                  \
                   + "        </style>\n"                                                               \
                   + "    </head>\n"                                                                    \
                   + "    <body>\n"                                                                     \
                   + "        <h1>%s</h1>\n" % Location                                                 \
                   + "        <div id=\"WXFeedsMap\"></div>\n"                                          \
                   + "        <script defer src=\"%s\">\n" % GoogleMapsMapsURL                          \
                   + "        </script>\n"                                                              \
                   + "    </body>\n"                                                                    \
                   + "</html>\n"

            # Write the HTML file.

            try:
                with open( HTMLFilePathAbsolute, 'w' ) as f:
                    f.write( HTML )
            except:
               self.log.error( "CreateGoogleMapsMap: Writing Google Map HTML File: %s Failed!" % HTMLFilePathAbsolute )

               ReturnStatus = False

        if ReturnStatus == True and         \
           JavaScriptIsFileStale == True:
            # No errors so far and the Java Script file is stale.
            #
            # Create the marker list and the Java Script file.
            #
            # Markers could be:
            # - a marker for the Location specified or found by Google Maps Places,
            # - a marker for this station and/or 
            # - a list of markers specified by the caller.

            if Location != "" and           \
               PositionSpecified == False:
                # We have a Location but not position.
                # Look up position (latitude and longitude) for Location.

                if self.GoogleMapsPlacesUse == True:
                    ReturnedStatus      = False
                    ReturnedName        = "* Not Found *"
                    ReturnedLatitude    = BAD_LATITUDE
                    ReturnedLongitude   = BAD_LONGITUDE
                    ReturnedIcon        = ""

                    if DEBUG == True:
                        self.log.debug( "CreateGoogleMapsMap: Location: %s" % Location )

                    try:
                        ReturnedStatus, ReturnedName, ReturnedLatitude, ReturnedLongitude, ReturnedIcon = self.GoogleMapsPlaceSearch( Location )
                    except:
                        self.log.error( "CreateGoogleMapsMap: GoogleMapsPlaceSearch() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

                        ReturnedStatus = False

                    if ReturnedStatus == True:
                        # Use the returned information.

                        if DEBUG == True:
                            self.log.debug( "CreateGoogleMapsMap: GoogleMapsPlaceSearch() returned: %s, %s, %f, %f, %s for %s." %   \
                                            (
                                              ReturnedStatus,
                                              ReturnedName,
                                              ReturnedLatitude,
                                              ReturnedLongitude,
                                              ReturnedIcon,
                                              Location
                                            )
                                          )

                        Latitude            = ReturnedLatitude
                        Longitude           = ReturnedLongitude
                        PositionSpecified   = True
                    else:
                        self.log.warning( "CreateGoogleMapsMap: GoogleMapsPlaceSearch() returned %s for '%s'!" % ( ReturnedStatus, Location ) )
                else:
                    self.log.warning( "CreateGoogleMapsMap: PositionSpecified = %s but no way to look up position!" % PositionSpecified )

            OurMarkers = []

            StationMarkerPositionUsed = False

            # Create a marker for Location.

            if PositionSpecified == True:
                # If we have position we can create a marker for Location with correct position.

                if DEBUG == True:
                    self.log.debug( "CreateGoogleMapsMap: Adding marker for %s." % Location )

                OurMarkers.append( GoogleMapsMarker(
                                                     MarkerType,
                                                     Latitude,
                                                     Longitude,
                                                     Location,
                                                     MarkerInfoWindowContent
                                                   )
                                 )
            else:
                # Use this station's position to create a marker indicating no position.

                self.log.warning("GreateGoogleMapsMap: Position for '%s' not known - will use station position." % Location)

                if self.StationLatitude != BAD_LATITUDE and     \
                   self.StationLongitude != BAD_LONGITUDE:

                    Latitude            = self.StationLatitude
                    Longitude           = self.StationLongitude
                    PositionSpecified   = True

                    OurMarkers.append( GoogleMapsMarker(
                                                         MarkerType,
                                                         Latitude,
                                                         Longitude,
                                                         "!*** %s Using Station Position ***!" % Location,
                                                         MarkerInfoWindowContent
                                                       )
                                     )

                    StationMarkerPositionUsed = True
                else:
                    # No position for Location or this station - we must be in the middle of nowhere!

                    self.log.warning("GreateGoogleMapsMap: Latitude and longitude not set for this station!")

                    # From Wikipedia: "The most remote island is Bouvet Island, a small, uninhabited island in 
                    # the South Atlantic Ocean that is a dependency of Norway. It lies at coordinates 54°26′S 3°24′E. 
                    # The nearest land is the uninhabited Queen Maud Land, Antarctica, over 1,600 km (994 mi) to 
                    # the south. The nearest inhabited lands are Gough Island, 1,845 km (1,146 mi) away, Tristan 
                    # da Cunha, 2,260 km (1,404 mi) away, and the coast of South Africa, 2,580 km (1,603 mi) away.

                    # But!  While testing this I learned that weewx will abort if [Station] latitude and 
                    # longitude are nonexistent, so this code will probably never execute unless 
                    # longitude and latitude are set to "bad" values.  On well . . .

                    Latitude            = -54 + (26/60)
                    Longitude           = 3 + (3/60)
                    PositionSpecified   = True

                    OurMarkers.append( GoogleMapsMarker(
                                                         MarkerType,
                                                         Latitude,
                                                         Longitude,
                                                         "We Should Be At %s But We're Lost!" % Location,
                                                         MarkerInfoWindowContent
                                                       )
                                     )

            # Create a marker for this station but don't create it if 
            # we used the station's position for Location.

            if StationMarkerPositionUsed == False and       \
               DisplayStationMarker == True:
                ReturnedStatus, StationMarker = self.StationMarker()

                if DEBUG == True:
                    self.log.debug( "CreateGoogleMapsMap: Adding station marker." )

                assert StationMarker != None

                if DEBUG == True:
                    self.log.debug( "CreateGoogleMapsMap: StationMarker() returned: %s" % ReturnedStatus )
                    self.log.debug( "CreateGoogleMapsMap: StationMarker.Icon: %s" % StationMarker.Type )
                    self.log.debug( "CreateGoogleMapsMap: StationMarker.Latitude: %f" % StationMarker.Latitude )
                    self.log.debug( "CreateGoogleMapsMap: StationMarker.Longitude: %f" % StationMarker.Longitude )
                    self.log.debug( "CreateGoogleMapsMap: StationMarker.Text: %s" % StationMarker.Text )
                    self.log.debug( "CreateGoogleMapsMap: StationMarker.InfoWindowContent: %s" % StationMarker.InfoWindowContent )

                if ReturnedStatus == True:
                    OurMarkers.append( StationMarker )
                else:
                    self.log.warning( "CreateGoogleMapsMap: StationMarker() returned False." )

            # Copy Markers.

            if Markers != None:
                for Marker in Markers:
                    OurMarkers.append( Marker )

            # Create the Java Script file.

            JavaScript =    "\n"                                                                                                \
                         + "// ****** WXFeeds JavaScript for Google Maps ****** //\n"                                           \
                         + "\n"                                                                                                 \
                         + "//  Initialize and add the map.\n"                                                                  \
                         + "\n"                                                                                                 \
                         + "function initMap() {\n"                                                                             \
                         + "\n"                                                                                                 \
                         + "    createMap( %f, %f, %d )\n" % ( Latitude, Longitude, Zoom )                                      \
                         + "}\n"                                                                                                \
                         + "\n"                                                                                                 \
                         + "function createMap( Latitude, Longitude, Zoom ) {\n"                                                \
                         + "\n"                                                                                                 \
                         + "    //  The location of interest\n"                                                                 \
                         + "\n"                                                                                                 \
                         + "    var MapCenter = { lat: Latitude, lng: Longitude };\n"                                           \
                         + "\n"                                                                                                 \
                         + "    //  The map, centered at location\n"                                                            \
                         + "\n"                                                                                                 \
                         + "    var map = new google.maps.Map(\n"                                                               \
                         + "                                     document.getElementById('WXFeedsMap')\n"                       \
                         + "                                   , {\n"                                                           \
                         + "                                         zoom: Zoom\n"                                              \
                         + "                                       , center: MapCenter\n"                                       \
                         + "                                     }\n"                                                           \
                         + "                                 );\n"                                                              \
                         + "\n"                                                                                                 \
                         + "    //  Icons\n"                                                                                    \
                         + "\n"                                                                                                 \
                         + "    var IconBase = '" + self.WXFeedsIconBase + "/';\n"                                              \
                         + "\n"                                                                                                 \
                         + "    var icons = {\n"                                                                                \
                         + "                    misc:               { icon: IconBase + 'WXFeeds-map-marker.png'}\n"             \
                         + "                  , current_planned:    { icon: IconBase + 'WXFeeds-current-planned-cone.png' }\n"  \
                         + "                  , future_planned:     { icon: IconBase + 'WXFeeds-future-planned-cone.png' }\n"   \
                         + "                  , incident:           { icon: IconBase + 'WXFeeds-incident.png' }\n"              \
                         + "                  , road_condition:     { icon: IconBase + 'WXFeeds-road-condition.png' }\n"        \
                         + "                  , weather_station:    { icon: IconBase + 'WXFeeds-weather-station.png' }\n"       \
                         + "                  , parking:            { icon: IconBase + 'parking_lot_maps.png'}\n"               \
                         + "                  , library:            { icon: IconBase + 'library_maps.png' }\n"                  \
                         + "                  , info:               { icon: IconBase + 'info-i_maps.png' }\n"                   \
                         + "                };\n"                                                                               \
                         + "\n"                                                                                                 \
                         + "    //  Markers\n"                                                                                  \
                         + "\n"

            poi = 0

            for Marker in OurMarkers:
                InfoWindowContent = ""

                if Marker.InfoWindowContent != "":
                    # Remove characters that cause JavaScript to report an error.

                    InfoWindowContent = Marker.InfoWindowContent
                    InfoWindowContent = InfoWindowContent.replace( "\n", "<br/>" )
                    InfoWindowContent = InfoWindowContent.replace( "'", "&apos;" )
#                    self.log.debug( "len(Info), Info: %d, %s" % ( len( InfoWindowContent ), InfoWindowContent ) )

                JavaScript +=   "        var marker%d = new google.maps.Marker(\n" % poi                                                \
                              + "                                                 {\n"                                                  \
                              + "                                                     position: new google.maps.LatLng( %f, %f )\n" %   \
                                                                                      (
                                                                                          Marker.Latitude
                                                                                        , Marker.Longitude
                                                                                      )                                                 \
                              + "                                                   , icon: icons.%s.icon\n" % Marker.Type              \
                              + "                                                   , map: map\n"                                       \
                              + "                                                   , title: \"%s\"\n" % Marker.Text                    \
                              + "                                                 }\n"                                                  \
                              + "                                             );\n"                                                     \
                              + "\n"

                if InfoWindowContent != "":
                    JavaScript +=   "        var infowindow%d = new google.maps.InfoWindow( { content: \'%s\' } );\n" % ( poi, InfoWindowContent )          \
                                  + "\n"                                                                                                                    \
                                  + "        marker%d.addListener( 'click', function() { infowindow%d.open( map, marker%d ); } );\n" % ( poi, poi, poi )    \
                                  + "\n"

                poi += 1

            JavaScript +=  "}\n"

            # Write the Java Script file.

            try:
                with open( JavaScriptFilePathAbsolute, 'wt' ) as f:
                    f.write( JavaScript )
            except:
               self.log.error( "CreateGoogleMapsMap: Writing Java Script File: %s Failed!" % JavaScriptFilePathAbsolute )

               ReturnStatus = False

        if ReturnStatus == False:
            HTMLFileName = ""

        return ReturnStatus, HTMLFileName, HTMLFilePathAbsolute

    def CreateGoogleMapsJavaScript(
                                    self,
                                    FilePathStemPrefix        : str,
                                    Markers                   : dict,
                                    AddStationMarker = False
                                  ) -> ( bool,         str ):
                                  #      ReturnStatus, JavaScriptFilePathRelative

        ReturnStatus                = False
        JavaScriptFilePathRelative  = ""

        assert FilePathStemPrefix != ""

        #DEBUG = True
        DEBUG = False

        # We create a Java Script file for any markers.
        #
        # If this function is called we assume Markers are not static.
        #
        # The JS file is created every time - no caching of any kind!

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsJavaScript: FilePathStemPrefix: %s" % FilePathStemPrefix )
            self.log.debug( "CreateGoogleMapsJavaScript: Markers: %d" % len( Markers ) )
            self.log.debug( "CreateGoogleMapsJavaScript: AddStationMarker: %s" % AddStationMarker )
 
        JavaScriptFileName,             \
        JavaScriptFilePathRelative,     \
        JavaScriptFilePathAbsolute,     \
        JavaScriptCacheFilePathAbsolute = self.GetMapJavaScriptPaths( FilePathStemPrefix )

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsJavaScript: JavaScriptFileName: %s" % JavaScriptFileName )
            self.log.debug( "CreateGoogleMapsJavaScript: JavaScriptFilePathRelative: %s" % JavaScriptFilePathRelative )
            self.log.debug( "CreateGoogleMapsJavaScript: JavaScriptFilePathAbsolute: %s" % JavaScriptFilePathAbsolute )
            self.log.debug( "CreateGoogleMapsJavaScript: JavaScriptCacheFilePathAbsolute: %s" % JavaScriptCacheFilePathAbsolute )

        # Create marker list.
        #
        # Our Markers could be:
        # - a marker for this station and/or 
        # - a list of markers specified by the caller.

        OurMarkers = []

        # Add a marker for this station?

        if AddStationMarker == True:
            ReturnedStatus, StationMarker = self.StationMarker()

            if DEBUG == True:
                self.log.debug( "CreateGoogleMapsJavaScript: Adding station marker." )

            assert StationMarker != None

            if ReturnedStatus == True:
                OurMarkers.append( StationMarker )
            else:
                self.log.warning( "CreateGoogleMapsJavaScript: StationMarker() returned False." )

        # Copy Markers.

        if Markers != None:
            for Marker in Markers:
                OurMarkers.append( Marker )

        # Create the Java Script file.

        JavaScript =    "\n"                                                                                                    \
                      + "// ****** WXFeeds JavaScript for Google Maps ****** //\n"                                              \
                      + "\n"                                                                                                    \
                      + "//  Create the map.\n"                                                                                 \
                      + "\n"                                                                                                    \
                      + "function createMap( Latitude, Longitude, Zoom, EncodedPolyline ) {\n"                                  \
                      + "\n"                                                                                                    \
                      + "    //  The location of interest\n"                                                                    \
                      + "\n"                                                                                                    \
                      + "    var MapCenter = { lat: Latitude, lng: Longitude };\n"                                              \
                      + "\n"                                                                                                    \
                      + "    //  The map, centered at location\n"                                                               \
                      + "\n"                                                                                                    \
                      + "    var map = new google.maps.Map(\n"                                                                  \
                      + "                                     document.getElementById('WXFeedsMap')\n"                          \
                      + "                                   , {\n"                                                              \
                      + "                                         zoom: Zoom\n"                                                 \
                      + "                                       , center: MapCenter\n"                                          \
                      + "                                     }\n"                                                              \
                      + "                                 );\n"                                                                 \
                      + "\n"                                                                                                    \
                      + "    //  Icons\n"                                                                                       \
                      + "\n"                                                                                                    \
                      + "    var IconBase = '" + self.WXFeedsIconBase + "/';\n"                                                 \
                      + "\n"                                                                                                    \
                      + "    var icons = {\n"                                                                                   \
                      + "                    misc:               { icon: IconBase + 'WXFeeds-map-marker.png'}\n"                \
                      + "                  , current_planned:    { icon: IconBase + 'WXFeeds-current-planned-cone.png' }\n"     \
                      + "                  , future_planned:     { icon: IconBase + 'WXFeeds-future-planned-cone.png' }\n"      \
                      + "                  , incident:           { icon: IconBase + 'WXFeeds-incident.png' }\n"                 \
                      + "                  , road_condition:     { icon: IconBase + 'WXFeeds-road-condition.png' }\n"           \
                      + "                  , weather_station:    { icon: IconBase + 'WXFeeds-weather-station.png' }\n"          \
                      + "                  , parking:            { icon: IconBase + 'parking_lot_maps.png'}\n"                  \
                      + "                  , library:            { icon: IconBase + 'library_maps.png' }\n"                     \
                      + "                  , info:               { icon: IconBase + 'info-i_maps.png' }\n"                      \
                      + "                };\n"                                                                                  \
                      + "\n"                                                                                                    \
                      + "    //  Markers\n"                                                                                     \
                      + "\n"

        poi = 0

        for Marker in OurMarkers:
            InfoWindowContent = ""

            if Marker.InfoWindowContent != "":
                # Remove characters that cause JavaScript to report an error.

                InfoWindowContent = Marker.InfoWindowContent
                InfoWindowContent = InfoWindowContent.replace( "\n", "<br/>" )
                InfoWindowContent = InfoWindowContent.replace( "'", "&apos;" )
#                self.log.debug( "len(Info), Info: %d, %s" % ( len( InfoWindowContent ), InfoWindowContent ) )

            JavaScript +=   "    var marker%d = new google.maps.Marker(\n" % poi                                                \
                          + "                                           {\n"                                                    \
                          + "                                             position: new google.maps.LatLng( %f, %f ),\n" %      \
                                                                                (
                                                                                  Marker.Latitude,
                                                                                  Marker.Longitude
                                                                                )                                               \
                          + "                                             icon: icons.%s.icon,\n" % Marker.Type                 \
                          + "                                             map: map,\n"                                          \
                          + "                                             title: \"%s\"\n" % Marker.Text                        \
                          + "                                           }\n"                                                    \
                          + "                                         );\n"                                                     \
                          + "\n"

            if InfoWindowContent != "":
                JavaScript +=   "    var infowindow%d = new google.maps.InfoWindow( { content: \'%s\' } );\n" %                     \
                                                                                        (
                                                                                          poi,
                                                                                          InfoWindowContent
                                                                                        )                                           \
                              + "\n"                                                                                                \
                              + "    marker%d.addListener( 'click', function() { infowindow%d.open( map, marker%d ); } );\n" %      \
                                                                                                         ( poi, poi, poi )          \
                              + "\n"

            poi += 1

        JavaScript +=   "    //  Encoded Polyline\n"                                                                                    \
                      + "\n"                                                                                                            \
                      + "    if ( EncodedPolyline !== \"\" ) {\n"                                                                       \
                      + "        const polyline = new google.maps.Polyline(\n"                                                          \
                      + "                                   {\n"                                                                        \
                      + "                                     path: google.maps.geometry.encoding.decodePath( EncodedPolyline ),\n"     \
                      + "                                     strokeColor: '#FF0000',\n"                                                \
                      + "                                     strokeOpacity: 0.8,\n"                                                    \
                      + "                                     strokeWeight: 2\n"                                                        \
                      + "                                   }\n"                                                                        \
                      + "                                                 );\n"                                                         \
                      + "\n"                                                                                                            \
                      + "        polyline.setMap( map );\n"                                                                             \
                      + "    }\n"

        JavaScript +=  "}\n"

        # Write the Java Script file.

        try:
            with open( JavaScriptFilePathAbsolute, 'wt' ) as f:
                f.write( JavaScript )
            ReturnStatus = True
        except:
           self.log.error( "CreateGoogleMapsMap: Writing Java Script File: %s Failed!" % JavaScriptFilePathAbsolute )

           ReturnStatus = False

        if ReturnStatus == False:
            JavaScriptFilePathRelative = ""

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsMap: Returning: %s, %s." % ( ReturnStatus, JavaScriptFilePathRelative ) )

        return ReturnStatus, JavaScriptFilePathRelative

    def CreateGoogleMapsMapHTML(
                                 self,
                                 FilePathStemPrefix         = "",
                                 ForceStaleFile             = False,
                                 MapHeader                  = "",
                                 Latitude                   = BAD_LATITUDE,
                                 Longitude                  = BAD_LONGITUDE,
                                 JavaScriptFilePathRelative = "",
                                 Zoom                       = 12,
                                 EncodedPolyline            = ""
                               ) -> ( bool,         str,                  str ):
                               #      ReturnStatus, HTMLFileName, HTMLFilePathAbsolute

        ReturnStatus = False

        assert FilePathStemPrefix != ""
        assert MapHeader != ""
        assert Latitude != BAD_LATITUDE
        assert Longitude != BAD_LONGITUDE
        assert JavaScriptFilePathRelative != ""

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsMapHTML: FilePathStemPrefix: %s" % FilePathStemPrefix )
            self.log.debug( "CreateGoogleMapsMapHTML: ForceStaleFile: %s" % ForceStaleFile )
            self.log.debug( "CreateGoogleMapsMapHTML: MapHeader: %s" % MapHeader )
            self.log.debug( "CreateGoogleMapsMapHTML: Latitude: %s" % Latitude )
            self.log.debug( "CreateGoogleMapsMapHTML: Longitude: %s" % Longitude )
            self.log.debug( "CreateGoogleMapsMapHTML: JavaScriptFilePathRelative: %s" % JavaScriptFilePathRelative )
            self.log.debug( "CreateGoogleMapsMapHTML: Zoom: %s" % Zoom )

        HTMLFileName, HTMLFilePathRelative, HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute =       \
                self.GetMapHTMLPaths( FilePathStemPrefix )

        if DEBUG == True:
            self.log.debug( "CreateGoogleMapsMapHTML: HTMLFileName: %s" % HTMLFileName )
            self.log.debug( "CreateGoogleMapsMapHTML: HTMLFilePathRelative: %s" % HTMLFilePathRelative )
            self.log.debug( "CreateGoogleMapsMapHTML: HTMLFilePathAbsolute: %s" % HTMLFilePathAbsolute )
            self.log.debug( "CreateGoogleMapsMapHTML: HTMLCacheFilePathAbsolute: %s" % HTMLCacheFilePathAbsolute )

        HTMLIsFileStale = False

        if ForceStaleFile == True:
            HTMLIsFileStale = True
        else:
            HTMLIsFileStale = self.IsFileStale( HTMLFilePathAbsolute, HTMLCacheFilePathAbsolute, self.WXFeedsServerUse )

        if DEBUG == True:
            if HTMLIsFileStale == True:
                self.log.debug( "CreateGoogleMapsMapHTML: HTML file is stale." )
            else:
                self.log.debug( "CreateGoogleMapsMapHTML: HTML file is NOT stale." )

        if HTMLIsFileStale == False:
            # HTML is NOT stale so we don't need to create it.

            return True, HTMLFileName, HTMLFilePathAbsolute

        # The HTML file is stale so create it.

        GoogleMapsMapsURL = GOOGLE_MAPS_MAPS_SERVICE_URL

        if self.GoogleMapsMapsServerUse == True:
            GoogleMapsMapsURL = GoogleMapsMapsURL.replace( "?", "%3F" )
            GoogleMapsMapsURL = GoogleMapsMapsURL.replace( "&", "%26")

            GoogleMapsMapsURL = self.WXFeedsServerURL + "/WXFeeds?Service=GoogleMapsMaps&url=" + GoogleMapsMapsURL
            GoogleMapsMapsURL += "&accessToken=" + self.WXFeedsServerAccessToken
        else:
            assert self.GoogleMapsAPIKey != ""

            GoogleMapsMapsURL += "&key=" + self.GoogleMapsAPIKey

        HTML =   "<!DOCTYPE html>\n"                                                                                \
               + "<html lang=\"en-CA\">\n"                                                                          \
               + "    <head>\n"                                                                                     \
               + "        <title>%s</title>\n" % MapHeader                                                          \
               + "        <script>\n"                                                                               \
               + "            function initMap() {\n"                                                               \
               + "\n"                                                                                               \
               + "                const EncodedPolyline = \"%s\"\n" % EncodedPolyline                               \
               + "\n"                                                                                               \
               + "                createMap( %f, %f, %d, EncodedPolyline )\n" % ( Latitude, Longitude, Zoom )       \
               + "            }\n"                                                                                  \
               + "        </script>\n"                                                                              \
               + "        <script src=\'%s\'></script>\n" % JavaScriptFilePathRelative                              \
               + "        <style>\n"                                                                                \
               + "            /* Set the size of the div element that contains the map */\n"                        \
               + "            #WXFeedsMap {\n"                                                                      \
               + "                height: 400px;  /* The height is 400 pixels */\n"                                 \
               + "                width: 100%;    /* The width is the width of the web page */\n"                   \
               + "            }\n"                                                                                  \
               + "        </style>\n"                                                                               \
               + "    </head>\n"                                                                                    \
               + "    <body>\n"                                                                                     \
               + "        <h1>%s</h1>\n" % MapHeader                                                                \
               + "        <div id=\"WXFeedsMap\"></div>\n"                                                          \
               + "        <script defer src=\"%s\">\n" % GoogleMapsMapsURL                                          \
               + "        </script>\n"                                                                              \
               + "    </body>\n"                                                                                    \
               + "</html>\n"

        # Write the HTML file.

        ReturnStatus = True     # Assume success.

        try:
            with open( HTMLFilePathAbsolute, 'w' ) as f:
                f.write( HTML )
        except:
           self.log.error( "CreateGoogleMapsMapHTML: Writing Google Map HTML File: %s Failed!" % HTMLFilePathAbsolute )

           ReturnStatus = False

#        if ReturnStatus == True:
#            # Write the cache file.
#
#            try:
#                with open( HTMLCacheFilePathAbsolute, 'w' ) as f:
#                    f.write( "" )
#            except:
#                self.log.error( "CreateGoogleMapsMapHTML: Writing HTML Cache File: %s Failed!" % HTMLCacheFilePathAbsolute )
#
#                ReturnStatus = False

        if ReturnStatus == False:
            HTMLFilePathRelative = ""

        return ReturnStatus, HTMLFileName, HTMLFilePathAbsolute

    def StationMarker( self ) -> ( bool, GoogleMapsMarker ):

        ReturnStatus = False    # Assume failure!

        Marker = GoogleMapsMarker(
                                   "",
                                   BAD_LATITUDE,
                                   BAD_LONGITUDE,
                                   "",
                                   ""
                                 )

        if self.StationLocation != "" and               \
           self.StationLatitude != BAD_LATITUDE and     \
           self.StationLongitude != BAD_LONGITUDE:

#            self.log.debug( "StationMarker: StationLocation: %s" % self.StationLocation )
#            self.log.debug( "StationMarker: StationURL: %s" % self.StationURL )

            InfoWindowContent = ""
            if self.StationURL != "":
                InfoWindowContent = "<a href=\"%s\">%s</a>" % ( self.StationURL, self.StationLocation )
            else:
                self.log.warning( "StationMarker: Station URL NOT set!" )

            Marker = GoogleMapsMarker(
                                       "weather_station",
                                       self.StationLatitude,
                                       self.StationLongitude,
                                       self.StationLocation,
                                       InfoWindowContent
                                     )

            ReturnStatus = True
        else:
            self.log.warning( "StationMarker: Station location and/or position NOT set!" )

        return ReturnStatus, Marker

    def AirPortCodesGetAirportInfo(
                                    self,
                                    IATACode : str
                                  ) -> ( bool,         str,  float,    float,     float,     str ):
                                  #      ReturnStatus, Name, Latitude, Longitude, Elevation, Website

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "AirPortCodesGetAirportInfo: In AirPortCodesGetAirportInfo(). . ." )

        Name        = "!!Not Found By Air-Port-Codes!!"
        Latitude    = BAD_LATITUDE
        Longitude   = BAD_LONGITUDE
        Elevation   = BAD_ELEVATION
        Website     = ""

        AirPortCodesURL = AIR_PORT_CODES_AIRPORT_INFO_SERVICE_URL
        AirPortCodesURL += "?iata=" + IATACode

        if self.AirPortCodesServerUse == False:
            assert self.AirPortCodesAPIKey != ""
            assert self.AirPortCodesAPISecret != ""

            if AirPortCodesURL.find( "?" ) < 0:
                AirPortCodesURL += "?"
            else:
                AirPortCodesURL += "&"

            AirPortCodesURL += "key=" + self.AirPortCodesURL
            AirPortCodesURL += "&secret=" + self.AirPortCodesAPISecret

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        AirportInfoData = {}

        if DEBUG == True:
            self.log.debug( "AirPortCodesGetAirportInfo: AirPortCodesURL: %s" % AirPortCodesURL )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesLocalCacheUse: %s" % self.AirPortCodesLocalCacheUse )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesLocalCacheExpiry_minutes: %d" % self.AirPortCodesLocalCacheExpiry_minutes )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesLocalCacheForceMiss: %s" % self.AirPortCodesLocalCacheForceMiss )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesServerUse: %s" % self.AirPortCodesServerUse )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesRemoteCacheUse: %s" % self.AirPortCodesRemoteCacheUse )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesRemoteCacheExpiry_minutes: %d" % self.AirPortCodesRemoteCacheExpiry_minutes )
            self.log.debug( "AirPortCodesGetAirportInfo: self.AirPortCodesRemoteCacheForceMiss: %s" % self.AirPortCodesRemoteCacheForceMiss )

        try:
            ReturnedStatus, CacheHit, StatusCode, AirportInfoData = self.GetEndpointData(
                                                                                          EndpointURL               = AirPortCodesURL,
                                                                                          ContentType               = CONTENT_JSON,
                                                                                          ValidateDataFunction      = self.AirPortCodesGetAirportInfoValidateData,
                                                                                          Service                   = "AirPortCodesAirportInfo",
                                                                                          CacheKey                  = IATACode,
                                                                                          LocalCacheUse             = self.AirPortCodesLocalCacheUse,
                                                                                          LocalCacheExpiry_minutes  = self.AirPortCodesLocalCacheExpiry_minutes,
                                                                                          LocalCacheForceMiss       = self.AirPortCodesLocalCacheForceMiss,
                                                                                          ServerUse                 = self.AirPortCodesServerUse,
                                                                                          RemoteCacheUse            = self.AirPortCodesRemoteCacheUse,
                                                                                          RemoteCacheExpiry_minutes = self.AirPortCodesRemoteCacheExpiry_minutes,
                                                                                          RemoteCacheForceMiss      = self.AirPortCodesRemoteCacheForceMiss
                                                                                        )
        except:
            self.log.error( "AirPortCodesGetAirportInfo: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "AirPortCodesGetAirportInfo: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" %
                                ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "AirPortCodesGetAirportInfo: GetEndpointData() returned False for Station: %s" % IATACode )

            ReturnStatus = False
        else:
            # We should have valid Air-Port-Codes information (in json format).

            Airport = AirportInfoData[ 'airport' ]

            Name        = Airport[ 'name' ]
            Latitude    = float( Airport[ 'latitude' ] )
            Longitude   = float( Airport[ 'longitude' ] )
            Elevation   = float( Airport[ 'elevation' ] )
            Website     = Airport[ 'website' ]

            ReturnStatus = True

        if DEBUG == True:
            self.log.debug( "AirPortCodesGetAirportInfo: Status = %d, Name = %s, Latitude = %f, Longitude = %f, Elevation = %f, Website = %s" %  \
                            ( 
                              ReturnStatus,
                              Name,
                              Latitude,
                              Longitude,
                              Elevation,
                              Website
                            )
                          )
            self.log.debug( "AirPortCodesGetAirportInfo: Exiting AirPortCodesGetAirportInfo()." )

        return ReturnStatus, Name, Latitude, Longitude, Elevation, Website

    def AirPortCodesGetAirportInfoValidateData(
                                                self,
                                                StatusCode        : str,
                                                AirportInfoData   : dict
                                              ) -> bool:
                                              #    ReturnStatus

        ReturnStatus = False

        if AirportInfoData[ 'status' ] != None and          \
           AirportInfoData[ 'status' ] == True and          \
           AirportInfoData[ 'statusCode' ] != None and      \
           AirportInfoData[ 'statusCode' ] == 200 and       \
           AirportInfoData[ 'airport' ] != None and         \
           AirportInfoData[ 'term' ] != None:
#            self.log.debug( "AirPortCodesGetAirportInfoValidateData:AirportInfoData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "AirPortCodesGetAirportInfoValidateData:AirportInfoData is NOT valid!" )

            pass

        return ReturnStatus

    def ECCityAlert(
                     self,
                     City              : str,
                     FeedURL           : str,
                     Latitude          : float,
                     Longitude         : float,
                     PositionSpecified : bool,
                     Zoom              : int
                   ) -> bool:

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        AlertFeedURL = ""

        if FeedURL != None and    \
           FeedURL != "":
            TargetString = "city"
            x = FeedURL.find(TargetString)

            if x >= 0:
                AlertFeedURL = FeedURL[0:x] + "warning" + FeedURL[x + len(TargetString):len(FeedURL)]
#                self.log.debug("ECCityAlert: FeedURL: %s" % FeedURL)
#                self.log.debug("ECCityAlert: AlertFeedURL: %s" % AlertFeedURL)
            else:
                # Cannot get alert URL from FeedURL - we're done!

                self.log.warning( "ECCityAlert: Cannot get Alert URL for '%s' from '%s'!" % ( City, FeedURL ) )

                ReturnStatus = False
                return ReturnStatus

        else:
            self.log.warning( "ECCityAlert: FeedURL for '%s' was not specified!" % City )

            ReturnStatus = False
            return ReturnStatus

        assert AlertFeedURL != ""

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        ECCityAlertData = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            ECCityAlertData = self.GetEndpointData(
                                                    EndpointURL               = AlertFeedURL,
                                                    ContentType               = CONTENT_XML,
                                                    ValidateDataFunction      = self.ECCityAlertValidateData,
                                                    Service                   = "ECCityAlert",
                                                    CacheKey                  = City,
                                                    LocalCacheUse             = self.ECCityAlertsLocalCacheUse,
                                                    LocalCacheExpiry_minutes  = self.ECCityAlertsLocalCacheExpiry_minutes,
                                                    LocalCacheForceMiss       = self.ECCityAlertsLocalCacheForceMiss,
                                                    ServerUse            = self.ECCityAlertsServerUse,
                                                    RemoteCacheUse            = self.ECCityAlertsRemoteCacheUse,
                                                    RemoteCacheExpiry_minutes = self.ECCityAlertsRemoteCacheExpiry_minutes,
                                                    RemoteCacheForceMiss      = self.ECCityAlertsRemoteCacheForceMiss
                                                  )
        except:
            self.log.error( "ECCityAlert: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "ECCityAlert: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "ECCityAlert: GetEndpointData() returned False for Location: %s" % City )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "ECCityAlert: Received %d entries for '%s'." % ( len( ECCityAlertData.entries ), City ) )

#            EntryIndex = 0
#
#            for Entry in ReturnedFeed.entries:
#                self.log.debug("EC Alert Entry %d: Entry.tags[0].term = %s" % (EntryIndex, Entry.tags[0].term))
#                self.log.debug("EC Alert Entry %d: %s" % (EntryIndex, Entry))
#                entryIndex += 1

        assert len( ECCityAlertData.entries ) > 0

        MapHTMLFileName = ""

        if self.GoogleMapsMapsUse == True:
            # Get links to the alert.

            LinkToAlert = ""

            for Entry in ECCityAlertData.entries:
                for Tag in Entry.tags:
                    if Tag.term == WARNINGS_AND_WATCHES:
                        LinkToAlert = Entry.link
                        break

#            self.log.debug( "LinkToAlert: %s" % LinkToAlert )                            

            MarkerInfoWindowContent = ""
            if LinkToAlert != "":
                MarkerInfoWindowContent += "<a href=\"%s\">Alert</a>" % LinkToAlert

            ReturnedStatus,         \
            MapHTMLFileName,        \
            MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                FilePathStemPrefix        = "ECCityAlert-" + City,
                                                                ForceStaleFiles           = not CacheHit,
                                                                Location                  = City,
                                                                Latitude                  = Latitude,
                                                                Longitude                 = Longitude,
                                                                PositionSpecified         = PositionSpecified,
                                                                Zoom                      = Zoom,
                                                                MarkerType                = "misc",
                                                                MarkerInfoWindowContent   = MarkerInfoWindowContent,
                                                                DisplayStationMarker      = True
                                                              )
            if ReturnedStatus == True:
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                MapHTMLFilePathRelative = ""

        HTML = self.GetDivTag(ENVIRONMENT_CANADA, "Alert", City)
        HTML += ("<p class=\"WXFeedsP1\"><b>%s</b>\n" % City)
#        HTML += ( "<p class=\"WXFeedsP1\">\n" )
        self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
#            HTML += "&nbsp;\n<a href=\"%s\"><span class=\"icon-google-map\"></span></a>\n" % MapHTMLFilePathRelative

            HTML = "&nbsp;<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        HTML = "</p>\n"
        self.WriteGeneratedFiles( HTML )

        # There could be multipe warnings and/or watches in the returned feed.

        for Entry in ECCityAlertData.entries:
#            self.log.debug("EC Alerts Entry: %s" % Entry)
#            self.log.debug("EC Alerts Entry.tags[0].term: %s" % Entry.tags[0].term)

            if Entry.tags[0].term == WARNINGS_AND_WATCHES:
#                self.log.debug("EC Alerts \"WARNINGS_AND_WATCHES\" Entry")
#                self.log.debug("EC Alerts Entry: %s" % Entry)

                if Entry.title.find(NO_WATCHES_OR_WARNINGS_IN_EFFECT) >= 0:
                    # There are no watches or warnings in effect.

#                    self.log.debug("There are NO watches or warnings in effect: %s" % Entry.title)

                    HTML =    "<p class=\"WXFeedsP2\">"                                                     \
                            +     "<b>"                                                                     \
                            +         "<span class=\"WXFeedsNoWatchesOrWarnings\">"                         \
                            +             "&nbsp;&nbsp;"                                                    \
                            +                 "No Alerts in effect"                                         \
                            +             "&nbsp;&nbsp;"                                                    \
                            +         "</span>"                                                             \
                            +     "</b>"                                                                    \
                            +     "&nbsp;\n"                                                                \
                            +     "<a href=\"%s\">Public Alerts</a>\n" % Entry.link                         \
                            +     "&nbsp;\n"                                                                \
                            +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link     \
                            + "</p>\n"
                else:
                    # There are watches or warnings in effect.

#                    self.log.debug("There are watches or warnings in effect: %s" % Entry.title)

                    # Get the alert without the city.

                    Alert = ""

                    x = Entry.title.find(", ")
                    if x >= 0:
                        Alert = Entry.title[0:x]

                    Alert = Alert.strip()
                    assert Alert != ""

                    # Warning, Watch, Special Air Quality Statement or Special Weather Statement?

                    AlertClass = ""

                    if Alert.find( WATCH ) >= 0:
                        # Watch

                        AlertClass = "WXFeedsWatch"
                    elif Alert.find( WARNING ) >= 0:
                        # Warning

                        AlertClass = "WXFeedsWarning"
                    elif Alert.find( SPECIAL_AIR_QUALITY_STATEMENT ) >= 0:
                        # Special Air Quality

                        AlertClass = "WXFeedsSpecialAirQualityStatement"
                    elif Alert.find( SPECIAL_WEATHER_STATEMENT ) >= 0:
                        # Special Weather Statement

                        AlertClass = "WXFeedsSpecialWeatherStatement"
                    else:
                        # Don't know!

                        self.log.error( "ECCityAlert: Unknown EC Alert: %s" % Entry.title )

                        AlertClass = "WXFeedsUnkownAlert"

                    assert AlertClass != ""

                    HTML =   "<p class=\"WXFeedsP2\">"                                                      \
                           +     "<b>"                                                                      \
                           +         "<span class=\"%s\">" % AlertClass                                     \
                           +             "&nbsp;&nbsp;"                                                     \
                           +             "<span class=\"icon-warning-sign\"></span>"                        \
                           +             "&nbsp;&nbsp;"                                                     \
                           +               Alert                                                            \
                           +             "&nbsp;&nbsp;"                                                     \
                           +             "<span class=\"icon-warning-sign\"></span>"                        \
                           +             "&nbsp;&nbsp;"                                                     \
                           +         "</span>"                                                              \
                           +     "</b>"                                                                     \
                           +     "&nbsp;\n"                                                                 \
                           +     "<a href=\"%s\">Full Alert</a>\n" % Entry.link                             \
                           +     "&nbsp;\n"                                                                 \
                           +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link      \
                           +     "<br/>\n"                                                                  \
                           +     Entry.summary                                                              \
                           + "</p>\n"
            else:
                self.log.error("Unexpected Entry Type Returned In EC Alerts Feed: %s" % Entry.tags[0].term)

        HTML += "</div>\n"
        self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        ReturnedFeed    = None
        Markers         = None

        return ReturnStatus

    def ECCityAlertValidateData(
                                 self,
                                 StatusCode            : int,
                                 ECCityAlertsData      : str
                               ) -> bool:
                               #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "ECCityAlertsValidateData: ECCityAlertsData: %s" % ECCityAlertsData )

        if StatusCode == 200 and                    \
           ECCityAlertsData.entries != None:
#            self.log.debug( "ECCityAlertsData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "ECCityAlertsData is NOT valid!" )

            pass

        return ReturnStatus

    def ECCityWeather(
                       self,
                       City                 : str,
                       FeedURL              : str,
                       InlineForecast       : bool,
                       Latitude             : float,
                       Longitude            : float,
                       PositionSpecified    : bool,
                       Zoom                 : int
                     ) -> bool:
                     #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if FeedURL == None or   \
           FeedURL == "":
            self.log.warning( "ECCityWeather: FeedURL for '%s' was not specified!" % City )

            ReturnStatus = False
            return ReturnStatus

        ReturnedStatus      = False
        CacheHit            = False
        StatusCode          = -1
        ECCityWeatherData   = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            ECCityWeatherData = self.GetEndpointData(
                                                      EndpointURL               = FeedURL,
                                                      ContentType               = CONTENT_XML,
                                                      ValidateDataFunction      = self.ECCityWeatherValidateData,
                                                      Service                   = "ECCityWeather",
                                                      CacheKey                  = City,
                                                      LocalCacheUse             = self.ECCityWeatherLocalCacheUse,
                                                      LocalCacheExpiry_minutes  = self.ECCityWeatherLocalCacheExpiry_minutes,
                                                      LocalCacheForceMiss       = self.ECCityWeatherLocalCacheForceMiss,
                                                     ServerUse            = self.ECCityWeatherServerUse,
                                                      RemoteCacheUse            = self.ECCityWeatherRemoteCacheUse,
                                                      RemoteCacheExpiry_minutes = self.ECCityWeatherRemoteCacheExpiry_minutes,
                                                      RemoteCacheForceMiss      = self.ECCityWeatherRemoteCacheForceMiss
                                                    )
        except:
            self.log.error( "ECCityWeather: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "ECCityWeather: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "ECCityWeather: GetEndpointData() returned False for Location: %s" % City )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "ECCityWeather: Received %d entries for '%s'." % ( len( ECCityWeatherData.entries ), City ) )

        assert len( ECCityWeatherData.entries ) > 0

        MapHTMLFileName = ""

        if self.GoogleMapsMapsUse == True:
            # Get links to any alert (even if it's a non alert) and to the forecast.

            LinkToAlert     = ""
            LinkToForecast  = ""

            for Entry in ECCityWeatherData.entries:
                for Tag in Entry.tags:
                    if Tag.term == WARNINGS_AND_WATCHES:
                        LinkToAlert = Entry.link
                    else:
                        if Tag.term == WEATHER_FORECASTS:
                            LinkToForecast = Entry.link

                    if LinkToAlert != "" and    \
                       LinkToForecast != "":
                        break

#            self.log.debug( "LinkToAlert: %s" % LinkToAlert )                            
#            self.log.debug( "LinkToForecast: %s" % LinkToForecast )                            

            MarkerInfoWindowContent = ""
            if LinkToAlert != "":
                MarkerInfoWindowContent += "<a href=\"%s\">Alert</a>" % LinkToAlert
            if LinkToForecast != "":
                if MarkerInfoWindowContent != "":
                    MarkerInfoWindowContent += "<br/><br/>"

                MarkerInfoWindowContent += "<a href=\"%s\">Forecast</a>" % LinkToForecast
        
            ReturnedStatus,         \
            MapHTMLFileName,        \
            MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                FilePathStemPrefix      = "ECCityWeather-" + City,
                                                                ForceStaleFiles         = not CacheHit,
                                                                Location                = City,
                                                                Latitude                = Latitude,
                                                                Longitude               = Longitude,
                                                                PositionSpecified       = PositionSpecified,
                                                                Zoom                    = Zoom,
                                                                MarkerType              = "misc",
                                                                MarkerInfoWindowContent = MarkerInfoWindowContent,
                                                                DisplayStationMarker    = True,
                                                                MarkersAreStatic        = False
                                                              )
            if ReturnedStatus == True:
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                MapHTMLFileName = ""

        AlertSummarys               = ""
        CurrentConditionsSummary    = ""
        FullWeatherReportURL        = ""

        CurrentConditionsObserved = False

        ForecastIssuedDateTime  = ""
        Forecast                = ""

        HTML = self.GetDivTag(ENVIRONMENT_CANADA, "Weather", City)
        HTML += "<p class=\"WXFeedsP1\"><b>%s</b>\n" % City
#            HTML += "<p class=\"WXFeedsP1\">\n"
        self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
#                HTML += "&nbsp;\n<a href=\"%s\"><span class=\"icon-google-map\"></span></a>\n" % MapHTMLFilePathRelative

            HTML = "&nbsp;<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        HTML = "</p>\n"
        self.WriteGeneratedFiles( HTML )

        for Entry in ECCityWeatherData.entries:
#                self.log.debug("EC Weather Entry: %s" % Entry)

            AlertClass = ""

            if Entry.tags[0].term == WARNINGS_AND_WATCHES:
#                    self.log.debug("W&W Entry: %s" % Entry)

                if Entry.title.find(NO_WATCHES_OR_WARNINGS_IN_EFFECT) >= 0:
                    # There are no watches or warnings in effect.

#                        self.log.debug("There are NO watches or warnings in effect: %s" % Entry.title)

                    AlertSummarys +=   "<p class=\"WXFeedsP2\">"                                                       \
                                        +   "<b>"                                                                      \
                                        +     "<span class=\"WXFeedsNoWatchesOrWarnings\">"                            \
                                        +       "&nbsp;&nbsp;"                                                         \
                                        +       "No Alerts in effect"                                                  \
                                        +       "&nbsp;&nbsp;"                                                         \
                                        +     "</span>"                                                                \
                                        +   "</b>\n"                                                                   \
                                        +   "&nbsp;\n"                                                                 \
                                        +   "<a href=\"%s\">Public Alerts</a>\n" % Entry.link                          \
                                        +   "&nbsp;\n"                                                                 \
                                        +   "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link      \
                                        + "</p>\n"
                else:
                    # There are watches or warnings in effect.

#                        self.log.debug("There are watches or warnings in effect: %s" % Entry.title)

                    # Get the alert without the city.

                    Alert = ""

                    x = Entry.title.find(", ")
                    if x >= 0:
                        Alert = Entry.title[0:x]

                    Alert = Alert.strip()

                    assert Alert != ""

                    # Warning, Watch, Special Air Quality Statement or Special Weather Statement?

                    AlertClass = ""

                    if Alert.find( WATCH ) >= 0:
                        # Watch

                        AlertClass = "WXFeedsWatch"
                    elif Alert.find( WARNING ) >= 0:
                        # Warning.

                        AlertClass = "WXFeedsWarning"
                    elif Alert.find( SPECIAL_AIR_QUALITY_STATEMENT ) >= 0:
                        # Special Air Quality Statement.

                        AlertClass = "WXFeedsSpecialAirQualityStatement"
                    elif Alert.find( SPECIAL_WEATHER_STATEMENT ) >= 0:
                        # Special Weather Statement.

                        AlertClass = "WXFeedsSpecialWeatherStatement"
                    else:
                        # Don't know!

                        self.log.error( "ECCityWeather: Unknown EC Alert: %s" % Entry.title )

                        AlertClass = "WXFeedsUnkownAlert"

                    assert AlertClass != ""

                    AlertSummarys +=   "<p class=\"WXFeedsP2\">\n"                                                         \
                                        +      "<b>\n"                                                                     \
                                        +          "<span class=\"%s\">" % AlertClass                                      \
                                        +              "&nbsp;&nbsp;"                                                      \
                                        +              "<span class=\"icon-warning-sign\"></span>"                         \
                                        +              "&nbsp;&nbsp;\n"                                                    \
                                        +              Alert                                                               \
                                        +              "\n&nbsp;&nbsp;"                                                    \
                                        +              "<span class=\"icon-warning-sign\"></span>"                         \
                                        +              "&nbsp;&nbsp;"                                                      \
                                        +          "</span>\n"                                                             \
                                        +      "</b>\n"                                                                    \
                                        +      "&nbsp;\n"                                                                  \
                                        +      "<a href=\"%s\">Full Alert</a>\n" % Entry.link                              \
                                        +      "&nbsp;\n"                                                                  \
                                        +      "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link       \
                                        +      "<br/>\n"                                                                   \
                                        +      Entry.summary                                                               \
                                        +  "</p>\n"
                continue

            if Entry.tags[0].term == CURRENT_CONDITIONS:
#                    self.log.debug("EC Weather Current Conditions Entry: %s" % Entry)

                CurrentConditionsObserved = True
                CurrentConditionsSummary = Entry.summary
                FullWeatherReportURL = Entry.link
                continue

            if Entry.tags[0].term == WEATHER_FORECASTS:
#                    self.log.debug("Forecast Entry: \"%s\"" % Entry)

                # If there is no Current Conditons entry there won't yet be 
                # a full weather report URL, so get it from a Weather Forecast 
                # entry.

                if CurrentConditionsObserved == False:
                    FullWeatherReportURL = Entry.link

                # And get the forecast issued date/time while we have a forecast entry.

                ForecastIssuedDateTime = ""

                TargetString = "Forecast issued "
                x = Entry.summary.find(TargetString)
                assert x >= 0

                if x >= 0:
                    ForecastIssuedDateTime = Entry.summary[x:len(Entry.summary)]
#                        self.log.debug("ForecastIssuedDateTime:\"%s\"" % ForecastIssuedDateTime)
                else:
                    self.log.error("Didn't find \"%s\" in forecast entry: %s" %(TargetString, Entry.summary))

                assert ForecastIssuedDateTime != ""

                break    # We're done after receiving a forecast entry.

        HTML =   "<p class=\"WXFeedsP2\"><b>Current Conditions</b>"                                         \
                +     "&nbsp;\n"                                                                            \
                +     "<a href=\"%s\">Full Report</a>\n" % FullWeatherReportURL                             \
                +     "&nbsp;\n"                                                                            \
                +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % FullWeatherReportURL       \
                + "</p>\n"                                                                                  \
                + AlertSummarys

        if CurrentConditionsObserved == True:
            HTML += "<p class=\"WXFeedsP2\">" + CurrentConditionsSummary + "</p>\n"
        else:
            HTML += "<p class=\"WXFeedsP2\">Current Conditions Not Observed</p>\n"

        if InlineForecast == True:

            # Display Environment Canada's forecast for this city.

            # Get the URL for the hourly forecast.

#                self.log.debug("FeedURL: \"%s\"" % FeedURL)

            HourlyForecastURL = ""

            TargetString = "/rss/city/"

            x = FeedURL.find(TargetString)
            assert x >= 0

            if x >= 0:
                CityID = FeedURL[x+len(TargetString):len(FeedURL)]
                assert CityID != ""
#                    self.log.debug("CityID: \"%s\"" % CityID)

                TargetString = "_e.xml"
                y = CityID.find(TargetString)
                assert y >= 0

                if y >= 0:
                    CityID = CityID[0:y]
                    assert CityID != ""
#                       self.log.debug("CityID: \"%s\"" % CityID)

                HourlyForecastURL = FeedURL[0:x] + "/forecast/hourly/" + CityID + "_metric_e.html"
#                    self.log.debug("HourlyForecastURL: \"%s\"" % HourlyForecastURL)
            else:
                self.log.error("Unable To Get EC Hourly Forecast URL! FeedURL = %s" % FeedURL)

            assert ForecastIssuedDateTime != ""
            assert HourlyForecastURL != ""

            # Put the forecast into a 2 column table.

            HTML += "<p class=\"WXFeedsP2\"><b>" + ForecastIssuedDateTime + "</b>"

            if HourlyForecastURL != "":
                HTML +=   "&nbsp;\n"                                                                      \
                        + "<a href=\"%s\">Hourly Forecast</a>\n" % HourlyForecastURL                      \
                        + "&nbsp;\n"                                                                      \
                        + "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % HourlyForecastURL

            HTML += "</p>\n"

            HTML +=   "<table class=\"WXFeedsT2\">\n<tbody>\n"

#                        +     "<thead>"                                           \
#                        +         "<tr>"                                          \
#                        +             "<th></th>"    \
#                        +             "<th></th>"    \
#                        +         "</tr>"                                         \
#                        +     "</thead>"                                          \

#                       +             "<th style=\"vertical-align:top\"></th>"    \
#                       +             "<th style=\"vertical-align:top\"></th>"    \

            Forecast = ""

            for Entry in ECCityWeatherData.entries:
#                    self.log.debug("EC Weather Entry: %s" % Entry)

                if Entry.tags[0].term == WARNINGS_AND_WATCHES:
#                        self.log.debug("W&W Entry: %s" % Entry)

                    if Entry.title.find(NO_WATCHES_OR_WARNINGS_IN_EFFECT) >= 0:
                        # There are no watches or warnings in effect - do nothing.

                        continue
                    else:
                        # There are watches or warnings in effect.

                        # Get the alert without the city.

                        Alert = ""

                        x = Entry.title.find(", ")
                        if x >= 0:
                            Alert = Entry.title[0:x]

                        Alert = Alert.strip()

                        assert Alert != ""

                        # Warning, Watch, Special Air Quality Statement or Special Weather Statement?

                        if Alert.find( WATCH ) >= 0:
                            # Watch

                            AlertClass = "WXFeedsWatch"
                        elif Alert.find( WARNING ) >= 0:
                            # Warning.

                            AlertClass = "WXFeedsWarning"
                        elif Alert.find( SPECIAL_AIR_QUALITY_STATEMENT ) >= 0:
                            # Special Air Quality Statement.

                            AlertClass = "WXFeedsSpecialAirQualityStatement"
                        elif Alert.find( SPECIAL_WEATHER_STATEMENT ) >= 0:
                            # Special Weather Statement.

                            AlertClass = "WXFeedsSpecialWeatherStatement"
                        else:
                            # Don't know!

                            self.log.error( "ECCityWeather: Unknown EC Alert: %s" % Entry.title )

                            AlertClass = "WXFeedsUnkownAlert"

                    assert AlertClass != ""

                    HTML +=   "<tr>\n"                                                                          \
                            +      "<td colspan=2>\n"                                                           \
                            +          "<b>\n"                                                                  \
                            +              "<span class=\"%s\">" % AlertClass                                   \
                            +              "&nbsp;&nbsp;"                                                       \
                            +              "<span class=\"icon-warning-sign\"></span>"                          \
                            +              "&nbsp;&nbsp;"                                                       \
                            +              Alert                                                                \
                            +              "&nbsp;&nbsp;"                                                       \
                            +              "<span class=\"icon-warning-sign\"></span>"                          \
                            +              "&nbsp;&nbsp;"                                                       \
                            +              "</span>\n"                                                          \
                            +          "</b>\n"                                                                 \
                            +          "&nbsp;\n"                                                               \
                            +          "<a href=\"%s\">Full Alert</a>\n" % Entry.link                           \
                            +          "&nbsp;\n"                                                               \
                            +          "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link    \
                            +      "</td>\n"                                                                    \
                            +  "</tr>\n"
                elif Entry.tags[0].term == CURRENT_CONDITIONS:

                    continue
                elif Entry.tags[0].term == WEATHER_FORECASTS:
#                        self.log.debug("EC Weather Forecast Entry Title: %s" % Entry.title)
#                        self.log.debug("EC Weather Forecast Entry Summary: %s" % Entry.summary)

                    # Get day/night from Entry.title.

                    DayNight = ""
                    TargetString = ":"
                    x = Entry.title.find(TargetString)

                    if x >= 0:
                        DayNight = Entry.title[0:x]
#                            self.log.debug("DayNight: \"%s\"" % DayNight)

                    # Get the forecast from Entry.summary.

                    TargetString = "Forecast issued "
                    x = Entry.summary.find(TargetString)
#                        self.log.debug("x = %d" % x)

                    if x >= 0:
                        Forecast = Entry.summary[0:x-1]
#                            self.log.debug("Forecast: \"%s\"" % Forecast)
                    else:
                        self.log.error("Unable To Get Forecast From Forecast EntryEC Hourly: %s" % Entry.summary)

                    HTML += ("<tr>\n<td>\n<b>\n%s\n</b>\n</td>\n<td>\n%s\n</td>\n</tr>\n" % (DayNight, Forecast))

            # Close the table.

            HTML += "</tbody>\n</table>\n"

        HTML += "</div>\n"
        self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def ECCityWeatherValidateData(
                                   self,
                                   StatusCode           : int,
                                   ECCityWeatherData    : str
                                 ) -> bool:
                                 #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "ECCityWeatherValidateData: ECCityWeatherData: %s" % ECCityWeatherData )

        if StatusCode == 200 and                    \
           ECCityWeatherData.entries != None:
#            self.log.debug( "ECCityWeatherValidateData: ECCityWeatherData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "ECCityWeatherValidateData: ECCityWeatherData is NOT valid!" )

            pass

        return ReturnStatus

    def ECMarineForecast(
                          self,
                          MarineForecastArea   : str,
                          FeedURL              : str,
                          Latitude             : float,
                          Longitude            : float,
                          PositionSpecified    : bool,
                          Zoom                 : int
                        ) -> bool:
                        #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "ECMarineForecast: MarineForecastArea: '%s'" % MarineForecastArea )
            self.log.debug( "ECMarineForecast: FeedURL: '%s'" % FeedURL )
            self.log.debug( "ECMarineForecast: Latitude: %f" % Latitude )
            self.log.debug( "ECMarineForecast: Longitude: %f" % Longitude )
            self.log.debug( "ECMarineForecast: PositionSpecified: %s" % PositionSpecified )
            self.log.debug( "ECMarineForecast: Zoom: %d" % Zoom )

        assert FeedURL != ""

        ReturnedStatus          = False
        CacheHit                = False
        StatusCode              = -1
        ECMarineForecastData    = {}
        
        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            ECMarineForecastData = self.GetEndpointData(
                                                         EndpointURL               = FeedURL,
                                                         ContentType               = CONTENT_XML,
                                                         ValidateDataFunction      = self.ECMarineForecastValidateData,
                                                         Service                   = "ECMarineForecast",
                                                         CacheKey                  = MarineForecastArea,
                                                         LocalCacheUse             = self.ECMarineForecastsLocalCacheUse,
                                                         LocalCacheExpiry_minutes  = self.ECMarineForecastsLocalCacheExpiry_minutes,
                                                         LocalCacheForceMiss       = self.ECMarineForecastsLocalCacheForceMiss,
                                                         ServerUse            = self.ECMarineForecastsServerUse,
                                                         RemoteCacheUse            = self.ECMarineForecastsRemoteCacheUse,
                                                         RemoteCacheExpiry_minutes = self.ECMarineForecastsRemoteCacheExpiry_minutes,
                                                         RemoteCacheForceMiss      = self.ECMarineForecastsRemoteCacheForceMiss
                                                       )
        except:
            self.log.error( "ECMarineForecast: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "ECMarineForecast: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "ECMarineForecast: GetEndpointData() returned False For: '%s'" % MarineForecastArea )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "ECMarineForecast: Received %d entries for '%s'." % ( len( ECMarineForecastData.entries ), MarineForecastArea ) )

        HTML = ""

#        EntryIndex = 0
#
#        for Entry in ReturnedFeed.entries:
#            self.log.debug("EC Weather Entry %d: entry.tags[0].term = %s" % (EntryIndex, Entry.tags[0].term))
#            self.log.debug("EC Weather Entry %d: %s" % (EntryIndex, Entry))
#            EntryIndex += 1

        #
        # A marine area feed may contain information for than 1 area.  For example, the 
        # Juan de Fuca Strait feed includes information for the west entrance, the central 
        # strait and the east entrance.  So we must filter based on 'MarineForecastArea'.
        #
        # For each area, there could be an (or more?) entry for:
        # - Marine Warnings and Watches
        # - Marine Forecast for today, tonight and the next day
        # - Extended Marine Forecast
        #
        # We have to check 'term' and maybe the contents of 'title' to 
        # identify what area an entry is for.
        #

        MapHTMLFileName = ""

        if self.GoogleMapsMapsUse == True:
            # There could be three links we'd like to display in the map marker - find 'em!

            AlertLink               = ""
            ForecastLink            = ""
            ExtendedForecastLink    = ""

            for Entry in ECMarineForecastData.entries:
#                if DEBUG == True:
#                    self.log.debug ("ECMarineForecast: EC Marine Area Entry: %s" % Entry )

                if Entry.title.find( MarineForecastArea ) >= 0:
                    if DEBUG == True:
                        self.log.debug("ECMarineForecast: Entry For: %s: %s" % ( MarineForecastArea, Entry ) )

                    # This entry is for the area of interest.

#                    if DEBUG == True:
#                        self.log.debug( "ECMarineForecast: Entry found for: %s" % MarineForecastArea )

                    assert len( Entry.tags ) == 1

                    if Entry.tags[0].term == MARINE_WARNINGS_AND_WATCHES:
                        if DEBUG == True:
                            self.log.debug("ECMarineForecast: Alert entry found for: %s" % MarineForecastArea)

                        assert AlertLink == ""  # We expect only one alert.

                        AlertLink = Entry.link
                    elif Entry.tags[0].term == MARINE_FORECASTS:
#                        if DEBUG == True:
#                            self.log.debug("ECMarineForecast: Marine forecast entry found for: %s" % Entry)

                        # Get the forecast period without the Marine Forecast Area name.

                        ForecastPeriod = ""

                        x = Entry.title.find(SHORT_TERM_MARINE_FORECAST_PHRASE)
                        if x == 0:
                            # Short term forecast.

                            if DEBUG == True:
                                self.log.debug("ECMarineForecast: Short-term forecast entry found for: %s" % Entry)

                            assert ForecastLink == ""   # We expect only one short-term forecast.

                            ForecastLink = Entry.link
                        else:
                            # This should be an extended forecast.

                            x = Entry.title.find(EXTENDED_MARINE_FORECAST_PHRASE)
                            assert x == 0

                            if DEBUG == True:
                                self.log.debug("ECMarineForecast: Extended forecast entry found for: %s" % Entry)

                            assert ExtendedForecastLink == ""

                            ExtendedForecastLink = Entry.link
                    else:    # Unknown entry type!
                        self.log.error("ECMarineForecast: Unknown Marine Area Entry Type For EC Marine Area: %s" % MarineForecastArea)

            # The forecast and extended forecast links are probably the same.
            # If they are, get rid of the extended forecast link.

            if DEBUG == True:
                self.log.debug("ECMarineForecast: AlertLink: %s" % AlertLink )
                self.log.debug("ECMarineForecast: ForecastLink: %s" % ForecastLink )
                self.log.debug("ECMarineForecast: ExtendedForecastLink: %s" % ExtendedForecastLink )

            if ForecastLink == ExtendedForecastLink:
                ExtendedForecastLink = ""

                if DEBUG == True:
                    self.log.debug("ECMarineForecast: AlertLink: %s" % AlertLink )
                    self.log.debug("ECMarineForecast: ForecastLink: %s" % ForecastLink )
                    self.log.debug("ECMarineForecast: ExtendedForecastLink: %s" % ExtendedForecastLink )

            MarkerInfoWindowContent = ""
            if AlertLink != "":
                MarkerInfoWindowContent += "<a href=\"%s\">Alert</a>" %  AlertLink
            if ForecastLink != "":
                if MarkerInfoWindowContent != "":
                    MarkerInfoWindowContent += "<br/><br/>"
                MarkerInfoWindowContent += "<a href=\"%s\">Forecast</a>" %  ForecastLink
            if ExtendedForecastLink != "":
                if MarkerInfoWindowContent != "":
                    MarkerInfoWindowContent += "<br/><br/n"
                MarkerInfoWindowContent += "<a href=\"%s\">Extended Forecast</a>" %  ExtendedForecastLink

            ReturnedStatus, MapHTMLFileName, MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                                                 FilePathStemPrefix      = "ECMarineForecast-" + MarineForecastArea,
                                                                                                 ForceStaleFiles         = False,
                                                                                                 Location                = MarineForecastArea,
                                                                                                 Latitude                = Latitude,
                                                                                                 Longitude               = Longitude,
                                                                                                 PositionSpecified       = PositionSpecified,
                                                                                                 Zoom                    = Zoom,
                                                                                                 MarkerType              = "misc",
                                                                                                 MarkerInfoWindowContent = MarkerInfoWindowContent,
                                                                                                 DisplayStationMarker    = True,
                                                                                                 MarkersAreStatic        = False
                                                                                               )
            if ReturnedStatus == True:
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                MapHTMLFileName = ""

        HTML = self.GetDivTag( ENVIRONMENT_CANADA, "MarineForecastArea", MarineForecastArea )
        HTML += "<p class=\"WXFeedsP1\"><b>%s</b>\n" % MarineForecastArea
        self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
            HTML = "&nbsp;\n<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        HTML = "</p>\n"
        self.WriteGeneratedFiles( HTML )

        for Entry in ECMarineForecastData.entries:
            if DEBUG == True:
                self.log.debug("ECMarineForecast: EC Marine Area Entry: %s" % Entry)

            if Entry.title.find(MarineForecastArea) >= 0:
                if DEBUG == True:
                    self.log.debug("ECMarineForecast: EC Weather Entry For: %s: %s" % ( MarineForecastArea, Entry ) )

                # This entry is for the entity of interest.

                if DEBUG == True:
                    self.log.debug("ECMarineForecast: Entry found for: %s" % MarineForecastArea)

                if Entry.tags[0].term == MARINE_WARNINGS_AND_WATCHES:
                    if DEBUG == True:
                        self.log.debug("ECMarineForecast: Alert entry found for: %s" % MarineForecastArea)

                    # Get the alert without the Marine Area name.

                    Alert = ""
                    x = Entry.title.find( MarineForecastArea )
                    Alert = Entry.title[0:x-2]
                    assert Alert != ""

                    AlertInEffect = False
                    AlertClass = ""

                    if Alert.find( WATCH ) >= 0:
                         # A watch.

                        AlertInEffect = True
                        AlertClass = "WXFeedsWatch"

                        if Alert.find( ENDED ) >= 0:
                             # A watch ended.

                            AlertInEffect = False
                            AlertClass = "WXFeedsNoWatchesOrWarnings"

                    else:
                        # Should be a warning.

                        assert Alert.find( WARNING ) >= 0

                        AlertInEffect = True
                        AlertClass = "WXFeedsWarning"

                        if Alert.find( ENDED ) >= 0:
                             # A warning ended.

                            AlertInEffect = False
                            AlertClass = "WXFeedsNoWatchesOrWarnings"

                    assert AlertClass != ""

                    HTML =   "<p class=\"WXFeedsP2\">\n"                    \
                           +     "<b>\n"                                    \
                           +         "<span class=\"%s\">" % AlertClass     \
                           +             "&nbsp;&nbsp;"

                    if AlertInEffect == True:

                        HTML +=   "<span class=\"icon-warning-sign\"></span>"    \
                                + "&nbsp;&nbsp;"

                    HTML +=   Alert             \
                            + "&nbsp;&nbsp;"

                    if AlertInEffect == True:

                        HTML +=   "<span class=\"icon-warning-sign\"></span>"    \
                                + "&nbsp;&nbsp;"

                    HTML +=           "</span>"                                                             \
                            +     "</b>\n"                                                                  \
                            +     "&nbsp;\n"                                                                \
                            +     "<a href=\"%s\">Full Alert</a>\n" % Entry.link                            \
                            +     "&nbsp;\n"                                                                \
                            +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link     \
                            +     "<br/>\n"                                                                 \
                            +     self.CleanupECText( Entry.summary )                                       \
                            + "</p>\n"

                elif Entry.tags[0].term == MARINE_FORECASTS:
                    if DEBUG == True:
                        self.log.debug("ECMarineForecast: Marine forecast entry found for: %s" % Entry)

                    # Get the forecast period without the Marine Area name.

                    ForecastPeriod = ""

                    x = Entry.title.find(SHORT_TERM_MARINE_FORECAST_PHRASE)
                    if x == 0:
                       # Short term forecast.

                       # Separator between forecast period and Marine Area is ". - ".

                       x = Entry.title.find( ". - ")
                       assert x > 0

                       ForecastPeriod = Entry.title[0:x]

                    else:
                        # This should be an extended forecast.

                        x = Entry.title.find(EXTENDED_MARINE_FORECAST_PHRASE)
                        assert x == 0

                        # Separator between extended forecast period and Marine Area is " - ".

                        x = Entry.title.find(" - ")
                        assert x > 0

                        ForecastPeriod = Entry.title[0:x]

                    assert ForecastPeriod != ""

                    HTML += "<p class=\"WXFeedsP2\"><b>" + ForecastPeriod + "</b>"                          \
                         +      "&nbsp;\n"                                                                  \
                         +      "<a href=\"%s\">Full Forecast</a>\n" % Entry.link                           \
                         +      "&nbsp;\n"                                                                  \
                         +      "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link       \
                         +      "<br/>\n"                                                                   \
                         +      self.CleanupECText( Entry.summary )                                         \
                         +  "</p>\n"

                else:    # Unknown entry type!
                    self.log.error("ECMarineForecast: Unknown Marine Area Entry Type For EC Marine Area: %s" % MarineForecastArea)

        HTML += "</div>\n"
        self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def ECMarineForecastValidateData(
                                      self,
                                      StatusCode               : int,
                                      ECMarineForecastData    : str
                                    ) -> bool:
                                    #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "ECMarineForecastValidateData: ECMarineForecastData: %s" % ECMarineForecastData )

        if StatusCode == 200 and                    \
           ECMarineForecastData.entries != None:
#            self.log.debug( "ECMarineForecastValidateData: ECMarineForecastData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "ECMarineForecastValidateData: ECMarineForecastData is NOT valid!" )

            pass

        return ReturnStatus

    def ECPublicAlert(
                       self,
                       PublicAlertArea      : str,
                       FeedURL              : str,
                       Latitude             : float,
                       Longitude            : float,
                       PositionSpecified    : bool,
                       Zoom                 : int
                     ) -> bool:
                     #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        assert FeedURL != ""

        ReturnedStatus      = False
        CacheHit            = False
        StatusCode          = -1
        ECPublicAlertData   = {}
        
        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            ECPublicAlertData = self.GetEndpointData(
                                                      EndpointURL                   = FeedURL,
                                                      ContentType                   = CONTENT_XML,
                                                      ValidateDataFunction          = self.ECPublicAlertValidateData,
                                                      Service                       = "ECPublicAlert",
                                                      CacheKey                      = PublicAlertArea,
                                                      LocalCacheUse                 = self.ECPublicAlertsLocalCacheUse,
                                                      LocalCacheExpiry_minutes      = self.ECPublicAlertsLocalCacheExpiry_minutes,
                                                      LocalCacheForceMiss           = self.ECPublicAlertsLocalCacheForceMiss,
                                                      ServerUse                     = self.ECPublicAlertsServerUse,
                                                      RemoteCacheUse                = self.ECPublicAlertsRemoteCacheUse,
                                                      RemoteCacheExpiry_minutes     = self.ECPublicAlertsRemoteCacheExpiry_minutes,
                                                      RemoteCacheForceMiss          = self.ECPublicAlertsRemoteCacheForceMiss
                                                    )
        except:
            self.log.error( "ECPublicAlert: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "ECPublicAlert: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "ECPublicAlert: GetEndpointData() returned False For: '%s'" % PublicAlertArea )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "ECPublicAlert: Received %d entries for '%s'." % ( len( ECPublicAlertData.entries ), PublicAlertArea ) )

        # There can be multiple alerts per area and the name of the town, city, area, etc. 
        # may not match the name of the area we were called with, and there doesn't appear 
        # to be a good way to select a particular alert ('cause province is not specified and 
        # there could be duplication of names without it), so we just use all entries.

        HTML = ""

#        EntryIndex = 0
#
#        for Entry in ReturnedFeed.entries:
#            self.log.debug("EC Weather Entry %d: entry.tags[0].term = %s" % (EntryIndex, Entry.tags[0].term))
#            self.log.debug("EC Weather Entry %d: %s" % (EntryIndex, Entry))
#            EntryIndex += 1

        MapHTMLFileName = ""

        if self.GoogleMapsMapsUse == True:

            MarkerInfoWindowContent = ""

            if len( ECPublicAlertData.entries ) >= 1:
                MarkerInfoWindowContent = "<a href=\"%s\">Alert</a>" %  ECPublicAlertData.entries[ 0 ].link
       
            ReturnedStatus, MapHTMLFileName, MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                                                 FilePathStemPrefix        = "ECPublicAlert-" + PublicAlertArea,
                                                                                                 ForceStaleFiles           = not CacheHit,
                                                                                                 Location                  = PublicAlertArea,
                                                                                                 Latitude                  = Latitude,
                                                                                                 Longitude                 = Longitude,
                                                                                                 PositionSpecified         = PositionSpecified,
                                                                                                 Zoom                      = Zoom,
                                                                                                 MarkerType                = "misc",
                                                                                                 MarkerInfoWindowContent   = MarkerInfoWindowContent,
                                                                                                 DisplayStationMarker      = True,
                                                                                                 MarkersAreStatic          = False
                                                                                               )
            if ReturnedStatus == True:
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                MapHTMLFileName = ""

        HTML = self.GetDivTag( ENVIRONMENT_CANADA, "PublicAlert", PublicAlertArea )
        HTML += "<p class=\"WXFeedsP1\"><b>%s</b>\n" % PublicAlertArea
        self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
            HTML = "&nbsp;\n<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        HTML = "</p>\n"
        self.WriteGeneratedFiles( HTML )

        for Entry in ECPublicAlertData.entries:
            if DEBUG == True:
                self.log.debug( "ECPublicAlert Entry: %s" % Entry )

            # Alert could be a Watch, a Warning or "No Alerts. . .".
            #
            # Highlight a Watch or a Warning.

            # Assume no alert in effect.

            AlertInEffect = False
            AlertClass = ""

            if Entry.title.find( NO_ALERTS_IN_EFFECT ) >= 0:
                # No alerts in effect.

                AlertInEffect = False
                AlertClass = "WXFeedsNoWatchesOrWarnings"
            elif Entry.title.find( WATCH ) >= 0:
                # Is a Watch.

                AlertInEffect = True
                AlertClass = "WXFeedsWatch"
            elif Entry.title.find( WARNING ) >= 0:
                # Is a Warning.

                AlertInEffect = True
                AlertClass = "WXFeedsWarning"
            elif Entry.title.find( SPECIAL_AIR_QUALITY_STATEMENT ) >= 0:
                # Is a Special Air Quality Statement.

                AlertInEffect = True
                AlertClass = "WXFeedsSpecialAirQualityStatement"
            elif Entry.title.find( SPECIAL_WEATHER_STATEMENT ) >= 0:
                # Is a Special Weather Statement.

                AlertInEffect = True
                AlertClass = "WXFeedsSpecialWeatherStatement"
            else:
                # Don't know!

                self.log.error( "ECPublicAlert: Unknown EC Alert: %s" % Entry.title )

                AlertInEffect = False
                AlertClass = "WXFeedsUnkownAlert"

            assert AlertClass != ""

            HTML =   "<p class=\"WXFeedsP2\">"                          \
                   +     "<b>\n"                                        \
                   +            "<span class=\"%s\">" % AlertClass      \
                   +                "&nbsp;&nbsp;"

            if AlertInEffect == True:
                HTML +=   "<span class=\"icon-warning-sign\"></span>"    \
                        + "&nbsp;&nbsp;"

            HTML +=   Entry.title       \
                    + "&nbsp;&nbsp;"

            if AlertInEffect == True:
                HTML +=    "<span class=\"icon-warning-sign\"></span>"    \
                         + "&nbsp;&nbsp;"

            HTML +=   "</span>\n"                                                               \
                    + "</b>\n"                                                                  \
                    + "&nbsp;\n"                                                                \
                    + "<a href=\"%s\">Full Alert</a>\n" % Entry.link                            \
                    + "&nbsp;\n"                                                                \
                    + "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link     \
                    + "<br/>\n"

            if AlertInEffect == True:

                 HTML += Entry.summary

            HTML +=   "</p>\n"

        HTML += "</div>\n"
        self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def ECPublicAlertValidateData(
                                   self,
                                   StatusCode           : int,
                                   ECPublicAlertData   : str
                                 ) -> bool:
                                 #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "ECCityPublicAlertValidateData: ECPublicAlertData: %s" % ECPublicAlertData )

        if StatusCode == 200 and                    \
           ECPublicAlertData.entries != None:
#            self.log.debug( "ECCityPublicAlertValidateData: ECPublicAlertData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "ECCityPublicAlertValidateData: ECPublicAlertData is NOT valid!" )

            pass

        return ReturnStatus

    def CleanupECText(
                       self,
                       Summary : str
                     ) -> str:

        # Environment Canada uses some curious if not invalid constructs in 
        # their text strings.

#        self.log.debug( "Summary: '%s'" % Summary )

        s = Summary

        s = Summary.replace( "\n ", " " )
        s = s.replace( "\n", " " )
        s = s.replace( "\r ", " " )
        s = s.replace( "\r", " " )
        s = s.replace( "<br />", "<br/>" )

#        self.log.debug( "s: '%s'" % s )

        return s

    def DriveBCParseUpdateDateTime(
                                    self,
                                    UpdateDateTime  : str,
                                    Published_dt    : datetime
                                  ) -> bool:

        # Returns:
        #
        #   ReturnStatus: True or False.
        #
        #   UpdateDateTimeUTC_dt: UpdateDateTime parsed and converted to UTC.

        ReturnStatus = True     # Assume success!

        #DEBUG = True
        DEBUG = False

        UpdateDateTimeLocal_dt  = None
        UpdateDateTimeUTC_dt    = None

        Months = dict( Jan = 1, Feb = 2, Mar = 3, Apr = 4, May = 5, Jun = 6, Jul = 7, Aug = 8, Sep = 9, Oct = 10, Nov = 11, Dec = 12 )

        # Published_dt will have a time zone offset but it is not time zone aware.

        if DEBUG == True:
            self.log.debug( "DriveBCParseUpdateDateTime: Published_dt (z): %s" % Published_dt.strftime( "%Y/%m/%d %H:%M:%S %z" ) )
            self.log.debug( "DriveBCParseUpdateDateTime: Published_dt (Z): %s" % Published_dt.strftime( "%Y/%m/%d %H:%M:%S %Z" ) )
            self.log.debug( "DriveBCParseUpdateDateTime: Published_dt UTCOffset: %s" % Published_dt.utcoffset() )

        # Create time zone aware published datetime.

        PublishedUTC_dt = Published_dt.replace(tzinfo=datetime.timezone.utc) - Published_dt.utcoffset()
        if DEBUG == True:
            self.log.debug( "DriveBCParseUpdateDateTime: Published_dt (NOT TZ Aware): %s, PublishedUTC_dt (TZ Aware): %s" % ( Published_dt, PublishedUTC_dt ) )

        # Format of date/time in a Drive BC Event:
        #
        # Expected Normal Format: 'Wed Nov 25 at 9:12 AM PST'
        #
        # But could be:           'Mon May 31, 2021' or
        #                         'Fri Jan 8, 2021 at 5:00 PM PST'!
        #
        # Drive BC update date/times (last update and next update) are not elements in 
        # the Drive BC feed - they are commentary in 'summary' and 'summary.detail.value'.
        #
        # Not all above expected formats include a year!  So we will use the year 
        # from 'Published_dt' on the assumption that the alert was updated and published 
        # in the same year.  We will verify the year by checking the date's day name 
        # against what is in 'UpdateDateTime'.

        if DEBUG == True:
            self.log.debug( "DriveBCParseUpdateDateTime: UpdateDateTime: %s" % UpdateDateTime )

        DateTimeList = UpdateDateTime.strip().split( " " )
        #self.log.debug( "DateTimeList: %s" % DateTimeList )

        if len( DateTimeList ) >= 3 and len( DateTimeList ) <= 8:
 
            Year        = 0
            MonthName   = ""
            MonthNumber = 0
            DayName     = ""
            DayNumber   = 0
            Hour        = 0
            Minute      = 0
            Second      = 0
            AMPM        = ""
            TimeZone    = ""

            i = 0

            if len( DateTimeList ) >= 3: # 3 entries suggests date/time has at least a date.
                DayName     = DateTimeList[ 0 ]
                MonthName   = DateTimeList[ 1 ]
                MonthNumber = Months[ MonthName ]

                if DateTimeList[ 2 ].find( "," ) >= 0:
                    # This should be a day number followed by the year.

                    DayNumber = int( DateTimeList[ 2 ].replace( ",", "" ) )

                    assert len( DateTimeList ) >= 4     # 4 entries suggests date/time has year or 
                                                        # date/time does not have year but has time 
                                                        # after 'at'.

                    Year = int( DateTimeList[ 3 ] )

                    i = 4   # Index of next element.
                else:
                    # This should be a day number.

                    DayNumber   = int( DateTimeList[ 2 ] )
                    Year        = Published_dt.year
                    #Year        -= 1                # Only for testing Published_dt in prior year.

                    i = 3   # Index of next element

                if len( DateTimeList ) == i + 4 and \
                   DateTimeList[ i ] == "at":           # Should have time.
                    HoursMinutes = DateTimeList[ i + 1 ]

                    if len( HoursMinutes ) == 4:    # Assume time = 'H:MM' - no leading zero for hours.
                        assert HoursMinutes[ 1 ] == ":"

                        Hour    = int( HoursMinutes[ 0 ] )
                        Minute  = int( HoursMinutes[ 2 : 4 ] )
                    else:                           # Assume time = 'HH:MM'.
                        assert len( HoursMinutes ) == 5
                        assert HoursMinutes[ 2 ] == ":"

                        Hour    = int( HoursMinutes[ 0 : 2 ] )
                        Minute  = int( HoursMinutes[ 3 : 5 ] )

                    assert DateTimeList[ i + 2 ] == "AM" or DateTimeList[ i + 2 ] == "PM"

                    AMPM = DateTimeList[ i + 2 ]

                    assert DateTimeList[ i + 3 ] != ""

                    TimeZone = DateTimeList[ i + 3 ]
                else:
                    # No time.

                    pass

                if DEBUG == True:
                    self.log.debug( "DriveBCParseUpdateDateTime: UpdateDateTime (Parsed): Date: %d %d %d Time: %d %d %d TZ: %s" %   \
                                        (
                                          Year,
                                          MonthNumber,
                                          DayNumber,
                                          Hour,
                                          Minute,
                                          Second,
                                          TimeZone
                                        )
                                  )

                UpdateDateTimeLocal_dt = datetime.datetime( Year, MonthNumber, DayNumber, Hour, Minute, Second )
                if DEBUG == True:
                    self.log.debug( "DriveBCParseUpdateDateTime: <1> UpdateDateTimeLocal_dt (12-hour time): %s %s" % ( UpdateDateTimeLocal_dt, AMPM  ) )

                # Check that we have the right year - see comment above.
                #
                # We check using 12-hour time because if we convert 12:00 to 24-hour 
                # time the day will increment!

                OurDayName = UpdateDateTimeLocal_dt.strftime( "%a" )

                if OurDayName == DayName:
                    # Day Check 1 Passed!

                    if DEBUG == True:
                        self.log.debug( "DriveBCParseUpdateDateTime: Day Name Check <1> PASSED!" )
                else:
                    # Day Check 1 Failed!

                    if DEBUG == True:
                        self.log.debug( "DriveBCParseUpdateDateTime: Day Name Check <1> FAILED! Input: %s, UpdateDateTimeLocal_dt: %s" % (UpdateDateTime, UpdateDateTimeLocal_dt ) )
                        self.log.debug( "DriveBCParseUpdateDateTime:                            Input Day: %s, Check Day: %s" % (DayName, OurDayName ) )

                    # Adjust year by + 1 and try again.
 
                    UpdateDateTimeLocal_dt = datetime.datetime( Year + 1, MonthNumber, DayNumber, Hour, Minute, Second )
                    if DEBUG == True:
                        self.log.debug( "DriveBCParseUpdateDateTime: <2> UpdateDateTimeLocal_dt (12-hour time): %s %s" % ( UpdateDateTimeLocal_dt, AMPM  ) )

                    OurDayName = UpdateDateTimeLocal_dt.strftime( "%a" )

                    if OurDayName == DayName:
                        # Day Check 2 Passed!

                        if DEBUG == True:
                            self.log.debug( "DriveBCParseUpdateDateTime: Day Name Check <2> PASSED!" )
                    else:
                        # Day Check 2 Failed!

                        self.log.error( "DriveBCParseUpdateDateTime: Day Name Check <2> FAILED! Input: %s, UpdateDateTimeLocal_dt: %s" % (UpdateDateTime, UpdateDateTimeLocal_dt ) )
                        self.log.error( "DriveBCParseUpdateDateTime:                            Input Day: %s, Check Day: %s" % (DayName, OurDayName ) )

                        ReturnStatus = False

                if ReturnStatus == True:
                    # No errors so far.

                    if AMPM == "PM":   # Adjust for PM.
                        UpdateDateTimeLocal_dt += datetime.timedelta( hours = 12 )
                    if DEBUG == True:
                        self.log.debug( "DriveBCParseUpdateDateTime: UpdateDateTimeLocal_dt (24-hour time): %s" % UpdateDateTimeLocal_dt )

                        # Adjust for time zone.
                        #
                        # Published_dt is always Pacific Time so we have to adjust if our time zone is Mountain Time.

                    if TimeZone == "MST" or     \
                       TimeZone == "MDT":
                        UpdateDateTimeLocal_dt -= datetime.timedelta( hours = 1 )
                    else:
                        if TimeZone != "":
                            assert TimeZone == "PST" or TimeZone == "PDT"

                            if TimeZone != "PST" and TimeZone != "PDT":
                                self.log.error( "DriveBCParseUpdateDateTime: Unexpected Time Zone in '%s'" % UpdateDateTime )

                    # Adjust UpdateDateTimeLocal_dt to UTC.

                    UpdateDateTimeUTC_dt = UpdateDateTimeLocal_dt.replace(tzinfo=datetime.timezone.utc) - Published_dt.utcoffset()
                    if DEBUG == True:
                        self.log.debug( "DriveBCParseUpdateDateTime: UpdateDateTimeUTC_dt: %s" % UpdateDateTimeUTC_dt )
            else:
                self.log.error( "DriveBCParseUpdateDateTime: Error - No date! UpdateDateTime = >%s<" % UpdateDateTime )

                ReturnStatus = False
        else:
            self.log.error( "DriveBCParseUpdateDateTime: Error - No or invalid date/time! UpdateDateTime = >%s<" % UpdateDateTime )

            ReturnStatus = False

        if ReturnStatus == True:
            assert UpdateDateTimeUTC_dt != None

        return ReturnStatus, UpdateDateTimeUTC_dt

    def DriveBCEventGetUpdatesDateTime(
                                        self,
                                        Event : str
                                      ) -> ( bool, bool, datetime.datetime, bool, datetime.datetime ):
                                      #      ReturnStatus, LastUpdadeValid, LastUpdated_dt, NextUpdateValid, NextUpdate_dt

        ReturnStatus = False

        LastUpdadeValid = False
        LastUpdated_dt  = 0
        NextUpdateValid = False
        NextUpdate_dt   = 0

        #DEBUG = True
        DEBUG = False

        # 'next update' and 'last updated' date/times are not elements in the Drive BC 
        # feed - they are commentary in 'summary' and 'summary.detail.value'.
        # They do not include a year!  So we will use the year from 'published' 
        # on the assumption that the alert was updated and published in the same year.  We 
        # will verify the year by checking the date's name against what is in 'summary'.
        #
        # This may seem like a lot of work but if we don't do this, dates just before or 
        # after the end of a calendar year could be (will be?) wrong.
        #

        # From Drive BC sample:
        #
        #       'summary': 'Big Bar Reaction Ferry. Seasonal shutdown at Big Bar Reaction Ferry (Big Bar Reaction Ferry). Arial 
        #                   Tram available to foot traffic only. Next update time Mon May 31, 2021. Last updated Wed Nov 25 at 
        #                   9:12 AM PST. (DBC-24168)', 
        #
        #       'published': 'Wed, 25 Nov 2020 09:12:34 -0800', 

        DateTimeFormat = "%a %b %-d at %-I:%M %p %Z"

        if DEBUG == True:
            self.log.debug( "DriveBCEventGetUpdatesDateTime: Event.summary: %s" % Event.summary )

        LastUpdatedStartString  = "Last updated "
        LastUpdatedEndString    = ". "

        NextUpdateStartString   = "Next update time "
        NextUpdateEndString     = ". "

        LastUpdated = ""
        NextUpdate  = ""

        Start   = Event.summary.find( LastUpdatedStartString )
        End     = Event.summary.find( LastUpdatedEndString, Start )

        if Start >= 0 and End >= 0:
            LastUpdated = Event.summary[ Start + len( LastUpdatedStartString ) : End ]
            if DEBUG == True:
                self.log.debug( "DriveBCEventGetUpdatesDateTime: LastUpdated (from summary) = >%s<" % LastUpdated )
        else:
            # We couldn't find a last updated date/time in 'summary' so 
            # create one using current date/time (which we then parse 
            # later as if it came from 'summary'.

            LastUpdated = self.GetTimeLocalAware().strftime( DateTimeFormat )
            if DEBUG == True:
                self.log.debug( "DriveBCEventGetUpdatesDateTime: LastUpdated (from now()) = >%s<" % LastUpdated )

        assert LastUpdated != ""

        Start   = Event.summary.find( NextUpdateStartString )
        End     = Event.summary.find( NextUpdateEndString, Start )

        if Start >= 0 and End >= 0:
            NextUpdate = Event.summary[ Start + len( NextUpdateStartString ) : End ]
            if DEBUG == True:
                self.log.debug( "DriveBCEventGetUpdatesDateTime: NextUpdate (from summary) = >%s<" % NextUpdate )
        else:
            # We couldn't find a next update date/time in 'summary' so 
            # create one using current date/time (which we then parse 
            # later as if it came from 'summary'.

            NextUpdate = self.GetTimeLocalAware().strftime( DateTimeFormat )
            if DEBUG == True:
                self.log.debug( "DriveBCEventGetUpdatesDateTime: NextUpdate (from now()) = >%s<" % NextUpdate )

        assert NextUpdate != ""

#        if DEBUG == True:
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Start of TESTING! ******" )
#
#            # Testing for format without a date:
#            #           
#            # Expected Normal Format: 'Wed Nov 25 at 9:12 AM PST'
#            #
#            # We will use LastUpdate for testing.
#
#            # Test 1: Actual dates from a real event - should NOT fail:
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 1 - Actual Data - Should PASS <1> ******" )
#
#            LastUpdatedTest = "Wed Nov 25 at 9:12 AM PST"   # 25 9:12 AM PST = 25 09:12 PST = 2020-11-25 17:12 UTC 
#            PublishedTest   = "25 Nov 2020 09:12:34 -0800"  # 2020-11-25 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 2: Publish and Update in January 2020 - should NOT fail.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 2 - AM - After New Year - Should PASS <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 AM PST"    # 1 9:12 AM PST = 1 09:12 PST = 1 17:12 UTC = 
#                                                            # 2020-01-01 17:12 UTC
#            PublishedTest   = "1 Jan 2020 09:12:34 -0800"   # 2020-01-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 3: Publish in December 2019 and Update in January 2020 - should fail Day Check <1>.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 3 - AM - Across Years - Should FAIL <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 AM PST"    # 1 9:12 AM PST = 1 09:12 PST = 1 17:12 UTC = 
#                                                            # 2020-01-01 17:12 UTC
#            PublishedTest   = "1 Dec 2019 09:12:34 -0800"   # 2019-12-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#             # Test 4: Publish and Update in January 2020 - should NOT fail.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 4 - PM -After New Year - Should PASS <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 PM PST"    # 1 9:12 PM PST = 1 21:12 PST = 2 05:12 UTC = 
#                                                            # 2020-01-02 05:12 UTC
#            PublishedTest   = "1 Jan 2020 09:12:34 -0800"   # 2020-01-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 5: Publish in December 2019 and Update in January 2020 - should fail Day Check <1>.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 5 - PM - Across Years - Should FAIL <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 PM PST"    # 1 9:12 PM PST = 1 21:12 PST = 2 05:12 UTC = 
#                                                            # 2020-01-02 05:12 UTC 
#            PublishedTest   = "1 Dec 2019 09:12:34 -0800"   # 2019-12-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 6: MST: Publish and Update in January 2020 - should NOT fail.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 6 - MST AM - After New Year - Should PASS <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 AM MST"    # 1 9:12 AM MST = 1 08:12 AM PST = 1 08:12 PST = 1 16:12 UTC = 
#                                                            # 2020-01-01 16:12 UTC
#            PublishedTest   = "1 Jan 2020 09:12:34 -0800"   # 2020-01-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 7: MST: Publish in December 2019 and Update in January 2020 - should fail Day Check <1>.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 7 - MST AM - Across Years - Should FAIL <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 AM MST"    # 1 9:12 AM MST = 1 08:12 AM PST = 1 08:12 PST = 1 16:12 UTC = 
#                                                            # 2020-01-01 16:12 UTC
#            PublishedTest   = "1 Dec 2019 09:12:34 -0800"   # 2019-12-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 8: MST: Publish and Update in January 2020 - should NOT fail.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 8 - MST PM - After New Year - Should PASS <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 PM MST"    # 1 9:12 PM MST = 1 08:12 PM PST = 1 20:12 PST = 2 04:12 UTC = 
#                                                            # 2020-01-02 04:12 UTC
#            PublishedTest   = "1 Jan 2020 09:12:34 -0800"   # 2020-01-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            # Test 9: MST: Publish in December 2019 and Update in January 2020 - should fail Day Check <1>.
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** Test 9 - MST PM - Across Years - Should FAIL <1> ******" )
#
#            LastUpdatedTest = "Wed Jan 1 at 9:12 PM MST"    # 1 9:12 PM MST = 1 08:12 PM PST = 1 20:12 PST = 2 04:12 UTC = 
#                                                            # 2020-01-02 04:12 UTC
#            PublishedTest   = "1 Dec 2019 09:12:34 -0800"   # 2019-12-01 17:12:34 UTC
#
#            PublishedTest_dt = datetime.datetime.strptime( PublishedTest, "%d %b %Y %H:%M:%S %z" )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest: %s" % PublishedTest )
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: TEST! PublishedTest_dt : %s" % PublishedTest_dt )
#
#            ReturnedStatus, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdatedTest, PublishedTest_dt )
#            assert ReturnedStatus == True
#
#            self.log.debug( "DriveBCEventGetUpdatesDateTime: ****** End of TESTING! ******" )
#
#            assert 1 != 1

        Published_dt = parser.parse( Event.published )
        if DEBUG == True:
            self.log.debug( "DriveBCEventGetUpdatesDateTime: Published_dt: %s" % Published_dt )

        LastUpdatedValid, LastUpdated_dt = self.DriveBCParseUpdateDateTime( LastUpdated, Published_dt )
        if DEBUG == True and    \
           LastUpdatedValid == False:
            self.log.debug( "DriveBCEventGetUpdatesDateTime: LastUpdatedValid: %s is NOT Valid!" % LastUpdated )

        NextUpdateValid, NextUpdate_dt = self.DriveBCParseUpdateDateTime( NextUpdate, Published_dt )
        if DEBUG == True and    \
           NextUpdateValid == False:
            self.log.debug( "DriveBCEventGetUpdatesDateTime: NextUpdateValid: %s is NOT Valid!" % LastUpdated )

        ReturnStatus = LastUpdatedValid and LastUpdated_dt

        return ReturnStatus, LastUpdatedValid, LastUpdated_dt, NextUpdateValid, NextUpdate_dt

    def DriveBCAreas(
                      self,
                      Areas                     : str,
                      FeedURLs                  : str,
                      Latitude                  : float,
                      Longitude                 : float,
                      PositionSpecified         : bool,
                      Zoom                      : int,
                      DisplayLinksToDriveBC     : str,
                      DistrictFilter            : str,
                      RouteFilter               : str,
                      TypeFilter                : str,
                      SeverityFilter            : str,
                      DisplayAttributes         : bool,
                      DisplayAttributeLists     : bool
                   ) -> bool:

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        HTMLFilePathStemPrefix          = "DriveBCEvent-"
        JavaScriptFilePathStemPrefix    = "DriveBCEvents-"

        # Convert filter strings to  lists.

        DistrictFilterList  = []
        RouteFilterList     = []
        TypeFilterList      = []
        SeverityFilterList  = []

        if DistrictFilter != "":
            DistrictFilterList = DistrictFilter.split( "|" )

            for x in range( len( DistrictFilterList ) ):
                DistrictFilterList[ x ] = DistrictFilterList[ x ].strip()

        if RouteFilter != "":
            RouteFilterList = RouteFilter.split( "|" )

            for x in range( len( RouteFilterList ) ):
                RouteFilterList[ x ] = RouteFilterList[ x ].strip()

        if TypeFilter != "":
            TypeFilterList = TypeFilter.split( "|" )

            for x in range( len( TypeFilterList ) ):
                TypeFilterList[ x ] = TypeFilterList[ x ].strip()

        if SeverityFilter != "":
            SeverityFilterList = SeverityFilter.split( "|" )

            for x in range( len( SeverityFilterList ) ):
                SeverityFilterList[ x ] = SeverityFilterList[ x ].strip()

        # These lists are intended to make it easier to see what values 
        # exist in each of the event attributes.  The idea is to 
        # occasionally display the attribute lists so that values 
        # can be used in attribute filters.

        DBCDistrict     = []
        DBCRoute        = []
        DBCType         = []
        DBCSeverity     = []

        # We also build a list of markers for Google Maps.

        Markers         = []

        MarkerType              = ""
        MarkerLatitude          = BAD_LATITUDE
        MarkerLongitude         = BAD_LONGITUDE
        MarkerText              = ""

        # List of HTML map files that currently exist, as in 
        # HTML map files that exist before we process Areas.

        CurrentDriveBCHTMLMapFiles = []

        Events          = []        # Store selected events in a list.
        SelectedEvents  = []        # Store key data about an event in a list.
        EventIDs        = set()     # Store event IDs in a set so we can detect duplicates.

        # Split Areas and FeedURLs into lists.

        if Areas != "":
            AreaList = Areas.split( "|" )

            for x in range( len( AreaList ) ):
                AreaList[ x ] = AreaList[ x ].strip()

        if FeedURLs != "":
            FeedURLList = FeedURLs.split( "|" )

            for x in range( len( FeedURLList ) ):
                FeedURLList[ x ] = FeedURLList[ x ].strip()

        assert len( AreaList ) >0
        assert len( FeedURLList ) > 0
        assert len( AreaList ) == len( FeedURLList ) > 0

        TotalReturnedEvents = 0
        
        # Note if a feed returns zero entries.

        AFeedReturnedZeroEntries = False

        for i in range( len( AreaList ) ):
            if DEBUG == True:
                self.log.debug( "DriveBCAreas: Area: %s, FeedURL: %s" % ( AreaList[ i ], FeedURLList[ i ] ) )

            ReturnedStatus      = False
            CacheHit            = False
            StatusCode          = -1
            DriveBCData         = {}

            try:
                ReturnedStatus,         \
                CacheHit,               \
                StatusCode,             \
                DriveBCData = self.GetEndpointData(
                                                    EndpointURL                 = FeedURLList[ i ],
                                                    ContentType                 = CONTENT_XML,
                                                    ValidateDataFunction        = self.DriveBCValidateData,
                                                    Service                     = "DriveBCEvents",
                                                    CacheKey                    = AreaList[ i ],
                                                    LocalCacheUse               = self.DriveBCLocalCacheUse,
                                                    LocalCacheExpiry_minutes    = self.DriveBCLocalCacheExpiry_minutes,
                                                    LocalCacheForceMiss         = self.DriveBCLocalCacheForceMiss,
                                                    ServerUse                   = self.DriveBCServerUse,
                                                    RemoteCacheUse              = self.DriveBCRemoteCacheUse,
                                                    RemoteCacheExpiry_minutes   = self.DriveBCRemoteCacheExpiry_minutes,
                                                    RemoteCacheForceMiss        = self.DriveBCRemoteCacheForceMiss
                                                  )
            except:
                self.log.error( "DriveBCAreas: GetEndpointData() threw an exception!  Unexpected error: %s" % sys.exc_info()[0] )

                ReturnedStatus = False 

            if DEBUG == True:
                self.log.debug( "DriveBCAreas: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

            if ReturnedStatus == False:
                # GetEndpointData() returned False so we're done!

                self.log.warning( "DriveBCAreas: GetEndpointData() returned False for '%s'." % AreaList[ i ] )

                ReturnStatus = False
                return ReturnStatus

            if DEBUG == True:
                self.log.debug( "DriveBCAreas: GetEndpointData() returned successully for URL: %s" % AreaList[ i ] )

            if StatusCode != 200:
                if DEBUG == True:
                    self.log.debug( "DriveBCAreas: GetEndpointData() returned BAD StatusCode for URL: %s" % AreaList[ i ] )

                ReturnStatus = False
                return ReturnStatus

            self.log.info( "DriveBCAreas: Received %d entries for '%s'." % ( len( DriveBCData.entries ), AreaList[ i ] ) )

            TotalReturnedEvents += len( DriveBCData.entries )

            # We occasionaly get a good return with 0 entries - maybe Drive BC is updating their data, don't know.
            #
            # If we get 0 entries, proceed as usual but we won't clean up map files 'cause we quite likely want many 
            # of them, and if we deleted them, we'd just have to recreate them when we do get entries (in a cycle or two).

            if len( DriveBCData.entries ) == 0:
                AFeedReturnedZeroEntries = True

            for Event in DriveBCData.entries:
#                self.log.debug( "DriveBC Event: %s" % Event )

                # Get the DriveBC event identifier.

                x = Event.id.find( "~DBC" )
                assert x >= 0

                EventID = ""
                if x >= 0:
                    EventID = Event.id[ x + 1 : len( Event.id ) ]

                if DEBUG == True:
                    self.log.debug( "DriveBCAreas: EventID = %s" % EventID )
                assert EventID != ""

                # Get attributes for this event.

                District    = Event[ 'dbc_district' ]
                Route       = Event[ 'dbc_route' ]
                Type        = Event[ 'dbc_type' ] 
                Severity    = Event[ 'dbc_severity' ]

                # Maintain lists of event attributes.

                if District not in DBCDistrict:
                    DBCDistrict.append( District )
                if Route not in DBCRoute:
                    DBCRoute.append( Route )
                if Type not in DBCType:
                    DBCType.append( Type )
                if Severity not in DBCSeverity:
                    DBCSeverity.append( Severity )

                # If we are going to create maps, we create a marker for every event.
                # We do this so the user sees all events on a map even if they've 
                # filtered them out of the list.
                # But we only want to create one marker per event so only create a
                # marker if this event is not in our set.

                if self.GoogleMapsMapsUse == True and    \
                   EventID not in EventIDs:
                    # Create a marker for this event.

                    if Type == "Current Planned":
                        MarkerType = "current_planned"
                    elif Type == "Future Planned":
                        MarkerType = "future_planned"
                    elif Type == "Incident":
                        MarkerType = "incident"
                    elif Type == "Road Condition":
                        MarkerType = "road_condition"
                    else:
                        self.log.error( "Unknown DriveBC Event Type = %s" % Type )

                        MarkerType = "misc"

                    MarkerText = Type

                    # Add short desription from first sentence in 'summary'.

                    x = Event.summary.find( "." )
                    if x >=0:
                        MarkerText += ": " + Event.summary[ 0 : x ]

                    if MarkerText == "":
                        MarkerText = "Unknown"

                    MarkerInfoWindowContent = Type + ": " + Event[ "summary" ]

                    Markers.append( 
                                    GoogleMapsMarker(
                                                      MarkerType,
                                                      float( Event[ "geo_lat" ] ),
                                                      float( Event[ "geo_long" ] ),
                                                      MarkerText,
                                                      MarkerInfoWindowContent
                                                     )
                                      )

                UseThisEvent = False

                if len( DistrictFilterList ) > 0:
                    # We have a district filter list so use it.

                    for FilterDistrict in DistrictFilterList:
                        if FilterDistrict == District:
                            UseThisEvent = True
                            break
                else:
                    # No district filter list so use this event.

                    UseThisEvent = True

                if UseThisEvent == True:
                    # We're interested in this event so far.
                    #
                    # Next filter. . .

                    UseThisEvent = False

                    if len( RouteFilterList ) > 0:
                        # We have a route filter list so use it.

                        for FilterRoute in RouteFilterList:
                            if FilterRoute == Route:
                                UseThisEvent = True
                                break
                    else:
                        # No route filter list so use this event.

                        UseThisEvent = True

                if UseThisEvent == True:
                    # We're interested in this event so far.
                    #
                    # Next filter. . .

                    UseThisEvent = False

                    if len( TypeFilterList ) > 0:
                        # We have a type filter list so use it.

                        for FilterType in TypeFilterList:
                            if FilterType == Type:
                                UseThisEvent = True
                                break
                    else:
                        # No type filter list so use this event.

                        UseThisEvent = True

                if UseThisEvent == True:
                    # We're interested in this event so far.
                    #
                    # Next filter. . .

                    UseThisEvent = False

                    if len( SeverityFilterList ) > 0:
                        # We have a severity filter list so use it.

                        for FilterSeverity in SeverityFilterList:
                            if FilterSeverity == Severity:
                                UseThisEvent = True
                                break
                    else:
                        # No route filter list so use this event.

                        UseThisEvent = True

                if UseThisEvent == True:
                    # We're interested in this event.

                    # But only if it's not already in the selected event list.

                    if EventID not in EventIDs:
                        if DEBUG == True:
                            self.log.debug( "DriveBCAreas: Event: %s NOT in EventIDs - will add it." % ( EventID ) )

                        EventIDs.add( EventID )
                        Events.append( Event )
                        EventsIndex = len( Events ) - 1
 
                        # Get last updates and next update date/time.
                        #
                        # I intended to use LastUpdated_dt and NextUpdate_dt to optimize 
                        # the creation of maps but we do nothing other than store them 
                        # at this time: Dec 21, 2020.  Well, actually, LastUpdated_dt 
                        # is used to sort SelectedEvents.
    
#                        self.log.debug( "DriveBCAreas: Selected Event: %s" % Event )

                        ReturnedStatus      = False
                        LastUpdatedValid    = False
                        LastUpdated_dt      = 0
                        NextUpdateValid     = False
                        NextUpdate_dt       = 0

                        ReturnedStatus, LastUpdatedValid, LastUpdated_dt, NextUpdateValid, NextUpdate_dt = self.DriveBCEventGetUpdatesDateTime( Event )

                        # I think these checks aren't needed because DriveBCEventGetUpdatesDateTime() and the 
                        # functions it calls will substitude an invalid date/time basded on now() or nowtuc().

                        if ReturnedStatus == False:
                            self.log.error( "DriveBCAreas: DriveBCEventGetUpdatesDateTime() returned False!" )

                        if ReturnedStatus == True and       \
                           LastUpdatedValid == False:
                            self.log.error( "DriveBCAreas: DriveBCEventGetUpdatesDateTime() returned False for LastUpdated!" )

                            LastUpdated_dt = GetCurrentUTCAware()

                        if ReturnedStatus == True and       \
                           NextUpdateValid == False:
                            self.log.error( "DriveBCAreas: DriveBCEventGetUpdatesDateTime() returned False for NextUpdate!" )

                            NextUpdate_dt = GetCurrentUTCAware()

                        if DEBUG == True:
                            self.log.debug( "DriveBCAreas: LastUpdated_dt: %s, NextUpdate_dt: %s" % ( LastUpdated_dt, NextUpdate_dt ) )

                        TypeIndex = -1

                        if   Type == "Incident":
                            TypeIndex = 0
                        elif Type == "Road Condition":
                            TypeIndex = 1
                        elif Type == "Current Planned":
                            TypeIndex = 2
                        elif Type == "Future Planned":
                            TypeIndex = 3
                        else:
                            self.log.error( "Unknown DriveBC Event Type: %s" % Type )

                        SelectedEvents.append(
                                               SelectedEventsItem(
                                                                   EventID,
                                                                   Severity,
                                                                   Route,
                                                                   TypeIndex,
                                                                   LastUpdated_dt,
                                                                   NextUpdate_dt,
                                                                   Latitude     = float( Event.get( "geo_lat" ) ),
                                                                   Longitude    = float( Event.get( "geo_long" ) ),
                                                                   CacheHit     = CacheHit,
                                                                   Index        = EventsIndex
                                                                )
                                             )

        # We now have a list of events we are interested in.

        if DEBUG == True:
            self.log.debug( "DriveBCAreas: AFeedReturnedZeroEntries: %s" % AFeedReturnedZeroEntries )

        if AFeedReturnedZeroEntries == False:
            # We received entries for every feed so we will proceed with file cleanup.

            # Get a list of current Drive BC maps.

            # Create a pattern for DriveBC HTML files.
            # We use GetMapHTMLPaths() for consistency.  We won't be 
            # using HTMLFileName, HTMLFilePathRelative or HTMLCacheFilePathAbsolute
            #
            # We use FAKE_ASTERISK and not "*" 'cause GetMapHTMLPaths() parses out 
            # '*' characters 'cause it's not valid in filenames.  We'll replace 
            # FAKE_ASTERISK with '*' later.

            HTMLFileName, HTMLFilePathRelative, DriveBCHTMLFilesPattern, HTMLCacheFilePathAbsolute =       \
                    self.GetMapHTMLPaths( HTMLFilePathStemPrefix + FAKE_ASTERISK )
            assert DriveBCHTMLFilesPattern != ""
            if DEBUG == True:
                self.log.debug( "DriveBCAreas: DriveBCHTMLFilesPattern: %s" % DriveBCHTMLFilesPattern )

            DriveBCHTMLFilesPattern = DriveBCHTMLFilesPattern.replace( FAKE_ASTERISK, "*" )
            if DEBUG == True:
                self.log.debug( "DriveBCAreas: DriveBCHTMLFilesPattern: %s" % DriveBCHTMLFilesPattern )

            DriveBCFilesPath = glob.glob( DriveBCHTMLFilesPattern )
            assert DriveBCFilesPath != None

            for FilePath in DriveBCFilesPath:
                if DEBUG == True:
                    self.log.debug("FilePath: %s" % FilePath )

                CurrentDriveBCHTMLMapFiles.append( FilePath )

            if DEBUG == True:
                self.log.debug( "DriveBCAreas: There are %d items in the currently existing HTML map file list." % len( CurrentDriveBCHTMLMapFiles ) )

#            for p in CurrentDriveBCHTMLMapFiles:
#                self.log.debug("p (FilePath): %s" % p )

        MapJavaScriptFilePath = ""

        if self.GoogleMapsMapsUse == True:
            # Create Java Script for Google Maps Map markers.

            ReturnedStatus = False

            ReturnedStatus, MapJavaScriptFilePath = self.CreateGoogleMapsJavaScript(
                                                                                     FilePathStemPrefix = JavaScriptFilePathStemPrefix + Areas,
                                                                                     Markers            = Markers,
                                                                                     AddStationMarker   = True
                                                                                   )
            assert ReturnedStatus == True       # CreateGoogleMapsJavaScript() should not fail!
            assert MapJavaScriptFilePath != ""

            if ReturnedStatus == False:
                # Something bad did happen so disable map creation.

                MapJavaScriptFilePath = ""

            if DEBUG == True:
                self.log.debug( "DriveBCAreas: MapJavaScriptFilePath: %s" % MapJavaScriptFilePath )

        if DEBUG == True:
            self.log.debug( "DriveBCAreas: %d Events were selected." % len( SelectedEvents) )

#        self.log.debug( "SelectedEvents BEFORE sort:" )
#        for SelectedEvent in SelectedEvents:
#            self.log.debug( "SelectedEvent: %s, %s, %d, %s, %d" % ( SelectedEvent.Severity,
#									  SelectedEvent.Route,
#									  SelectedEvent.TypeIndex,
#									  SelectedEvent.LastUpdated.strftime( "%Y-%m-%d %a %H:%M:%S PST" ), 
#									  SelectedEvent.Index ) )

        # We want to sort SelectedEvents by "severity" then "last updated" to mimic 
        # DriveBC's default presentation.  In fact, it looks like DriveBC's 
        # default sort order is "severity", "route", "last updated" so we'll 
        # do the same.
        #
	    # Descending sort for last updated date so that most recently updated events are first.

        SelectedEvents.sort( key = SelectedEventsItem.GetLastUpdated, reverse = True )
        SelectedEvents.sort( key = SelectedEventsItem.GetTypeIndex )
        SelectedEvents.sort( key = SelectedEventsItem.GetRoute )
        SelectedEvents.sort( key = SelectedEventsItem.GetSeverity )

#        self.log.debug( "SelectedEvents AFTER sort:" )
#        for SelectedEvent in SelectedEvents:
#            self.log.debug( "SelectedEvent: %s, %s, %s, %d" % ( SelectedEvent.Severity,
#								      SelectedEvent.Route,
#								      SelectedEvent.LastUpdated.strftime( "%Y-%m-%d %a %H:%M:%S PST" ), 
#								      SelectedEvent.Index ) )

#        self.log.debug( "DriveBC Events Length = %d, SelectedEvents Length = %d" % ( len( ReturnedFeed.entries ), len( SelectedEvents ) ) )

        # Create links to Drive BC.

        if PositionSpecified == True:

            DriveBCMapViewURL =   DRIVE_BC_MAP_VIEW_URL_BASE                    \
                                + "%f" % Latitude                               \
                                + DRIVE_BC_MAP_VIEW_URL_LAT_LON_SEPARATOR       \
                                + "%f" % Longitude                              \
                                + DRIVE_BC_MAP_VIEW_URL_ZOOM_SEPARATOR          \
                                + "%d" % Zoom
        else:
            DriveBCMapViewURL = DRIVE_BC_MAP_VIEW_URL

        DriveBCLinks =   "<p class=\"WXFeedsP1\">\n"                                                                                    \
                       +     "<b>\n"                                                                                                    \
                       +         "Drive BC:"                                                                                            \
                       +         "&nbsp;"                                                                                               \
                       +         "<a href=\"%s\">Map View</a>\n" % DriveBCMapViewURL                                                    \
                       +         "&nbsp;"                                                                                               \
                       +         "<a href=\"%s\"><span class=\"icon-map\"></span></a>\n" % DriveBCMapViewURL                            \
                       +         "&nbsp; - &nbsp;"                                                                                      \
                       +         "<a href=\"%s\">Conditions & Events</a>\n" % DRIVE_BC_CONDITIONS_AND_EVENTS_URL                        \
                       +         "&nbsp;"                                                                                               \
                       +         "<a href=\"%s\"><span class=\"icon-checklist\"></span></a>\n" % DRIVE_BC_CONDITIONS_AND_EVENTS_URL     \
                       +     "</b>\n"                                                                                                   \
                       + "</p>\n"                                                                                                       \
                       + "<hr>\n"

        if DisplayLinksToDriveBC == BEFORE or    \
           DisplayLinksToDriveBC == BOTH:
            self.WriteGeneratedFiles( DriveBCLinks )

        HTML =   "<p class=\"WXFeedsP1\"><b>Drive BC:&nbsp;%s</b>\n" % Areas                                                \
                + "&nbsp;<a href=\"%s\">%s</a>" % ( DRIVE_BC_URL, "DriveBC Events" )                                        \
                + "&nbsp;<a href=\"%s\"><span class=\"icon-play\"></span></a><br/>\n" % DRIVE_BC_URL                        \
                + "Displaying %d of %d Events For Area(s): %s" % ( len( SelectedEvents ), TotalReturnedEvents, Areas )

        if DistrictFilter != "":
            HTML += "<br/><b>District Filter:</b> %s" % DistrictFilter

        if RouteFilter != "":
           HTML += "<br/><b>Route Filter:</b> %s" % RouteFilter

        if TypeFilter != "":
            HTML += "<br/><b>Type Filter:</b> %s" % TypeFilter

        if SeverityFilter != "":
           HTML += "<br/>S<b>everity Filter:</b> %s" % SeverityFilter

        HTML +=   "</p>\n"      \
                + "<hr>\n"
        self.WriteGeneratedFiles( HTML )

        for SelectedEvent in SelectedEvents:
            if self.GoogleMapsMapsUse == True:
                # Create a map for this event.

                assert MapJavaScriptFilePath != ""

                MapHTMLFileName = ""

                ReturnedStatus, MapHTMLFileName, MapHTMLFilePathAbsolute =       \
                    self.CreateGoogleMapsMapHTML(
                                                  FilePathStemPrefix            = HTMLFilePathStemPrefix + SelectedEvent.ID,
                                                  ForceStaleFile                = not SelectedEvent.CacheHit,
                                                  MapHeader                     = "DriveBC Event - " + SelectedEvent.ID,
                                                  Latitude                      = SelectedEvent.Latitude,
                                                  Longitude                     = SelectedEvent.Longitude,
                                                  Zoom                          = Zoom,
                                                  JavaScriptFilePathRelative    = MapJavaScriptFilePath
                                                )
                if ReturnedStatus == True:
                    assert MapHTMLFileName != ""
                    assert MapHTMLFilePathAbsolute != ""

                    # Remove this map from the list of currently existing maps.

                    try:
                        CurrentDriveBCHTMLMapFiles.remove( MapHTMLFilePathAbsolute )
                    except ValueError as e:
                        # If this is a new event there won't be an existing file so not a error.

                        if DEBUG == True:
                            self.log.debug( "DriveBCAreas: CurrentDriveBCHTMLMapFiles.remove() threw an exception: e = %s, sys.exc_type = %s" %
                                        ( e, sys.exc_info() ) )

                        pass
                else:
                    # Disable map linkage.

                    MapHTMLFileName = ""

                if DEBUG == True:
                    self.log.debug( "DriveBCAreas: MapHTMLFileName: %s" % MapHTMLFileName )

            SeverityClass = "WXFeedsDBCSeverityNormal"

            if SelectedEvent.Severity == "Major":
                SeverityClass = "WXFeedsDBCSeverityMajor"

            HTML =   "<p class=\"WXFeedsP2\">\n"                                                                                \
                   +     "<b>\n"                                                                                                \
                   +         Events[ SelectedEvent.Index ].title                                                                \
                   +     "</b>\n"                                                                                               \
                   +     "&nbsp;<a href=\"%s\">DriveBC Page</a>\n" % Events[ SelectedEvent.Index ].link                         \
                   +     "&nbsp;<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Events[ SelectedEvent.Index ].link
            self.WriteGeneratedFiles( HTML )

            if MapHTMLFileName != "":
#                HTML += "&nbsp;<a href=\"%s\"><span class=\"icon-google-map\"></span></a>\n" % MapHTMLFilePathRelative

                HTML = "&nbsp;<a href=\""
                self.WriteGeneratedFiles( HTML )
                self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
                HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
                self.WriteGeneratedFiles( HTML )

            HTML =   "<br/>\n"                                     \
                   + "<span class=\"%s\">" % SeverityClass         \
                   +     Events[ SelectedEvent.Index ].summary     \
                   + "</span>\n"

            if DisplayAttributes == True:
                HTML +=   "<br/>\n"                                                         \
                        +     "Attributes: "                                                \
                        +         Events[ SelectedEvent.Index ][ 'dbc_district' ] + " | "   \
                        +         Events[ SelectedEvent.Index ][ 'dbc_route' ] + " | "      \
                        +         Events[ SelectedEvent.Index ][ 'dbc_type' ] + " | "       \
                        +         Events[ SelectedEvent.Index ][ 'dbc_severity' ]

            HTML += "</p>\n<hr>\n"
            self.WriteGeneratedFiles( HTML )

#            if DEBUG == True:  # Uncomment and enable DEBUG to stop after 1 event,=.
#            break

#        self.log.debug( "dbc_district: %s" % DBCDistrict )
#        self.log.debug( "dbc_route: %s" % DBCRoute )
#        self.log.debug( "dbc_type: %s" % DBCType )
#        self.log.debug( "dbc_severity: %s" % DBCSeverity )

        if DisplayAttributeLists == True:
            DBCDistrict.sort()
            DBCRoute.sort()
            DBCType.sort()
            DBCSeverity.sort()

            HTML =   "<p class=\"WXFeedsP2\">\n"                        \
                   + "<b>DriveBC Event Attribues:</b><br/>\n"           \
                   + "<b>DBCDistrict:</b> %s<br/>\n" % DBCDistrict      \
                   + "<b>DBCRoute:</b> %s<br/>\n" % DBCRoute            \
                   + "<b>DBCType:</b> %s<br/>\n" % DBCType              \
                   + "<b>DBCSeverity:</b> %s<br/>\n" % DBCSeverity      \
                   + "</p>\n<hr>\n"
            self.WriteGeneratedFiles( HTML )

        if DisplayLinksToDriveBC == AFTER or    \
           DisplayLinksToDriveBC == BOTH:
            self.WriteGeneratedFiles( DriveBCLinks )

        # Delete any map files still in the currently existing map file list.

        if DEBUG == True:
            self.log.debug( "DriveBCAreas: There are %d items left in the currently existing HTML map file list." % len( CurrentDriveBCHTMLMapFiles ) )

        for HTMLMapFilePathAbsolute in CurrentDriveBCHTMLMapFiles:
            if DEBUG == True:
                self.log.debug( "DriveBCAreas: HTMLMapFilePathAbsolute: %s" % HTMLMapFilePathAbsolute )

            p = Path( HTMLMapFilePathAbsolute )
            assert p != None    # Should never fail!

            if DEBUG == True:
                self.log.debug( "DriveBCAreas: Deleting: %s" % HTMLMapFilePathAbsolute )

            try:
                p.unlink()
            except FileNotFoundError as e:
                # We created a list of map related files above so all files should still exist 
                # so we shouldn't get this error.

                self.log.error( "DriveBCAreas: Threw an exception while deleteing map file: e = %s, sys.exc_type = %s" %
                                        ( e, sys.exc_info() ) )

            FileName = p.name
            if DEBUG == True:
                self.log.debug( "DriveBCAreas: Map HTML Cache FileName: %s" % FileName )

            CacheFilePathAbsolute = self.GetCachePath( FileName )
            pc = Path( CacheFilePathAbsolute )

            if pc.exists() == True:     # We might not have a cache file for this map file.
                if DEBUG == True:
                    self.log.debug( "DriveBCAreas: Deleting: %s" % CacheFilePathAbsolute )

                try:
                    pc.unlink()
                except FileNotFoundError as e:
                    # We might not have a cache file for this map file - not an error.
                    # But we shouldn't get his error because we checked existence above.

                    self.log.error( "DriveBCAreas: Threw an exception while deleteing map cache file: e = %s, sys.exc_type = %s" %
                                          ( e, sys.exc_info() ) )

        ReturnStatus = True

        return ReturnStatus

    def DriveBCValidateData(
                             self,
                             StatusCode     : int,
                             DriveBCData    : dict
                           ) -> bool:
                           #    ReturnStatus

        ReturnStatus = False

#        self.log.debug( "DriveBCValidateData: DriveBCData: %s" % DriveBCData )

        if StatusCode == 200 and DriveBCData.entries != None:
#            self.log.debug( "DriveBCValidateData: DriveBCData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "DriveBCValidateData: DriveBCData is NOT valid!" )

            pass

        return ReturnStatus

    def GetCurrentUTCAware(
                             self
                          ):

        d = datetime.datetime.utcnow()
        d = d.replace(tzinfo=datetime.timezone.utc) # Convert it to an aware datetime object in UTC time.

        return d
        
    def GetTimeLocalAware(
                             self
                           , UTCAware = None
                         ):
                         
        if UTCAware == None:
            UTCAware = self.GetCurrentUTCAware()

        d = UTCAware.astimezone() # Convert local timezone (still aware).

        return d

    def NOAANWSForecast(
                         self,
                         Location               : str,
                         Latitude               : float,
                         Longitude              : float,
                         PositionSpecified      : bool,
                         Zoom                   : int,
                         RequestedUnits         : str,
                         GetForecast            : bool,
                         GetHourlyForecast      : bool,
                         HourlyForecastHours    : int
                       ) -> bool:

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        # The NOAA NWS Forecast service API requires the location of the forecast 
        # to be specified using a NOAA NWS unique grid system.
        #
        # NOAA NWS grid positions can be looked up using the NOAA NWS points endpoint.
        # If position (latitude and longitude) is specified, try to get the NOAA NWS grid position using the NOAA NWS points API.

        if PositionSpecified == False:
            # Latitude and longitude not specified so we're done!

            if DEBUG == True:
                self.log.error( "NOAANWSForecast: Latitude and Longitude NOT Specified for Location: %s" % Location )

            ReturnStatus = False
            return ReturnStatus

        # Get NOAA NWS Points.

        assert self.NOAANWSPointsUse == True

        NOAANWSPointsURL = NOAA_NWS_POINTS_SERVICE_URL
        NOAANWSPointsURL += "%f,%f" % ( Latitude, Longitude )

        ReturnedStatus      = False
        CacheHit            = False
        StatusCode          = -1
        NOAANWSPointData    = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            NOAANWSPointData = self.GetEndpointData(
                                                     EndpointURL                = NOAANWSPointsURL,
                                                     ContentType                = CONTENT_JSON,
                                                     ValidateDataFunction       = self.NOAANWSPointValidateData,
                                                     Service                    = "NOAANWSPoint",
                                                     CacheKey                   = Location,
                                                     LocalCacheUse              = self.NOAANWSPointsLocalCacheUse,
                                                     LocalCacheExpiry_minutes   = self.NOAANWSPointsLocalCacheExpiry_minutes,
                                                     LocalCacheForceMiss        = self.NOAANWSPointsLocalCacheForceMiss,
                                                     ServerUse                  = self.NOAANWSPointsServerUse,
                                                     RemoteCacheUse             = self.NOAANWSPointsRemoteCacheUse,
                                                     RemoteCacheExpiry_minutes  = self.NOAANWSPointsRemoteCacheExpiry_minutes,
                                                     RemoteCacheForceMiss       = self.NOAANWSPointsRemoteCacheForceMiss
                                                   )
        except:
            self.log.error( "NOAANWSForecast: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "NOAANWSForecast: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "NOAANWSForecast: GetEndpointData() returned False for Location: %s" % Location )

            ReturnStatus = False
            return ReturnStatus

        if StatusCode != 200:
            self.log.info( "NOAANWSForecast: GetEndpointData() returned successully with unsuccesful status_code: %s" % StatusCode )

            ReturnStatus = False
            return ReturnStatus

        # We should have valid NOAA NWS Points information (in json format).

        WFOTimeZone_str = NOAANWSPointData[ 'properties' ][ 'timeZone' ]
        if DEBUG == True:
            self.log.debug( "NOAANWSForecast: WFOTimeZone_str = >%s<" % WFOTimeZone_str )

        try:
            WFOTimeZone_tzinfo = timezone( WFOTimeZone_str )
        except pytz.exceptions.UnknownTimeZoneError:
            self.log.error( "NOAANWSForecast: NOAA NWS Points Information: Unknown TimeZone: %s - Will Use UTC" % WFOTimeZone_str )

            WFOTimeZone_tzinfo = timezone( "UTC" )

        assert WFOTimeZone_tzinfo != None
        if DEBUG == True:
            self.log.debug( "NOAANWSForecast: WFOTimeZone_tzinfo = %s" % WFOTimeZone_tzinfo )

        DateTimeFormat = ( "%Y-%m-%d %H:%M %Z" )

        Units = RequestedUnits

        if Units != UNITS_US and    \
           Units != UNITS_SI:
            self.log.warning( "NOAA NWS Forecast: Units must be \"%s\" or \"%s\", NOT \"%s\".  Will use \"%s\"." %    \
                                (
                                  UNITS_US,
                                  UNITS_SI,
                                  Units,
                                  UNITS_US
                                )
                            )
            Units = UNITS_US

        if GetForecast == True:
            ForecastURL = NOAANWSPointData[ 'properties' ][ 'forecast' ]
            ForecastURL += "?units=" + Units

            ReturnedStatus  = False
            CacheHit        = False
            StatusCode      = -1
            ForecastData    = {}

            if DEBUG == True:
                self.log.debug( "NOAANWSForecast: ForecastURL: %s" % ForecastURL )

            try:
                ReturnedStatus,         \
                CacheHit,               \
                StatusCode,             \
                ForecastData = self.GetEndpointData(
                                                     EndpointURL                = ForecastURL,
                                                     ContentType                = CONTENT_JSON,
                                                     ValidateDataFunction       = self.NOAANWSForecastValidateData,
                                                     Service                    = "NOAANWSForecast",
                                                     CacheKey                   = Location,
                                                     LocalCacheUse              = self.NOAANWSForecastsLocalCacheUse,
                                                     LocalCacheExpiry_minutes   = self.NOAANWSForecastsLocalCacheExpiry_minutes,
                                                     LocalCacheForceMiss        = self.NOAANWSForecastsLocalCacheForceMiss,
                                                     ServerUse                  = self.NOAANWSForecastsServerUse,
                                                     RemoteCacheUse             = self.NOAANWSForecastsRemoteCacheUse,
                                                     RemoteCacheExpiry_minutes  = self.NOAANWSForecastsRemoteCacheExpiry_minutes,
                                                     RemoteCacheForceMiss       = self.NOAANWSForecastsRemoteCacheForceMiss
                                                   )
            except:
                self.log.error( "NOAANWSForecast: Forecast GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

                ReturnedStatus = False 

            if DEBUG == True:
                self.log.debug( "NOAANWSForecast: Forecast GetEndpointData() returned: %s for:' %s'" % ( ReturnedStatus, Location ) ) 

            if ReturnedStatus == False:
                # GetEndpointData() returned False so we're done!

                self.log.error( "NOAANWSForecast: Forecast GetEndpointData() returned False for Location: %s" % Location )

                ReturnStatus = False
                return ReturnStatus

            # It's possible to get a successful return with an unsuccessful status_code.

            if StatusCode != 200 and      \
               StatusCode != 503:
                self.log.info( "NOAANWSForecast: Forecast GetEndpointData() returned successully with unsuccesful status_code: %s" % StatusCode )

                ReturnStatus = False
                return ReturnStatus

            # We should have a good forecast (200) or an expiry notification (503) - display it.

            assert ForecastData != None

            if DEBUG == True:
                self.log.debug( "NOAANWSForecast: ForecastData: %s" % ForecastData )

            if StatusCode == 200:
                # Should be a good forecast.

                MapHTMLFileName = ""
 
                if self.GoogleMapsMapsUse == True and        \
                   PositionSpecified == True:
                    MarkerInfoWindowContent = Location

                    ReturnedStatus,         \
                    MapHTMLFileName,        \
                    MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                        FilePathStemPrefix      = "NOAANWSForecast-" + Location,
                                                                        ForceStaleFiles         = not CacheHit,
                                                                        Location                = Location,
                                                                        Latitude                = Latitude,
                                                                        Longitude               = Longitude,
                                                                        PositionSpecified       = PositionSpecified,
                                                                        MarkerType              = "misc",
                                                                        MarkerInfoWindowContent = MarkerInfoWindowContent,
                                                                        Zoom                    = Zoom,
                                                                        DisplayStationMarker    = True
                                                                      )
                    if DEBUG == True:
                        self.log.debug( "NOAANWSForecast: CreateGoogleMapsMap() returned: %s." % ReturnedStatus )

                    if ReturnedStatus == True:
                        assert MapHTMLFileName != ""
                        assert MapHTMLFilePathAbsolute != ""

                    if ReturnedStatus == False:
                        MapHTMLFileName = ""

                ForecastProperties = {}
                ForecastProperties = ForecastData[ 'properties' ]
                assert ForecastProperties != None

                ForecastPeriods = []
                ForecastPeriods = ForecastProperties[ 'periods' ]
                assert ForecastPeriods != None

                self.log.info( "NOAANWSForecast: Forecast Received %d entries for '%s'." % ( len( ForecastPeriods ), Location ) )

                # Get valid period.

                # NOAA NWS Forecast feed times are in UTC - they all have UTC Offsets = "+00:00".
                #
                # We'll check just in case NOAA NWS changes their current practice.

                ValidTimes_str = ForecastProperties[ 'validTimes' ]
#                self.log.debug( "Forecast ValidTime_str = %s" % ValidTimes_str )

                x = ValidTimes_str.find( "+00:00" )
                if x < 0:
                    self.log.error( "NOAANWSForecast: Expected UTC Offset = \"+00:00\" Not Found!  ValidTime_str = %s" % ValidTime_str ) 

                x = ValidTimes_str.find( "/" )
                assert x >= 0

                ValidTimesDateTime_str = ValidTimes_str[ 0 : x ]
                ValidTimesDuration_str = ValidTimes_str[ x + 1 : len( ValidTimes_str ) ]
#                self.log.debug( "Forecast ValidTimesDateTime_str = >%s<, ValidTimesDuration_str = >%s<" % ( ValidTimesDateTime_str, ValidTimesDuration_str ) )

                ValidTimesStartDateTime_dt = isodate.parse_datetime( ValidTimesDateTime_str )
                ValidTimesDuration_td = isodate.parse_duration( ValidTimesDuration_str )

                ValidTimesEndDateTime_dt = ValidTimesStartDateTime_dt + ValidTimesDuration_td

#                self.log.debug( "Start UTC = %s, End UTC = %s" % (
#                                                                     ValidTimesStartDateTime_dt.astimezone( timezone( "UTC" ) ).strftime( DateTimeFormat )
#                                                                   , ValidTimesEndDateTime_dt.astimezone( timezone( "UTC" ) ).strftime( DateTimeFormat )
#                                                                 )
#                              )

#                self.log.debug( "Start LT = %s, End LT = %s" % (
#                                                                   ValidTimesStartDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )
#                                                                 , ValidTimesEndDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )
#                                                               )
#                              )

                ForecastUpdatedDateTime_dt = isodate.parse_datetime( ForecastProperties[ 'updated' ] )
#                self.log.debug( "Forecast updated UTC = %s, LT = %s" % (
#                                                                           ForecastUpdatedDateTime_dt.astimezone( timezone( "UTC" ) ).strftime( DateTimeFormat )
#                                                                         , ForecastUpdatedDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )
#                                                                       )
#                              )

                HTML = self.GetDivTag( NOAA_NWS_FORECASTS_HEADER, "Forecast", Location )
                HTML +=   "<p class=\"WXFeedsP1\"><b>%s&nbsp;-&nbsp;Forecast</b>" % Location
                self.WriteGeneratedFiles( HTML )

                if MapHTMLFileName != "":
#                    HTML += "&nbsp;<a href=\"%s\"><span class=\"icon-google-map\"></span></a>\n" % MapHTMLFilePathRelative

                    HTML = "&nbsp;<a href=\""
                    self.WriteGeneratedFiles( HTML )
                    self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
                    HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
                    self.WriteGeneratedFiles( HTML )

                HTML = "</p>\n"
                self.WriteGeneratedFiles( HTML )

                HTML =   "<p class=\"WXFeedsP2\">\n"                                                                                                    \
                       + "<b>Forecast Valid From %s to %s<br/>\n" % (
                                                                      ValidTimesStartDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat ),
                                                                      ValidTimesEndDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )
                                                                    )                                                                                              \
                       +     "Last Updated: %s</b>\n" % ForecastUpdatedDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )                     \
                       + "</p>\n"

                # Put the forecast into a 3 column table:
                #   Col 1: name
                #   Col 2: icon
                #   Col 3: detailedForecast

                HTML +=   "<table class=\"WXFeedsT2\">\n"                               \
                        +     "<thead>\n"                                               \
                        +         "<tr>\n"                                              \
                        +             "<th style=\"padding-right: 1em;\"></th>\n"       \
                        +             "<th></th>\n"                                     \
                        +             "<th style=\"padding-left: 1em;\"></th>\n"        \
                        +         "</tr>\n"                                             \
                        +     "</thead>\n"                                              \
                        + "<tbody>\n"

                for Period in ForecastPeriods:
#                    self.log.debug( "Period: %s" % Period )

                    HTML +=   "<tr>\n"                                                                                                                  \
                         +     "<td style=\"padding-right: 1em;\"><b>%s</b></td>\n" % Period[ "name" ]                                                  \
                         +     "<td><span class=\"icon-noaa-nws-forecast\" style=\"background-image: url('%s')\"></span></td>\n" % Period[ "icon" ]     \
                         +     "<td style=\"padding-left: 1em;\">%s</td>\n" % Period[ "detailedForecast" ]                                              \
                         + "</tr>\n"

                # Close the table.

                HTML += "</tbody>\n</table>\n<hr>\n"

                HTML += "</div>\n"

                self.WriteGeneratedFiles( HTML )
            else:
                # This should be an expiry.

                assert StatusCode == 503
                assert ForecastData[ 'status' ] == 503

                HTML = self.GetDivTag( NOAA_NWS_FORECASTS_HEADER, "Forecast", Location )

                HTML += "<p class=\"WXFeedsP1\"><b>%s&nbsp;-&nbsp;Forecast</b>" % Location
                self.WriteGeneratedFiles( HTML )

                if MapHTMLFileName != "":
#                    HTML += "&nbsp;<a href=\"%s\"><span class=\"icon-google-map\"></span></a>\n" % MapHTMLFilePathRelative

                    HTML = "&nbsp;<a href=\""
                    self.WriteGeneratedFiles( HTML )
                    self.WriteGeneratedFileRelativeLink( MapHTMLFileName )
                    HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
                    self.WriteGeneratedFiles( HTML )

                HTML = "</p>\n"
                self.WriteGeneratedFiles( HTML )

                HTML =   "<p class=\"WXFeedsP2\">\n"                        \
                       + "<b>%s</b><br/>\n" % ForecastData[ 'title' ]       \
                       + "%s\n" % ForecastData[ 'detail' ]                  \
                       + "</p>\n"                                           \
                       + "</div>\n"
                self.WriteGeneratedFiles( HTML )

        if GetHourlyForecast == True:
            HourlyForecastURL = NOAANWSPointData[ 'properties' ][ 'forecastHourly' ]

            HourlyForecastURL += "?units=" + Units
            if DEBUG == True:
                self.log.debug( "NOAANWSForecast: Hourly ForecastURL = %s" % HourlyForecastURL )

            ReturnedStatus      = False
            CacheHit            = False
            StatusCode          = -1
            HourlyForecastData  = {}

            try:
                ReturnedStatus,         \
                CacheHit,               \
                StatusCode,             \
                HourlyForecastData = self.GetEndpointData(
                                                           EndpointURL                  = HourlyForecastURL,
                                                           ContentType                  = CONTENT_JSON,
                                                           ValidateDataFunction         = self.NOAANWSHourlyForecastValidateData,
                                                           Service                      = "NOAANWSHourlyForecast",
                                                           CacheKey                     = Location,
                                                           LocalCacheUse                = self.NOAANWSHourlyForecastsLocalCacheUse,
                                                           LocalCacheExpiry_minutes     = self.NOAANWSHourlyForecastsLocalCacheExpiry_minutes,
                                                           LocalCacheForceMiss          = self.NOAANWSHourlyForecastsLocalCacheForceMiss,
                                                          ServerUse               = self.NOAANWSHourlyForecastsServerUse,
                                                           RemoteCacheUse               = self.NOAANWSHourlyForecastsRemoteCacheUse,
                                                           RemoteCacheExpiry_minutes    = self.NOAANWSHourlyForecastsRemoteCacheExpiry_minutes,
                                                           RemoteCacheForceMiss         = self.NOAANWSHourlyForecastsRemoteCacheForceMiss
                                                         )
            except:
                self.log.error( "NOAANWSForecast: Hourly Forecast GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

                ReturnedStatus = False 

            if DEBUG == True:
                self.log.debug( "NOAANWSForecast: Hourly Forecast GetEndpointData() returned: %s, %s" % ( ReturnedStatus, HourlyForecastData ) ) 

            if ReturnedStatus == False:
                # GetEndpointData() returned False so we're done!

                self.log.error( "NOAANWSForecast: Hourly Forecast GetEndpointData() returned False for Location: %s" % Location )

                ReturnStatus = False
                return ReturnStatus

            if StatusCode != 200:
                self.log.info( "NOAANWSForecast: Hourly Forecast GetEndpointData() returned successully with unsuccesful status_code: %s" % StatusCode )

                ReturnStatus = False
                return ReturnStatus

            # We should have a good forecast - display it.

            MapHTMLFileName = ""
 
            if self.GoogleMapsMapsUse == True and        \
               PositionSpecified == True:
                MarkerInfoWindowContent = Location

                ReturnedStatus,         \
                MapHTMLFileName,        \
                MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                    FilePathStemPrefix      = "NOAANWSHourlyForecast-" + Location,
                                                                    ForceStaleFiles         = not CacheHit,
                                                                    Location                = Location,
                                                                    Latitude                = Latitude,
                                                                    Longitude               = Longitude,
                                                                    PositionSpecified       = PositionSpecified,
                                                                    MarkerType              = "misc",
                                                                    MarkerInfoWindowContent = MarkerInfoWindowContent,
                                                                    Zoom                    = Zoom,
                                                                    DisplayStationMarker    = True
                                                                  )
                if DEBUG == True:
                    self.log.debug( "NOAANWSForecast: CreateGoogleMapsMap() returned: %s." % ReturnedStatus )

                if ReturnedStatus == True:
                    assert MapHTMLFileName != ""
                    assert MapHTMLFilePathAbsolute != ""

                if ReturnedStatus == False:
                    MapHTMLFileName = ""

            ForecastProperties = {}
            ForecastProperties = HourlyForecastData[ 'properties' ]
            assert ForecastProperties != None

            ForecastPeriods = []
            ForecastPeriods = ForecastProperties[ 'periods' ]
            assert ForecastPeriods != None

            self.log.info( "NOAANWSForecast: Hourly Forecast Received %d entries for '%s'." % ( len( ForecastPeriods ), Location ) )

            # Get valid period.

            ValidTimes_str = ForecastProperties[ 'validTimes' ]
#            self.log.debug( "Hourly Forecast ValidTimes_str = %s" % ValidTimes_str )
            
            x = ValidTimes_str.find( "/" )
            assert x >= 0

            ValidTimesDateTime_str = ValidTimes_str[ 0 : x ]
            ValidTimesDuration_str = ValidTimes_str[ x + 1 : len( ValidTimes_str ) ]
#            self.log.debug( "ValidTimesDateTime_str = >%s<, ValidTimesDuration_str = >%s<" % ( ValidTimesDateTime_str, ValidTimesDuration_str ) )

            ValidTimesStartDateTime_dt = isodate.parse_datetime( ValidTimesDateTime_str )
            ValidTimesDuration_td = isodate.parse_duration( ValidTimesDuration_str )

            ValidTimesEndDateTime_dt = ValidTimesStartDateTime_dt + ValidTimesDuration_td

#            self.log.debug( "Start =%s, Duration = %d, End = %s" % (
#                                                                       ValidTimesStartDateTime_dt.strftime( "%Y-%m-%d %H:%M:%S" )
#                                                                     , ValidTimesDuration_td.total_seconds()
#                                                                     , ValidTimesEndDateTime_dt.strftime( "%Y-%m-%d %H:%M:%S" )
#                                                                   ) )

            ForecastUpdatedDateTime_dt = isodate.parse_datetime( ForecastProperties[ 'updated' ] )
#            self.log.debug( "Hourly Forecast updated = %s" % ForecastProperties[ 'updated' ] )

            HTML = self.GetDivTag( NOAA_NWS_FORECASTS_HEADER, "HourlyForecast", Location )

            HTML +=   "<p class=\"WXFeedsP1\"><b>%s&nbsp;-&nbsp;Hourly Forecast</b>" % Location
            self.WriteGeneratedFiles( HTML )

            if MapHTMLFileName != "":
                HTML = "&nbsp;<a href=\""
                self.WriteGeneratedFiles( HTML )
                self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
                HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
                self.WriteGeneratedFiles( HTML )

            HTML = "</p>\n"
            self.WriteGeneratedFiles( HTML )

            HTML =   "<p class=\"WXFeedsP2\">\n"                                                                                                                    \
                   + "<b>Hourly Forecast Valid From %s to %s<br/>\n" % (
                                                                          ValidTimesStartDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat ),
                                                                          ValidTimesEndDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )
                                                                        )                                                                                           \
                   +     "Last Updated: %s</b>\n" % ForecastUpdatedDateTime_dt.astimezone( WFOTimeZone_tzinfo ).strftime( DateTimeFormat )                          \
                   + "</p>\n"

            # Put the forecast into a 3 column table:
            #   Col 1: startTime
            #   Col 2: icon
            #   Col 3: temperature with temperatureUnit
            #   Col 4: windSpeed with windDirection
            #   Col 5: shortForecast

            HTML +=   "<table class=\"WXFeedsT2\">\n"                                       \
                    +     "<thead>\n"                                                       \
                    +         "<tr>\n"                                                      \
                    +             "<th style=\"padding-right: 1em;\">Hour</th>\n"           \
                    +             "<th></th>\n"                                             \
                    +             "<th style=\"padding-left: 1em;\">Temperature</th>\n"     \
                    +             "<th>Wind</th>\n"                                         \
                    +             "<th>Forecast</th>\n"                                     \
                    +         "</tr>\n"                                                     \
                    +     "</thead>\n"                                                      \
                    + "<tbody>\n"

            NumForecastHours = len( ForecastPeriods )

            assert HourlyForecastHours >= 0

            if HourlyForecastHours < NumForecastHours:
                NumForecastHours = HourlyForecastHours

            for p in range( NumForecastHours ):
#                self.log.debug( "Period: %s" % Period )

                ForecastPeriodStartTime_dt = isodate.parse_datetime( ForecastPeriods[ p ][ 'startTime' ] )

                DateTime_str = ""

                DateTimeHour_str = ForecastPeriodStartTime_dt.strftime( "%H" )

                if p == 0 or    \
                   DateTimeHour_str == "00":
                    DateTimeDay_str = ForecastPeriodStartTime_dt.strftime( "%A" )

                    HTML +=   "<tr>\n"                                                                      \
                            +     "<td style=\"padding-right: 1em;\"><b>%s</b></td>\n" % DateTimeDay_str    \
                            +     "<td></td>\n"                                                             \
                            +     "<td style=\"padding-left: 1em;\"></td>\n"                                \
                            +     "<td></td>\n"                                                             \
                            +     "<td></td>\n"                                                             \
                            + "</tr>\n"
          
                HTML +=   "<tr>\n"                                                                                                                                         \
                +     "<td style=\"padding-right: 1em;\"><b>%s</b></td>\n" % DateTimeHour_str                                                                              \
                +     "<td><span class=\"icon-noaa-nws-hourly-forecast\" style=\"background-image: url('%s')\"></span></td>\n" % ForecastPeriods[ p ][ 'icon' ]            \
                +     "<td style=\"padding-left: 1em;\">%s &nbsp;&deg;%s</td>\n" % ( ForecastPeriods[ p ][ 'temperature' ], ForecastPeriods[ p ][ 'temperatureUnit' ] )    \
                +     "<td>%s %s</td>\n" % ( ForecastPeriods[ p ][ 'windSpeed' ], ForecastPeriods[ p ][ 'windDirection' ] )                                                \
                +     "<td>%s</td>\n" % ForecastPeriods[ p ][ 'shortForecast' ]                                                                                            \
                + "</tr>\n"

            # Close the <table> and <div>.

            HTML += "</tbody>\n</table>\n<hr>\n"
            HTML += "</div>\n"
            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def NOAANWSPointValidateData(
                                  self,
                                  StatusCode        : int,
                                  NOAANWSPointData  : dict
                                ) -> bool:
                                #    ReturnStatus

        ReturnStatus = False

        # 'id': 'https://api.weather.gov/points/48.9391,-119.4356', 

        if StatusCode == 200 and                    \
           NOAANWSPointData[ 'id' ] != None and    \
           NOAANWSPointData[ 'id' ].find( "https://api.weather.gov/points/" ) == 0:
#            self.log.debug( "NOAANWSPointData is valid." )

            ReturnStatus = True
        else:
#            self.log.debug( "NOAANWSPointData is NOT valid!" )

            pass

        return ReturnStatus

    def NOAANWSForecastValidateData(
                                     self,
                                     StatusCode     : str,
                                     ForecastData   : dict
                                   ) -> bool:
                                   #    ReturnStatus

        ReturnStatus = False

        if StatusCode == 200 or     \
           StatusCode == 503:

            if StatusCode == 200:
                ForecastProperties = ForecastData[ 'properties' ]

                if ForecastProperties != None and           \
                   ForecastProperties[ 'periods' ] != None:
                    ReturnStatus = True
            else:
#{
#	'correlationId': '5352f1a', 
#	'title': 'Forecast Grid Expired', 
#	'type': 'https://api.weather.gov/problems/ForecastGridExpired', 
#	'status': 503, 
#	'detail': 'The requested forecast grid was issued 2020-10-22T09:51:19+00:00 and has expired.', 
#	'instance': 'https://api.weather.gov/requests/5352f1a'

                assert StatusCode == 503

                Title = Forecast[ 'title' ]

                if Title != None and        \
                   Title == "Forecast Grid Expired":
                    ReturnStatus = True

        return ReturnStatus

    def NOAANWSHourlyForecastValidateData(
                                           self,
                                           StatusCode           : str,
                                           HourlyForecastData   : dict
                                         ) -> bool:
                                         #    ReturnStatus

        ReturnStatus = False

        if StatusCode == 200:
            ForecastProperties = HourlyForecastData[ 'properties' ]

            if ForecastProperties[ 'periods' ] != None and  \
               ForecastProperties[ 'validTimes' ] != None:

                ReturnStatus = True

        return ReturnStatus

    def NOAANWSSPC(
                    self,
                    Location            : str,
                    FeedURL             : str,
                    Latitude            : float,
                    Longitude           : float,
                    PositionSpecified   : bool,
                    Zoom                : int
                  ) -> bool:
                  #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        EndpointURL = FeedURL

        if self.WXFeedsServerUse == True:
            assert self.WXFeedsServerAccessToken != ""

            EndpointURL = "/NOAANWSSPC/alerts?location=%s&url=%s" % ( Location.replace( " ", "+" ), EndpointURL )
            if DEBUG == True:
                self.log.debug( "NOAANWSSPC: EndpointURL: %s" % EndpointURL )

        ReturnedStatus      = False
        CacheHit            = False
        StatusCode          = -1
        NOAANWSSPCData      = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            NOAANWSSPCData = self.FOOGetEndpointData(
                                                   EndpointURL                      = EndpointURL,
                                                   ContentType                      = CONTENT_XML,
                                                   ValidateDataFunction             = self.NOAANWSSPCValidateData,
                                                   LocalCacheUse                    = self.NOAANWSSPCUseLocalCache,
                                                   LocalCacheStem                   = Location + "-NOAANWSSPCAlert",
                                                   LocalCacheExpiry_minutes         = self.NOAANWSSPCCacheExpiry_minutes,
                                                   WXFeedsServerUse            = self.WXFeedsServerUse,
                                                   WXFeedsServerForceCacheMiss = self.NOAANWSSPCForceCacheMiss
                                                 )
        except:
            self.log.error( "NOAANWSSPC: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "NOAANWSSPC: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "NOAANWSSPC: GetEndpointData() returned False for Location: %s" % Location )

            ReturnStatus = False
            return ReturnStatus

        if StatusCode != 200:
            self.log.info( "NOAANWSSPC: GetEndpointData() returned successully with unsuccesful status_code: %s" % StatusCode )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "NOAANWSSPC: Received %d entries for '%s'." % ( len( NOAANWSSPCData.entries ), Location ) )

        MapHTMLFileName = ""
 
        if self.GoogleMapsMapsUse == True and    \
           PositionSpecified == True:
            MarkerInfoWindowContent = Location

            ReturnedStatus,         \
            MapHTMLFileName,        \
            MapHTMLFilePathAbsolute = self.CreateGoogleMapsMap(
                                                                FilePathStemPrefix        = "NOAANWSSPCAlert-" + Location,
                                                                ForceStaleFiles           = not CacheHit,
                                                                Location                  = Location,
                                                                Latitude                  = Latitude,
                                                                Longitude                 = Longitude,
                                                                PositionSpecified         = PositionSpecified,
                                                                Zoom                      = Zoom,
                                                                MarkerType                = "misc",
                                                                MarkerInfoWindowContent   = MarkerInfoWindowContent,
                                                                DisplayStationMarker      = True
                                                              )
            if DEBUG == True:
                self.log.debug( "NOAANWSSPC: CreateGoogleMapsMap() returned: %s." % ReturnedStatus )
                self.log.debug( "NOAANWSSPC: MapHTMLFileName: '%s'" % MapHTMLFileName)
                self.log.debug( "NOAANWSSPC: MapHTMLFilePathAbsolute: '%s'" % MapHTMLFilePathAbsolute)
                    
                assert MapHTMLFileName != ""
                assert MapHTMLFilePathAbsolute != ""

            if ReturnedStatus == False:
                MapHTMLFileName = ""

        HTML = "<p class=\"WXFeedsP1\"><b>%s</b>\n" % Location
        self.WriteGeneratedFiles( HTML )

        if MapHTMLFileName != "":
            HTML = "&nbsp;\n<a href=\""
            self.WriteGeneratedFiles( HTML )
            self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
            HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
            self.WriteGeneratedFiles( HTML )

        # We received what looked like a good response but we received 0 entries and 
        # 'NOAANWSData.feed.title' caused an "AttributeError: object has no attribute 'title'"! so:
        #
        # - use "NOAANWSSPCData.feed[ 'title' ]" same as in validate function and 
        # - test and report error if it doesn't exist.

        if len( NOAANWSSPCData.entries ) == 0 or           \
           NOAANWSSPCData.feed[ 'title' ] == None or        \
           NOAANWSSPCData.feed[ 'title' ] == "":
            self.log.error( "NOAANWSSPC: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d: Have %d entries, or no 'title' in: %s" %     \
                                ( ReturnedStatus, CacheHit, StatusCode, len( NOAANWSSPCData.entries ), NOAANWSSPCData ) )

            ReturnStatus = False
            return ReturnStatus

        HTML = "<br/>%s</p>\n" % NOAANWSSPCData.feed[ 'title' ]
        self.WriteGeneratedFiles( HTML )

        for Entry in NOAANWSSPCData.entries:
#            self.log.debug( "SPC Entry: %s" % Entry )

            # We need to determine if there are, or are not, active alerts 
            # because if there are not, not all the elements we refer to 
            # will be present.

            ActiveAlerts = True    # Assume we will be getting alerts.
            AlertClass = ""        # But we don't know what kind of alert.

            if   Entry.title == NOAA_NWS_SPC_NO_ACTIVE_ALERTS:
#                 self.log.debug( "%s for %s" % (NOAA_NWS_SPC_NO_ACTIVE_ALERTS, Location ) )

                ActiveAlerts = False

            # We only report "Actual" alerts.

            if ActiveAlerts == True and    \
               Entry.cap_status != "Actual":
                continue

            if   ActiveAlerts == False:
                AlertClass = "WXFeedsNoActiveAlerts"
            elif Entry.cap_severity == "Minor":
                AlertClass = "WXFeedsMinor"
            elif Entry.cap_severity == "Moderate":
                AlertClass = "WXFeedsModerate"
            elif Entry.cap_severity == "Severe":
                AlertClass = "WXFeedsSevere"
            elif Entry.cap_severity == "Extreme":
                AlertClass = "WXFeedsExtreme"
            else:
                 # Unknown severity - actually Unknown or some other value.

                AlertClass = "WXFeedsExtreme"

            assert AlertClass != ""

            HTML =   "<p class=\"WXFeedsP2\">\n"                        \
                   +     "<b>\n"                                        \
                   +         "<span class=\"%s\">\n" % AlertClass       \
                   +             "&nbsp;\n"

            if ActiveAlerts == True:
                HTML +=   "<span class=\"icon-warning-sign\"></span>\n"    \
                        + "&nbsp;\n"

            HTML +=   Entry.title    \
                    + "\n&nbsp;\n"

            if ActiveAlerts == True:
                HTML +=   "<span class=\"icon-warning-sign\"></span>\n"    \
                        + "&nbsp;\n"

            HTML += "</span></b>\n"

            if ActiveAlerts == True:
                HTML +=   "&nbsp;\n"                                                                    \
                        + "<a href=\"%s\">Full Alert</a>\n" % Entry.link                                \
                        + "&nbsp;\n"                                                                    \
                        + "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link         \
                        + "<br/>\n"                                                                     \
                        + Entry.summary                                                                 \
                        + "</p>\n"                                                                      \
                        +     "<table class=\"WXFeedsT2\">\n"                                           \
                        +         "<thead>\n"                                                           \
                        +             "<tr>\n"                                                          \
                        +                 "<th>Urgency</th>\n"                                          \
                        +                 "<th>Severity</th>\n"                                         \
                        +                 "<th>Certainty</th>\n"                                        \
                        +             "</tr>\n"                                                         \
                        +         "</thead>\n"                                                          \
                        +         "<tbody>\n"                                                           \
                        +             "<tr>\n"                                                          \
                        +                 "<td>%s</td>\n" % Entry.cap_urgency                           \
                        +                 "<td>\n"                                                      \
                        +                     "<b>\n"                                                   \
                        +                         "<span class=\"%s\">\n" % AlertClass                  \
                        +                             "&nbsp;\n"                                        \
                        +                             "<span class=\"icon-warning-sign\"></span>\n"     \
                        +                             "&nbsp;\n"                                        \
                        +                             Entry.cap_severity                                \
                        +                             "\n&nbsp;\n"                                      \
                        +                             "<span class=\"icon-warning-sign\"></span>\n"     \
                        +                             "&nbsp;\n"                                        \
                        +                         "</span>\n"                                           \
                        +                     "</b>\n"                                                  \
                        +                 "</td>\n"                                                     \
                        +                 "<td>%s</td>\n" % Entry.cap_certainty                         \
                        +             "</tr>\n"                                                         \
                        +         "</tbody>\n"                                                          \
                        +     "</table>\n"
            else:
                HTML += "</p>\n"

            HTML += "<hr>"
            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def NOAANWSSPCValidateData(
                                self,
                                StatusCode          : str,
                                NOAANWSSPCData      : dict
                              ) -> bool:
                              #    ReturnStatus

        ReturnStatus = False

        if StatusCode == 200 or        \
           NOAANWSSPCData.feed[ 'title' ] != None:
            ReturnStatus = True
        else:
            ReturnStatus = False

        return ReturnStatus

    def NOAANWSNHCBasin(
                         self,
                         Basin      : str,
                         FeedURL    : str,
                       ) -> bool:
                       #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        EndpointURL = FeedURL

        if self.WXFeedsServerUse == True:
            assert self.WXFeedsServerAccessToken != ""

            EndpointURL = "/NOAANWSNHC?basin=%s&url=%s" % ( Basin.replace( " ", "+" ), EndpointURL )

        ReturnedStatus      = False
        CacheHit            = False
        StatusCode          = -1
        NOAANWSNHCBasinData = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            NOAANWSNHCBasinData = self.FOOGetEndpointData(
                                                        EndpointURL                        = EndpointURL,
                                                        ContentType                        = CONTENT_XML,
                                                        ValidateDataFunction               = self.NOAANWSNHCBasinValidateData,
                                                        LocalCacheUse                      = self.NOAANWSNHCUseLocalCache,
                                                        LocalCacheStem                     = Basin + "-NOAANWSHNCBasin",
                                                        LocalCacheExpiry_minutes           = self.NOAANWSNHCCacheExpiry_minutes,
                                                        WXFeedsServerUse              = self.WXFeedsServerUse,
                                                        WXFeedsServerForceCacheMiss   = self.NOAANWSNHCForceCacheMiss
                                                      )
        except:
            self.log.error( "NOAANWSNHCBasin: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "NOAANWSNHCBasin: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "NOAANWSNHCBasin: GetEndpointData() returned False for Basin: '%s'" % Basin )

            ReturnStatus = False
            return ReturnStatus

        if StatusCode != 200:
            self.log.info( "NOAANWSNHCBasin: GetEndpointData() returned successully with unsuccesful ReturnStatus: %s" % StatusCode )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "NOAANWSNHCBasin: Received %d entries for '%s'." % ( len( NOAANWSNHCBasinData.entries ), Basin ) )

        HTML =   "<p class=\"WXFeedsP1\"><b>%s</b><br/>\n" % Basin                   \
               + "%s<br/>%s</p>\n" % ( NOAANWSNHCBasinData.feed.title, NOAANWSNHCBasinData.feed.subtitle )
        self.WriteGeneratedFiles( HTML )

        for Entry in NOAANWSNHCBasinData.entries:
#            self.log.debug( "NHC Entry: %s" % Entry )

            HTML =   "<p class=\"WXFeedsP2\">\n"                                                  \
                   +     "<b>\n"                                                                  \
                   +         Entry.title                                                          \
                   +     "</b>\n"                                                                 \
                   +     "&nbsp;\n"                                                               \
                   +     "<a href=\"%s\">NOAA NWS - NHC Page</a>\n" % Entry.link                  \
                   +     "&nbsp;\n"                                                               \
                   +     "<a href=\"%s\"><span class=\"icon-play\"></span></a>\n" % Entry.link    \
                   +     "<br/>\n"                                                                \
                   +     Entry.summary                                                            \
                   + "</p>\n"                                                                     \
                   + "<hr>\n"
            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def NOAANWSNHCBasinValidateData(
                                     self,
                                     StatusCode             : str,
                                     NOAANWSNHCBasinData    : dict
                                   ) -> bool:
                                   #    ReturnStatus

        ReturnStatus = False

        if StatusCode == 200 or        \
           len( NOAANWSNHCBasinData.entries ) >= 0:
            ReturnStatus = True
        else:
            ReturnStatus = False

        return ReturnStatus

    def WSDOTBorderCrossings(
                              self,
                              Zoom: int
                            ) -> bool:

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        FeedURL = ""

        # Get feed URL from WS DOT dictionary.

        FeedURL = self.WSDOTDictionary.get( WS_DOT_BORDER_CROSSINGS_KEY, "" )

        if FeedURL == "":
            self.log.warning( "WSDOTBorderCrossings: WS-DOT Traveler Information Subject NOT in WS-DOT Dictionary!" )
            return ReturnStatus

        if DEBUG == True:
            self.log.debug( "WSDOTBorderCrossings: FeedURL: %s" % FeedURL )

        EndpointURL = FeedURL

        if self.WXFeedsServerUse == True:
            assert self.WXFeedsServerAccessToken != ""

            EndpointURL = "/WSDOTServices/borderCrossings?url=%s" % EndpointURL
        else:
            assert self.WSDOTServicesAccessCode != ""

            EndpointURL += "?AccessCode=" + self.WSDOTServicesAccessCode

        if DEBUG == True:
            self.log.debug( "WSDOTBorderCrossings: EndpointURL: %s" % EndpointURL )

        ReturnedStatus              = False
        CacheHit                    = False
        StatusCode                  = -1
        WSDOTBorderCrossingsData    = {}

        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            WSDOTBorderCrossingsData = self.FOOGetEndpointData(
                                                             EndpointURL                        = EndpointURL,
                                                             ContentType                        = CONTENT_JSON,
                                                             ValidateDataFunction               = self.NOAANWSSPCValidateData,
                                                             LocalCacheUse                      = self.WSDOTBorderCrossingsUseLocalCache,
                                                             LocalCacheStem                     = "WSDOT-BorderCrossingsData",
                                                             LocalCacheExpiry_minutes           = self.WSDOTBorderCrossingsCacheExpiry_minutes,
                                                             WXFeedsServerUse              = self.WXFeedsServerUse,
                                                             WXFeedsServerForceCacheMiss   = self.WSDOTBorderCrossingsForceCacheMiss
                                                           )
        except:
            self.log.error( "WSDOTBorderCrossings: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 

        if DEBUG == True:
            self.log.debug( "WSDOTBorderCrossings: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "WSDOTBorderCrossings: GetEndpointData() returned False!" )

            ReturnStatus = False
            return ReturnStatus

        if StatusCode != 200:
            self.log.info( "WSDOTBorderCrossings: GetEndpointData() returned successully with unsuccesful status_code: %s" % StatusCode )

            ReturnStatus = False
            return ReturnStatus

        self.log.info( "WSDOTBorderCrossings: Received %d entries." % len( WSDOTBorderCrossingsData ) )

        # Get the reading time from the first Border Crossing entry.

        BorderCrossing = WSDOTBorderCrossingsData[ 0 ]
        ReadingTime = BorderCrossing.get( "Time" )

        # 'Time': '/Date(1598370300000-0700)/',

        ReadingTime_dt = datetime.datetime.fromtimestamp( int( ReadingTime[ 6 : 16 ] ) )
        ReadingTime = ReadingTime_dt.strftime( "%-I:%M %p %A, %B %-d, %Y" )
#        self.log.debug( "ReadingTime = %s" % ReadingTime )

        HTML =   "<p class=\"WXFeedsP1\"><b>Washington State - DOT - Border Crossings<br/>\n"    \
               +     "Northbound Border Wait Times</b><br/>\n"                                              \
               +     "&nbsp;" + ReadingTime + "\n"                                                          \
               + "</p>\n"

        self.WriteGeneratedFiles( HTML )

        # A list of markers for Google Maps.

        Markers = []

        MarkersMapHeader                = "WSDOT - Border Crossings"
        HTMLFilePathStemPrefix          = "WSDOT-BorderCrossing-"
        JavaScriptFilePathStemPrefix    = "WSDOT-BorderCrossings"

        GoogleMapsJavaScriptFileName = ""

        if self.GoogleMapsMapsUse == True:
            # We're going to use Google Maps.

            # Create a Google Maps marker for each BoorderCrossingLocation.
            # Add to Text for each CrossingName.

            MarkerType              = "misc"
            MarkerText              = ""
            MarkerInfoWindowContent = "<a href=\"%s\">%s</a>" % ( WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_URL, WASHINGTON_STATE_DEPARTMENT_OF_TRANSPORTATION_URL )

            BorderCrossingLocationOpen  = False
            MarkerCount                 = 0

            for BorderCrossing in WSDOTBorderCrossingsData:
                if DEBUG == True:
                    self.log.debug( "WSDOTBorderCrossings: BorderCrossing: %s" % BorderCrossing )

                BorderCrossingLocation = BorderCrossing[ 'BorderCrossingLocation' ]

                if DEBUG == True:
                    if BorderCrossingLocation != None:
                        self.log.debug( "WSDOTBorderCrossings: BorderCrossingLocation: %s" % BorderCrossingLocation )

                if BorderCrossingLocation != None:
                    # We want to create a marker for this border crossing location.
                    # But we might have to complete a marker for a previous crossing location.

                    if BorderCrossingLocationOpen == True:
                        # Complete marker for previous crossing location.

                        Markers.append( 
                                        GoogleMapsMarker(
                                                          MarkerType,
                                                          Latitude,
                                                          Longitude,
                                                          MarkerText,
                                                          MarkerInfoWindowContent
                                                        )
                                      )
                        MarkerCount += 1

                        BorderCrossingLocationOpen = False
                else:
                    # BorderCrossingLocation == None

                    pass

                if BorderCrossingLocationOpen == False and      \
                   BorderCrossingLocation != None:
                    # Initiate a marker for this border crossing location.

                    Latitude                = BorderCrossingLocation[ 'Latitude' ]
                    Longitude               = BorderCrossingLocation[ 'Longitude' ]
                    MarkerText              = BorderCrossingLocation[ 'Description' ]

                # Add CrossingName WaitTime.

                CrossingName = BorderCrossing[ 'CrossingName' ]
                assert CrossingName != None

                if DEBUG == True:
                    self.log.debug( "WSDOTBorderCrossings: CrossingName: %s" % CrossingName )

                # Insert the CrossingName information.

                WaitTime = BorderCrossing[ 'WaitTime' ]
                assert WaitTime != ""

                WaitTimeMessage = ""
                if WaitTime == -1:
                    WaitTimeMessage = "Closed"
                elif WaitTime <= 5:
                    WaitTimeMessage = "Less Than 5 Min."
                else:
                    WaitTimeMessage = str( WaitTime ) + " Min."
                assert WaitTimeMessage != ""
 
                MarkerText += "\\n  " + CrossingName + "\\n    " + WaitTimeMessage

                BorderCrossingLocationOpen = True

            # Finish an open border crossing location.

            if BorderCrossingLocationOpen == True:
                Markers.append( 
                                GoogleMapsMarker(
                                                  MarkerType,
                                                  Latitude,
                                                  Longitude,
                                                  MarkerText,
                                                  MarkerInfoWindowContent
                                                )
                                      )
                MarkerCount += 1

                BorderCrossingLocationOpen = False

            assert BorderCrossingLocationOpen == False

            if DEBUG == True:
                self.log.debug( "WSDOTBorderCrossings: MarkerCount: %d" % MarkerCount )

            MapJavaScriptFilePath = ""

            # Create Java Script for Google Maps Map markers.

            ReturnedStatus = False

            ReturnedStatus, MapJavaScriptFilePath = self.CreateGoogleMapsJavaScript(
                                                                                     FilePathStemPrefix = JavaScriptFilePathStemPrefix,
                                                                                     Markers            = Markers,
                                                                                     AddStationMarker   = True
                                                                                   )
            assert ReturnedStatus == True       # CreateGoogleMapsJavaScript() should not fail!
            assert MapJavaScriptFilePath != ""

            if ReturnedStatus == False:
                # Something bad did happen so disable map creation.

                MapJavaScriptFilePath = ""

            if DEBUG == True:
                self.log.debug( "WSDOTBorderCrossings: MapJavaScriptFilePath: %s" % MapJavaScriptFilePath )

        # Scan border crossings to create HTML.

        HTML                    = ""
        MapHTMLFilePathRelative = ""

        BorderCrossingLocationOpen  = False
        CrossingLocationCount       = 0
        CrossingCount               = 0

        # Put each Border Crossing into it's own table.

        for BorderCrossing in WSDOTBorderCrossingsData:
            if DEBUG == True:
                self.log.debug( "WSDOTBorderCrossings: BorderCrossing: %s" % BorderCrossing )

            BorderCrossingLocation = BorderCrossing.get( "BorderCrossingLocation", None )
            if DEBUG == True:
                self.log.debug( "WSDOTBorderCrossings: BorderCrossingLocation = %s" % BorderCrossingLocation )

            if BorderCrossingLocation != None:
                # This is a new border crossing location - start a new table.
                # But first we might have to finish a table for a previous 
                # crossing location.

                if BorderCrossingLocationOpen == True:
                    # Finish the open table.

                    HTML +=        "</tbody>\n"    \
                             + "</table>\n"        \
                             + "<hr>\n"
                    self.WriteGeneratedFiles( HTML )

                    CrossingLocationCount += 1

                    BorderCrossingLocationOpen = False
            else:
                # BorderCrossingLocation == None

                pass

            if BorderCrossingLocationOpen == False and      \
               BorderCrossingLocation != None:
                # Initiate a table for this border crossing location.

                Description = BorderCrossingLocation[ 'Description' ]
                Latitude    = BorderCrossingLocation[ 'Latitude' ]
                Longitude   = BorderCrossingLocation[ 'Longitude' ]

                HTML =   "<table class=\"WXFeedsT2\">\n"               \
                       +     "<thead>\n"                               \
                       +         "<tr>\n"                              \
                       +             "<th></th>\n"                     \
                       +             "<th></th>\n"                     \
                       +             "<th></th>\n"                     \
                       +         "</tr>\n"                             \
                       +     "</thead>\n"                              \
                       +     "<tbody>\n"
                self.WriteGeneratedFiles( HTML )

                BorderCrossingLocationOpen = True

                if self.GoogleMapsMapsUse == True:
                    # Create a Google Map.

                    assert MapJavaScriptFilePath != ""

                    MapHeader = "WSDOT - Border Crossing - %s" % Description

                    ReturnedStatus,         \
                    MapHTMLFileName,        \
                    MapHTMLFilePathAbsolute = self.CreateGoogleMapsMapHTML(
                                                                            FilePathStemPrefix         = MapHeader,
                                                                            ForceStaleFile             = not CacheHit,
                                                                            MapHeader                  = MapHeader,
                                                                            Latitude                   = Latitude,
                                                                            Longitude                  = Longitude,
                                                                            Zoom                       = Zoom,
                                                                            JavaScriptFilePathRelative = MapJavaScriptFilePath
                                                                          )
                    if ReturnedStatus == True:
                        assert MapHTMLFileName != ""
                        assert MapHTMLFilePathAbsolute != ""
                    else:
                        # Disable display of map icon.

                        MapHTMLFileName = ""

                    if DEBUG == True:
                        self.log.debug( "WSDOTBorderCrossings: MapHTMLFileName: %s" % MapHTMLFileName )

                if MapHTMLFileName != "":
                    # We're going to display a map icon so only display Description.

#                    HTML =   "<tr>\n"                                                                                                  \
#                           +     "<td><b>Description:&nbsp;</b></td>\n"                                                                \
#                           +     "<td>%s</td>\n" % Description                                                                         \
#                           +     "<td><a href=\"%s\"><span class=\"icon-google-map\"></span></a></td>\n" % MapHTMLFilePathRelative     \
#                           + "</tr>\n"

                    HTML =   "<tr>\n"                                                                                                  \
                           +     "<td><b>Description:&nbsp;</b></td>\n"                                                                \
                           +     "<td>%s</td>\n" % Description                                                                         \
                           +     "<td><a href=\""
                    self.WriteGeneratedFiles( HTML )
                    self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
                    HTML = "\"><span class=\"icon-google-map\"></span></a></td>\n</tr>\n"
                    self.WriteGeneratedFiles( HTML )

                else:
                    # We're not going to display a map icon so display Description, Latitude and Longitude.

                    HTML =   "<tr>\n"                                                                          \
                           +     "<td><b>Description:&nbsp;</b></td> <td>%s</td> <td></td>\n" % Description    \
                           + "</tr>\n"                                                                         \
                           + "<tr>\n"                                                                          \
                           +     "<td><b>Latitude:&nbsp;</b></td>    <td>%f</td> <td></td>\n" % Latitude       \
                           + "</tr>\n"                                                                         \
                           + "<tr>\n"                                                                          \
                           +     "<td><b>Longitude:&nbsp;</b></td>   <td>%f</td> <td></td>\n" % Longitude      \
                           + "</tr>\n"
                    self.WriteGeneratedFiles( HTML )

                CrossingCount = 0

            assert BorderCrossingLocationOpen == True

            if CrossingCount > 0:
                # Insert an empty row between crossings at the same location.

                HTML = "<tr><td>&nbsp;</td> <td>&nbsp;</td> <td>&nbsp;</td></tr>\n"
                self.WriteGeneratedFiles( HTML )

            # Insert the CrossingName information.

            CrossingName = BorderCrossing.get( "CrossingName" )
            WaitTime = BorderCrossing.get( "WaitTime" )

            WaitTimeMessage = ""
            if   WaitTime == -1:
                WaitTimeMessage = "Closed"
            elif WaitTime <= 5:
                WaitTimeMessage = "Less Than 5 Min."
            else:
                WaitTimeMessage = str( WaitTime ) + " Min."
            assert WaitTimeMessage != ""

            HTML =   "<tr>\n"                                                                                 \
                   +     "<td></td>  <td><b>Crossing Name:&nbsp;</b></td> <td>%s</td>\n" % CrossingName       \
                   + "</tr>\n"                                                                                \
                   +  "<tr>\n"                                                                                \
                   +      "<td></td> <td><b>Wait Time:&nbsp;</b></td>     <td>%s</td>\n" % WaitTimeMessage    \
                   +  "</tr>\n"
            self.WriteGeneratedFiles( HTML )

            CrossingCount += 1

        if BorderCrossingLocationOpen == True:
            # Finish an open table for a crossing location.

            HTML =        "</tbody>\n"    \
                   + "</table>\n"         \
                   + "<hr>\n"
            self.WriteGeneratedFiles( HTML )

            CrossingLocationCount += 1

            BorderCrossingLocationOpen = False

        assert BorderCrossingLocationOpen == False

        if DEBUG == True:
            self.log.debug( "WSDOTBorderCrossings: CrossingLocationCount: %d" % CrossingLocationCount )

        assert CrossingLocationCount == MarkerCount

        ReturnStatus = True

        return ReturnStatus

    def WSDOTBorderCrossingsValidateData(
                                self,
                                StatusCode          : str,
                                NOAANWSSPCData      : dict
                              ) -> bool:
                              #    ReturnStatus

        ReturnStatus = False

        if StatusCode == 200 or        \
           len( BorderCrossingsData.feed ) >= 0:
            ReturnStatus = True
        else:
            ReturnStatus = False

        return ReturnStatus

    def Alberta511Ferries(
                             self
                           , FeedURL : str
                         ):

        ReturnStatus = False

        try:
            Response = requests.get( FeedURL )
        except:
            self.log.error("Alberta511Ferries: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511Ferries: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        Ferries = Response.json()
#        self.log.debug( "Data for %d Ferries received." % len( Ferries ) )

        self.log.info( "Alberta511Ferries: Received %d entries." % len( Ferries ) )

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Ferries</b></p>\n"

        self.WriteGeneratedFiles( HTML )

        for Ferry in Ferries:
            # Put each Ferry into it's own table.

            FerryID = Ferry.get( "Id" )

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                                                       \
                   +     "<thead>\n"                                                                                                                                  \
                   +         "<tr>\n"                                                                                                                                 \
                   +             "<th></th>\n"                                                                                                                        \
                   +             "<th></th>\n"                                                                                                                        \
                   +         "</tr>\n"                                                                                                                                \
                   +     "</thead>\n"                                                                                                                                 \
                   +     "<tbody>\n"                                                                                                                                  \
                   +         "<tr>\n"                                                                                                                                 \
                   +             "<td><b>Name:&nbsp;</b></td>\n"                                                                                                      \
                   +             "<td>%s&nbsp;<a href=\"https://511.alberta.ca/map#FerryTerminals-%s\">511 Alberta Map</a></td>\n" % ( Ferry.get( "Name" ), FerryID ) \
                   +         "</tr>\n"                                                                                                                                \
                   +         "<tr><td><b>Description:&nbsp;</b></td>    <td>%s</td></tr>\n" % Ferry.get( "Description" )                                              \
                   +         "<tr><td><b>Status:&nbsp;</b></td>         <td>%s</td></tr>\n" % Ferry.get( "Status" )                                                   \
                   +         "<tr><td><b>Id:&nbsp;</b></td>             <td>%s</td></tr>\n" % FerryID                                                                 \
                   +         "<tr><td><b>Latitude:&nbsp;</b></td>       <td>%f</td></tr>\n" % Ferry.get( "Latitude" )                                                 \
                   +         "<tr><td><b>Longitude:&nbsp;</b></td>      <td>%f</td></tr>\n" % Ferry.get( "Longitude" )                                                \
                   +         "<tr><td><b>Address:&nbsp;</b></td>        <td>%s</td></tr>\n" % Ferry.get( "Address" )                                                  \
                   +     "</tbody>\n"                                                                                                                                 \
                   + "</table>\n"                                                                                                                                     \
                   + "<hr>\n"

            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def Alberta511Parks(
                           self
                         , FeedURL : str
                         , ParkNameFilterCondition : str
                         , ParkNameFilter : str
                       ):

        ReturnStatus = False

        try:
            Response = requests.get( FeedURL )
        except:
            self.log.error("Alberta511Parks: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511Parks: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        Parks = Response.json()
#        self.log.debug( "Data for %d Parks received." % len( Parks ) )

        self.log.info( "Alberta511Parks: Received %d entries." % len( Parks ) )

        # Select park entries of interest and then sort them by Name.

        # Convert park name filter to a list.

        ParkNameFilterList = []

        if ParkNameFilter != "":
            ParkNameFilterList = ParkNameFilter.split( "|" )

            for x in range( len( ParkNameFilterList ) ):
                ParkNameFilterList[ x ] = ParkNameFilterList[ x ].strip()

        SelectedParks = []

        for Park in Parks:
            ParkName = Park.get( "Name" )

            # Filter for parks of interest.

            UseThisEntry = False

            if len( ParkNameFilterList ) > 0:
                # We have a park name filter list so use it.

                for FilterParkName in ParkNameFilterList:
                    if ( ParkNameFilterCondition == EQUAL and       \
                         ParkName == FilterParkName ) or              \
                       ( ParkNameFilterCondition == CONTAINS and    \
                         ParkName.find( FilterParkName ) >= 0 ):
                        UseThisEntry = True
                        break
            else:
                # No park name filter list so use this entry.

                UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this one.

                SelectedParks.append( SortList( ParkName, Parks.index( Park ) ) )

#        self.log.debug( "Parks Length = %d, SelectedParks Length = %d" % ( len( Parks ), len( SelectedParks ) ) )

        SelectedParks.sort( key = SortList.GetKey )

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Parks</b></p>\n"

        self.WriteGeneratedFiles( HTML )

        for SortListItem in SelectedParks:
                # Put each Park into it's own table.

                ParkWebsite = Parks[ SortListItem.Index ].get( "Website" )

                HTML =   "<table class=\"WXFeedsT2\">\n"                                                                                            \
                       +     "<thead>\n"                                                                                                            \
                       +         "<tr>\n"                                                                                                           \
                       +             "<th></th>\n"                                                                                                  \
                       +             "<th></th>\n"                                                                                                  \
                       +         "</tr>\n"                                                                                                          \
                       +     "</thead>\n"                                                                                                           \
                       +     "<tbody>\n"                                                                                                            \
                       +         "<tr>\n"                                                                                                           \
                       +             "<td><b>Name:&nbsp;</b></td>\n"                                                                                \
                       +             "<td>%s</td>\n" % Parks[ SortListItem.Index ].get( "Name" )                                                    \
                       +         "</tr>\n"                                                                                                          \
                       +         "<tr>\n"                                                                                                           \
                       +             "<td><b>Website:&nbsp;</b></td>\n"                                                                             \
                       +             "<td><a href=\"%s\">%s</a></td>\n" % ( ParkWebsite, ParkWebsite )                                              \
                       +         "</tr>\n"                                                                                                          \
                       +         "<tr><td><b>Id:&nbsp;</b></td>             <td>%s</td></tr>\n" % Parks[ SortListItem.Index ].get( "Id" )           \
                       +         "<tr><td><b>Latitude:&nbsp;</b></td>       <td>%f</td></tr>\n" % Parks[ SortListItem.Index ].get( "Latitude" )     \
                       +         "<tr><td><b>Longitude:&nbsp;</b></td>      <td>%f</td></tr>\n" % Parks[ SortListItem.Index ].get( "Longitude" )    \
                       +     "</tbody>\n"                                                                                                           \
                       + "</table>\n"                                                                                                               \
                       + "<hr>\n"

                self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def Alberta511Cameras(
                           self,
                           FeedURL                  : str,
                           DisplayDisabledCameras   : bool,
                           RoadwayNameFilter        : str,
                         ) -> bool:

        ReturnStatus = False

        try:
            Session = requests.session()
            assert Session != None

            Response = Session.get( FeedURL )
            assert Response != None
        except:
            self.log.error("Alberta511Cameras: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511Camaeras: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        Cameras = Response.json()
#        self.log.debug( "Data for %d Cameras received." % len( Cameras ) )

        self.log.info( "Alberta511Cameras: Received %d entries." % len( Cameras ) )

##       'Id': 'ESS_A_011-06 NORDEGG.C1--1', 
##        'Organization': 'Alberta511', 
##        'RoadwayName': 'Hwy 11', 
##        'DirectionOfTravel': 'Westbound', 
##        'Latitude': 52.49038611, 
##        'Longitude': -116.0399333, 
##        'Name': 'Hwy 11 & Nordegg Ranger Stn.', 
##        'Url': 'https://511.alberta.ca/map/Cctv/ESS_A_011-06 NORDEGG.C1--1', 
##        'Status': 'Enabled', 
##        'Description': 'WB, North side of HWY', 
#        'HasRWISData': True, 
##        'AirTemperature': '16.6 °C', 
##        'PavementTemperature': '23.5 °C', 
##        'WindSpeed': '3.6 km/h', 
##        'WindDirection': 'NW', 
##        'RelativeHumidity': '52%', 
##        'WeatherUpdated': 'Aug 18 2020, 9:02 PM'

        # Select camera entries of interest and then sort them by RoadwayName.

        # Convert roadway name filter to a list.

        RoadwayNameFilterList = []

        if RoadwayNameFilter != "":
            RoadwayNameFilterList = RoadwayNameFilter.split( "|" )

            for x in range( len( RoadwayNameFilterList ) ):
                RoadwayNameFilterList[ x ] = RoadwayNameFilterList[ x ].strip()

        SelectedCameras = []

        for Camera in Cameras:
            RoadwayName = Camera.get( "RoadwayName" )

            if DisplayDisabledCameras == False and         \
               Camera.get( "Status" ) == "Disabled":
                continue

            # Filter for roadways of interest.

            UseThisEntry = False

            if len( RoadwayNameFilterList ) > 0:
                # We have a roadway name filter list so use it.

                for FilterRoadwayName in RoadwayNameFilterList:
                    if FilterRoadwayName == RoadwayName:
                        UseThisEntry = True
                        break
            else:
                # No area filter list so use this entry.

                UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this one.

                SelectedCameras.append( SortList( RoadwayName, Cameras.index( Camera ) ) )

#        self.log.debug( "Cameras Length = %d, SelectedCameras Length = %d" % ( len( Cameras ), len( SelectedCameras ) ) )

        SelectedCameras.sort( key = SortList.GetKey )

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Cameras</b></p>\n"

        self.WriteGeneratedFiles( HTML )

        for SortListItem in SelectedCameras:
            # Put each Camera into it's own table.

            CameraUrl = Cameras[ SortListItem.Index ].get( "Url" )
            CameraUrlhref = CameraUrl.replace( " ", "%20" )

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                                                \
                   +     "<thead>\n"                                                                                                                           \
                   +         "<tr>\n"                                                                                                                          \
                   +             "<th></th>\n"                                                                                                                 \
                   +             "<th></th>\n"                                                                                                                 \
                   +         "</tr>\n"                                                                                                                         \
                   +     "</thead>\n"                                                                                                                          \
                   +     "<tbody>\n"                                                                                                                           \
                   +         "<tr><td><b>Roadway Name:&nbsp;</b></td>        <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "RoadwayName" )          \
                   +         "<tr><td><b>Direction Of Travel:&nbsp;</b></td> <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "DirectionOfTravel" )    \
                   +         "<tr>\n"                                                                                                                          \
                   +             "<td><b>Name:&nbsp;</b></td>\n"                                                                                               \
                   +             "<td><a href=\"%s\">%s</a></td>\n" % ( CameraUrlhref, Cameras[ SortListItem.Index ].get( "Name" ) )                           \
                   +         "</tr>\n"                                                                                                                         \
                   +         "<tr><td><b>Description:&nbsp;</b></td>         <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Description" )          \
                   +         "<tr>\n"                                                                                                                          \
                   +             "<td><b>Url:&nbsp;</b></td>\n"                                                                                                \
                   +             "<td><a href=\"%s\">%s</a></td>\n" % ( CameraUrlhref, CameraUrl )                                                             \
                   +         "</tr>\n"                                                                                                                         \
                   +         "<tr><td><b>Id:&nbsp;</b></td>                  <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Id" )                   \
                   +         "<tr><td><b>Organization:&nbsp;</b></td>        <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Organization" )         \
                   +         "<tr><td><b>Status:&nbsp;</b></td>              <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Status" )               \
                   +         "<tr><td><b>Latitude:&nbsp;</b></td>            <td>%f</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Latitude" )             \
                   +         "<tr><td><b>Longitude:&nbsp;</b></td>           <td>%f</td></tr>\n" % Cameras[ SortListItem.Index ].get( "Longitude" )

            if Cameras[ SortListItem.Index ].get( "HasRWISData" ) == True:
                HTML +=   "<tr><td><b>Air Temperature:&nbsp;</b></td>      <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "AirTemperature" )         \
                        + "<tr><td><b>Pavement Temperature:&nbsp;</b></td> <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "PavementTemperature" )    \
                        + "<tr><td><b>Wind Speed:&nbsp;</b></td>           <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "WindSpeed" )              \
                        + "<tr><td><b>Wind Direction:&nbsp;</b></td>       <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "WindDirection" )          \
                        + "<tr><td><b>Relative Humidity:&nbsp;</b></td>    <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "RelativeHumidity" )       \
                        + "<tr><td><b>Weather Updated:&nbsp;</b></td>      <td>%s</td></tr>\n" % Cameras[ SortListItem.Index ].get( "WeatherUpdated" )

            HTML +=   "</tbody>\n"    \
                    + "</table>\n"    \
                    + "<hr>\n"

            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        Session.close()
        Response.close()

        return ReturnStatus

    def Alberta511GroupedCameras(
                                  self,
                                  FeedURL                   : str,
                                  DisplayDisabledCameras    : bool,
                                  RoadwayNameFilter         : str,
                                ) -> bool:

        ReturnStatus = False

        try:
            Session = requests.session()
            assert Session != None

            Response = Session.get( FeedURL )
            assert Response != None
        except:
            self.log.error("Alberta511GroupedCameras: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511GroupedCamaeras: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        GroupedCameras = Response.json()
#        self.log.debug( "Data for %d Grouped Cameras received." % len( GroupedCameras ) )

        self.log.info( "Alberta511GroupedCameras: Received %d entries." % len( GroupedCameras ) )

#	{
#		'Id': 'VTMS_AB_216-06 AHDNW HWY16A|1', 
#		'Organization': 'Alberta511', 
#		'RoadwayName': 'Hwy 216', 
#		'DirectionOfTravel': 'Northbound', 
#		'Latitude': 53.540778, 
#		'Longitude': -113.659822, 
#		'Views': 
#		[
#			{
#				'Id': 'VTMS_AB_216-06 AHDNW HWY16A.C1--1', 
#				'Name': 'Hwy 216: Anthony Henday Drive and Stony Plain Road West Interchange', 
#				'Url': 'https://511.alberta.ca/map/Cctv/VTMS_AB_216-06 AHDNW HWY16A.C1--1', 
#				'Status': 'Enabled', 
#				'Description': 'SB,  West side of HWY'
#			}, 
#			{
#				'Id': 'VTMS_AB_216-06 AHDNW HWY16A.C2--1', 
#				'Name': 'Hwy 216: Anthony Henday Drive and Stony Plain Road West Interchange', 
#				'Url': 'https://511.alberta.ca/map/Cctv/VTMS_AB_216-06 AHDNW HWY16A.C2--1', 
#				'Status': 'Enabled', 
#				'Description': 'SB,  West side of HWY'
#			}, 
#			{
#				'Id': 'VTMS_AB_216-06 AHDNW HWY16A.C3--1', 
#				'Name': 'Hwy 216: Anthony Henday Drive and Stony Plain Road West Interchange', 
#				'Url': 'https://511.alberta.ca/map/Cctv/VTMS_AB_216-06 AHDNW HWY16A.C3--1', 
#				'Status': 'Enabled', 
#				'Description': 'SB,  West side of HWY'
#			}
#		], 
#		'HasRWISData': False, 
#		'AirTemperature': None, 
#		'PavementTemperature': None, 
#		'WindSpeed': None, 
#		'WindDirection': None, 
#		'RelativeHumidity': None, 
#		'WeatherUpdated': None
#	}

        # Select grouped camera entries of interest and then sort them by RoadwayName.

        # Convert roadway name filter to a list.

        RoadwayNameFilterList = []

        if RoadwayNameFilter != "":
            RoadwayNameFilterList = RoadwayNameFilter.split( "|" )

            for x in range( len( RoadwayNameFilterList ) ):
                RoadwayNameFilterList[ x ] = RoadwayNameFilterList[ x ].strip()

        SelectedGroupedCameras = []

        for CameraGroup in GroupedCameras:
            RoadwayName = CameraGroup.get( "RoadwayName" )

            # Filter for roadways of interest.

            UseThisEntry = False

            if len( RoadwayNameFilterList ) > 0:
                # We have a roadway name filter list so use it.

                for FilterRoadwayName in RoadwayNameFilterList:
                    if FilterRoadwayName == RoadwayName:
                        UseThisEntry = True
                        break
            else:
                # No area filter list so use this entry.

                UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this one.

                SelectedGroupedCameras.append( SortList( RoadwayName, GroupedCameras.index( CameraGroup ) ) )

#        self.log.debug( "GroupedCameras Length = %d, SelectedGroupedCameras Length = %d" % ( len( GroupedCameras ), len( SelectedGroupedCameras ) ) )

        SelectedGroupedCameras.sort( key = SortList.GetKey )

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Grouped Cameras</b></p>\n"

        self.WriteGeneratedFiles( HTML )
        
        for SortListItem in SelectedGroupedCameras:
            # Put each Camera Group into it's own table.

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                                                                            \
                   +     "<thead>\n"                                                                                                                                            \
                   +         "<tr>\n"                                                                                                                                           \
                   +             "<th></th>\n"                                                                                                                                  \
                   +             "<th></th>\n"                                                                                                                                  \
                   +             "<th></th>\n"                                                                                                                                  \
                   +         "</tr>\n"                                                                                                                                          \
                   +     "</thead>\n"                                                                                                                                           \
                   +     "<tbody>\n"                                                                                                                                            \
                   +         "<tr><td><b>Roadway Name:&nbsp;</b></td>        <td>%s</td> <td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "RoadwayName" )           \
                   +         "<tr><td><b>Direction Of Travel:&nbsp;</b></td> <td>%s</td> <td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "DirectionOfTravel" )

            Views = GroupedCameras[ SortListItem.Index].get( "Views" )
            assert Views != None

            c = 0

            for View in Views:
                if DisplayDisabledCameras == False and         \
                   View.get( "Status" ) == "Disabled":
                    continue

                if c > 0:
                   # Insert an empty row in the table between views.

                   HTML += "<tr><td>&nbsp;</td><td>&nbsp;</td><td>&nbsp;</td></tr>\n"

                CameraUrl = View.get( "Url" )
                CameraUrlhref = CameraUrl.replace( " ", "%20" )

                HTML +=   "<tr>\n"                                                                          \
                        +     "<td></td>\n"                                                                 \
                        +     "<td><b>Name:&nbsp;</b></td>\n"                                               \
                        +     "<td><a href=\"%s\">%s</a></td>\n" % ( CameraUrlhref, View.get( "Name" ) )    \
                        + "</tr>\n"                                                                         \
                        + "<tr>\n"                                                                          \
                        +     "<td></td>\n"                                                                 \
                        +     "<td><b>Description:&nbsp;</b></td>\n"                                        \
                        +     "<td>%s</td>\n" % View.get( "Description" )                                   \
                        + "</tr>\n"                                                                         \
                        + "<tr>\n"                                                                          \
                        +     "<td></td>\n"                                                                 \
                        +     "<td><b>Url:&nbsp;</b></td>\n"                                                \
                        +     "<td><a href=\"%s\">%s</a></td>\n" % ( CameraUrlhref, CameraUrl )             \
                        + "</tr>\n"                                                                         \
                        + "<tr>\n"                                                                          \
                        +     "<td></td>\n"                                                                 \
                        +     "<td><b>Id:&nbsp;</b></td>\n"                                                 \
                        +     "<td>%s</td>\n"  % View.get( "Id" )                                           \
                        + "</tr>\n"                                                                         \
                        + "<tr>\n"                                                                          \
                        +     "<td></td>\n"                                                                 \
                        +     "<td><b>Status:&nbsp;</b></td>\n"                                             \
                        +     "<td>%s</td>\n"  % View.get( "Status" )                                       \
                        + "</tr>\n"

                c += 1

            HTML +=   "<tr><td><b>Id:&nbsp;</b></td>                  <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "Id" )              \
                    + "<tr><td><b>Organization:&nbsp;</b></td>        <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "Organization" )    \
                    + "<tr><td><b>Latitude:&nbsp;</b></td>            <td>%f</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "Latitude" )        \
                    + "<tr><td><b>Longitude:&nbsp;</b></td>           <td>%f</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "Longitude" )

            if GroupedCameras[ SortListItem.Index].get( "HasRWISData" ) == True:
                HTML +=   "<tr><td><b>Air Temperature:&nbsp;</b></td>      <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "AirTemperature" )         \
                        + "<tr><td><b>Pavement Temperature:&nbsp;</b></td> <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "PavementTemperature" )    \
                        + "<tr><td><b>Wind Speed:&nbsp;</b></td>           <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "WindSpeed" )              \
                        + "<tr><td><b>Wind Direction:&nbsp;</b></td>       <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "WindDirection" )          \
                        + "<tr><td><b>Relative Humidity:&nbsp;</b></td>    <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "RelativeHumidity" )       \
                        + "<tr><td><b>Weather Updated:&nbsp;</b></td>      <td>%s</td><td></td></tr>\n" % GroupedCameras[ SortListItem.Index].get( "WeatherUpdated" )

            HTML +=   "</tbody>\n"    \
                    + "</table>\n"    \
                    + "<hr>\n"

            self.WriteGeneratedFiles( HTML )
 
        ReturnStatus = True

        Session.close()
        Response.close()

        return ReturnStatus

    def Alberta511Events(
                            self
                          , FeedURL : str
                          , EventTypeFilter : str
                          , RoadwayNameFilter : str
             ):

        ReturnStatus = False

        try:
            Response = requests.get( FeedURL )
        except:
            self.log.error("Alberta511Events: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511Events: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        Events = Response.json()
#        self.log.debug( "Data for %d Events received." % len( Events ) )

        self.log.info( "Alberta511Events: Received %d entries." % len( Events ) )

##        'ID': 'ERS--11530', 
##        'Organization': 'ERS', 
##        'RoadwayName': 'HWY-3', 
##        'DirectionOfTravel': 'Both Directions', 
##        'Description': 'Overhead Power Line Installation on HWY-3 Both Directions from Range Road 303 to Range Road 302 .  Activities: Traffic congestion   ', 
##        'Reported': 1596290400, 
##        'LastUpdated': 1597915481, 
##        'StartDate': 1596290400, 
##        'PlannedEndDate': 1601506800, 
##        'LanesAffected': 'No Data', 
##        'Latitude': 49.558101, 
##        'Longitude': -114.059414, 
##        'LatitudeSecondary': 49.531252, 
##        'LongitudeSecondary': -113.980072, 
##        'EventType': 'roadwork'

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Events</b></p>\n"

        self.WriteGeneratedFiles( HTML )

        # Select event entries of interest and then sort them by RoadwayName.

        # Convert event type filter to a list.

        EventTypeFilterList = []

        if EventTypeFilter != "":
            EventTypeFilterList = EventTypeFilter.split( "|" )

            for x in range( len( EventTypeFilterList ) ):
                EventTypeFilterList[ x ] = EventTypeFilterList[ x ].strip()

        # Convert roadway name filter to a list.

        RoadwayNameFilterList = []

        if RoadwayNameFilter != "":
            RoadwayNameFilterList = RoadwayNameFilter.split( "|" )

            for x in range( len( RoadwayNameFilterList ) ):
                RoadwayNameFilterList[ x ] = RoadwayNameFilterList[ x ].strip()

        SelectedEvents = []

        for Event in Events:
            # Filter for event type of interest.

            EventType = Event.get( "EventType" )
            RoadwayName = Event.get( "RoadwayName" )

            UseThisEntry = False

            if len( EventTypeFilterList ) > 0:
                # We have an event type filter list so use it.

                for FilterEventType in EventTypeFilterList:
                    if FilterEventType == EventType:
                        UseThisEntry = True
                        break
            else:
                # No event type filter list so use this entry.

                UseThisEntry = True

            if UseThisEntry == True:
                # Event type is of interest or no event type filter.
                #
                # Is this roadway of interest?

                UseThisEntry = False
            
                if len( RoadwayNameFilterList ) > 0:
                    # We have a roadway name filter list so use it.#

                    for FilterRoadwayName in RoadwayNameFilterList:
                        if FilterRoadwayName == RoadwayName:
                            UseThisEntry = True
                            break
                else:
                    # No roadway name filter list so use this entry.

                    UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this one.

                SelectedEvents.append( SortList( RoadwayName, Events.index( Event ) ) )

#        self.log.debug( "Events Length = %d, SelectedEvents Length = %d" % ( len( Events ), len( SelectedEvents ) ) )

        SelectedEvents.sort( key = SortList.GetKey )

        d = self.GetCurrentUTCAware()
#        self.log.debug( "Now (UTC): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z"  ) )
        d = self.GetTimeLocalAware( d )
#        self.log.debug( "Now (LT): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z" ) )
                                
        LocalTimeZone = d.strftime( "%Z" )
#        self.log.debug( "Local Time Zone = %s", LocalTimeZone )
        
        DateTimeFormat = "%b %-d %Y, %H:%M " + LocalTimeZone

        for SortListItem in SelectedEvents:

            ReportedDateTime = ""
            Reported = Events[ SortListItem.Index ].get( "Reported" )
            if Reported != None: 
                ReportedDateTime = datetime.datetime.fromtimestamp( Reported ).strftime( DateTimeFormat )

            LastUpdatedDateTime = ""
            LastUpdated = Events[ SortListItem.Index ].get( "LastUpdated" )
            if LastUpdated != None: 
                LastUpdatedDateTime = datetime.datetime.fromtimestamp( LastUpdated ).strftime( DateTimeFormat )

            StartDateDateTime = ""
            StartDate = Events[ SortListItem.Index ].get( "StartDate" )
            if StartDate != None: 
                StartDateDateTime = datetime.datetime.fromtimestamp( StartDate ).strftime( DateTimeFormat )

            PlannedEndDateDateTime = ""
            PlannedEndDate = Events[ SortListItem.Index ].get( "PlannedEndDate" )
            if PlannedEndDate != None:
                PlannedEndDateDateTime = datetime.datetime.fromtimestamp( PlannedEndDate ).strftime( DateTimeFormat )

            # Put each Event into it's own table.

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                                                \
                   +     "<thead>\n"                                                                                                                           \
                   +         "<tr>\n"                                                                                                                          \
                   +             "<th></th>\n"                                                                                                                 \
                   +             "<th></th>\n"                                                                                                                 \
                   +         "</tr>\n"                                                                                                                         \
                   +     "</thead>\n"                                                                                                                          \
                   +     "<tbody>\n"                                                                                                                           \
                   +         "<tr><td><b>Event Type:&nbsp;</b></td>          <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "EventType" )             \
                   +         "<tr><td><b>Roadway Name:&nbsp;</b></td>        <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "RoadwayName" )           \
                   +         "<tr><td><b>Direction Of Travel:&nbsp;</b></td> <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "DirectionOfTravel" )     \
                   +         "<tr><td><b>Description:&nbsp;</b></td>         <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "Description" )           \
                   +         "<tr><td><b>Lanes Affected:&nbsp;</b></td>      <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "LanesAffected" )         \
                   +         "<tr><td><b>Reported:&nbsp;</b></td>            <td>%s</td></tr>\n" % ReportedDateTime                                            \
                   +         "<tr><td><b>Last Updated:&nbsp;</b></td>        <td>%s</td></tr>\n" % LastUpdatedDateTime                                         \
                   +         "<tr><td><b>Start Date:&nbsp;</b></td>          <td>%s</td></tr>\n" % StartDateDateTime                                           \
                   +         "<tr><td><b>Planned End Date:&nbsp;</b></td>    <td>%s</td></tr>\n" % PlannedEndDateDateTime                                      \
                   +         "<tr><td><b>ID:&nbsp;</b></td>                  <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "ID" )                    \
                   +         "<tr><td><b>Organization:&nbsp;</b></td>        <td>%s</td></tr>\n" % Events[ SortListItem.Index ].get( "Organization" )          \
                   +         "<tr><td><b>Latitude:&nbsp;</b></td>            <td>%f</td></tr>\n" % Events[ SortListItem.Index ].get( "Latitude" )              \
                   +         "<tr><td><b>Longitude:&nbsp;</b></td>           <td>%f</td></tr>\n" % Events[ SortListItem.Index ].get( "Longitude" )             \
                   +         "<tr><td><b>Latitude Secondary:&nbsp;</b></td>  <td>%f</td></tr>\n" % Events[ SortListItem.Index ].get( "LatitudeSecondary" )     \
                   +         "<tr><td><b>Longitude Secondary:&nbsp;</b></td> <td>%f</td></tr>\n" % Events[ SortListItem.Index ].get( "LongitudeSecondary" )

            HTML +=   "</tbody>\n"    \
                    + "</table>\n"    \
                    + "<hr>\n"

            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        Response        = None
        Events          = None
        SelectedEvents  = None

        return ReturnStatus

    def Alberta511Alerts(
                          self,
                          FeedURL : str,
                        ):

        ReturnStatus = False

        try:
            Response = requests.get( FeedURL )
        except:
            self.log.error("Alberta511Alerts: requests.get() returned error: %d!" % Response.status_code)
            ReturnStatus = False
            return ReturnStatus

#        self.log.info( "Alberta511Alerts: requests.get() successully returned for 511 Alberta: URL: %s" % FeedURL )
#        self.log.debug( "Response: %s" % Response )

        Alerts = Response.json()
#        self.log.debug( "Data for %d Alerts received." % len( Alerts ) )

        self.log.info( "Alberta511Alerts: Received %d entries." % len( Alerts ) )

        HTML = "<p class=\"WXFeedsP1\"><b>511 Alberta Alerts</b></p>\n"

        self.WriteGeneratedFiles( HTML )

        d = self.GetCurrentUTCAware()
#        self.log.debug( "Now (UTC): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z"  ) )
        d = self.GetTimeLocalAware( d )
#        self.log.debug( "Now (LT): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z" ) )
                                
        LocalTimeZone = d.strftime( "%Z" )
#        self.log.debug( "Local Time Zone = %s", LocalTimeZone )
        
        DateTimeFormat = "%b %-d %Y, %H:%M " + LocalTimeZone

        for Alert in Alerts:

            StartTimeDateTime = ""
            StartTime = Alert.get( "StartTime" )
            if StartTime != None: 
                StartTimeDateTime = datetime.datetime.fromtimestamp( StartTime ).strftime( DateTimeFormat )

            EndTimeDateTime = ""
            EndTime = Alert.get( "EndTime" )
            if EndTime != None: 
                EndTimeDateTime = datetime.datetime.fromtimestamp( EndTime ).strftime( DateTimeFormat )

            Regions = Alert.get( "Regions" )

            RegionsCSV = ""

            r = 0

            for Region in Regions:
#                self.log.debug( "Region: %s" % Region )

                if r > 0:
                    RegionsCSV += ", "

                RegionsCSV += Region

                r += 1

            # Put each Alert into it's own table.

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                     \
                   +     "<thead>\n"                                                                                                \
                   +         "<tr>\n"                                                                                               \
                   +             "<th></th>\n"                                                                                      \
                   +             "<th></th>\n"                                                                                      \
                   +         "</tr>\n"                                                                                              \
                   +     "</thead>\n"                                                                                               \
                   +     "<tbody>\n"                                                                                                \
                   +         "<tr><td><b>Message:&nbsp;</b></td>           <td>%s</td></tr>\n" % Alert.get( "Message" )             \
                   +         "<tr><td><b>Notes:&nbsp;</b></td>             <td>%s</td></tr>\n" % Alert.get( "Notes" )               \
                   +         "<tr><td><b>Start Time:&nbsp;</b></td>        <td>%s</td></tr>\n" % StartTimeDateTime                  \
                   +         "<tr><td><b>End Time:&nbsp;</b></td>          <td>%s</td></tr>\n" % EndTimeDateTime                    \
                   +         "<tr><td><b>Region(s):&nbsp;</b></td>         <td>%s</td></tr>\n" % RegionsCSV                         \
                   +         "<tr><td><b>High Importance:&nbsp;</b></td>   <td>%s</td></tr>\n" % Alert.get( "HighImportance" )      \
                   +         "<tr><td><b>Send Notification:&nbsp;</b></td> <td>%s</td></tr>\n" % Alert.get( "SendNotification" )

            HTML +=   "</tbody>\n"    \
                    + "</table>\n"    \
                    + "<hr>\n"

            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

        return ReturnStatus

    def Alberta511RoadConditions(
                                  self,
                                  FeedURL               : str,
                                  AreaNameFilter        : str,
                                  RoadwayNameFilter     : str,
                                  Zoom                  : int,
                                  DisplayAttributeLists : bool
                                ) -> bool:
                                #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        HTMLFilePathStemPrefix          = "511AlbertaRoadCondition-"
        JavaScriptFilePathStemPrefix    = "511AlbertaRoadConditions"

        # Convert filter strings to  lists.

        AreaNameFilterList      = []
        RoadwayNameFilterList   = []
    
        if AreaNameFilter != "":
            AreaNameFilterList = AreaNameFilter.split( "|" )

            for x in range( len( AreaNameFilterList ) ):
                AreaNameFilterList[ x ] = AreaNameFilterList[ x ].strip()

        if RoadwayNameFilter != "":
            RoadwayNameFilterList = RoadwayNameFilter.split( "|" )

            for x in range( len( RoadwayNameFilterList ) ):
                RoadwayNameFilterList[ x ] = RoadwayNameFilterList[ x ].strip()

        # This list is intended to make it easier to see what values 
        # exist for 'AreaName' and 'RoadwayName'.  The idea is to occasionally 
        # display the list so that values can be used to filter entries.

        AreaNames       = []
        RoadwayNames    = []

        # We also build a list of markers for Google Maps.

        Markers = []

        # List of HTML map files that currently exist, as in HTML map 
        # files that exist before we process this feed.

        Current511AlbertaHTMLMapFiles = []

        SelectedRoadConditions = []        # Store key data about a road condition in a list.

        EndpointURL = FeedURL

        if self.WXFeedsServerUse == True:
            assert self.WXFeedsServerAccessToken != ""

            EndpointURL = "/511Alberta?resource=RoadConditions&url=%s" % EndpointURL

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: EndpointURL: %s" % EndpointURL )

        ReturnedStatus  = False
        CacheHit        = False
        StatusCode      = -1
        RoadConditions  = {}
        
        try:
            ReturnedStatus,         \
            CacheHit,               \
            StatusCode,             \
            RoadConditions = self.FOOGetEndpointData(
                                                   EndpointURL                      = EndpointURL,
                                                   ContentType                      = CONTENT_JSON,
                                                   ValidateDataFunction             = self.Alberta511RoadConditionsValidateData,
                                                   LocalCacheUse                    = self.Alberta511RoadConditionsUseLocalCache,
                                                   LocalCacheStem                   = "511AlbertaRoadConditions",
                                                   LocalCacheExpiry_minutes         = self.Alberta511RoadConditionsCacheExpiry_minutes,
                                                   WXFeedsServerUse            = self.WXFeedsServerUse,
                                                   WXFeedsServerForceCacheMiss = self.Alberta511RoadConditionsForceCacheMiss
                                                 )
        except:
            self.log.error( "Alberta511RoadConditions: GetEndpointData() threw an exception: sys.exc_type = %s" % sys.exc_info()[0] )

            ReturnedStatus = False 
 
        #if DEBUG == True:
        self.log.debug( "Alberta511RoadConditions: GetEndpointData() returned: ReturnedStatus: %s, CacheHit: %s, StatusCode: %d" % ( ReturnedStatus, CacheHit, StatusCode ) ) 

        if ReturnedStatus == False:
            # GetEndpointData() returned False so we're done!

            self.log.error( "Alberta511RoadConditions: GetEndpointData() returned False!" )

            ReturnStatus = False
            return ReturnStatus
 
        if StatusCode != 200:
            # JSONDataCache returned bad StatusCode so we're done!

            self.log.error( "Alberta511RoadConditions: GetEndpointData() returned bad StatusCode: %d!" % StatusCode )

            ReturnStatus = False
            return ReturnStatus
 
        self.log.info( "Alberta511RoadConditions: Received %d entries." % len( RoadConditions ) )

        ID_KEY                  = "ID"
        VARIABLES_KEY           = "Variables"
        CACHE_HIT_KEY           = "CacheHit"
        LAST_UPDATED_KEY        = "LastUpdated"
        LATITUDE_KEY            = "Latitude"
        LONGITUDE_KEY           = "Longitude"
        POSITION_SPECIFIED_KEY  = "PositionSpecified"

        # If the cache is empty or we received new data (CacheHit == False), initialize the cache.
        # If the cache is empty and we received cached data (CacheHit == True) it means the 
        # data was received and cached before WeeWX was (re-)started - we treat it as if it's new data.

        if len( self.Alberta511RoadConditionCache ) == 0 or     \
           CacheHit == False:

            if DEBUG == True:
                self.log.debug( "len( self.Alberta511RoadConditionCache ) = %d, CacheHit = %s" % ( len( self.Alberta511RoadConditionCache ), CacheHit ) )
        
            for RoadCondition in RoadConditions:
#                if DEBUG == True:
#                    self.log.debug( "Alberta511RoadConditions: RoadCondition: %s" % RoadCondition )

                ID = str( RoadCondition[ 'Id'] )
                assert ID != ""

                if ID not in self.Alberta511RoadConditionCache:
                    # Road Condition Cache Miss - put this Road Condition into the cache.

#                    if DEBUG == True:
#                        self.log.debug( "Alberta511RoadConditions: NEW Road Condition: %s" % ID )

                    CacheHit            = False     # New data.

                    LastUpdated         = RoadCondition[ 'LastUpdated' ]
                    PositionSpecified   = False
                    Latitude            = BAD_LATITUDE
                    Longitude           = BAD_LONGITUDE

                    self.Alberta511RoadConditionCache[ ID ] = {
                                                                VARIABLES_KEY : 
                                                                                {
                                                                                  CACHE_HIT_KEY             : CacheHit,
                                                                                  LAST_UPDATED_KEY          : LastUpdated,
                                                                                  POSITION_SPECIFIED_KEY    : PositionSpecified,
                                                                                  LATITUDE_KEY              : Latitude,
                                                                                  LONGITUDE_KEY             : Longitude,
                                                                                }
                                                              }
                else:
                    # Road Condition Cache Hit - update existing cache entry.

#                    if DEBUG == True:
#                        self.log.debug( "Alberta511RoadConditions: POSSIBLE UPDATED Road Condition: %s" % ID )

                    # If this is an updated Road Condition reset variables so that maps are updated.

                    LastUpdated = self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ LAST_UPDATED_KEY ]

                    if RoadCondition[ 'LastUpdated' ] > LastUpdated:
                        if DEBUG == True:
                            self.log.debug( "Alberta511RoadConditions: Going to UPDATE Road Condition: %s" % self.Alberta511RoadConditionCache[ ID ] )

                        self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ CACHE_HIT_KEY ]       = False
                        self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ LAST_UPDATED_KEY ]    = LastUpdated

                        if DEBUG == True:
                            self.log.debug( "Alberta511RoadConditions: UPDATED Road Condition: %s" % self.Alberta511RoadConditionCache[ ID ] )

        # Select road condition entries of interest and then sort them by AreaName and RoadwayName.

        # Calculating the position of a Road Condition can take a long time 
        # so only process a fraction of the returned entries per cycle.

        c_max   = int( 0.33333 * len( RoadConditions) ) + 1
#        c_max   = 1000
        c       = 0
#        self.log.debug( "c: %d, c_max: %d" % ( c, c_max ) )

        if DEBUG == True:
            n = 0

        for RoadCondition in RoadConditions:

            ID = str( RoadCondition[ 'Id'] )
            assert ID != ""

            AreaName = RoadCondition[ 'AreaName' ]
            assert AreaName != ""
            RoadwayName = RoadCondition[ 'RoadwayName' ]
            assert RoadwayName != ""

            # Maintain lists of area and roadway names.

            if AreaName not in AreaNames:
                AreaNames.append( AreaName )

            if RoadwayName not in RoadwayNames:
                RoadwayNames.append( RoadwayName )

            PositionSpecified   = False
            Latitude            = BAD_LATITUDE
            Longitude           = BAD_LONGITUDE

            # If we are going to create maps, we create a marker for every road condition.
            # We do this so that the user sees all road conditions on a map even if they've 
            # filtered them out.

            if self.GoogleMapsMapsUse == True:
                # Find the midpoint of the polyline for the road condition.

                if ID in self.Alberta511RoadConditionCache:
                    # Cache Hit!
                    #
                    # We loaded the Road Condition cache above so we should never get a cache miss.

#                    if DEBUG == True:
#                        self.log.debug( "Alberta511RoadConditions: Cache Hit . . ." )

                    PositionSpecified   = self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY][ POSITION_SPECIFIED_KEY ]

                    if PositionSpecified == True:
                        Latitude            = self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY][ LATITUDE_KEY ]
                        Longitude           = self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY][ LONGITUDE_KEY ]
                    else:
                        if c < c_max:
                            DeclodedPolylineOK  = False

#                           if DEBUG == True:
#                                self.log.debug( "Alberta511RoadConditions: EncodedPolyline: '%s'" % RoadCondition[ 'EncodedPolyline' ] )

                            try:
                                DecodedPolyline = polyline.decode( RoadCondition[ 'EncodedPolyline' ] )
#                                if DEBUG == True:
#                                    self.log.debug( "Alberta511RoadConditions: DecodedPolyline: '%s'" % DecodedPolyline )

                                DecodedPolylineOK = True
                            except:
                                self.log.error( "Alberta511RoadConditions: polyline.decode() threw an Exception!" )
                                assert 1 != 1

                            if DecodedPolylineOK == True:
                                MidpointIndex = self.PolylineMidpoint( DecodedPolyline )

                                if MidpointIndex >= 0:
                                    PositionSpecified   = True
                                    Latitude            = DecodedPolyline[ MidpointIndex ][ 0 ]
                                    Longitude           = DecodedPolyline[ MidpointIndex ][ 1 ]

#                                    if DEBUG == True:
#                                        self.log.debug( "Alberta511RoadConditions: Updating Road Condition Cache for '%s' with: Latitude: %f, Longitude: %f" % ( ID, Latitude, Longitude ) )
#                                        self.log.debug( "Alberta511RoadConditions: Updating Road Condition Cache: %s" % self.Alberta511RoadConditionCache[ ID ] )

                                    self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ POSITION_SPECIFIED_KEY ]  = PositionSpecified
                                    self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ LATITUDE_KEY ]            = Latitude
                                    self.Alberta511RoadConditionCache[ ID ][ VARIABLES_KEY ][ LONGITUDE_KEY ]           = Longitude

#                                    if DEBUG == True:
#                                        self.log.debug( "Alberta511RoadConditions: Updated Road Condition Cache for '%s': %s" % ( ID, self.Alberta511RoadConditionCache[ ID ] ) )

                                c += 1
                else:
                    # Cache Miss!

                    self.log.error( "Alberta511RoadConditions: Road Condition Cach for ID: '%s'!" % ID )
                    assert 1 != 1

#                if DEBUG == True:
#                    self.log.debug( "Alberta511RoadConditions: Final Position: %s, %f, %f" % ( PositionSpecified, Latitude, Longitude ) )

                if PositionSpecified == True:
                    # We have the position for this road condition so create a marker.

                    Type = "Road Condition"
                    MarkerType = "road_condition"

                    LocationDescription = RoadCondition[ 'LocationDescription' ]

                    Conditions = RoadCondition[ 'Primary Condition' ]

                    SecondaryConditions = RoadCondition[ 'Secondary Conditions' ]

                    sc = 0

                    for SecondaryCondition in SecondaryConditions:
                        if sc == 0:
                            Conditions += ": "
                    
                        if sc > 0:
                            Conditions += ", "

                        Conditions += SecondaryCondition

                        sc += 1

                    Visibility = RoadCondition[ 'Visibility' ]

                    MarkerText = RoadwayName + "\\n" + LocationDescription + "\\n" + Conditions

                    if Visibility != None and   \
                       Visibility != "":
                        MarkerText += "\\n" + "Visibility: %s" % Visibility

                    MarkerInfoWindowContent = "<a href=\"%s\">511 Alberta</a>" % ALBERTA_511_URL

                    Markers.append( 
                                    GoogleMapsMarker(
                                                      MarkerType,
                                                      Latitude,
                                                      Longitude,
                                                      MarkerText,
                                                      MarkerInfoWindowContent
                                                     )
                                  )

#            if DEBUG == True:
#                self.log.debug( "Alberta511RoadConditions: '%s'" % RoadCondition )

            # Filter for entries of interest.

            UseThisEntry = False

            if len( AreaNameFilterList ) > 0:
                # We have an area name filter list so use it.

                for FilterAreaName in AreaNameFilterList:
                    if FilterAreaName == AreaName:
                        UseThisEntry = True
                        break
            else:
                # No area name filter list so use this entry.

                UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this entry so far.
                #
                # Next filter. . .

                UseThisEntry = False

                if len( RoadwayNameFilterList ) > 0:
                    # We have a roadway name filter list so use it.

                    for FilterRoadwayName in RoadwayNameFilterList:
                        if FilterRoadwayName == RoadwayName:
                            UseThisEntry = True
                            break
                else:
                    # No roadway filter list so use this entry.

                    UseThisEntry = True

            if UseThisEntry == True:
                # We're interested in this one.

                SelectedRoadConditions.append( Selected511AlbertaRoadConditionItem(
                                                                                    ID,
                                                                                    AreaName,
                                                                                    RoadwayName,
                                                                                    RoadConditions.index( RoadCondition )
                                                                                  )
                                             )

#            if DEBUG == True:
#                n += 1
#                if n == 2:
#                    break

        # We now have a list if road conditions we are interested in.
        # 
        # Get a list of current 511 Alberta Road Condition maps.
        # 
        # Create a pattern for 511 Alberta Road Conditiona HTML files.
        # We use GetMapHTMLPaths() for consistency.  We won't be 
        # using HTMLFileName, HTMLFilePathRelative or HTMLCacheFilePathAbsolute
        #
        # We use FAKE_ASTERISK and not "*" 'cause GetMapHTMLPaths() parses out 
        # '*' characters 'cause it's not valid in filenames.  We'll replace 
        # FAKE_ASTERISK with '*' later.

        CurrentAlberta511HTMLMapFiles = []

        HTMLFileName, HTMLFilePathRelative, Alberta511HTMLFilesPattern, HTMLCacheFilePathAbsolute =       \
                self.GetMapHTMLPaths( HTMLFilePathStemPrefix + FAKE_ASTERISK )
        assert Alberta511HTMLFilesPattern != ""
        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: Alberta511HTMLFilesPattern: %s" % Alberta511HTMLFilesPattern )

        Alberta511HTMLFilesPattern = Alberta511HTMLFilesPattern.replace( FAKE_ASTERISK, "*" )
        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: Alberta511HTMLFilesPattern: %s" % Alberta511HTMLFilesPattern )

        Alberta511FilesPath = glob.glob( Alberta511HTMLFilesPattern )
        assert Alberta511FilesPath != None

        for FilePath in Alberta511FilesPath:
#            if DEBUG == True:
#                self.log.debug("Alberta511RoadConditions: FilePath: %s" % FilePath )

            CurrentAlberta511HTMLMapFiles.append( FilePath )

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: There are %d items in the currently existing HTML map file list." % len( CurrentAlberta511HTMLMapFiles ) )

#        for p in CurrentAlberta511HTMLMapFiles:
#            self.log.debug("p (FilePath): %s" % p )

        MapJavaScriptFilePath = ""

        if self.GoogleMapsMapsUse == True:
            # Create Java Script for Google Maps Map markers.

#            if DEBUG == True:
#                self.log.debug( "Alberta511RoadConditions: Creating JavaScript. . ." )

            ReturnedStatus = False

            ReturnedStatus, MapJavaScriptFilePath = self.CreateGoogleMapsJavaScript(
                                                                                     FilePathStemPrefix = JavaScriptFilePathStemPrefix,
                                                                                     Markers            = Markers,
                                                                                     AddStationMarker   = True
                                                                                   )
            assert ReturnedStatus == True       # CreateGoogleMapsJavaScript() should not fail!
            assert MapJavaScriptFilePath != ""

            if ReturnedStatus == False:
                # Something bad did happen so disable map creation.

                MapJavaScriptFilePath = ""

            if DEBUG == True:
                self.log.debug( "Alberta511RoadConditions: MapJavaScriptFilePath: %s" % MapJavaScriptFilePath )

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: %d Road Conditions were selected." % len( SelectedRoadConditions) )

#        self.log.debug( "SelectedRoadConditions BEFORE Sort:" )
#        for SelectedRoadCondition in SelectedRoadConditions:
#            self.log.debug( "Key: %s, %s, Index: %d" % ( SelectedRoadCondition.AreaName, SelectedRoadCondition.RoadwayName, SelectedRoadCondition.Index ) )

#        self.log.warning( "SelectedRoadConditions: Sort . . ." )

        SelectedRoadConditions.sort( key = Selected511AlbertaRoadConditionItem.GetRoadwayName )
        SelectedRoadConditions.sort( key = Selected511AlbertaRoadConditionItem.GetAreaName )

#        self.log.debug( "SelectedRoadConditions AFTER Sort:" )
#        for SelectedRoadCondition in SelectedRoadConditions:
#            self.log.debug( "Key: %s, %s, Index: %d" % ( SelectedRoadCondition.AreaName, SelectedRoadCondition.RoadwayName, SelectedRoadCondition.Index ) )

        HTML =   "<p class=\"WXFeedsP1\"><b>511 Alberta Road Conditions</b>\n"                                          \
                + "&nbsp;<a href=\"%s\">%s</a>" % ( ALBERTA_511_URL, "511 Alberta" )                                    \
                + "&nbsp;<a href=\"%s\"><span class=\"icon-play\"></span></a><br/>\n" % ALBERTA_511_URL                 \
                + "Displaying %d of %d Road Conditions" % ( len( SelectedRoadConditions ), len( RoadConditions ) )

        if AreaNameFilter != "":
            HTML += "<br/><b>Area Name Filter:</b> %s" % AreaNameFilter

        if RoadwayNameFilter != "":
           HTML += "<br/><b>Roadway Name Filter:</b> %s" % RoadwayNameFilter

        HTML +=   "</p>\n"      \
                + "<hr>\n"
        self.WriteGeneratedFiles( HTML )

        d = self.GetCurrentUTCAware()
#        self.log.debug( "Now (UTC): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z"  ) )
        d = self.GetTimeLocalAware( d )
#        self.log.debug( "Now (LT): %s" % d.strftime( "%b %-d %Y, %H:%M:%S %Z" ) )
                                
        LocalTimeZone = d.strftime( "%Z" )
#        self.log.debug( "Local Time Zone = %s", LocalTimeZone )
        
        DateTimeFormat = "%b %-d %Y, %H:%M " + LocalTimeZone

        for SelectedRoadCondition in SelectedRoadConditions:
#            RoadCondition = RoadConditions[ SelectedRoadCondition.Index ]
#            assert RoadCondition != None
#            self.log.debug( "RoadCondition: %s" % RoadCondition )

            MapHTMLFileName = ""

#            if DEBUG == True:
#                self.log.debug( "Alberta511RoadConditions: Create Maps: %s, Position Specified: %s, JavaScript FN: %s" % 
#                                   ( self.GoogleMapsMapsUse, SelectedRoadCondition.PositionSpecified, MapJavaScriptFilePath) )

            if self.GoogleMapsMapsUse == True and                    \
               MapJavaScriptFilePath != "" and                          \
               self.Alberta511RoadConditionCache[ SelectedRoadCondition.ID ][ VARIABLES_KEY][ POSITION_SPECIFIED_KEY ] == True:
                ReturnedStatus,         \
                MapHTMLFileName,        \
                MapHTMLFilePathAbsolute = self.CreateGoogleMapsMapHTML(
                                                                        FilePathStemPrefix            = HTMLFilePathStemPrefix + SelectedRoadCondition.ID,
                                                                        ForceStaleFile                = not self.Alberta511RoadConditionCache[ SelectedRoadCondition.ID ][ VARIABLES_KEY][ CACHE_HIT_KEY ],
                                                                        MapHeader                     = "511 Alberta Road Condition: "        \
                                                                                                            + RoadConditions[ SelectedRoadCondition.Index ][ 'LocationDescription' ],
                                                                        Latitude                      = self.Alberta511RoadConditionCache[ SelectedRoadCondition.ID ][ VARIABLES_KEY][ LATITUDE_KEY ],
                                                                        Longitude                     = self.Alberta511RoadConditionCache[ SelectedRoadCondition.ID ][ VARIABLES_KEY][ LONGITUDE_KEY ],
                                                                        Zoom                          = Zoom,
                                                                        EncodedPolyline               = RoadConditions[ SelectedRoadCondition.Index ][ 'EncodedPolyline' ].replace( "\\", "\\\\" ),
                                                                        JavaScriptFilePathRelative    = MapJavaScriptFilePath
                                                                      )
                if ReturnedStatus == True:
                    assert MapHTMLFileName != ""
                    assert MapHTMLFilePathAbsolute != ""

                    # Set CacheHit so that map is not forced to be stale on next cycle.

                    self.Alberta511RoadConditionCache[ SelectedRoadCondition.ID ][ VARIABLES_KEY][ CACHE_HIT_KEY ] = True

                   # Remove this map from the list of currently existing maps.

                    try:
                        CurrentAlberta511HTMLMapFiles.remove( MapHTMLFilePathAbsolute )
                    except ValueError as e:
                        # If this is a new road condition there won't be an existing file so not a error.

#                        if DEBUG == True:
#                            self.log.debug( "Alberta511RoadConditions: CurrentAlberta511HTMLMapFiles.remove() threw an exception: e = %s, sys.exc_type = %s" %
#                                        ( e, sys.exc_info() ) )

                        pass
                else:
                    # Disable map linkage.

                    MapHTMLFileName = ""
                 
            LastUpdatedDateTime = ""
            LastUpdated = RoadConditions[ SelectedRoadCondition.Index ][ 'LastUpdated' ]
            if LastUpdated != None: 
                LastUpdatedDateTime = datetime.datetime.fromtimestamp( LastUpdated ).strftime( DateTimeFormat )

            Conditions = RoadConditions[ SelectedRoadCondition.Index ][ 'Primary Condition' ]

            SecondaryConditions = RoadConditions[ SelectedRoadCondition.Index ][ 'Secondary Conditions' ]
#            self.log.debug( "SecondaryConditions: %s" % SecondaryConditions )

            sc = 0

            for SecondaryCondition in SecondaryConditions:
#                self.log.debug( "Secondary Condition: %s" % SecondaryCondition )

                if sc == 0:
                    Conditions += ": "
                    
                if sc > 0:
                    Conditions += ", "

                Conditions += SecondaryCondition

                sc += 1

             # Put each Road Condition into it's own table.

            HTML =   "<table class=\"WXFeedsT2\">\n"                                                                    \
                   +     "<thead>\n"                                                                                    \
                   +         "<tr>\n"                                                                                   \
                   +             "<th></th>\n"                                                                          \
                   +             "<th></th>\n"                                                                          \
                   +         "</tr>\n"                                                                                  \
                   +     "</thead>\n"                                                                                   \
                   +     "<tbody>\n"                                                                                    \
                   +         "<tr>\n"                                                                                   \
                   +             "<td><b>Roadway Name:&nbsp;</b></td>\n"                                                \
                   +             "<td>%s" % RoadConditions[ SelectedRoadCondition.Index ][ 'RoadwayName' ]
            self.WriteGeneratedFiles( HTML )

            if MapHTMLFileName != "":
                HTML = "&nbsp;<a href=\""
                self.WriteGeneratedFiles( HTML )
                self.WriteGeneratedFilesRelativeLink( MapHTMLFileName )
                HTML = "\"><span class=\"icon-google-map\"></span></a>\n"
                self.WriteGeneratedFiles( HTML )
                  
            HTML =   "</td>\n</tr>\n"                                                                                        \
                   +      "<tr>\n"                                                                                           \
                   +          "<td><b>Location Description:&nbsp;</b></td>\n"                                                \
                   +          "<td>%s</td>\n" % RoadConditions[ SelectedRoadCondition.Index ][ 'LocationDescription' ]       \
                   +      "</tr>\n"                                                                                          \
                   +      "<tr>\n"                                                                                           \
                   +          "<td><b>Condition(s):&nbsp;</b></td>\n"                                                        \
                   +          "<td>%s</td>\n" % Conditions                                                                   \
                   +      "</tr>\n"

            Visibility = RoadConditions[ SelectedRoadCondition.Index ][ 'Visibility' ]
#            self.log.debug( "Visibility: %s" % Visibility )

            if Visibility != None and   \
               Visibility != "":
                HTML +=   "<tr>\n"                                      \
                        +     "<td><b>Visibility:&nbsp;</b></td>\n"     \
                        +     "<td>%s</td>\n" % Visibility              \
                        + "</tr>\n"

            HTML +=   "<tr>\n"                                                                                                  \
                    +             "<td><b>Area Name - Id:&nbsp;</b></td>\n"                                                     \
                    +             "<td>%s - %s</td>\n" % (
                                                           RoadConditions[ SelectedRoadCondition.Index ][ 'AreaName' ],
                                                           RoadConditions[ SelectedRoadCondition.Index ][ 'Id' ]
                                                         )                                                                      \
                    +         "</tr>\n"                                                                                         \
                    +         "<tr>\n"                                                                                          \
                    +             "<td><b>Last Updated:&nbsp;</b></td>\n"                                                       \
                    +             "<td>%s</td>\n" % LastUpdatedDateTime                                                         \
                    +         "</tr>\n"

            HTML +=   "</tbody>\n"    \
                    + "</table>\n"    \
                    + "<hr>\n"
            self.WriteGeneratedFiles( HTML )

        if DisplayAttributeLists == YES:
            AreaNames.sort()
            RoadwayNames.sort()

            HTML =   "<p class=\"WXFeedsP2\">\n"                                \
                   + "<b>511 Alberta Road Condition Attribues:</b><br/>\n"      \
                   + "<b>Area Names:</b> %s<br/>\n" % AreaNames                 \
                   + "<b>Roadway Names:</b> %s<br/>\n" % RoadwayNames           \
                   + "</p>\n<hr>\n"
            self.WriteGeneratedFiles( HTML )

        ReturnStatus = True

       # Delete any map files still in the currently existing map file list.

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditions: There are %d items left in the currently existing HTML map file list." % len( CurrentAlberta511HTMLMapFiles ) )

        for HTMLMapFilePathAbsolute in CurrentAlberta511HTMLMapFiles:
            if DEBUG == True:
                self.log.debug( "Alberta511RoadConditions: HTMLMapFilePathAbsolute: %s" % HTMLMapFilePathAbsolute )

            p = Path( HTMLMapFilePathAbsolute )
            assert p != None    # Should never fail!

            if DEBUG == True:
                self.log.debug( "Alberta511RoadConditions: Deleting: %s" % HTMLMapFilePathAbsolute )

            try:
                p.unlink()
            except FileNotFoundError as e:
                # We created a list of map related files above so all files should still exist 
                # so we shouldn't get this error.

                self.log.error( "Alberta511RoadConditions: Threw an exception while deleteing map file: e = %s, sys.exc_type = %s" %
                                        ( e, sys.exc_info() ) )

            FileName = p.name
#            if DEBUG == True:
#                self.log.debug( "Alberta511RoadConditions: Map HTML Cache FileName: %s" % FileName )

            CacheFilePathAbsolute = self.GetCachePath( FileName )
            pc = Path( CacheFilePathAbsolute )

            if pc.exists() == True:     # We might not have a cache file for this map file.
#                if DEBUG == True:
#                    self.log.debug( "Alberta511RoadConditions: Deleting: %s" % CacheFilePathAbsolute )

                try:
                    pc.unlink()
                except FileNotFoundError as e:
                    # We shouldn't get his error because we checked existence above.

                    self.log.error( "Alberta511RoadConditions: Threw an exception while deleteing map cache file: e = %s, sys.exc_type = %s" %
                                          ( e, sys.exc_info() ) )

        # Purge the Road Codition cache of stale entries.
        #
        # A stale entry is one that does not have an entry in 
        # the current list of Road Conditions.

        # Sort RoadConditions by 'Id' so that we can use a binary search.

        # Zero fill 'Id' so that Road Conditions sort correctly.
    
        for RoadCondition in RoadConditions:
            RoadCondition[ 'Id' ] = str( RoadCondition[ 'Id' ] ).zfill( 10 )

        RoadConditions.sort( key = lambda s: s[ 'Id' ] )

        AllFound = True

        #for ID in self.Alberta511RoadConditionCache:
        for ID in list( self.Alberta511RoadConditionCache.keys() ):
            TargetID = ID.zfill( 10 )

#            self.log.debug( "TargetID: %s, ID: %s, Entry: %s" % ( TargetID, ID, self.Alberta511RoadConditionCache[ ID ] ) )

            Index   = -1
            Lower   = 0
            Upper   = len( RoadConditions ) - 1

            while Lower <= Upper:
                Middle = ( Lower + Upper ) // 2

                MiddleID = str( RoadConditions[ Middle ][ 'Id'] )
#                self.log.debug( "MiddleID: '%s'" % MiddleID )
            
                if MiddleID == TargetID:
                    Index = Middle
                    break

                if MiddleID < TargetID:
                    Lower = Middle + 1
                else:
                    Upper = Middle - 1

            if Index < 0:
                self.log.debug( "RoadCondition *NOT FOUND* For Position. TargetID: '%s'.  Deleting. . ." % TargetID )
#                self.log.debug( "len( RoadConditions ) BEFORE Delete: %d" % len( self.Alberta511RoadConditionCache ) )
                del self.Alberta511RoadConditionCache[ ID ]
#                self.log.debug( "len( RoadConditions ) AFTER Delete: %d" % len( self.Alberta511RoadConditionCache ) )

                AllFound = False

        if DEBUG == True:
            if AllFound == True:
                self.log.debug( "All Road Conditions Were Found." )
            else:
                self.log.debug( "All Road Conditions Were *NOT* Found." )

        if len( self.Alberta511RoadConditionCache ) > len( RoadConditions ):
            self.log.error( "Alberta511RoadConditions: Road Condition Cache Error - More Cache Entries Than Road Conditions!" )

        return ReturnStatus

    def Alberta511RoadConditionsValidateData(
                                              self,
                                              StatusCode            : int,
                                              RoadConditionsData    : dict
                                            ) -> bool:
                                            #    ReturnStatus

        ReturnStatus = False

        #DEBUG = True
        DEBUG = False

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditionsValidateData. . ." )

        # The json (RoadConditions) is just entries, no explicit status elements, so check the first entry, if there is one.

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditionsValidateData: StatusCode: %d" % StatusCode )

        if StatusCode == 200 and            \
           len( RoadConditionsData ) == 0:
            ReturnStatus = True

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditionsValidateData: len( RoadConditionsData ) == %d" % len( RoadConditionsData ) )
            
        if len( RoadConditionsData ) > 0:
            for RoadCondition in RoadConditionsData:
                ID = RoadCondition[ 'Id' ]

                if ID != "":
                    # Got 'Id'!

                    if DEBUG == True:
                        self.log.debug( "Alberta511RoadConditionsValidateData: json is valid!" )

                    ReturnStatus = True

                break

        if DEBUG == True:
            self.log.debug( "Alberta511RoadConditionsValidateData: Exiting. . ." )

        return ReturnStatus

    def PolylineMidpoint(
                          self,
                          Points    : []
                        ) -> int:
                        #    MidpointIndex

        MidpointIndex = -1

#        self.log.debug( "PolylineMidpoint: Points: %s" % Points )

#        self.log.debug( "PolylineMidpoint: len( Points ): %d" % ( len( Points ) ) )

        SegmentDistances = []

#        for Point in Points:
#            self.log.debug( "PolylineMidpoint: lat: %f, lon: %f" % ( Point[ 0 ], Point[ 1 ] ) )

        # Calculate TotalDistance while saving SegementDistances.

        TotalDistance = 0.0

        for i in range( len( Points ) - 1 ):
            SegmentDistance = haversine.haversine( Points[ i ], Points[ i + 1 ], unit=haversine.Unit.METERS )
            #self.log.debug( "PolylineMidpoint: SegmentDistance: %f" % ( SegmentDistance ) )

            TotalDistance += SegmentDistance

            SegmentDistances.append( SegmentDistance )

#        self.log.debug( "PolylineMidpoint: TotalDistance: %f" % ( TotalDistance ) )

        # Find the midpoint.

        MidpointDistance = TotalDistance / 2
#        self.log.debug( "PolylineMidpoint: MidpointDistance: %f" % ( MidpointDistance ) )

        DistanceFromStart = 0.0

        for SegmentDistance in SegmentDistances:

            DistanceFromStart += SegmentDistance

            if DistanceFromStart > MidpointDistance:
                MidpointIndex = SegmentDistances.index( SegmentDistance )
                break

#        self.log.debug( "PolylineMidpoint: MidpointIndex: %d" % ( MidpointIndex ) )

        return MidpointIndex

    def WXFeedsServerLogin(
                            self,
                            ServerURL   : str,
                            Username    : str,
                            Password    : str
                          ) -> ( bool,         str,         int ):
                          #      ReturnStatus, AccessToken, AccessTokenExpiry_seconds

        assert ServerURL != ""
        assert Username != ""
        assert Password != ""

        ReturnStatus = False

        AccessToken                 = ""
        AccessTokenExpiry_seconds   = -1

        #DEBUG = True
        DEBUG = False
        
        user = { "username" : self.WXFeedsServerUsername, "password" : self.WXFeedsServerPassword }
        if DEBUG == True:
            self.log.debug( "WXFeedsServerLogin: user: '%s'" % user )

        URL = ServerURL + "/login"
        if DEBUG == True:
            self.log.debug( "WXFeedsServerLogin: WXFeeds Server Login URL: '%s'" % URL )

        ResponseOK = True   # Assume success.

        try:
            response = requests.post( URL, json=user )
#                    self.log.debug( "WXFeeds Server response: %s" % response )
#                    self.log.debug( "WXFeeds Server response.status_code: %d" % response.status_code )
#                    self.log.debug( "WXFeeds Server response.text: %s" % response.text )
#                    self.log.debug( "WXFeeds Server response.json(): %s" % response.json() )
        except requests.exceptions.RequestException as e:
            self.log.error( "WXFeedsServerLogin: requests.get() returned RequestException!  URL = %s, e = %s, sys.exc_type = %s" % ( URL, e, sys.exc_info() ) )

            ResponseOK = False

        if ResponseOK == True and response.status_code == 200:
            response_json = response.json()

            if response_json[ 'loginStatus' ] == "Login Successful":
                # Login successful.

                if DEBUG == True:
                    self.log.debug( "WXFeedsServerLogin: Server Login Successful." )

                AccessToken = response_json[ 'accessToken' ]
                if DEBUG == True:
                    self.log.debug( "WXFeedsServerLogin: AccessToken: %s" % AccessToken )

                AccessTokenExpiry_seconds = int( response_json[ 'tokenExpiry' ] )
                if DEBUG == True:
                    self.log.info( "WXFeedsServerLogin: AccessTokenExpiry_seconds: %d" % AccessTokenExpiry_seconds )

                # Verify the access token that was returned above.

                # The access token really should be put into and Authorization Bearer header but we simply use 
                # URL query parameters for all data to be sent to a server including access tokens.  The 
                # access token is set to expire so the risk is limited.

                URL = ServerURL + "/verify?accessToken=" + AccessToken
                if DEBUG == True:
                    self.log.debug( "WXFeedsServerLogin: WXFeeds Server Verification URL: '%s'" % URL )

                ResponseOK = True   # Assume success.

                try:
                    response = requests.get( URL )
#                        self.log.debug( "WXFeeds: WXFeeds Server response: %s" % response )
#                        self.log.debug( "WXFeeds: WXFeeds Server response.status_code: %d" % response.status_code )
                except requests.exceptions.RequestException as e:
                    self.log.error( "WXFeedsServerLogin: requests.get() returned RequestException!  URL = %s, e = %s, sys.exc_type = %s" % ( URL, e, sys.exc_info() ) )

                    ResponseOK = False

                if ResponseOK == True and response.status_code == 200:
                    if DEBUG == True:
                        self.log.debug( "WXFeedsServerLogin: Server Login Verfication Successful." )

                    ReturnStatus = True
                else:
                    if DEBUG == True:
                        self.log.debug( "WXFeeWXFeedsServerLoginds: Login Verification NOT Successful!" )

                    ReturnStatus = False
        else:
            if DEBUG == True:
                self.log.debug( "WXFeedsServerLogin: WXFeeds Login NOT Successful!" )

            ReturnStatus = False
        
        return ReturnStatus, AccessToken, AccessTokenExpiry_seconds
